Package: scRepertoire
Title: A toolkit for single-cell immune receptor profiling
Version: 2.9.0
Authors@R: c(
    person(given = "Nick", family = "Borcherding", role = c("aut", "cre"), email = "ncborch@gmail.com"), 
    person(given = "Qile", family = "Yang", role = c("aut"), email = "qile.yang@berkeley.edu"), 
    person(given = "Ksenia", family = "Safina", role = c("aut"), email = "safina@broadinstitute.org"),
    person(given = "Justin", family = "Reimertz", role = c("ctb"), email = "justin.reimertz@childrens.harvard.edu"))
Description: scRepertoire is a toolkit for processing and analyzing single-cell T-cell receptor (TCR) and immunoglobulin (Ig). The scRepertoire framework supports use of 10x, AIRR, BD, MiXCR, TRUST4, and WAT3R single-cell formats. The functionality includes basic clonal analyses, repertoire summaries, distance-based clustering and interaction with the popular Seurat and SingleCellExperiment/Bioconductor R single-cell workflows.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
biocViews: Software, ImmunoOncology, SingleCell, Classification,
        Annotation, Sequencing
Depends: ggplot2, R (>= 4.0)
Imports: dplyr, evmix, ggalluvial, ggdendro, ggraph, grDevices, igraph,
        immApex, iNEXT, Matrix, quantreg, Rcpp, rjson, rlang,
        S4Vectors, SeuratObject, SingleCellExperiment,
        SummarizedExperiment, tidygraph, purrr, lifecycle, methods
Suggests: BiocManager, BiocStyle, circlize, knitr, Peptides, rmarkdown,
        scales, scater, Seurat, spelling, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Language: en-US
LinkingTo: Rcpp
URL: https://www.borch.dev/uploads/scRepertoire/
BugReports: https://github.com/BorchLab/scRepertoire/issues
Roxygen: list(markdown = TRUE)
git_url: https://git.bioconductor.org/packages/scRepertoire
git_branch: devel
git_last_commit: 7b864c4
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-29 01:06:49 UTC; biocbuild
Author: Nick Borcherding [aut, cre],
  Qile Yang [aut],
  Ksenia Safina [aut],
  Justin Reimertz [ctb]
Maintainer: Nick Borcherding <ncborch@gmail.com>

# Generated by roxygen2: do not edit by hand

S3method(quietBCRgenes,Seurat)
S3method(quietBCRgenes,default)
S3method(quietTCRgenes,Seurat)
S3method(quietTCRgenes,default)
export(StartracDiversity)
export(addVariable)
export(alluvialClones)
export(annotateInvariant)
export(clonalAbundance)
export(clonalBias)
export(clonalBin)
export(clonalCluster)
export(clonalCompare)
export(clonalDiversity)
export(clonalHomeostasis)
export(clonalLength)
export(clonalNetwork)
export(clonalOccupy)
export(clonalOverlap)
export(clonalOverlay)
export(clonalProportion)
export(clonalQuant)
export(clonalRarefaction)
export(clonalScatter)
export(clonalSizeDistribution)
export(combineBCR)
export(combineExpression)
export(combineTCR)
export(createHTOContigList)
export(exportClones)
export(expression2List)
export(getCirclize)
export(getContigDoublets)
export(highlightClones)
export(loadContigs)
export(percentAA)
export(percentGeneUsage)
export(percentGenes)
export(percentKmer)
export(percentVJ)
export(positionalEntropy)
export(positionalProperty)
export(quietBCRgenes)
export(quietTCRgenes)
export(quietVDJgenes)
export(subsetClones)
export(vizCirclize)
export(vizGenes)
import(ggplot2)
importFrom(Matrix,sparseMatrix)
importFrom(Rcpp,sourceCpp)
importFrom(S4Vectors,DataFrame)
importFrom(SeuratObject,Embeddings)
importFrom(SeuratObject,Idents)
importFrom(SingleCellExperiment,colData)
importFrom(SingleCellExperiment,reducedDim)
importFrom(SingleCellExperiment,reducedDimNames)
importFrom(SummarizedExperiment,"colData<-")
importFrom(SummarizedExperiment,colData)
importFrom(dplyr,"%>%")
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,coalesce)
importFrom(dplyr,count)
importFrom(dplyr,desc)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,select)
importFrom(dplyr,slice_max)
importFrom(dplyr,summarise)
importFrom(dplyr,summarize)
importFrom(dplyr,summarize_all)
importFrom(dplyr,ungroup)
importFrom(evmix,dgpd)
importFrom(evmix,fgpd)
importFrom(evmix,pgpd)
importFrom(evmix,qgpd)
importFrom(evmix,rgpd)
importFrom(ggalluvial,StatStratum)
importFrom(ggalluvial,geom_alluvium)
importFrom(ggalluvial,geom_flow)
importFrom(ggalluvial,geom_stratum)
importFrom(ggalluvial,to_lodes_form)
importFrom(ggdendro,dendro_data)
importFrom(ggdendro,label)
importFrom(ggdendro,segment)
importFrom(ggplot2,ggplot)
importFrom(ggraph,circle)
importFrom(ggraph,geom_edge_bend)
importFrom(ggraph,geom_node_point)
importFrom(ggraph,ggraph)
importFrom(ggraph,guide_edge_colourbar)
importFrom(ggraph,scale_edge_colour_gradientn)
importFrom(grDevices,hcl.colors)
importFrom(grDevices,hcl.pals)
importFrom(graphics,plot)
importFrom(iNEXT,ggiNEXT)
importFrom(iNEXT,iNEXT)
importFrom(igraph,"E<-")
importFrom(igraph,"V<-")
importFrom(igraph,E)
importFrom(igraph,V)
importFrom(igraph,`V<-`)
importFrom(igraph,as_adjacency_matrix)
importFrom(igraph,as_data_frame)
importFrom(igraph,cluster_edge_betweenness)
importFrom(igraph,cluster_fast_greedy)
importFrom(igraph,cluster_leiden)
importFrom(igraph,cluster_louvain)
importFrom(igraph,cluster_spinglass)
importFrom(igraph,cluster_walktrap)
importFrom(igraph,components)
importFrom(igraph,graph_from_data_frame)
importFrom(igraph,graph_from_edgelist)
importFrom(igraph,membership)
importFrom(igraph,set_vertex_attr)
importFrom(immApex,ace_richness)
importFrom(immApex,buildNetwork)
importFrom(immApex,calculateEntropy)
importFrom(immApex,calculateFrequency)
importFrom(immApex,calculateGeneUsage)
importFrom(immApex,calculateMotif)
importFrom(immApex,chao1_richness)
importFrom(immApex,d50_dom)
importFrom(immApex,getIR)
importFrom(immApex,gini_coef)
importFrom(immApex,gini_simpson)
importFrom(immApex,hill_q)
importFrom(immApex,inv_simpson)
importFrom(immApex,norm_entropy)
importFrom(immApex,pielou_evenness)
importFrom(immApex,shannon_entropy)
importFrom(lifecycle,deprecated)
importFrom(methods,is)
importFrom(methods,slot)
importFrom(quantreg,rqss)
importFrom(rjson,fromJSON)
importFrom(rlang,"%||%")
importFrom(rlang,":=")
importFrom(rlang,.data)
importFrom(rlang,sym)
importFrom(stats,as.dist)
importFrom(stats,as.formula)
importFrom(stats,ave)
importFrom(stats,hclust)
importFrom(stats,integrate)
importFrom(stats,mad)
importFrom(stats,na.omit)
importFrom(stats,optim)
importFrom(stats,pgamma)
importFrom(stats,qgamma)
importFrom(stats,qt)
importFrom(stats,quantile)
importFrom(stats,reshape)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,xtabs)
importFrom(tidygraph,activate)
importFrom(tidygraph,as_tbl_graph)
importFrom(utils,combn)
importFrom(utils,getFromNamespace)
importFrom(utils,read.csv)
importFrom(utils,read.delim)
importFrom(utils,stack)
importFrom(utils,write.csv)
useDynLib(scRepertoire, .registration = TRUE)

# scRepertoire: A toolkit for single-cell immune profiling
<!-- badges: start -->
[![BioC status](http://www.bioconductor.org/shields/build/release/bioc/scRepertoire.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/scRepertoire)
[![Bioc Devel Build](http://www.bioconductor.org/shields/build/devel/bioc/scRepertoire.svg)](http://www.bioconductor.org/checkResults/devel/bioc-LATEST/scRepertoire)
[![R-CMD-check](https://github.com/BorchLab/scRepertoire/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/BorchLab/scRepertoire/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/BorchLab/scRepertoire/branch/devel/graph/badge.svg)](https://app.codecov.io/gh/BorchLab/scRepertoire?branch=devel)
[![Bioc Downloads](http://www.bioconductor.org/shields/downloads/release/scRepertoire.svg)](http://bioconductor.org/packages/stats/bioc/scRepertoire/)
<!-- badges: end -->

<img align="right" src="https://www.borch.dev/uploads/screpertoire/reference/figures/screpertoire_hex.png" width="305" height="352">

### Introduction

Single-cell sequencing is an emerging technology in the field of immunology and oncology that allows researchers to couple RNA quantification and other modalities, like immune cell receptor profiling at the level of an individual cell. Unlike the transcriptomic field, there is a lack of options for software that allow for single-cell immune receptor profiling. Enabling users to easily combine RNA and immune profiling, the scRepertoire framework supports 10x, AIRR, BD, MiXCR, TRUST4, and WAT3R single-cell clonal formats and interaction with popular R-based single-cell data pipelines.

### Working with scRepertoire

scRepertoire has a comprehensive [website](https://www.borch.dev/uploads/screpertoire/) for detailed tutorials and function information.

### Quick Start

```R
library(scRepertoire)

# Load and combine contigs into clones
combined.TCR <- combineTCR(contig_list,
                           samples = c("P17B", "P17L", "P18B", "P18L",
                                       "P19B", "P19L", "P20B", "P20L"))

# Visualize clonal frequency
clonalQuant(combined.TCR, clone.call = "strict", chain = "both")

# Attach clonal data to a single-cell object
scRep_example <- combineExpression(combined.TCR, scRep_example)
```

### Installation

#### Development Branch
[immApex](https://github.com/BorchLab/immApex) is a required dependency. If not using Bioconductor, install both:

```R
remotes::install_github(c("BorchLab/immApex", "BorchLab/scRepertoire@devel"))
```

#### Bioconductor
The current stable version is available on [Bioconductor](https://www.bioconductor.org/packages/release/bioc/html/scRepertoire.html):

```R
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("scRepertoire")
```

#### Legacy Version 1
```R
devtools::install_github("BorchLab/scRepertoire@v1")
```

### Getting Data

GitHub limits the size of individual files. To access the full Seurat object paired with scRepertoire, download the .rda from [Zenodo](https://zenodo.org/records/18187313). A smaller version of the cohort is built into scRepertoire as **scRep_example**.

### Deep Learning Extensions

scRepertoire is compatible with [Trex](https://github.com/BorchLab/Trex) for deep-learning-based autoencoding of T cell receptors and [Ibex](https://github.com/BorchLab/Ibex) for B cell receptors. For building custom deep-learning models with immune receptors, see [immApex](https://github.com/BorchLab/immApex).

### Citation

* **Version 2**: Yang, Q, & Safina, K., Nguyen, K., Tuong, Z.K., & Borcherding, N. (2025). "scRepertoire 2: Enhanced and efficient toolkit for single-cell immune profiling." *PLoS Computational Biology* https://doi.org/10.1371/journal.pcbi.1012760
* **Version 1**: Borcherding, Nicholas, Nicholas L. Bormann, and Gloria Kraus. "scRepertoire: An R-based toolkit for single-cell immune receptor analysis." *F1000Research* https://doi.org/10.12688/f1000research.22139.2

If you are building your own tool based on scRepertoire, reach out, we are happy to help and make things compatible.

### Bug Reports / Feature Requests

- Submit issues on [GitHub](https://github.com/BorchLab/scRepertoire/issues). If possible, include a [reproducible example](https://reprex.tidyverse.org/) using the built-in **scRep_example** and **contig_list** data.
- [Pull requests](https://github.com/BorchLab/scRepertoire/pulls) are welcome. Please target the "devel" branch.

#' Plot powerTCR Clustering Based on Clonal Size
#'
#' This function produces a hierarchical clustering of clones by sample 
#' using discrete gamma-GPD spliced threshold model. If using this 
#' model please read and cite powerTCR (more info available at 
#' [PMID: 30485278](https://pubmed.ncbi.nlm.nih.gov/30485278/)).
#' 
#' @details
#' The probability density function (pdf) for the **Generalized Pareto Distribution (GPD)** is given by:
#'  \deqn{f(x|\mu, \sigma, \xi) = \frac{1}{\sigma} \left( 1 + \xi \left( \frac{x - \mu}{\sigma} \right) \right)^{-\left( \frac{1}{\xi} + 1 \right)}}
#' 
#' Where:
#' \itemize{
#'   \item{\eqn{\mu} is a location parameter}
#'   \item{\eqn{\sigma > 0} is a scale parameter}
#'   \item{\eqn{\xi} is a shape parameter}
#'   \item{\eqn{x \ge \mu} if \eqn{\xi \ge 0} and \eqn{\mu \le x \le \mu - \sigma/\xi} if \eqn{\xi < 0}}
#' }
#'               
#' The probability density function (pdf) for the **Gamma Distribution** is given by:
#' \deqn{f(x|\alpha, \beta) = \frac{x^{\alpha-1} e^{-x/\beta}}{\beta^\alpha \Gamma(\alpha)}}
#' 
#' Where:
#' \itemize{
#'   \item{\eqn{\alpha > 0} is the shape parameter}
#'   \item{\eqn{\beta > 0} is the scale parameter}
#'   \item{\eqn{x \ge 0}}
#'   \item{\eqn{\Gamma(\alpha)} is the gamma function of \eqn{\alpha}}
#' }
#' 
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list,
#'                        samples = c("P17B", "P17L", "P18B", "P18L",
#'                                    "P19B","P19L", "P20B", "P20L"))
#' 
#' # Using clonalSizeDistribution()
#' clonalSizeDistribution(combined,
#'                        clone.call = "strict",
#'                        method="ward.D2")
#'
#' @param input.data The product of [combineTCR()],
#' [combineBCR()], or [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param threshold Numerical vector containing the thresholds
#' the grid search was performed over.
#' @param method The clustering parameter for the dendrogram.
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed as
#' by list element or active identity in the case of single-cell objects.
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals].
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @importFrom ggdendro dendro_data segment label
#' @importFrom stats hclust optim pgamma as.dist as.formula qgamma runif xtabs
#' @export
#' @concept Visualizing_Clones
#' @return A ggplot object visualizing dendrogram of clonal size distribution
#'  or a data.frame if `exportTable = TRUE`.
#' @author Hillary Koch
#'
clonalSizeDistribution <- function(input.data,
                                   clone.call = NULL,
                                   chain = "both",
                                   method = "ward.D2",
                                   threshold = 1,
                                   group.by = NULL,
                                   export.table = NULL,
                                   palette = "inferno",
                                   # Deprecated arguments
                                   cloneCall = NULL,
                                   exportTable = NULL,
                                   ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "clonalSizeDistribution", default = "strict")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "clonalSizeDistribution", default = FALSE)

  x <- y <- xend <- yend <- NULL
  input.data <- .dataWrangle(input.data,
                             group.by,
                             .theCall(input.data, clone.call,
                                      check.df = FALSE, silent = TRUE),
                             chain)
  clone.call <- .theCall(input.data, clone.call)
  sco <- .is.seurat.or.se.object(input.data)
  if(!is.null(group.by) & !sco) {
    input.data <- .groupList(input.data, group.by)
  }
  data <- bind_rows(input.data)
  unique_df<- unique(data[,clone.call])

  # Create long-format summary table
  summary_df <- dplyr::bind_rows(input.data, .id = "sample") %>%
    dplyr::group_by(sample, .data[[clone.call]]) %>%
    dplyr::summarise(Freq = dplyr::n(), .groups = 'drop')
  wide_matrix <- xtabs(as.formula(paste("Freq ~", clone.call, "+ sample")),
                       data = summary_df)
  Con.df <- as.data.frame.matrix(wide_matrix)
  Con.df[[clone.call]] <- rownames(Con.df)
  rownames(Con.df) <- NULL
  Con.df <- Con.df[, c(clone.call, setdiff(names(Con.df), clone.call))]
  
  # Fit models
  list <- lapply(seq_len(ncol(Con.df))[-1], function(x) {
    suppressWarnings(.fdiscgammagpd(Con.df[,x], useq = threshold))
  })
  
  names(list) <- names(input.data)
  grid <- 0:10000
  distances <- .get_distances(list, grid, modelType="Spliced")
  mat_melt <- dendro_data(hclust(as.dist(distances), method = method), 
                          type = "rectangle")
  
  #Plotting
  plot <- ggplot() + 
            geom_segment(data = segment(mat_melt), 
                         aes(x = x, 
                             y = y, 
                             xend = xend, 
                             yend = yend)) +
            geom_text(data = label(mat_melt), 
                              aes(x = x, y = -0.02, 
                                  label = label, 
                                  hjust = 0), 
                      size = 4) +
            geom_point(data = label(mat_melt), 
                              aes(x = x, 
                                  y = -0.01, 
                                  color = as.factor(label)), 
                       size = 2) + 
            coord_flip() +
            scale_y_reverse(expand = c(0.2, 0)) + 
            scale_color_manual(values = .colorizer(palette, nrow(label(mat_melt)))) + 
            .themeRepertoire(..., grid_lines = "none") + 
            guides(color = "none") + 
            theme(axis.title = element_blank(), 
                  axis.ticks.y = element_blank(), 
                  axis.text.y = element_blank()) 
  
  if (export.table) {
    return(distances)
  }
  return(plot)
}

#################################################################
# Section 1: Main Fitting Engine
#################################################################
# Fit a Spliced Gamma and GPD Model
#' @importFrom evmix fgpd
#' @importFrom methods is
.fdiscgammagpd <- function(x, useq, shift = NULL, pvector=NULL,
                          std.err = TRUE, method = "Nelder-Mead", ...){
  if(!is(x, "numeric")){
    stop("x must be numeric.")
  }
  
  if(!is.null(shift)){
    if(!is(shift, "numeric")){
      stop("shift must be numeric.")
    }
    if(shift != round(shift)){
      stop("shift must be an integer.")
    }
  }
  
  if(!is(useq, "numeric")){
    stop("useq must be numeric.")
  }
  
  if(any(x != round(x))){
    stop("all elements in x must be integers.")
  }
  
  if(any(useq != round(useq))){
    stop("all elements in useq must be integers.")
  }
  
  if(!is.null(pvector) & !(length(pvector) == 5)){
    stop("pvector must contain 5 elements.")
  }
  
  if(!(is.logical(std.err))){
    stop("std.err must be TRUE or FALSE.")
  }
  
  if(is.null(shift)){
    shift <- min(x)
  }
  
  if (is.null(pvector)) {
    pvector <- rep(NA,5)
    s <- log(mean(x+0.5))-mean(log(x+0.5))
    k <- (3-s + sqrt((s-3)^2 + 24*s))/12/s
    pvector[1] <- k
    pvector[2] <- k/mean(x)
    pvector[3] <- as.vector(quantile(x, 0.9))
    
    xu <- x[x>=pvector[3]]
    initfgpd <- fgpd(xu, min(xu)-10^(-5))
    pvector[4] <- initfgpd$mle[1]
    pvector[5] <- initfgpd$mle[2]
  }
  
  bulk <- lapply(seq_along(useq),
                 function(idx,x,useq) x < useq[idx], x=x, useq=useq)
  tail <- lapply(seq_along(useq),
                 function(idx,x,useq) x >= useq[idx], x=x, useq=useq)
  phiu <- lapply(seq_along(useq),
                 function(idx,tail) mean(tail[[idx]]), tail=tail)
  
  gammfit <- list()
  gpdfit <- list()
  nllhu <- rep(NA, length(useq))
  for(i in seq_along(useq)){
    gammfit[[i]] <- tryCatch(expr = .fdiscgamma(pvector[1:2],x[bulk[[i]]],
                                               useq[i],
                                               phiu[[i]],
                                               shift,
                                               method="Nelder-Mead"),
                             error = function(err) NA)
    gpdfit[[i]] <- tryCatch(expr = .fdiscgpd(pvector[4:5],
                                            x[tail[[i]]],
                                            useq[i],
                                            phiu[[i]],
                                            method="Nelder-Mead"),
                            error = function(err) {
                              pvec3 <- as.vector(quantile(x,1-phiu[[i]]))
                              xu <- x[x>=pvec3]
                              initfgpd.adj <-
                                fgpd(x, min(xu)-10^(-5))
                              pvec4 <- initfgpd.adj$mle[1]
                              pvec5 <- initfgpd.adj$mle[2]
                              tryCatch(expr = .fdiscgpd(c(pvec4,pvec5),
                                                       x[tail[[i]]],
                                                       useq[i],
                                                       phiu[[i]],
                                                       method="Nelder-Mead"),
                                       error = function(err2) NA)
                            })
    nllhu[i] <- tryCatch(expr = gammfit[[i]]$value + gpdfit[[i]]$value,
                         error = function(err) NA)
  }
  
  bestfit <- which.min(nllhu)
  fit.out <- list(gammfit[[bestfit]], gpdfit[[bestfit]])
  names(fit.out) <- c("bulk","tail")
  mle <- c(mean(x >= useq[bestfit]),
           exp(fit.out$bulk$par),
           useq[bestfit],
           exp(fit.out$tail$par[1]),
           fit.out$tail$par[2])
  names(mle) <- c("phi","shape","rate","thresh","sigma","xi")
  if(std.err){
    H <- fit.out$bulk$hessian %>% rbind(matrix(rep(0,4),nrow = 2)) %>%
      cbind(rbind(matrix(rep(0,4),nrow = 2),fit.out$tail$hessian))
    fisherInf <- tryCatch(expr = solve(H), error = function(err) NA)
    out <- list(x = as.vector(x), shift = shift, init = as.vector(pvector),
                useq = useq, nllhuseq = nllhu,
                optim = fit.out, nllh = nllhu[bestfit],
                mle=mle, fisherInformation = fisherInf)
  } else{
    out <- list(x = as.vector(x), shift = shift, init = as.vector(pvector),
                useq = useq, nllhuseq = nllhu,
                optim = fit.out, nllh = nllhu[bestfit],
                mle=mle)
  }
  out
}

#################################################################
# Section 2: Core Modeling Functions
#################################################################

# Discrete Gamma Probability Mass Function
#' @importFrom methods is
.ddiscgamma <- function(x, shape, rate, thresh, phiu, shift = 0, log = FALSE){
  if(any(x != floor(x))){
    stop("x must be an integer")
  }
  
  out <- rep(0, length(x))
  
  up <- pgamma(x+1-shift, shape=shape, rate=rate)
  down <- pgamma(x-shift, shape=shape, rate=rate)
  
  if(!log){
    b <- pgamma(thresh-shift, shape=shape, rate=rate)
    out[x < thresh] <- ((1-phiu)*(up-down)/b)[x < thresh]
  } else{
    b <- pgamma(thresh-shift, shape=shape, rate=rate, log.p = TRUE)
    out[x < thresh] <- (log(1-phiu)+log(up-down) - b)[x < thresh]
  }
  out
}

.pdiscgamma <- function(q, shape, rate, thresh, phiu, shift = 0){
  probs <- .ddiscgamma(0:q, shape, rate, thresh, phiu, shift)
  sum(probs)
}

.qtruncgamma_internal <- function(p, shape, rate, a=0, b=Inf) {
  Fa <- pgamma(a, shape, rate)
  Fb <- pgamma(b, shape, rate)
  p_adjusted <- p * (Fb - Fa) + Fa
  qgamma(p_adjusted, shape, rate)
}

.qdiscgamma <- function(p, shape, rate, thresh, phiu, shift = 0){
  .qtruncgamma_internal(p/(1-phiu), 
                        shape=shape, 
                        rate=rate, 
                        b=thresh-shift) %>% 
    floor() + shift
}

.rtruncgamma_internal <- function(n, shape, rate, a = 0, b = Inf) {
  Fa <- pgamma(a, shape, rate)
  Fb <- pgamma(b, shape, rate)
  u <- runif(n, min = Fa, max = Fb)
  qgamma(u, shape, rate)
}

.rdiscgamma <- function(n, shape, rate, thresh, shift = 0){
  .rtruncgamma_internal(n, 
                        shape = shape, 
                        rate = rate, 
                        b = thresh - shift) %>% 
    floor() + shift
}

# Discrete GPD Probability Mass Function
#' @importFrom evmix pgpd
.ddiscgpd <- function(x, thresh, sigma, xi, phiu, log = FALSE){
  up <- pgpd(x+1, u=thresh, sigmau=sigma, xi=xi)
  down <- pgpd(x, u=thresh, sigmau=sigma, xi=xi)
  
  if(!log){
    phiu*(up-down)
  } else{
    log(phiu) + log(up-down)
  }
}

.pdiscgpd <- function(q, thresh, sigma, xi, phiu){
  probs <- .ddiscgpd(thresh:q, thresh, sigma, xi, phiu)
  sum(probs)
}

#' @importFrom evmix qgpd
.qdiscgpd <- function(p, thresh, sigma, xi, phiu){
  qgpd(p/phiu, u=thresh, sigmau = sigma, xi=xi) %>% floor
}

#' @importFrom evmix rgpd
.rdiscgpd <- function(n, thresh, sigma, xi){
  rgpd(n, u=thresh, sigmau=sigma, xi=xi) %>% floor
}

# Spliced Gamma-GPD Probability Mass Function
.ddiscgammagpd <- function(x, fit=NULL, shape, rate, u, sigma, xi,
                           phiu=NULL, shift = 0, log = FALSE){
  if(!is.null(fit)){
    if(!all(c("x", "shift", "init", "useq", "nllhuseq", "nllh",
              "optim", "mle") %in% names(fit))){
      stop("\"fit\" is not of the correct structure. It must be one model
                 fit from fdiscgammagpd.")
    }
    phiu <- fit$mle['phi']
    shape <- fit$mle['shape']
    rate <- fit$mle['rate']
    u <- fit$mle['thresh']
    sigma <- fit$mle['sigma']
    xi <- fit$mle['xi']
    shift <- fit$shift
  }
  if(!is(x, "numeric")){
    stop("x must be numeric.")
  }
  
  if(!is(shift, "numeric")){
    stop("shift must be numeric.")
  }
  
  if(any(x != floor(x))){
    stop("x must be an integer")
  }
  
  if(shift != round(shift)){
    stop("shift must be an integer.")
  }
  
  if(any(c(shape, rate, sigma) <= 0)){
    stop("shape, rate, and sigma must all be positive.")
  }
  
  if(!is.null(phiu)){
    if(phiu < 0 | phiu > 1){
      stop("phiu must be in [0,1].")
    }
  }
  
  if(is.null(phiu)){
    phiu <- 1-.pdiscgamma(u-1, shape=shape, rate=rate,
                          thresh=Inf, phiu = 0, shift=shift)
  }
  
  out <- rep(NA, length(x))
  
  if(sum(x>=u) != 0){
    out[x>=u] <- .ddiscgpd(x[x>=u], u, sigma, xi, phiu, log=log)
  }
  
  if(sum(x<u) != 0){
    out[x<u] <- .ddiscgamma(x[x<u], shape, rate, u, phiu, shift, log=log)
  }
  out
}

# --- Negative Log-Likelihood (NLL) Functions ---

# NLL for Discrete Gamma
.discgammanll <- function(param, dat, thresh, phiu, shift=0){
  shape <- exp(param[1])
  rate <- exp(param[2])
  
  if(any(dat > thresh-1)){ warning("data must be less than the threshold") }
  
  ll <- log(.ddiscgamma(dat, shape, rate, thresh, phiu, shift))
  
  sum(-ll)
}

# NLL for Discrete GPD
.discgpdnll <- function(param, dat, thresh, phiu){
  sigma <- exp(param[1])
  xi <- param[2]
  ll <- log(.ddiscgpd(dat, thresh, sigma, xi, phiu))
  
  sum(-ll)
}

# --- Fitting Wrappers ---

# Wrapper for Gamma Optimization
.fdiscgamma <- function(param, dat, thresh, phiu, shift = 0, method, ...){
  opt <- optim(log(param), .discgammanll, dat=dat, thresh=thresh,
               phiu=phiu, shift=shift, method=method, hessian = TRUE, ...)
  opt
}

# Wrapper for GPD Optimization
.fdiscgpd <- function(param, dat, thresh, phiu, method, ...){
  opt <- optim(c(log(param[1]),param[2]), .discgpdnll, dat=dat, thresh=thresh,
               phiu=phiu, method=method, hessian = TRUE, ...)
  opt
}

#################################################################
# Section 3: Distance Calculation
#################################################################

# Calculate Pairwise Distances
#' @importFrom methods is
.get_distances <- function(fits, grid, modelType = "Spliced"){
  if(!is(grid, "numeric")){
    stop("grid must be numeric.")
  }
  if(any(grid != round(grid))){
    stop("all elements in grid must be integers.")
  }
  if(!(modelType %in% c("Spliced", "Desponds"))){
    stop("modelType must be either \"Spliced\" or \"Desponds\".")
  }
  
  distances <- matrix(rep(0, length(fits)^2), nrow = length(fits))
  if(!is.null(names(fits))){
    rownames(distances) <- colnames(distances) <- names(fits)
  }
  
  for(i in seq_len((length(fits)-1))){
    for(j in (i+1):length(fits)){
      distances[i,j] <- .JS_dist(fits[[i]],
                                 fits[[j]],
                                 grid,
                                 modelType = modelType)
    }
  }
  distances <- distances + t(distances)
  distances
}

.JS_dist <- function(fit1, fit2, grid, modelType = "Spliced"){
  if(!(modelType %in% c("Spliced", "Desponds"))){
    stop("modelType must be either \"Spliced\" or \"Desponds\".")
  }
  
  if(modelType == "Spliced"){
    if(!all(c("x", "init", "useq", "nllhuseq", "nllh",
              "optim", "mle") %in% names(fit1))){
      stop("\"fit1\" is not of the correct structure. It must be a model
                 fit from fdiscgammagpd.")
    }
    if(!all(c("x", "init", "useq", "nllhuseq", "nllh",
              "optim", "mle") %in% names(fit2))){
      stop("\"fit2\" is not of the correct structure. It must be a model
                 fit from fdiscgammagpd.")
    }
    shiftp <- fit1$shift
    shiftq <- fit2$shift
    phip <- fit1$mle['phi']
    phiq <- fit2$mle['phi']
    shapep <- fit1$mle['shape']
    shapeq <- fit2$mle['shape']
    ratep <- fit1$mle['rate']
    rateq <- fit2$mle['rate']
    threshp <- fit1$mle['thresh']
    threshq <- fit2$mle['thresh']
    sigmap <- fit1$mle['sigma']
    sigmaq <- fit2$mle['sigma']
    xip <- fit1$mle['xi']
    xiq <- fit2$mle['xi']
    
    out <- .JS_spliced(grid, shiftp, shiftq, phip, phiq, shapep, shapeq,
                       ratep, rateq, threshp, threshq, sigmap, sigmaq,
                       xip, xiq)
  } else if(modelType == "Desponds"){
    if(!all(c("min.KS", "Cmin", "powerlaw.exponent",
              "pareto.alpha") == names(fit1))){
      stop("\"fit1\" is not of the correct structure. It must be a model
                 fit from fdesponds.")
    }
    if(!all(c("min.KS", "Cmin", "powerlaw.exponent",
              "pareto.alpha") == names(fit2))){
      stop("\"fit2\" is not of the correct structure. It must be a model
                 fit from fdesponds.")
    }
    Cminp <- fit1['Cmin']
    Cminq <- fit2['Cmin']
    alphap <- fit1['pareto.alpha']
    alphaq <- fit2['pareto.alpha']
    out <- .JS_desponds(grid, Cminp, Cminq, alphap, alphaq)
  }
  out
}

# Jensen-Shannon Distance Calculation
#' @importFrom evmix dgpd
#' @importFrom methods is
.JS_spliced <- function(grid, shiftp, shiftq, phip, phiq, shapep, shapeq, ratep,
                        rateq, threshp, threshq, sigmap, sigmaq, xip, xiq){
  if(!is(grid, "numeric")){
    stop("grid must be numeric.")
  }
  
  if(any(grid != round(grid))){
    stop("all elements in grid must be integers.")
  }
  
  if(any(!is(c(shiftp, shiftq, phip, phiq, shapep, shapeq,
               ratep, rateq, threshp, threshq,
               sigmap, sigmaq, xip, xiq), "numeric"))){
    stop("shiftp, shiftq, phip, phiq, shapep, shapeq, ratep, rateq,
              threshp, threshq, sigmap, sigmaq, xip, and xiq must be numeric.")
  }
  
  if(shiftp != round(shiftp) | shiftq != round(shiftq)){
    stop("shiftp and shiftq must be integers.")
  }
  
  if(any(c(shapep, shapeq, ratep, rateq, sigmap, sigmaq) <= 0)){
    stop("shapep, shapeq, ratep, rateq, sigmap, and sigmaq must be 
             greater than 0.")
  }
  
  if(any(c(phip, phiq) > 1) | any(c(phip, phiq) < 0)){
    stop("phip and phiq must be in [0,1].")
  }
  
  if(ratep <= 0 | rateq <= 0){
    stop("ratep and rateq must be greater than 0.")
  }
  
  if(shapep <= 0 | shapeq <= 0){
    stop("shapep and shapeq must be greater than 0.")
  }
  
  if(threshp != round(threshp) | threshq != round(threshq)){
    stop("threshp and threshq must be integers.")
  }
  
  K <- max(grid)
  
  P <- .ddiscgammagpd(min(grid):K, shape = shapep, rate = ratep,
                      u=threshp, sigma = sigmap,
                      xi = xip, phiu = phip, shift=shiftp,
                      log = FALSE)
  adjp <- which(P == 0)
  if(length(adjp) != 0){
    P[adjp] <- dgpd(adjp+0.5, u=threshp,
                    sigmau = sigmap, xi = xip, phiu = phip)
  }
  
  Q <- .ddiscgammagpd(min(grid):K, shape = shapeq, rate = rateq,
                      u=threshq, sigma = sigmaq,
                      xi = xiq, phiu = phiq, shift=shiftq,
                      log = FALSE)
  adjq <- which(Q == 0)
  if(length(adjq) != 0){
    Q[adjq] <- dgpd(adjq+0.5, u=threshq,
                    sigmau = sigmaq, xi = xiq, phiu = phiq)
  }
  
  M <- 0.5*(P+Q)
  pzero <- which(P == 0)
  qzero <- which(Q == 0)
  
  sum1 <- sum2 <- rep(NA, length(grid))
  sum1 <- P*(log(P) - log(M))
  sum2 <- Q*(log(Q) - log(M))
  
  if(length(intersect(pzero, qzero)) != 0){
    sum1[intersect(pzero, qzero)] <- 0
    sum2[intersect(pzero, qzero)] <- 0
  }
  if(length(setdiff(pzero, qzero)) != 0){
    sum1[setdiff(pzero, qzero)] <- 0
  }
  if(length(setdiff(qzero, pzero)) != 0){
    sum2[setdiff(qzero, pzero)] <- 0
  }
  
  out <- sqrt(0.5*(sum(sum1) + sum(sum2)))
  out
}

# Core JS Distance Logic
#' @importFrom stats integrate
#' @importFrom methods is
.JS_desponds <- function(grid, Cminp, Cminq, alphap, alphaq){
  if(!is(grid, "numeric")){
    stop("grid must be numeric.")
  }
  
  if(any(grid != round(grid))){
    stop("all elements in grid must be integers.")
  }
  
  if(any(!is(c(Cminp, Cminq, alphap, alphaq), "numeric"))){
    stop("Cminp, Cminq, alphap, and alphaq must be numeric.")
  }
  
  if(Cminp != round(Cminp) | Cminq != round(Cminq)){
    stop("Cminp and Cminq must be integers.")
  }
  
  if(alphap <= 0 | alphaq <= 0){
    stop("alphap and alphaq must be greater than 0.")
  }
  
  lower <- min(grid)
  upper <- max(grid)
  
  out <- integrate(.eval_desponds,
                   lower = lower, upper = upper,
                   Cminp = Cminp, Cminq = Cminq,
                   alphap = alphap, alphaq = alphaq)$value %>% sqrt
  out
}

.dpareto_internal <- function(x, scale, shape) {
  ifelse(x < scale, 0, (shape * (scale^shape)) / (x^(shape + 1)))
}

.eval_desponds <- function(t, Cminp, Cminq, alphap, alphaq){
  M <- 0.5*(.dpareto_internal(t, scale=Cminp, shape=alphap) +
              .dpareto_internal(t, scale=Cminq, shape=alphaq))
  
  one <- .dpareto_internal(t, scale=Cminp, shape=alphap)
  two <- .dpareto_internal(t, scale=Cminq, shape=alphaq)
  
  if(one == 0 & two == 0){
    out <- 0
  } else if(one == 0 & two != 0){
    out <- .dpareto_internal(t, scale=Cminq, shape=alphaq) *
      (log(.dpareto_internal(t, scale=Cminq, shape=alphaq))-log(M))
  } else if(one != 0 & two == 0){
    out <- .dpareto_internal(t, scale=Cminp, shape=alphap) *
      (log(.dpareto_internal(t, scale=Cminp, shape=alphap))-log(M))
  } else{
    out <- .dpareto_internal(t, scale=Cminp, shape=alphap) *
      (log(.dpareto_internal(t, scale=Cminp, shape=alphap))-log(M)) +
      .dpareto_internal(t, scale=Cminq, shape=alphaq) *
      (log(.dpareto_internal(t, scale=Cminq, shape=alphaq))-log(M))
  }
  out
}

# test script for clonalSizeDistribution.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list,
                      samples = c("P17B", "P17L", "P18B", "P18L",
                                  "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined,
                        variable.name = "Type",
                        variables = rep(c("B", "L"), 4))

test_that("Output formats (distance matrix vs. plot) are correct", {
  # Test that export.table = TRUE returns a matrix
  dist_matrix <- clonalSizeDistribution(combined[1:2], export.table = TRUE)
  expect_true(is.matrix(dist_matrix))
  
  # Test that the default behavior returns a ggplot object
  plot_output <- clonalSizeDistribution(combined[1:2])
  expect_s3_class(plot_output, "ggplot")
})

test_that("Distance matrix has correct properties", {
  dist_matrix <- clonalSizeDistribution(combined[1:3], export.table = TRUE)
  num_samples <- length(combined[1:3])
  
  # Matrix should be square and have correct dimensions
  expect_equal(nrow(dist_matrix), num_samples)
  expect_equal(ncol(dist_matrix), num_samples)
  
  # Should have correct row and column names
  expect_equal(rownames(dist_matrix), names(combined)[1:3])
  expect_equal(colnames(dist_matrix), names(combined)[1:3])
  
  # Diagonal should be zero
  expect_true(all(diag(dist_matrix) == 0))
  
  # Matrix should be symmetric
  expect_equal(dist_matrix, t(dist_matrix))
  
  # All values should be non-negative
  expect_true(all(dist_matrix >= 0))
})

test_that("`group.by` parameter correctly aggregates data", {
  dist_matrix_grouped <- clonalSizeDistribution(combined, group.by = "Type", export.table = TRUE)
  num_groups <- 2
  expect_equal(nrow(dist_matrix_grouped), num_groups)
  expect_equal(ncol(dist_matrix_grouped), num_groups)
  expect_true(all(rownames(dist_matrix_grouped) %in% c("B", "L")))
})

test_that("`method` parameter is passed to hclust correctly", {
  expect_silent(clonalSizeDistribution(combined[1:2], method = "ward.D2"))
  expect_silent(clonalSizeDistribution(combined[1:2], method = "complete"))
  expect_silent(clonalSizeDistribution(combined[1:2], method = "single"))
})

test_that("`clone.call` and `chain` parameters execute correctly", {
  expect_silent(
    res_aa <- clonalSizeDistribution(combined[1:2], clone.call = "aa", export.table = TRUE)
  )
  expect_true(is.matrix(res_aa))
  
  expect_silent(
    res_tra <- clonalSizeDistribution(combined[1:2], chain = "TRA", export.table = TRUE)
  )
  expect_true(is.matrix(res_tra))
})

test_that("Plotting output has the correct dendrogram components", {
  plot_output <- clonalSizeDistribution(combined[5:6])
  expect_s3_class(plot_output$layers[[1]]$geom, "GeomSegment")
  expect_s3_class(plot_output$layers[[2]]$geom, "GeomText")
  expect_s3_class(plot_output$layers[[3]]$geom, "GeomPoint")
  expect_s3_class(plot_output$coordinates, "CoordFlip")
  expect_true(plot_output$scales$scales[[1]]$trans$name == "reverse")
})

#################################################################
# Section 2: Testing Core Modeling Functions
#################################################################

test_that(".fdiscgammagpd returns a correctly structured list", {
  # Generate sample data that looks like clonal counts
  set.seed(42)
  x <- rpois(100, lambda = 2) + 1 
  
  fit <- .fdiscgammagpd(x, useq = 5)
  
  expect_type(fit, "list")
  expect_named(fit, c("x", "shift", "init", "useq", "nllhuseq", "optim", "nllh", "mle", "fisherInformation"))
  expect_type(fit$mle, "double")
  expect_equal(length(fit$mle), 6)
  expect_named(fit$mle, c("phi", "shape", "rate", "thresh", "sigma", "xi"))
})

test_that("NLL functions return a single positive number", {
  # Mock parameters and data
  params <- c(1.5, 2.5) # Log-scale parameters for shape, rate
  data <- c(1, 2, 3, 4)
  
  nll_gamma <- .discgammanll(log(params), dat = data, thresh = 5, phiu = 0.1, shift = 0)
  expect_type(nll_gamma, "double")
  expect_length(nll_gamma, 1)
  expect_true(nll_gamma > 0)
  
  # Mock parameters for GPD
  gpd_params <- c(log(1), 0.1) # log(sigma), xi
  gpd_data <- c(6, 7, 8)
  
  nll_gpd <- .discgpdnll(gpd_params, dat = gpd_data, thresh = 5, phiu = 0.1)
  expect_type(nll_gpd, "double")
  expect_length(nll_gpd, 1)
  expect_true(nll_gpd > 0)
})

test_that("Discrete distribution functions work as expected", {
  # Test that the probability mass functions are non-negative
  ddg <- .ddiscgamma(x = 1:5, shape = 1, rate = 1, thresh = 6, phiu = 0.5, shift = 0)
  expect_true(all(ddg >= 0))
  
  ddgpd <- .ddiscgpd(x = 6:10, thresh = 5, sigma = 1, xi = 0.1, phiu = 0.5)
  expect_true(all(ddgpd >= 0))
  
  # Test the random number generators
  set.seed(42)
  rands <- .rdiscgamma(100, shape = 2, rate = 1, thresh = 10, shift = 0)
  expect_length(rands, 100)
  expect_true(all(rands < 10)) # All values should be below the threshold
  expect_true(all(rands == floor(rands))) # All values should be integers
})

#################################################################
# Section 3: Testing Distance Calculation Functions
#################################################################


test_that(".JS_dist returns a single non-negative number", {
  set.seed(1)
  fit1 <- .fdiscgammagpd(rpois(100, 4) + 1, useq = 10)
  fit2 <- .fdiscgammagpd(rpois(100, 5) + 1, useq = 10)
  
  jsd <- .JS_dist(fit1, fit2, grid = 0:1000)
  
  expect_type(jsd, "double")
  expect_length(jsd, 1)
  expect_gte(jsd, 0) 
})

#' Visualize Distribution of Clonal Frequency
#'
#' This function allows the user to visualize the clonal expansion by overlaying the 
#' cells with specific clonal frequency onto the dimensional reduction plots in Seurat.
#' Credit to the idea goes to Drs Andreatta and Carmona and their work with
#' [ProjectTIL](https://github.com/carmonalab/ProjecTILs).
#'
#' @examples
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' # Getting a sample of a Seurat object
#' scRep_example <- get(data("scRep_example"))
#' 
#' # Using combineExpresion()
#' scRep_example <- combineExpression(combined, 
#'                                    scRep_example)
#' 
#' # Using clonalOverlay()
#' clonalOverlay(scRep_example, 
#'               reduction = "umap", 
#'               cutpoint = 3, 
#'               bins = 5) 
#' 
#' @param sc.data The single-cell object after [combineExpression()].
#' @param reduction The dimensional reduction to visualize.
#' @param cut.category Meta data variable of the single-cell object to use for 
#' filtering.
#' @param cutpoint The overlay cut point to include, this corresponds to the 
#' cut.category variable in the meta data of the single-cell object.
#' @param bins The number of contours to the overlay
#' @param pt.size The point size for plotting (default is 0.5)
#' @param pt.alpha The alpha value for plotting (default is 1)
#' @param facet.by meta data variable to facet the comparison
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @importFrom SeuratObject Embeddings
#' @export
#' @concept SC_Functions
#' @author Francesco Mazziotta, Nick Borcherding
#' 
#' @return A ggplot object visualizing distributions of clones along a 
#' dimensional reduction within the single-cell object

clonalOverlay <- function(sc.data, 
                          reduction = NULL, 
                          cut.category = "clonalFrequency",
                          cutpoint = 30, 
                          bins = 25, 
                          pt.size = 0.5,
                          pt.alpha = 1,
                          facet.by = NULL,
                          ...) {
  .checkSingleObject(sc.data)

  #Forming the data frame to plot
  tmp <- data.frame(.grabMeta(sc.data), .getCoord(sc.data, reduction))
  
  if(!cut.category %in% colnames(tmp)) {
    stop("If filtering the data using a cutpoint, ensure the cut.category correspond to a variable in the meta data.")
  }
  #Add facet variable if present
  if (!is.null(facet.by)) { 
    facet.by <- tmp[,facet.by]
    tmp <- data.frame(facet.by, tmp)
  }
  #If using cut.category for filtering
  if(!is.null(cut.category) & !is.null(cutpoint)) {
    tmp$include <- ifelse(tmp[,cut.category] >= cutpoint, "Yes", NA)
    tmp2 <- subset(tmp, include == "Yes")
  }
  
  #Plotting
  plot <- ggplot(tmp2, mapping = aes(x = tmp2[,(ncol(tmp2)-2)], 
                                     y = tmp2[,(ncol(tmp2)-1)])) +
    geom_point(tmp, mapping = aes(x = as.numeric(tmp[,(ncol(tmp)-2)]), 
                                  y = as.numeric(tmp[,(ncol(tmp)-1)]), 
                                  color = tmp[,"ident"]), 
               size = pt.size,
               alpha = pt.alpha) +
    geom_density_2d(color = "black", lwd=0.25, bins = bins) + 
    .themeRepertoire() + 
    labs(color = "Active Identity") +
    xlab("Dimension 1") + 
    ylab("Dimension 2")
  if (!is.null(facet.by)) {
    plot <- plot + facet_wrap(~facet.by) 
  }
  return(plot)
}

# test script for clonalOverlay.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list, samples = c("P17B", "P17L", "P18B", "P18L", "P19B", "P19L", "P20B", "P20L"))
scRep_example <- combineExpression(combined, scRep_example)

test_that("Input validation works correctly", {
  # Test for error when reduction is missing
  expect_error(clonalOverlay(scRep_example, reduction = "nonexistent_reduction"), 
               regexp = "Reduction nonexistent_reduction")
  
  # Test for error when cut.category is missing
  expect_error(clonalOverlay(scRep_example, reduction = "umap", cut.category = "nonexistent_category"),
               "If filtering the data using a cutpoint, ensure the cut.category correspond to a variable in the meta data.")
})

test_that("Data filtering with `cutpoint` is correct", {
  cutpoint_val <- 25
  manual_subset_rows <- rownames(scRep_example@meta.data[scRep_example@meta.data$clonalFrequency >= cutpoint_val, ])
  plot_obj <- clonalOverlay(scRep_example, reduction = "umap", cutpoint = cutpoint_val)
  density_data <- layer_data(plot_obj, 2) # Layer 2 should be the density plot
  all_coords <- scRep_example@reductions$umap@cell.embeddings
  filtered_data_internal_size <- nrow(scRep_example@meta.data[scRep_example@meta.data$clonalFrequency >= cutpoint_val,])
  expect_equal(filtered_data_internal_size, length(manual_subset_rows))
  plot_empty <- clonalOverlay(scRep_example, reduction = "umap", cutpoint = 100)
  density_data_empty <- layer_data(plot_empty, 2)
  expect_equal(nrow(density_data_empty), 0)
})

test_that("Plot structure is correct", {
  plot_obj <- clonalOverlay(scRep_example, reduction = "umap", cutpoint = 20, bins = 10)
  expect_length(plot_obj$layers, 2)
  expect_s3_class(plot_obj$layers[[1]]$geom, "GeomPoint")
  expect_s3_class(plot_obj$layers[[2]]$geom, "GeomDensity2d")
  expect_equal(plot_obj$layers[[2]]$stat_params$bins, 10)
  expect_equal(plot_obj$labels$x, "Dimension 1")
  expect_equal(plot_obj$labels$y, "Dimension 2")
  expect_equal(plot_obj$labels$colour, "Active Identity")
})

test_that("`facet.by` parameter works correctly", {
  plot_facet <- clonalOverlay(scRep_example, reduction = "umap", facet.by = "orig.ident")
  expect_s3_class(plot_facet$facet, "FacetWrap")
  facet_vars <- names(plot_facet$facet$params$facets)
  expect_equal(facet_vars, "facet.by")
})

test_that("Function handles Seurat objects without error", {
  expect_silent(clonalOverlay(scRep_example, reduction = "umap"))
})

#' Examining the clonal overlap between groups or samples
#'
#' This functions allows for the calculation and visualizations of 
#' various overlap metrics for clones. The methods include overlap 
#' coefficient (`overlap`), Morisita's overlap index 
#' (`morisita`), Jaccard index (`jaccard`), cosine 
#' similarity (`cosine`) or the exact number of clonal 
#' overlap (`raw`).
#' 
#' @details
#' The formulas for the indices are as follows:
#' 
#' **Overlap Coefficient:**
#' \deqn{overlap = \frac{\sum \min(a, b)}{\min(\sum a, \sum b)}}  
#' 
#' **Raw Count Overlap:**
#' \deqn{raw = \sum \min(a, b)}
#' 
#' **Morisita Index:**
#' \deqn{morisita = \frac{\sum a b}{(\sum a)(\sum b)}}  
#' 
#' **Jaccard Index:**
#' \deqn{jaccard = \frac{\sum \min(a, b)}{\sum a + \sum b - \sum \min(a, b)}}  
#' 
#' **Cosine Similarity:**
#' \deqn{cosine = \frac{\sum a b}{\sqrt{(\sum a^2)(\sum b^2)}}}  
#' 
#' Where:  
#' \itemize{  
#'   \item{\eqn{a} and \eqn{b} are the abundances of species \eqn{i} in groups A and B, respectively.}
#' }

#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' 
#' # Using clonalOverlap()
#' clonalOverlap(combined,
#'               clone.call = "aa",
#'               method = "jaccard")
#'
#' @param input.data The product of [combineTCR()],
#' [combineBCR()], or [combineExpression()]
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param method The method to calculate the `overlap`, `morisita`,
#' `jaccard`, `cosine` indices or `raw` for the base numbers
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups
#' automatically.
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals]
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#' 
#' @importFrom stats quantile
#' @export
#' @concept Visualizing_Clones
#' @return A ggplot object visualizing clonal overlap or a data.frame if
#'`exportTable = TRUE`.
clonalOverlap <- function(input.data,
                          clone.call = NULL,
                          method = c("overlap", "morisita", "jaccard", "cosine", "raw"),
                          chain = "both",
                          group.by = NULL,
                          order.by = NULL,
                          export.table = NULL,
                          palette = "inferno",
                          # Deprecated arguments
                          cloneCall = NULL,
                          exportTable = NULL,
                          ...){

    # Handle deprecated arguments
    clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                                 "clonalOverlap", default = "strict")
    export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                   "clonalOverlap", default = FALSE)

    method <- match.arg(method)
    if(method == "morisita") {
      return_type <- "freq"
    } else {
      return_type <- "unique"
    }
    input.data <- .dataWrangle(input.data,
                               group.by,
                               .theCall(input.data, clone.call,
                                        check.df = FALSE, silent = TRUE),
                               chain)
    if(!is.null(order.by)) {
      if(length(order.by) == 1 && order.by == "alphanumeric") {
        input.data <- input.data[.alphanumericalSort(names(input.data))]

      } else {
        input.data <- input.data[order.by]
      }
    }

    clone.call <- .theCall(input.data, clone.call)

    sco <- .is.seurat.or.se.object(input.data)
    if(!is.null(group.by) & !sco) {
      input.data <- .groupList(input.data, group.by)
    }

    num_samples <- length(input.data[])
    names_samples <- names(input.data)
    length <- seq_len(num_samples)

    #Selecting Index Function
    indexFunc <- switch(method,
                        "morisita" = .morisitaCalc,
                        "jaccard"  = .jaccardCalc,
                        "raw"      = .rawCalc,
                        "overlap"  = .overlapCalc,
                        "cosine"  =  .cosineCalc,
                        stop("Invalid method provided"))

    #Calculating Index
    coef_matrix <- data.frame(matrix(NA, num_samples, num_samples))
    coef_matrix <- .calculateIndex(input.data,
                                   length,
                                   clone.call,
                                   coef_matrix,
                                   indexFunc,
                                   return_type)

    #Data manipulation
    colnames(coef_matrix) <- names_samples
    rownames(coef_matrix) <- names_samples

    if (export.table) {
      return(coef_matrix)
    }
    
    coef_matrix_as_matrix <- as.matrix(coef_matrix)
    
    mat_melt <- data.frame(
      Var1 = rep(rownames(coef_matrix_as_matrix), ncol(coef_matrix_as_matrix)),
      Var2 = rep(colnames(coef_matrix_as_matrix), each = nrow(coef_matrix_as_matrix)),
      value = as.vector(coef_matrix_as_matrix)
    )
    
    mean_value <- mean(na.omit(mat_melt[,"value"]))
    
    plot <- ggplot(mat_melt, aes(x=.data[["Var1"]], y=.data[["Var2"]], fill=.data[["value"]])) +
                geom_tile() + 
                geom_tile(data = mat_melt[!is.na(mat_melt[,"value"]),], 
                          fill = NA, 
                          lwd= 0.25, 
                          color = "black") +
                labs(fill = .toCapitilize(method)) +
                geom_text(aes(label = round(value, digits = 3), 
                              color = ifelse(value <= mean_value,
                                             "white", "black")), 
                          na.rm = TRUE) +
                scale_fill_gradientn(colors = .colorizer(palette, 7), na.value = "white") +
                scale_color_identity() +
                .themeRepertoire() + 
                theme(axis.title = element_blank())
    return(plot) 
}

# Helper function to prepare data
.prepareDataFrame <- function(df, 
                              cloneCall, 
                              return_type = "unique") {
  if (return_type == "unique") {
    return(unique(df[, cloneCall]))
  } else if (return_type == "freq") {
    temp_df <- data.frame(table(df[, cloneCall]))
    colnames(temp_df) <- c(cloneCall, 'Count')
    temp_df[, 2] <- as.numeric(temp_df[, 2])
    return(temp_df)
  }
}

# Helper function for common loop and conditional structure
.calculateIndex <- function(df, 
                            length, 
                            cloneCall, 
                            coef_matrix, 
                            indexFunc, 
                            return_type = "unique") {
  for (i in seq_along(length)) {
    df_i <- .prepareDataFrame(df[[i]], cloneCall, return_type)
    for (j in seq_along(length)) {
      if (i >= j) { next }
      df_j <- .prepareDataFrame(df[[j]], cloneCall, return_type)
      coef_matrix[i, j] <- indexFunc(df_i, df_j)
    }
  }
  return(coef_matrix)
}

# Morisita Index calculation function
.morisitaCalc <- function(df_i, df_j) {
  merged <- merge(df_i, df_j, by = names(df_i)[1], all = TRUE)
  merged[is.na(merged)] <- 0
  
  X <- sum(merged[, 2])
  Y <- sum(merged[, 3])
  
  num <- 2 * sum(merged[, 2] * merged[, 3])
  den <- ((sum(df_i[, 2]^2) / (X^2)) + (sum(df_j[, 2]^2) / (Y^2))) * X * Y
  
  return(num / den)
}

# Jaccard Index calculation function
.jaccardCalc <- function(df_i, df_j) {
  overlap <- length(intersect(df_i, df_j))
  return(overlap / (length(df_i) + length(df_j) - overlap))
}

# Raw Index calculation function
.rawCalc <- function(df_i, df_j) {
  return(length(intersect(df_i, df_j)))
}

# Overlap Index calculation function
.overlapCalc <- function(df_i, df_j) {
  overlap <- length(intersect(df_i, df_j))
  return(overlap / min(length(df_i), length(df_j)))
}

# Overlap Index calculation function
.cosineCalc <- function(df_i, df_j) {
  all_species <- unique(c(df_i, df_j))
  vector_location1 <- as.integer(all_species %in% df_i)
  vector_location2 <- as.integer(all_species %in% df_j)
  
  return(sum(vector_location1 * vector_location2) / 
           (sqrt(sum(vector_location1^2)) * sqrt(sum(vector_location2^2))))
}

# test script for clonalOverlap.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined, 
                        variable.name = "Type", 
                        variables = rep(c("B", "L"), 4))

test_that("Input validation for `method` works", {
  expect_error(clonalOverlap(combined, method = "invalid_method"))
})

test_that("Output formats (data.frame vs. plot) are correct", {
  # Test that export.table = TRUE returns a data.frame
  df_output <- clonalOverlap(combined, method = "raw", export.table = TRUE)
  expect_true(is.data.frame(df_output))
  
  # Test that the default behavior returns a ggplot object
  plot_output <- clonalOverlap(combined, method = "raw")
  expect_s3_class(plot_output, "ggplot")
})

methods <- c("raw", "jaccard", "overlap", "morisita", "cosine")
for (m in methods) {
    # Create a unique test name for each method
    test_that(paste("Method '", m, "' produces a valid matrix"), {
      mat <- clonalOverlap(combined, method = m, export.table = TRUE)
      
      # Matrix should be square
      expect_equal(nrow(mat), ncol(mat))
      expect_equal(nrow(mat), length(combined))
      
      # Diagonal and lower triangle should be NA
      expect_true(all(is.na(mat[lower.tri(mat, diag = TRUE)])))
      
      # upper triangle should be numeric
      expect_true(is.numeric(mat[upper.tri(mat)]))
      
      # Check value ranges for indices
      if (m %in% c("jaccard", "overlap", "morisita", "cosine")) {
        expect_true(all(na.omit(mat) >= 0 & na.omit(mat) <= 1))
      } else if (m == "raw") {
        expect_true(all(na.omit(mat) >= 0))
      }
    })
}



methods <- c("raw", "jaccard", "morisita") 
for (m in methods) {
    test_that(paste("Grouping works for method '", m, "'"), {
      mat_grouped <- clonalOverlap(combined, group.by = "Type", method = m, export.table = TRUE)
      
      # Matrix dimensions should match the number of groups
      expect_equal(nrow(mat_grouped), 2)
      expect_equal(ncol(mat_grouped), 2)
      
      # Check row and column names
      expect_equal(rownames(mat_grouped), c("B", "L"))
      expect_equal(colnames(mat_grouped), c("B", "L"))
    })
}


test_that("Calculations are correct (sanity check)", {
  # Create a simple, predictable list of data frames
  list_for_testing <- list(
    A = data.frame(CTaa = c("a", "b", "c")),
    B = data.frame(CTaa = c("b", "c", "d", "e")),
    C = data.frame(CTaa = c("x", "y"))
  )
  
  # Test raw count
  mat_raw <- clonalOverlap(list_for_testing, clone.call = "aa", method = "raw", export.table = TRUE)
  expect_equal(mat_raw["A", "B"], 2) # Overlap is "b", "c"
  expect_equal(mat_raw["A", "C"], 0) # No overlap
  
  # Test Jaccard index
  mat_jaccard <- clonalOverlap(list_for_testing, clone.call = "aa", method = "jaccard", export.table = TRUE)
  # Jaccard(A,B) = |A intersect B| / |A union B| = 2 / (3 + 4 - 2) = 2/5 = 0.4
  expect_equal(mat_jaccard["A", "B"], 0.4)
})

test_that("Plotting output has the correct components", {
  plot_output <- clonalOverlap(combined, method = "jaccard")
  
  # Check for correct geometry
  expect_s3_class(plot_output$layers[[1]]$geom, "GeomTile")
  
  # Check labels
  expect_equal(plot_output$labels$fill, "Jaccard")
})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom lifecycle deprecated
#' @importFrom Rcpp sourceCpp
#' @useDynLib scRepertoire, .registration = TRUE
#'
#' @import ggplot2
#' @importFrom dplyr %>% select mutate filter arrange bind_rows group_by n
#' @importFrom dplyr summarise summarize
## usethis namespace: end
NULL

#' Plot cloneSize by Variable in Single-Cell Objects
#'
#' View the count of clones frequency group in Seurat or SCE object
#' meta data after [combineExpression()]. The visualization
#' will take the new meta data variable `cloneSize` and
#' plot the number of cells with each designation using a secondary
#' variable, like cluster. Credit to the idea goes to Drs. Carmona
#' and Andreatta and their work with [ProjectTIL](https://github.com/carmonalab/ProjecTILs).
#'
#' @examples
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list,
#'                         samples = c("P17B", "P17L", "P18B", "P18L",
#'                                     "P19B","P19L", "P20B", "P20L"))
#'
#' # Getting a sample of a Seurat object
#' scRep_example <- get(data("scRep_example"))
#'
#' # Using combineExpresion()
#' scRep_example <- combineExpression(combined, scRep_example)
#'
#' # Using clonalOccupy
#' clonalOccupy(scRep_example, x.axis = "ident")
#' table <- clonalOccupy(scRep_example, x.axis = "ident", export.table = TRUE)
#'
#' @param sc.data The single-cell object after [combineExpression()]
#' @param x.axis The variable in the meta data to graph along the x.axis.
#' @param label Include the number of clone in each category by x.axis variable
#' @param facet.by The column header used for faceting the graph
#' @param order.by A character vector defining the desired order of elements
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups
#' automatically.
#' @param proportion Convert the stacked bars into relative proportion
#' @param na.include Visualize NA values or not
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals]
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @importFrom dplyr count
#' @export
#' @concept SC_Functions
#' @return Stacked bar plot of counts of cells by clone frequency group

clonalOccupy <- function(sc.data,
                         x.axis = "ident",
                         label = TRUE,
                         facet.by = NULL,
                         order.by = NULL,
                         proportion = FALSE,
                         na.include = FALSE,
                         export.table = NULL,
                         palette = "inferno",
                         # Deprecated arguments
                         exportTable = NULL,
                         ...) {

  # Handle deprecated arguments
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "clonalOccupy", default = FALSE)
  .checkSingleObject(sc.data)
  meta <- .grabMeta(sc.data)
  
  meta <- meta %>%
            group_by(meta[,x.axis], meta[,facet.by], cloneSize) %>%
            count() %>%
            as.data.frame()
  colnames(meta)[1] <- x.axis
  
  if(!is.null(order.by)) {
    meta <- .orderingFunction(vector = order.by,
                               group.by = x.axis, 
                               data.frame = meta)
  } else {
    meta[,1] <- as.factor(meta[,1])
  }
  
  colnames(meta)[1] <- x.axis
  if(!is.null(facet.by)) {
    colnames(meta)[2] <- facet.by
  }
 
  #Check for NAs
  if (!na.include) {
    meta <- na.omit(meta)
  }
  meta <- meta[meta$n != 0,]
  
  #Convert to proportion
  if(proportion) {
    meta <- meta %>%
      group_by(meta[,1]) %>%
      mutate(total = sum(n), 
             prop = n/total)
    meta <- as.data.frame(meta)
  }
  if (export.table) {
    return(meta)
  }
  #Plotting
  col <- length(unique(meta[,"cloneSize"]))
  if(proportion) {
    plot <- ggplot(meta, aes(x = meta[,x.axis], y = prop, fill = cloneSize)) + 
              geom_bar(stat = "identity", color = "black", lwd = 0.25) 
    lab <- "Proportion of Cells"
    
  } else {
    plot <- ggplot(meta, aes(x = meta[,x.axis], y = n, fill = cloneSize)) + 
              geom_bar(stat = "identity", color = "black", lwd = 0.25) 
    lab <- "Single Cells"
    
  } 
  plot <- plot + 
            theme(axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.5)) +
            scale_fill_manual(values = rev(c(.colorizer(palette,col)))) + 
            ylab(lab) + 
            .themeRepertoire(...) + 
            theme(axis.title.x = element_blank())
  if (!is.null(facet.by)) {
    plot <- plot + 
              facet_grid(.~meta[,facet.by])
  }
  if (label) {
    plot <- plot + 
              geom_text(aes(label = n), position = position_stack(vjust = 0.5))
  }
  plot
}

# test script for clonalOccupy.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list, samples = c("P17B", "P17L", "P18B", "P18L", "P19B", "P19L", "P20B", "P20L"))
scRep_example <- combineExpression(combined, scRep_example)
scRep_example$Patient <- substring(scRep_example$orig.ident, 1, 3)
scRep_example$Type <- substring(scRep_example$orig.ident, 4, 4)

test_that("Output formats (table vs. plot) are correct", {
  # Test that export.table = TRUE returns a data.frame
  table_output <- clonalOccupy(scRep_example, x.axis = "seurat_clusters", export.table = TRUE)
  expect_s3_class(table_output, "data.frame")
  
  # Test that the default behavior returns a ggplot object
  plot_output <- clonalOccupy(scRep_example, x.axis = "seurat_clusters")
  expect_s3_class(plot_output, "ggplot")
})

test_that("Data aggregation and counting are correct", {
  table_output <- clonalOccupy(scRep_example, x.axis = "seurat_clusters", export.table = TRUE)
  c1_single_count <- table_output[table_output$seurat_clusters == "1" & table_output$cloneSize == "Large (0.01 < X <= 0.1)", "n"]
  c1_small_count <- table_output[table_output$seurat_clusters == "1" & table_output$cloneSize == "Medium (0.001 < X <= 0.01)", "n"]
  expect_equal(c1_single_count, 34)
  expect_equal(c1_small_count, 41)

})

test_that("`proportion` parameter works correctly", {
  table_prop <- clonalOccupy(scRep_example, x.axis = "seurat_clusters", proportion = TRUE, export.table = TRUE)
  expect_true(all(c("total", "prop") %in% names(table_prop)))
  total_props <- table_prop %>%
    group_by(seurat_clusters) %>%
    summarise(sum_prop = sum(prop))
  expect_true(all(abs(total_props$sum_prop - 1) < 1e-9))
  plot_prop <- clonalOccupy(scRep_example, x.axis = "seurat_clusters", proportion = TRUE)
  expect_equal(plot_prop$labels$y, "Proportion of Cells")
})

test_that("`facet.by` parameter works correctly", {
  # Test table output with facet
  table_facet <- clonalOccupy(scRep_example, x.axis = "seurat_clusters", facet.by = "Type", export.table = TRUE)
  expect_true("Type" %in% names(table_facet))
  
  # Test plot output with facet
  plot_facet <- clonalOccupy(scRep_example, x.axis = "seurat_clusters", facet.by = "Type")
  expect_s3_class(plot_facet$facet, "FacetGrid")
})

test_that("`label` parameter correctly adds geom_text", {
  # Test with label = TRUE (default)
  plot_labeled <- clonalOccupy(scRep_example, x.axis = "seurat_clusters", label = TRUE)
  expect_true("GeomText" %in% sapply(plot_labeled$layers, function(l) class(l$geom)[1]))
  
  # Test with label = FALSE
  plot_unlabeled <- clonalOccupy(scRep_example, x.axis = "seurat_clusters", label = FALSE)
  expect_false("GeomText" %in% sapply(plot_unlabeled$layers, function(l) class(l$geom)[1]))
})

test_that("`na.include` parameter works correctly", {
  # Test with na.include = FALSE (default)
  table_no_na <- clonalOccupy(scRep_example, x.axis = "seurat_clusters", export.table = TRUE, na.include = FALSE)
  expect_false(any(is.na(table_no_na$cloneSize)))
  
  # Test with na.include = TRUE
  table_with_na <- clonalOccupy(scRep_example, x.axis = "seurat_clusters", export.table = TRUE, na.include = TRUE)
  expect_true(any(is.na(table_with_na$cloneSize)))
  na_row <- table_with_na[table_with_na$seurat_clusters == "1" & is.na(table_with_na$cloneSize), ]
  expect_equal(nrow(na_row), 1)
  expect_equal(na_row$n, 7)
})

#' Scatter Plot Comparing Clones Across Two Samples
#'
#' This function produces a scatter plot directly comparing 
#' the specific clones between two samples. The clones will 
#' be categorized by counts into singlets or expanded, either 
#' exclusive or shared between the selected samples. 
#'
#' @examples
#' #Making combined contig data
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' 
#' # Using clonalScatter()
#' clonalScatter(combined, 
#'               x.axis = "P17B", 
#'               y.axis = "P17L",
#'               graph = "proportion")
#' 
#' @param input.data The product of [combineTCR()],
#' [combineBCR()], or [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param x.axis name of the list element to appear on the x.axis.
#' @param y.axis name of the list element to appear on the y.axis.
#' @param dot.size either total or the name of the list element to
#' use for size of dots.
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed
#' by list element or active identity in the case of single-cell objects.
#' @param graph graph either the clonal "proportion" or "count".
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals].
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @export
#' @concept Visualizing_Clones
#' @return A ggplot object visualizing clonal dynamics between two groupings or
#'  a data.frame if `exportTable = TRUE`.

clonalScatter <- function(input.data,
                          clone.call = NULL,
                          x.axis = NULL,
                          y.axis = NULL,
                          chain = "both",
                          dot.size = "total",
                          group.by = NULL,
                          graph = "proportion",
                          export.table = NULL,
                          palette = "inferno",
                          # Deprecated arguments
                          cloneCall = NULL,
                          exportTable = NULL,
                          ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "clonalScatter", default = "strict")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "clonalScatter", default = FALSE)

  sco <- .is.seurat.or.se.object(input.data)
  input.data <- .dataWrangle(input.data,
                             group.by,
                             .theCall(input.data, clone.call,
                                      check.df = FALSE, silent = TRUE),
                             chain)
  clone.call <- .theCall(input.data, clone.call)
  
  if(!is.null(group.by) & !sco) {
    input.data <- .groupList(input.data, group.by)
  }

  axes <- which(names(input.data) %in% c(x.axis, y.axis, dot.size))

  #Making new data frame and adding variables to graph
  x.df <- as.data.frame(table(input.data[[x.axis]][,clone.call]))
  colnames(x.df)[2] <- x.axis
  y.df <- as.data.frame(table(input.data[[y.axis]][,clone.call]))
  colnames(y.df)[2] <- y.axis
  mat <- merge(x.df, y.df, by = "Var1", all = TRUE)

  if (dot.size != "total") {
    if (!dot.size %in% colnames(mat)) {
      size.df <- as.data.frame(table(input.data[[dot.size]][,clone.call]))
      colnames(size.df)[2] <- dot.size
      mat <- merge(mat, size.df, by = "Var1", all = TRUE) }
    mat[is.na(mat)] <- 0
    mat[,paste0("size", ".fraction")] <- mat[,dot.size]/sum(mat[,dot.size])
    labeling <- unique(c(x.axis, y.axis, dot.size))
  } else {
    mat[is.na(mat)] <- 0
    labeling <- unique(c(x.axis, y.axis)) }
  
  mat[,"class"] <- NA
  mat[,"sum"] <- rowSums(mat[,labeling])
  
  #Assigning class based on x and y-axis counts
  for (i in seq_along(labeling)) {
    if (length(labeling) > 2) {
      mat[,"class"] <- ifelse(mat[,labeling[i]] == 1 & rowSums(mat[,labeling[-i]]) == 0, 
                                      paste0(labeling[i], ".singlet"), #if
                                      mat[,"class"]) #else
      mat[,"class"] <- ifelse(mat[,labeling[i]] > 1 & rowSums(mat[,labeling[-i]]) == 0, 
                                      paste0(labeling[i], ".expanded"), #if
                                      mat[,"class"]) #else
    } else if (length(labeling) == 2) {
      mat[,"class"] <- ifelse(mat[,labeling[i]] == 1 & mat[,labeling[-i]] == 0, 
                                      paste0(labeling[i], ".singlet"), #if
                                      mat[,"class"]) #else
      mat[,"class"] <- ifelse(mat[,labeling[i]] > 1 & mat[,labeling[-i]] == 0, 
                                      paste0(labeling[i], ".expanded"), #if
                                      mat[,"class"]) #else
    }
  }
  #Adding dual-expanded class
  mat[,"class"] <- ifelse(mat[,y.axis] >= 1 & mat[,x.axis] >= 1, paste0("dual.expanded"), mat[,"class"])
  
  #Calculating relative proportion
  mat[,paste0(x.axis, ".fraction")] <- mat[,x.axis]/sum(mat[,x.axis])
  mat[,paste0(y.axis, ".fraction")] <- mat[,y.axis]/sum(mat[,y.axis])
  
  #Altering the graphing parameters
  if (graph == "proportion") {
    x <- mat[,paste0(x.axis, ".fraction")]
    y <- mat[,paste0(y.axis, ".fraction")]
  } else if (graph == "count") {
    x <- mat[,x.axis]
    y <- mat[,y.axis] }
  if (dot.size != "total") {
    size <- mat[,dot.size]
  } else {
    mat[,"size"] <- mat[,"sum"]
  }
  if (export.table) {
    return(mat)
  }
  
  plot <- ggplot(mat, aes(x=x, y = y, fill = class)) + 
                .themeRepertoire(...) + 
                scale_fill_manual(values = .colorizer(palette,length(unique(mat$class)))) + 
                xlab(x.axis) + 
                ylab(y.axis) + 
                labs(size = "Total n")
  if (graph == "proportion") {
    plot <- plot + 
            geom_abline(slope = 1, 
                        intercept = 0, 
                        alpha = 0.4, 
                        lty=2)  + 
            scale_y_sqrt() + 
            scale_x_sqrt() 
  } else if (graph == "count") {
    plot <- plot + 
            ylim(0, max(x,y)) + 
            xlim(0, max(x,y)) 
  }
  plot <- plot + 
          geom_point(aes(size = size), 
                     shape = 21, 
                     color = "black", 
                     stroke = 0.25)
  
  return(plot)  
}

# test script for clonalScatter.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined,
                        variable.name = "Type",
                        variables = rep(c("B", "L"), 4))

test_that("Output formats (table vs. plot) are correct", {
  # Test that export.table = TRUE returns a data.frame
  table_output <- clonalScatter(combined, x.axis = "P17B", y.axis = "P17L", export.table = TRUE)
  expect_s3_class(table_output, "data.frame")
  
  # Test that the default behavior returns a ggplot object
  plot_output <- clonalScatter(combined, x.axis = "P17B", y.axis = "P17L")
  expect_s3_class(plot_output, "ggplot")
})

test_that("`graph` parameter correctly sets up data and plot", {
  plot_prop <- clonalScatter(combined, x.axis = "P17B", y.axis = "P17L", graph = "proportion")
  table_prop <- clonalScatter(combined, x.axis = "P17B", y.axis = "P17L", graph = "proportion", export.table = TRUE)
  expect_true(all(c("P17B.fraction", "P17L.fraction") %in% names(table_prop)))
  expect_true(any(sapply(plot_prop$layers, function(x) inherits(x$geom, "GeomAbline"))))
  expect_s3_class(plot_prop$scales$scales[[2]], "ScaleContinuous") # x-axis scale
  expect_s3_class(plot_prop$scales$scales[[3]], "ScaleContinuous") # y-axis scale
  expect_equal(plot_prop$scales$scales[[2]]$trans$name, "sqrt")
  
  # Test with graph = "count"
  plot_count <- clonalScatter(combined, x.axis = "P17B", y.axis = "P17L", graph = "count")
  table_count <- clonalScatter(combined, x.axis = "P17B", y.axis = "P17L", graph = "count", export.table = TRUE)
  expect_true(all(c("P17B.fraction", "P17L.fraction") %in% names(table_count)))
  expect_false(any(sapply(plot_count$layers, function(x) inherits(x$geom, "GeomAbline"))))
})

test_that("Clone classification logic is correct", {
  list_for_testing <- list(
    X1 = data.frame(CTaa = c("A", "B", "C", "C", "C")), # A=1, B=1, C=3
    X2 = data.frame(CTaa = c("B", "D", "D")),          # B=1, D=2
    X3 = data.frame(CTaa = c("E"))
  )
  
  df <- clonalScatter(list_for_testing, clone.call = "aa", x.axis = "X1", y.axis = "X2", export.table = TRUE)
  clone_A <- df[df$Var1 == "A", ]
  clone_B <- df[df$Var1 == "B", ]
  clone_C <- df[df$Var1 == "C", ]
  clone_D <- df[df$Var1 == "D", ]
  expect_equal(clone_A$class, "X1.singlet")   # Count of 1 in X1, 0 in X2
  expect_equal(clone_C$class, "X1.expanded")  # Count > 1 in X1, 0 in X2
  expect_equal(clone_D$class, "X2.expanded")  # Count > 1 in X2, 0 in X1
  expect_equal(clone_B$class, "dual.expanded") # Present in both
})

test_that("`dot.size` parameter works correctly", {
  # Test with default dot.size = "total"
  df_total <- clonalScatter(combined, x.axis = "P17B", y.axis = "P17L", dot.size = "total", export.table = TRUE)
  expect_true("size" %in% names(df_total))
  expect_equal(df_total$size, df_total$P17B + df_total$P17L)
  
  # Test with a third sample for dot size
  df_third <- clonalScatter(combined, x.axis = "P17B", y.axis = "P17L", dot.size = "P18B", export.table = TRUE)
  expect_true("P18B" %in% names(df_third))
})

test_that("Plotting output has the correct components", {
  plot_output <- clonalScatter(combined, x.axis = "P17B", y.axis = "P17L")
  
  # Check for correct geometry
  expect_s3_class(plot_output$layers[[1]]$geom, "GeomAbline") # First layer is abline for proportion
  expect_s3_class(plot_output$layers[[2]]$geom, "GeomPoint")
  
  # Check labels
  expect_equal(plot_output$labels$x, "P17B")
  expect_equal(plot_output$labels$y, "P17L")
  expect_equal(plot_output$labels$size, "Total n")
})

#' Annotate invariant T cells (MAIT or iNKT) in single-cell TCR data
#' 
#' The [annotateInvariant()] function identifies potential mucosal-associated 
#' invariant T (`MAIT``) cells or invariant natural killer T (`iNKT`) cells from 
#' single-cell sequencing datasets based on their characteristic  TCR usage. 
#' It extracts TCR chain information from the provided single-cell 
#' data, checks it against known invariant T-cell receptor criteria for either 
#' MAIT or iNKT cells, and returns a score indicating the presence (1) or 
#' absence (0) of these invariant cell populations for each individual cell. 
#' The function supports data from mouse and human samples, providing a 
#' convenient method to annotate specialized T-cell subsets within single-cell 
#' analyses.
#'
#' @param input.data The product of [combineTCR()] or [combineExpression()].
#' @param type Character specifying the type of invariant cells to
#'annotate (`MAIT` or `iNKT`).
#' @param species Character specifying the species 
#' ('mouse' or 'human').
#' 
#' @examples
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' 
#' # Getting a sample of a Seurat object
#' scRep_example <- get(data("scRep_example"))
#' 
#' # Using combineExpresion()
#' scRep_example <- combineExpression(combined, scRep_example)
#' 
#' # Using annotateInvariant()
#' annotateInvariant(input.data = scRep_example, type = "MAIT", species = "human")
#' annotateInvariant(input.data = scRep_example, type = "iNKT", species = "human")
#' 
#' @importFrom immApex getIR
#' @importFrom  rlang %||%
#' @export
#' @return A single-cell object or list with the corresponding annotation 
#' scores (0 or 1) added.
annotateInvariant <- function(input.data, 
                              type = c("MAIT", "iNKT"), 
                              species = c("mouse", "human")) {
  
  type <- match.arg(type)
  species <- match.arg(species)
  
  if(!.is.seurat.or.se.object(input.data) & !inherits(input.data, "list")) {
    stop("Please use the output of combineTCR() or combineExpression() as input.data")
  }

  
  TCRS <- lapply(c("TRA", "TRB"), function(x) {
    tmp <- getIR(input.data, chains = x, sequence.type = "aa")
    tmp
  })
  
  criteria <- switch(type,
                     "MAIT" = .MAIT.criteria,
                     "iNKT" = .iNKT.criteria)
  
  species.criteria <- criteria[[species]]
  
  TRA.data <- TCRS[[1]]
  TRB.data <- TCRS[[2]]
  
  barcode.ids <- unique(TRA.data$barcode)
  
  scores <- vapply(barcode.ids, function(barcode) {
    
    TRA.subset <- TRA.data[TRA.data$barcode == barcode,]
    TRB.subset <- TRB.data[TRB.data$barcode == barcode,]
    
    if (nrow(TRA.subset) == 0 | nrow(TRB.subset) == 0) return(0)
    
    TRA.v <- TRA.subset$v
    TRA.j <- TRA.subset$j
    TRB.v <- TRB.subset$v
    TRA.length <- nchar(TRA.subset$cdr3_aa)
    
    if (is.null(TRA.v) | is.null(TRA.j) | is.null(TRB.v) | is.null(TRA.length)) return(0)
    
    TRA.v.match <- grepl(paste(species.criteria$TRA.v, collapse = "|"), TRA.v)
    TRA.j.match <- grepl(paste(species.criteria$TRA.j, collapse = "|"), TRA.j)
    TRA.length.match <- any(TRA.length %in% species.criteria$TRA.length)
    TRB.v.match <- ifelse(is.null(species.criteria$TRB.v), TRUE, grepl(paste(species.criteria$TRB.v, collapse = "|"), TRB.v))
    
    as.integer(TRA.v.match & TRA.j.match & TRA.length.match & TRB.v.match)
    
  }, integer(1))
  
  output <- data.frame(row.names = barcode.ids, score = scores)
  new.variable.name <- paste0(type, ".score")
  colnames(output)[1] <- new.variable.name
  
  if(.is.seurat.or.se.object(input.data)) {
    if (.is.seurat.object(input.data)) { 
      input.data[[new.variable.name]] <- output
    } else {
      combined_col_names <- unique(c(colnames(colData(input.data)), new.variable.name))
      full_data <- merge(colData(input.data), output, by = "row.names", all.x = TRUE)
      # at this point, the rows in full_data are shuffled. match back with the original colData
      full_data <- full_data[match(colnames(input.data), full_data[,1]), ]
      rownames(full_data) <- full_data[, 1]
      full_data  <- full_data[, -1]
      colData(input.data) <- DataFrame(full_data[, combined_col_names])  
    }
  } else {
    input.data <- lapply(input.data, function(x) {
      merged_x <- merge(x, output, by.x = "barcode", by.y = "row.names", all.x = TRUE)  
      return(merged_x)
    })
  }
  
  return(input.data)
}

# Internal criteria definitions
.MAIT.criteria <- list(
  mouse = list(TRA.v = "TRAV1", 
               TRA.j = "TRAJ33", 
               TRA.length = 12, 
               TRB.v = c("TRBV13", "TRBV19")), 
  human = list(TRA.v = "TRAV1-2", 
               TRA.j = c("TRAJ33", "TRAJ20", "TRAJ12"), 
               TRA.length = 12,
               TRB.v = c("TRBV6", "TRBV20"))
)

.iNKT.criteria <- list(
  mouse = list(TRA.v = "TRAV11", 
               TRA.j = "TRAJ18", 
               TRA.length = 15,
               TRB.v = NULL),
  human = list(TRA.v = "TRAV10", 
               TRA.j = "TRAJ18",
               TRA.length = c(14, 15, 16),
               TRB.v = "TRBV25-1")
)

# test script for annotateInvariant.R - testcases are NOT comprehensive!

# Data to use
combined_real <- combineTCR(contig_list, samples = c("P17B", "P17L", "P18B", "P18L"))
test_obj <- suppressMessages(combineExpression(combined_real, scRep_example, clone.call = "aa"))

test_that("annotateInvariant() handles correct input format", {
  # Mocking input data with required structure
  
  result <- annotateInvariant(test_obj, type = "MAIT", species = "human")
  
  expect_true(is.data.frame(result[[]]))
  expect_true(any(colnames(result[[]]) ==  "MAIT.score"))
})

test_that("annotateInvariant() handles species and type argument matching correctly", {
  expect_error(
    annotateInvariant(mock_input, type = "INVALID", species = "human"),
    regexp = "should be one of .*MAIT.*iNKT.*"
  )
  
  expect_error(annotateInvariant(mock_input, type = "MAIT", species = "INVALID"),
               regexp = "should be one of .*mouse.*human.*")
})

test_that("annotateInvariant() fails with incorrect input format", {
  expect_error(
    annotateInvariant("wrong_input", type = "MAIT", species = "human"),
    regexp = "Please use the output of combineTCR\\(\\) or combineExpression\\(\\)"
  )
})

test_that("annotateInvariant() returns zero scores for non-matching cells", {
  result <- annotateInvariant(test_obj, type = "MAIT", species = "human")
  
  expect_equal(sum(result$MAIT.score), 1) 
})

test_that("annotateInvariant() correctly integrates with Seurat object", {
  if (requireNamespace("Seurat", quietly = TRUE)) {
    result <- annotateInvariant(test_obj, type = "MAIT", species = "human")
    
    expect_true("MAIT.score" %in% colnames(result@meta.data))
  }
})

#' Plot Clonal Homeostasis of the Repertoire
#'
#' This function calculates the space occupied by clone proportions.
#' The grouping of these clones is based on the parameter `clone.size`,
#' at default, `clone.size` will group the clones into bins of Rare = 0
#' to 0.0001, Small = 0.0001 to 0.001, etc. To adjust the proportions,
#' change the number or labeling of the clone.size parameter. If a matrix
#' output for the data is preferred, set `export.table` = TRUE.
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list,
#'                         samples = c("P17B", "P17L", "P18B", "P18L",
#'                                     "P19B","P19L", "P20B", "P20L"))
#' clonalHomeostasis(combined, clone.call = "gene")
#'
#' @param input.data The product of [combineTCR()], [combineBCR()], or
#' [combineExpression()].
#' @param clone.size The cut points of the proportions.
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups
#' automatically.
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals].
#' @param cloneSize \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.size` instead.
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @export
#' @importFrom utils stack
#' @importFrom stats reshape
#' @concept Visualizing_Clones
#' @return A ggplot object visualizing clonal homeostasis, or a data.frame if
#' `export.table = TRUE`.
clonalHomeostasis <- function(input.data,
                              clone.size = NULL,
                              clone.call = NULL,
                              chain = "both",
                              group.by = NULL,
                              order.by = NULL,
                              export.table = NULL,
                              palette = "inferno",
                              # Deprecated arguments
                              cloneSize = NULL,
                              cloneCall = NULL,
                              exportTable = NULL,
                              ...) {

    # Handle deprecated arguments
    clone.size <- .deprecate_arg(cloneSize, clone.size, "cloneSize", "clone.size",
                                 "clonalHomeostasis",
                                 default = c(Rare = .0001, Small = .001, Medium = .01, Large = .1, Hyperexpanded = 1))
    clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                                 "clonalHomeostasis", default = "strict")
    export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                   "clonalHomeostasis", default = FALSE)

    clone.size <- c(None = 0, clone.size)
    input.data <- .dataWrangle(input.data,
                               group.by,
                               .theCall(input.data, clone.call,
                                        check.df = FALSE, silent = TRUE),
                               chain)
    clone.call <- .theCall(input.data, clone.call)
    sco <- .is.seurat.or.se.object(input.data)
    if(!is.null(group.by) & !sco) {
      input.data <- .groupList(input.data, group.by)
    }
    
    #Generating data matrix to store value
    mat <- matrix(0, length(input.data), length(clone.size) - 1,
                dimnames = list(names(input.data),
                names(clone.size)[-1]))

    #Assigning the clonal grouping
    input.data <- lapply(input.data, '[[', clone.call)
    input.data <- lapply(input.data, na.omit)
    fun <- function(x) { table(x)/length(x) }
    input.data <- lapply(input.data, fun)
    for (i in 2:length(clone.size)) {
        mat[,i-1] <- vapply(input.data, function (x)
                              sum(x[x > clone.size[i-1] & x <= clone.size[i]]), FUN.VALUE = numeric(1))
        colnames(mat)[i-1] <- paste0(names(clone.size[i]), ' (',
                                    clone.size[i-1], ' < X <= ',
                                    clone.size[i], ')') }
    if (export.table) {
      return(mat)
    }
    
    #Plotting
    mat_df <- as.data.frame(mat) 
    mat_df$Var1 <- rownames(mat_df)
    
    varying_cols <- names(mat_df)[grep("<", names(mat_df))]
    
    mat_melt <- reshape(mat_df,
                        varying = varying_cols,
                        v.names = "value",
                        timevar = "category",
                        times = varying_cols,
                        idvar = "Var1",
                        direction = "long")
    
    if(!is.null(order.by)) {
      mat_melt <- .orderingFunction(vector = order.by,
                                    group.by = "Var1", 
                                    data.frame = mat_melt)
    }
    
    col <- length(unique(mat_melt$category))
    plot <- ggplot(mat_melt, aes(x=as.factor(.data[["Var1"]]), y=.data[["value"]], fill=.data[["category"]])) +
        geom_bar(stat = "identity", position="fill", 
                    color = "black", lwd= 0.25) +
        scale_fill_manual(values = rev(.colorizer(palette,col))) +
        labs(x = "Samples", 
             y = "Relative Abundance", 
             fill = "Clonal Group") +
        .themeRepertoire(...)
    return(plot)
}

# test script for clonalHomeostasis.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined,
                        variable.name = "Type",
                        variables = rep(c("B", "L"), 4))

test_that("Output formats (matrix vs. plot) are correct", {
  mat_output <- clonalHomeostasis(combined, export.table = TRUE)
  expect_true(is.matrix(mat_output))
  plot_output <- clonalHomeostasis(combined)
  expect_s3_class(plot_output, "ggplot")
})

test_that("The returned matrix has correct dimensions and properties", {
  mat_output <- clonalHomeostasis(combined, export.table = TRUE)
  expect_equal(nrow(mat_output), length(combined)) # Rows should equal number of samples
  expect_equal(ncol(mat_output), 5)               # Default has 5 bins
  expect_true(all(abs(rowSums(mat_output) - 1) < 1e-9))
  expected_colnames <- c("Rare (0 < X <= 1e-04)",
                         "Small (1e-04 < X <= 0.001)",
                         "Medium (0.001 < X <= 0.01)",
                         "Large (0.01 < X <= 0.1)",
                         "Hyperexpanded (0.1 < X <= 1)")
  expect_equal(colnames(mat_output), expected_colnames)
})

test_that("Custom `cloneSize` parameter works as expected", {
  custom_bins <- c(Small = 0.01, Medium = 0.2, Large = 1)
  mat_custom <- clonalHomeostasis(combined, clone.size = custom_bins, export.table = TRUE)
  expect_equal(ncol(mat_custom), 3)
  expected_custom_colnames <- c("Small (0 < X <= 0.01)",
                                "Medium (0.01 < X <= 0.2)",
                                "Large (0.2 < X <= 1)")
  expect_equal(colnames(mat_custom), expected_custom_colnames)
  expect_true(all(abs(rowSums(mat_custom) - 1) < 1e-9))
})

test_that("`group.by` parameter correctly aggregates the data", {
  mat_grouped <- clonalHomeostasis(combined, group.by = "Type", export.table = TRUE)
  expect_equal(nrow(mat_grouped), 2)
  expect_true(all(rownames(mat_grouped) %in% c("B", "L")))
  expect_true(all(abs(rowSums(mat_grouped) - 1) < 1e-9))
})

test_that("`clone.call` and `chain` parameters execute correctly", {
  mat_aa <- clonalHomeostasis(combined, clone.call = "aa", export.table = TRUE)
  expect_true(is.matrix(mat_aa))
  expect_equal(nrow(mat_aa), length(combined))
  mat_trb <- clonalHomeostasis(combined, chain = "TRB", export.table = TRUE)
  expect_true(is.matrix(mat_trb))
  expect_equal(nrow(mat_trb), length(combined))
})

test_that("Plotting output has the correct components", {
  plot_output <- clonalHomeostasis(combined)
  
  # Check for the correct geometry and position
  expect_s3_class(plot_output$layers[[1]]$geom, "GeomBar")
  expect_s3_class(plot_output$layers[[1]]$position, "PositionFill")
  
  # Check the labels
  expect_equal(plot_output$labels$y, "Relative Abundance")
  expect_equal(plot_output$labels$x, "Samples")
  expect_equal(plot_output$labels$fill, "Clonal Group")
})

#' Load Immune Receptor Sequencing Contigs
#'
#' @description
#' This function loads and processes contig data from various single-cell 
#' immune receptor sequencing formats. It reads data from a directory 
#' (recursively) or from an already loaded list/data frame, transforms it to a 
#' common structure, and returns a list of contigs ready for downstream analysis 
#' with [combineTCR()] or [combineBCR()].
#'
#' Supported file formats and their expected file names:
#' - `10X`: "filtered_contig_annotations.csv"
#' - `AIRR`: "airr_rearrangement.tsv"
#' - `BD`: "Contigs_AIRR.tsv"
#' - `Dandelion`: "all_contig_dandelion.tsv"
#' - `Immcantation`: "_data.tsv" (or similar)
#' - ``JSON`: ".json"
#' - `ParseBio`: "barcode_report.tsv"
#' - `MiXCR`: "clones.tsv"
#' - `TRUST4`: "barcode_report.tsv"
#' - `WAT3R`: "barcode_results.csv"
#'
#' @param input A directory path containing contig files or a list/data frame 
#' of pre-loaded contig data.
#' @param format A string specifying the data format. Must be one of:
#' `auto`, `10X`, `AIRR`, `BD`, `Dandelion`, `JSON`, `MiXCR`, `ParseBio`, 
#' `TRUST4`, `WAT3R`, or `Immcantation`. If "auto", the function attempts 
#' automatic format detection.
#'
#' @return A list of contigs formatted for use with [combineTCR()] or 
#' [combineBCR()]. Rows containing only NA values (aside from the barcode) 
#' are dropped.
#'
#' @examples
#' TRUST4 <- read.csv("https://www.borch.dev/uploads/contigs/TRUST4_contigs.csv")
#' contig.list <- loadContigs(TRUST4, format = "TRUST4")
#'
#' BD <- read.csv("https://www.borch.dev/uploads/contigs/BD_contigs.csv")
#' contig.list <- loadContigs(BD, format = "BD")
#'
#' WAT3R <- read.csv("https://www.borch.dev/uploads/contigs/WAT3R_contigs.csv")
#' contig.list <- loadContigs(WAT3R, format = "WAT3R")
#'
#' @importFrom utils read.csv read.delim
#' @importFrom rjson fromJSON
#' @export
loadContigs <- function(input, 
                        format = "10X") {
  valid_formats <- c("auto", "10X", "AIRR", "BD", "Dandelion", "JSON",
                     "MiXCR", "ParseBio", "TRUST4", "WAT3R", "Immcantation")
  
  # Define mapping from formats to file patterns (for directory inputs)
  format_patterns <- list(
    "WAT3R"       = "barcode_results.csv",
    "10X"         = "filtered_contig_annotations.csv",
    "AIRR"        = "airr_rearrangement.tsv",
    "Dandelion"   = "all_contig_dandelion.tsv",
    "Immcantation"= "_data.tsv",
    "MiXCR"       = "clones.tsv",
    "JSON"        = ".json",
    "TRUST4"      = "barcode_report.tsv",
    "BD"          = "Contigs_AIRR.tsv",
    "ParseBio"    = "barcode_report.tsv"
  )
  
  # Determine data input type and detect format if necessary
  if (is.character(input)) {
    # Input is a directory; use file name patterns for detection if "auto"
    if (format == "auto") {
      candidates <- c()
      for (fmt in names(format_patterns)) {
        pattern <- format_patterns[[fmt]]
        pattern_regex <- paste0(".*(", pattern, ")$")
        files_found <- list.files(
          path = input,
          pattern = pattern_regex,
          recursive = TRUE,
          full.names = TRUE
        )
        if (length(files_found) > 0) {
          candidates <- c(candidates, fmt)
        }
      }
      if (length(candidates) == 0) {
        warning("No contig files found in the specified directory using any known patterns.")
        return(list())
      } else if (length(unique(candidates)) > 1) {
        warning("Ambiguous format detection. Multiple candidate formats found: ",
                paste(unique(candidates), collapse = ", "),
                ". Defaulting to first candidate: ", unique(candidates)[1])
        detected_format <- unique(candidates)[1]
      } else {
        detected_format <- unique(candidates)[1]
        message("Automatically detected format: ", detected_format)
      }
    } else {
      detected_format <- format
    }
    
    # Use the detected or specified format to list files
    file_pattern <- format_patterns[[detected_format]]
    pattern_regex <- paste0(".*(", file_pattern, ")$")
    contig_files <- list.files(
      path = input,
      pattern = pattern_regex,
      recursive = TRUE,
      full.names = TRUE
    )
    
    if (length(contig_files) == 0) {
      warning("No contig files found in the specified directory with pattern ", file_pattern, ".")
      return(list())
    }
    
    # Choose a reader based on format
    reader <- switch(detected_format,
                     "JSON" = function(x) as.data.frame(fromJSON(x)),
                     "10X" = read.csv,
                     "WAT3R" = read.csv,
                     read.delim)
    rawDataDfList <- lapply(contig_files, reader)
  } else {
    # Input is pre-loaded data; ensure it is a list of data frames
    rawDataDfList <- .checkList(input)
    if (format == "auto") {
      detected_format <- .detectFormatFromData(rawDataDfList)
      message("Automatically detected format from data: ", detected_format)
    } else {
      detected_format <- format
    }
  }
  
  # Select the appropriate parser function based on the detected format
  loadFunc <- switch(detected_format,
                     "10X" = .parse10x,
                     "AIRR" = .parseAIRR,
                     "Dandelion" = .parseDandelion,
                     "JSON" = .parseJSON,
                     "MiXCR" = .parseMiXCR,
                     "TRUST4" = .parseTRUST4,
                     "BD" = .parseBD,
                     "WAT3R" = .parseWAT3R,
                     "Immcantation" = .parseImmcantation,
                     "ParseBio" = .parseParse)
  
  # Process and clean contig data, dropping rows that are entirely NA (except for barcode)
  cleaned_list <- loadFunc(rawDataDfList)
  .rmAllNaRowsFromLoadContigs(cleaned_list)
}

# Auto-detect format from a pre-loaded data object
.detectFormatFromData <- function(dfList) {
  # Use the first data frame for detection
  df <- if (is.data.frame(dfList)) dfList else dfList[[1]]
  cn <- colnames(df)
  
  if (all(c("chain", "productive", "cdr3") %in% cn)) {
    return("10X")
  } else if (all(c("cell_id", "locus", "consensus_count", "v_call", "d_call", "j_call", "c_call", "junction", "junction_aa") %in% cn)) {
    return("AIRR")
  } else if (all(c("cell_id", "locus", "v_call", "d_call", "j_call", "c_call", "cdr3", "cdr3_aa", "consensus_count", "productive") %in% cn)) {
    # Could be BD or Dandelion – defaulting to BD here.
    return("BD")
  } else if (all(c("tagValueCELL", "topChains", "readCount", "allVHitsWithScore") %in% cn)) {
    return("MiXCR")
  } else if (all(c("barcode", "chain1", "chain2") %in% cn)) {
    return("TRUST4")
  } else if (all(c("BC", "TRBV", "TRBD", "TRBJ") %in% cn)) {
    return("WAT3R")
  } else if (all(c("sequence_id", "locus", "consensus_count", "v_call", "d_call", "j_call", "cdr3") %in% cn)) {
    if ("productive" %in% cn) return("Immcantation")
  } else if (all(c("Barcode", "TRA_V") %in% cn) || all(c("Barcode", "IGH_V") %in% cn)) {
    return("ParseBio")
  }
  stop("Could not auto-detect format from the provided data structure.")
}

# Remove rows with all NA values (ignoring the "barcode" column)
.rmAllNaRowsFromLoadContigs <- function(dfList) {
  cols <- setdiff(colnames(dfList[[1]]), "barcode")
  lapply(dfList, function(df) {
    df[rowSums(!is.na(df[, cols, drop = FALSE])) > 0, , drop = FALSE]
  })
}

# Helper function to replace empty strings with NA
.sanitize_empty <- function(df) {
  df[df == ""] <- NA
  df
}

# Helper function to order a data frame by "reads" and "chain" (if present)
.order_df <- function(df) {
  if (all(c("reads", "chain") %in% colnames(df))) {
    df[order(df$reads, df$chain), ]
  } else {
    df
  }
}

## --- Parsing Functions ---
.parseTRUST4 <- function(df_list) {
  split_and_pad <- function(x, n = NULL) {
    parts <- unlist(strsplit(x, ","))
    if (is.null(n)) return(parts)
    if (length(parts) < n) {
      return(c(parts, rep("", n - length(parts))))
    } else if (length(parts) > n) {
      return(parts[seq_len(n)])
    } else {
      return(parts)
    }
  }
  
  processChain <- function(data, chain_col) {
    if (all(is.na(data[[chain_col]]))) {
      chain <- matrix(NA, ncol = 7, nrow = length(data[[chain_col]]))
    } else {
      chain <- t(vapply(
        data[[chain_col]], 
        split_and_pad, 
        FUN.VALUE = character(7), 
        n = 7,
        USE.NAMES = FALSE
      ))
      chain[chain == "*"] <- "None"
    }
    colnames(chain) <- c("v_gene", "d_gene", "j_gene", "c_gene", "cdr3_nt", "cdr3", "reads")
    data.frame(barcode = data$barcode, chain, stringsAsFactors = FALSE)
  }
  
  formatted <- lapply(df_list, function(data) {
    colnames(data)[1] <- "barcode"
    data <- .sanitize_empty(data)
    chain1 <- processChain(data, "chain2")
    chain2 <- processChain(data, "chain1")
    combined_data <- rbind(chain1, chain2)
    
    combined_data <- .sanitize_empty(combined_data)
    combined_data$chain <- substr(combined_data$v_gene, 1, 3)
    combined_data$reads <- as.numeric(combined_data$reads)
    combined_data
  })
  return(formatted)
}

.parseWAT3R <- function(df_list) {
  lapply(df_list, function(df) {
    df <- .sanitize_empty(df)
    chain2 <- data.frame(
      barcode = df$BC,
      chain = "TRB",
      v_gene = df$TRBV,
      d_gene = df$TRBD,
      j_gene = df$TRBJ,
      c_gene = NA,
      cdr3_nt = df$TRB_CDR3nuc,
      cdr3 = df$TRB_CDR3,
      reads = df$TRB_nReads,
      umis = df$TRB_CDR3_UMIcount,
      stringsAsFactors = FALSE
    )
    chain1 <- data.frame(
      barcode = df$BC,
      chain = "TRA",
      v_gene = df$TRAV,
      d_gene = NA,
      j_gene = df$TRAJ,
      c_gene = NA,
      cdr3_nt = df$TRA_CDR3nuc,
      cdr3 = df$TRA_CDR3,
      reads = df$TRA_nReads,
      umis = df$TRA_CDR3_UMIcount,
      stringsAsFactors = FALSE
    )
    chain3 <- data.frame(
      barcode = df$BC,
      chain = "TRA",
      v_gene = df[["TRAV.2"]],
      d_gene = NA,
      j_gene = df[["TRAJ.2"]],
      c_gene = NA,
      cdr3_nt = df[["TRA.2_CDR3nuc"]],
      cdr3 = df[["TRA.2_CDR3"]],
      reads = df[["TRA.2_nReads"]],
      umis = df[["TRA.2_CDR3_UMIcount"]],
      stringsAsFactors = FALSE
    )
    combined <- rbind(chain1, chain2, chain3)
    .order_df(.sanitize_empty(combined))
  })
}

.parseAIRR <- function(df_list) {
  lapply(df_list, function(df) {
    df <- df[, c("cell_id", "locus", "consensus_count", "v_call", "d_call", "j_call", "c_call", "junction", "junction_aa")]
    colnames(df) <- c("barcode", "chain", "reads", "v_gene", "d_gene", "j_gene", "c_gene", "cdr3_nt", "cdr3")
    .order_df(df)
  })
}

.parse10x <- function(df_list) {
  lapply(df_list, function(df) {
    df <- subset(df, chain != "Multi")
    df <- subset(df, productive %in% c(TRUE, "TRUE", "True", "true"))
    if (nrow(df) == 0) {
      stop("There are 0 contigs after filtering. Check the contig list for issues with productive chains.", call. = FALSE)
    }
    df <- subset(df, cdr3 != "None")
    df <- .sanitize_empty(df)
    .order_df(df)
  })
}

.parseBD <- function(df_list) {
  lapply(df_list, function(df) {
    df <- df[, c("cell_id", "locus", "v_call", "d_call", "j_call", "c_call", "cdr3", "cdr3_aa", "consensus_count", "productive")]
    colnames(df) <- c("barcode", "chain", "v_gene", "d_gene", "j_gene", "c_gene", "cdr3_nt", "cdr3", "reads", "productive")
    .order_df(df)
  })
}

.parseJSON <- function(df_list) {
  lapply(df_list, function(df) {
    df <- do.call(rbind, df)
    df <- .sanitize_empty(df)
    df <- as.data.frame(df, stringsAsFactors = FALSE)
    df <- df[, c("cell_id", "locus", "consensus_count", "v_call", "d_call", "j_call", "c_call", "junction", "junction_aa")]
    colnames(df) <- c("barcode", "chain", "reads", "v_gene", "d_gene", "j_gene", "c_gene", "cdr3_nt", "cdr3")
    df
  })
}

.parseMiXCR <- function(df_list) {
  lapply(df_list, function(df) {
    df <- .sanitize_empty(df)
    df <- as.data.frame(df, stringsAsFactors = FALSE)
    df <- df[, c("tagValueCELL", "topChains", "readCount", "allVHitsWithScore",
                 "allDHitsWithScore", "allJHitsWithScore", "allCHitsWithScore",
                 "nSeqCDR3", "aaSeqCDR3")]
    colnames(df) <- c("barcode", "chain", "reads", "v_gene", "d_gene",
                      "j_gene", "c_gene", "cdr3_nt", "cdr3")
    df
  })
}

.parseImmcantation <- function(df_list) {
  lapply(df_list, function(df) {
    df <- .sanitize_empty(df)
    df <- as.data.frame(df, stringsAsFactors = FALSE)
    if ("c_call" %in% colnames(df)) {
      df <- df[, c("sequence_id", "locus", "consensus_count", "v_call", "d_call", "j_call", "c_call", "cdr3", "cdr3_aa", "productive")]
      colnames(df) <- c("barcode", "chain", "reads", "v_gene", "d_gene", "j_gene", "c_gene", "cdr3_nt", "cdr3", "productive")
    } else {
      df <- df[, c("sequence_id", "locus", "consensus_count", "v_call", "d_call", "j_call", "cdr3", "cdr3_aa", "productive")]
      colnames(df) <- c("barcode", "chain", "reads", "v_gene", "d_gene", "j_gene", "cdr3_nt", "cdr3", "productive")
      df$c_gene <- NA
    }
    df$barcode <- vapply(strsplit(df$barcode, "_"), `[`, 1, FUN.VALUE = character(1))
    df
  })
}

.parseParse <- function(df_list) {
  lapply(df_list, function(df) {
    df <- .sanitize_empty(df)
    df[df %in% c("NaN", "nan")] <- NA
    df <- as.data.frame(df, stringsAsFactors = FALSE)
    
    isTcell <- any(c("TRA_V", "TRA_D", "TRA_J") %in% colnames(df))
    if (isTcell) {
      TRA1 <- df[, c("Barcode", "TRA_V", "TRA_D", "TRA_J", "TRA_C", "TRA_cdr3_aa", "TRA_read_count", "TRA_transcript_count")]
      TRA2 <- df[, c("Barcode", "secondary_TRA_V", "secondary_TRA_D", "secondary_TRA_J", "secondary_TRA_C", "secondary_TRA_cdr3_aa", "secondary_TRA_read_count", "secondary_TRA_transcript_count")]
      colnames(TRA1) <- colnames(TRA2) <- as.character(1:8)
      TRA <- rbind(TRA1, TRA2)
      TRA$chain <- "TRA"
      
      TRB1 <- df[, c("Barcode", "TRB_V", "TRB_D", "TRB_J", "TRB_C", "TRB_cdr3_aa", "TRB_read_count", "TRB_transcript_count")]
      TRB2 <- df[, c("Barcode", "secondary_TRB_V", "secondary_TRB_D", "secondary_TRB_J", "secondary_TRB_C", "secondary_TRB_cdr3_aa", "secondary_TRB_read_count", "secondary_TRB_transcript_count")]
      colnames(TRB1) <- colnames(TRB2) <- as.character(1:8)
      TRB <- rbind(TRB1, TRB2)
      TRB$chain <- "TRB"
      
      combined <- rbind(TRA, TRB)
    } else {
      IGH1 <- df[, c("Barcode", "IGH_V", "IGH_D", "IGH_J", "IGH_C", "IGH_cdr3_aa", "IGH_read_count", "IGH_transcript_count")]
      IGH2 <- df[, c("Barcode", "secondary_IGH_V", "secondary_IGH_D", "secondary_IGH_J", "secondary_IGH_C", "secondary_IGH_cdr3_aa", "secondary_IGH_read_count", "secondary_IGH_transcript_count")]
      colnames(IGH1) <- colnames(IGH2) <- as.character(1:8)
      IGH <- rbind(IGH1, IGH2)
      IGH$chain <- "IGH"
      
      IGK1 <- df[, c("Barcode", "IGK_V", "IGK_D", "IGK_J", "IGK_C", "IGK_cdr3_aa", "IGK_read_count", "IGK_transcript_count")]
      IGK2 <- df[, c("Barcode", "secondary_IGK_V", "secondary_IGK_D", "secondary_IGK_J", "secondary_IGK_C", "secondary_IGK_cdr3_aa", "secondary_IGK_read_count", "secondary_IGK_transcript_count")]
      colnames(IGK1) <- colnames(IGK2) <- as.character(1:8)
      IGK <- rbind(IGK1, IGK2)
      IGK$chain <- "IGK"
      
      IGL1 <- df[, c("Barcode", "IGL_V", "IGL_D", "IGL_J", "IGL_C", "IGL_cdr3_aa", "IGL_read_count", "IGL_transcript_count")]
      IGL2 <- df[, c("Barcode", "secondary_IGL_V", "secondary_IGL_D", "secondary_IGL_J", "secondary_IGL_C", "secondary_IGL_cdr3_aa", "secondary_IGL_read_count", "secondary_IGL_transcript_count")]
      colnames(IGL1) <- colnames(IGL2) <- as.character(1:8)
      IGL <- rbind(IGL1, IGL2)
      IGL$chain <- "IGL"
      
      combined <- rbind(IGH, IGK, IGL)
    }
    combined <- combined[rowSums(is.na(combined[, 2:8])) != 7, ]
    colnames(combined) <- c("barcode", "v_gene", "d_gene", "j_gene", "c_gene", "cdr3", "reads", "umis", "chain")
    combined$cdr3_nt <- NA
    combined <- combined[, c("barcode", "chain", "v_gene", "d_gene", "j_gene", "c_gene", "cdr3_nt", "cdr3", "reads", "umis")]
    .order_df(combined)
  })
}

.parseDandelion <- function(df_list) {
  lapply(df_list, function(df) {
    df <- df[, c("cell_id", "locus", "consensus_count", "v_call", "d_call", "j_call", "c_call", "cdr3", "cdr3_aa", "productive")]
    colnames(df) <- c("barcode", "chain", "reads", "v_gene", "d_gene", "j_gene", "c_gene", "cdr3_nt", "cdr3", "productive")
    df
  })
}

# test script for loadContigs.R - testcases are NOT comprehensive!

check_loadContigs_output <- function(loaded_data) {
  # Check if the output is a list containing a single data frame
  expect_type(loaded_data, "list")
  expect_length(loaded_data, 1)
  df <- loaded_data[[1]]
  expect_s3_class(df, "data.frame")
  
  # Check if the data frame is not empty
  expect_gt(nrow(df), 0)
  
  # Check for the presence of essential standardized columns
  expected_cols <- c("barcode", "chain", "reads", "v_gene", 
                     "d_gene", "j_gene", "c_gene", "cdr3_nt", "cdr3")
  expect_true(all(expected_cols %in% names(df)))
  
  # Check data types of key columns
  expect_type(df$chain, "character")
  expect_type(df$cdr3, "character")
  # Reads should be numeric/integer after parsing
  expect_true(is.numeric(df$reads) || is.integer(df$reads))
}


test_that("loadContigs correctly processes various formats from URL", {
  #TRUST4 format
  TRUST4 <- read.csv("https://www.borch.dev/uploads/contigs/TRUST4_contigs.csv")
  trial_trust4 <- loadContigs(TRUST4, format = "TRUST4")
  check_loadContigs_output(trial_trust4)
  
  # BD format
  BD <- read.csv("https://www.borch.dev/uploads/contigs/BD_contigs.csv")
  trial_bd <- loadContigs(BD, format = "BD")
  check_loadContigs_output(trial_bd)
  
  # WAT3R format
  WAT3R <- read.csv("https://www.borch.dev/uploads/contigs/WAT3R_contigs.csv")
  trial_wat3r <- loadContigs(WAT3R, format = "WAT3R")
  check_loadContigs_output(trial_wat3r)
  
  # 10X format (using pre-loaded contig_list data)
  data("contig_list")
  trial_10x <- loadContigs(contig_list[[1]], format = "10X")
  check_loadContigs_output(trial_10x)
  
  # MiXCR format
  MIXCR <- read.csv("https://www.borch.dev/uploads/contigs/MIXCR_contigs.csv")
  trial_mixcr <- loadContigs(MIXCR, format = "MiXCR")
  check_loadContigs_output(trial_mixcr)
  
  # Immcantation format
  Immcantation <- read.csv("https://www.borch.dev/uploads/contigs/Immcantation_contigs.csv")
  trial_immcantation <- loadContigs(Immcantation, format = "Immcantation")
  check_loadContigs_output(trial_immcantation)
  
  # ParseBio format
  Parse <- read.csv("https://www.borch.dev/uploads/contigs/Parse_contigs.csv")
  trial_parse <- loadContigs(Parse, format = "ParseBio")
  check_loadContigs_output(trial_parse)
  
  # Dandelion format
  Dandelion <- read.csv("https://www.borch.dev/uploads/contigs/Dandelion_contigs.csv")
  trial_dandelion <- loadContigs(Dandelion, format = "Dandelion")
  check_loadContigs_output(trial_dandelion)
})

test_that("loadContigs correctly auto-detects and processes various formats", {
  
  BD <- read.csv("https://www.borch.dev/uploads/contigs/BD_contigs.csv")
  trial_bd <- expect_message(
    loadContigs(BD, format = "auto"), 
    "Automatically detected format from data: AIRR"
  )
  
  WAT3R <- read.csv("https://www.borch.dev/uploads/contigs/WAT3R_contigs.csv")
  trial_wat3r <- expect_message(
    loadContigs(WAT3R, format = "auto"), 
    "Automatically detected format from data: WAT3R"
  )
  
  trial_10x <- expect_message(
      loadContigs(contig_list[[1]], format = "auto"), 
      "Automatically detected format from data: 10X"
  )
  
  MIXCR <- read.csv("https://www.borch.dev/uploads/contigs/MIXCR_contigs.csv")
  trial_mixcr <- expect_message(
    loadContigs(MIXCR, format = "auto"), 
    "Automatically detected format from data: MiXCR"
  )
  
  Immcantation <- read.csv("https://www.borch.dev/uploads/contigs/Immcantation_contigs.csv")
  trial_immcantation <- expect_message(
    loadContigs(Immcantation, format = "auto"), 
    "Automatically detected format from data: Immcantation"
  )
  
  Parse <- read.csv("https://www.borch.dev/uploads/contigs/Parse_contigs.csv")
  trial_parse <- expect_message(
    loadContigs(Parse, format = "auto"), 
    "Automatically detected format from data: ParseBio"
  )
})

test_that("loadContigs works with AIRR input (directory mode)", {
  # This test remains unchanged as it is self-contained and does not use internal data.
  # Create a temporary TSV file for AIRR format
  tmp_tsv <- tempfile(fileext = ".tsv")
  airr_data <- data.frame(
    cell_id = "cellA",
    locus = "TRB",
    consensus_count = 5,
    v_call = "TRBV1",
    d_call = "TRBD1",
    j_call = "TRBJ1",
    c_call = "TRBC1",
    junction = "ATGCGT",
    junction_aa = "ML",
    stringsAsFactors = FALSE
  )
  write.table(airr_data, tmp_tsv, sep = "\t", row.names = FALSE, quote = FALSE)
  
  # Copy to a temporary directory with proper filename
  tmp_dir_airr <- file.path(tempdir(), "airr_test")
  dir.create(tmp_dir_airr, showWarnings = FALSE)
  file.copy(tmp_tsv, file.path(tmp_dir_airr, "airr_rearrangement.tsv"), overwrite = TRUE)
  
  # Explicit format test
  result_airr <- loadContigs(tmp_dir_airr, format = "AIRR")
  expected_airr <- list(
    data.frame(
      barcode = "cellA",
      chain = "TRB",
      reads = as.integer(5),
      v_gene = "TRBV1",
      d_gene = "TRBD1",
      j_gene = "TRBJ1",
      c_gene = "TRBC1",
      cdr3_nt = "ATGCGT",
      cdr3 = "ML",
      stringsAsFactors = FALSE
    )
  )
  expect_identical(result_airr, expected_airr)
})

test_that("loadContigs returns empty list when no matching files are found in a directory", {
  empty_dir <- file.path(tempdir(), "empty_test_dir")
  dir.create(empty_dir, showWarnings = FALSE)
  file.remove(list.files(empty_dir, full.names = TRUE))
  expect_warning(res <- loadContigs(empty_dir, format = "10X"))
  expect_identical(res, list())
})

# Adding Global Variables
# data('v_gene','j_gene', 'c_gene', 'd_gene')
# note that currently the Rcpp internals have hardcoded column names so
# if some breaking change here is made, the Rcpp code will need to be updated,
# or functions need to be adjusted to intake expected column names that
# uses these variables
utils::globalVariables(c("v_gene", "j_gene", "c_gene", "d_gene", "chain"))

heavy_lines <- c("IGH", "cdr3_aa1", "cdr3_nt1", "vgene1")
light_lines <- c("IGLC", "cdr3_aa2", "cdr3_nt2", "vgene2")
l_lines <- c("IGLct", "cdr3", "cdr3_nt", "v_gene")
k_lines <- c("IGKct", "cdr3", "cdr3_nt", "v_gene")
h_lines <- c("IGHct", "cdr3", "cdr3_nt", "v_gene")
tcr1_lines <- c("TCR1", "cdr3_aa1", "cdr3_nt1")
tcr2_lines <- c("TCR2", "cdr3_aa2", "cdr3_nt2")
data1_lines <- c("TCR1", "cdr3", "cdr3_nt")
data2_lines <- c("TCR2", "cdr3", "cdr3_nt")
CT_lines <- c("CTgene", "CTnt", "CTaa", "CTstrict")

utils::globalVariables(c(
    "heavy_lines", "light_lines", "l_lines", "k_lines", "h_lines", "tcr1_lines",
    "tcr2_lines", "data1_lines", "data2_lines", "CT_lines"
))

#' @title Combine T Cell Receptor Contig Data
#'
#' @description This function consolidates a list of TCR sequencing results to
#' the level of  the individual cell barcodes. Using the `samples` and
#' `ID` parameters, the function will add the strings as prefixes to
#' prevent issues with repeated  barcodes. The resulting new barcodes will
#' need to match the Seurat or SCE object in order to use,
#' [combineExpression()]. Several levels of filtering exist -
#' `remove.na`, `remove.multi`, or `filter.multi` are parameters
#' that control how the function deals with barcodes with multiple chains
#' recovered.
#'
#' @examples
#' combined <- combineTCR(contig_list,
#'                         samples = c("P17B", "P17L", "P18B", "P18L",
#'                                     "P19B","P19L", "P20B", "P20L"))
#'
#' @param input.data List of filtered contig annotations or
#' outputs from [loadContigs()].
#' @param samples The labels of samples (recommended).
#' @param ID The additional sample labeling (optional).
#' @param remove.na This will remove any chain without values.
#' @param remove.multi This will remove barcodes with greater than 2 chains.
#' @param filter.multi This option will allow for the selection of the 2
#' corresponding chains with the highest expression for a single barcode.
#' @param filter.nonproductive This option will allow for the removal of
#' nonproductive chains if the variable exists in the contig data. Default
#' is set to TRUE to remove nonproductive contigs.
#' @param removeNA \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `remove.na` instead.
#' @param removeMulti \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `remove.multi` instead.
#' @param filterMulti \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `filter.multi` instead.
#' @param filterNonproductive \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `filter.nonproductive` instead.
#'
#' @export
#' @concept Loading_and_Processing_Contigs
#' @return List of clones for individual cell barcodes
#'
combineTCR <- function(input.data,
                       samples = NULL,
                       ID = NULL,
                       remove.na = NULL,
                       remove.multi = NULL,
                       filter.multi = NULL,
                       filter.nonproductive = NULL,
                       # Deprecated arguments
                       removeNA = NULL,
                       removeMulti = NULL,
                       filterMulti = NULL,
                       filterNonproductive = NULL) {

    # Handle deprecated arguments
    remove.na <- .deprecate_arg(removeNA, remove.na, "removeNA", "remove.na",
                                "combineTCR", default = FALSE)
    remove.multi <- .deprecate_arg(removeMulti, remove.multi, "removeMulti", "remove.multi",
                                   "combineTCR", default = FALSE)
    filter.multi <- .deprecate_arg(filterMulti, filter.multi, "filterMulti", "filter.multi",
                                   "combineTCR", default = FALSE)
    filter.nonproductive <- .deprecate_arg(filterNonproductive, filter.nonproductive,
                                           "filterNonproductive", "filter.nonproductive",
                                           "combineTCR", default = TRUE)

    input.data <- .checkList(input.data)
    input.data <- .checkContigs(input.data)
    out <- NULL
    final <- NULL
    for (i in seq_along(input.data)) {
        if(c("chain") %in% colnames(input.data[[i]])) {
          input.data[[i]] <- subset(input.data[[i]], chain != "Multi")
        }
        if(c("productive") %in% colnames(input.data[[i]]) & filter.nonproductive) {
          input.data[[i]] <- subset(input.data[[i]], productive %in% c(TRUE, "TRUE", "True", "true"))
        }
        input.data[[i]]$sample <- samples[i]
        input.data[[i]]$ID <- ID[i]
        if (filter.multi) {
          input.data[[i]] <- .filteringMulti(input.data[[i]])
        }
    }
    #Prevents error caused by list containing elements with 0 rows
    blank.rows <- which(unlist(lapply(input.data, nrow)) == 0)
    if(length(blank.rows) > 0) {
      input.data <- input.data[-blank.rows]
      if(!is.null(samples)) {
        samples <- samples[-blank.rows]
      }
      if(!is.null(ID)) {
        ID <- ID[-blank.rows]
      }
    }
    if (!is.null(samples)) {
      out <- .modifyBarcodes(input.data, samples, ID)
    } else {
      out <- input.data
    }
    for (i in seq_along(out)) {
        data2 <- .makeGenes(cellType = "T", out[[i]])
        Con.df <- .constructConDfAndParseTCR(data2)
        Con.df <- .assignCT(cellType = "T", Con.df)
        Con.df[Con.df == "NA_NA" | Con.df == "NA;NA_NA;NA"] <- NA
        data3 <- merge(data2[,-which(names(data2) %in% c("TCR1","TCR2"))],
            Con.df, by = "barcode")
      
        columns_to_include <- c("barcode")
        # Conditionally add columns based on user input
        if (!is.null(samples)) {
          columns_to_include <- c(columns_to_include, "sample")
        }
        if (!is.null(ID)) {
          columns_to_include <- c(columns_to_include, "ID")
        }
      
        # Add TCR and CT lines which are presumably always needed
        columns_to_include <- c(columns_to_include, tcr1_lines, tcr2_lines, CT_lines)
      
        # Subset the data frame based on the dynamically built list of columns
        data3 <- data3[, columns_to_include]
      
        final[[i]] <- data3
    }
    name_vector <- character(length(samples))
    for (i in seq_along(samples)) {
        if (!is.null(samples) && !is.null(ID)) {
            curr <- paste(samples[i], "_", ID[i], sep="")
        } else if (!is.null(samples) & is.null(ID)) {
            curr <- paste(samples[i], sep="")
        }
        name_vector[i] <- curr
    }
    names(final) <- name_vector
    for (i in seq_along(final)){
      final[[i]]<-final[[i]][!duplicated(final[[i]]$barcode),]
      final[[i]]<-final[[i]][rowSums(is.na(final[[i]])) < 10, ]
      final[[i]][final[[i]] == "NA"] <- NA
    }
    if (remove.na) {
      final <- .removingNA(final)
    }
    if (remove.multi) {
      final <- .removingMulti(final)
    }
    #Adding list element names to output if samples NULL
    if(is.null(samples)) {
      names(final) <- paste0("S", seq_len(length(final)))
    }
    final
}

#' Combine B Cell Receptor Contig Data
#'
#' This function consolidates a list of BCR sequencing results to the level
#' of the individual cell barcodes. Using the samples and ID parameters,
#' the function will add the strings as prefixes to prevent issues with
#' repeated barcodes. The resulting new barcodes will need to match the
#' Seurat or SCE object in order to use, [combineExpression()]. Unlike
#' [combineTCR()], combineBCR produces a column `CTstrict` based on the
#' edit distance clustering from [clonalCluster()]. The `CTstrict` column
#' is formatted as `Heavy_Light` (underscore-separated) for downstream
#' compatibility. Connected clones are labeled with `cluster.X`, while
#' unconnected clones (singlets) are labeled with the V gene and CDR3
#' sequence (e.g., `IGHV3-64.CAKSYS..._IGKV3-15.CQQYSN...`).
#'
#' @examples
#' # Data derived from the 10x Genomics intratumoral NSCLC B cells
#' BCR <- read.csv("https://www.borch.dev/uploads/contigs/b_contigs.csv")
#' combined <- combineBCR(BCR,
#'                        samples = "Patient1",
#'                        threshold = 0.85)
#'
#' @param input.data List of filtered contig annotations or outputs from
#' [loadContigs()].
#' @param samples A character vector of sample labels. Must be the same length
#' as the input list.
#' @param ID An optional character vector for additional sample identifiers.
#' @param call.related.clones Logical. If `TRUE`, uses `clonalCluster()` to
#' identify related clones based on sequence similarity. If `FALSE`, defines
#' clones by the exact V-gene and CDR3 amino acid sequence.
#' @param group.by The column header used for to group clones.
#' If (`NULL``), clusters will be calculated across samples.
#' @param threshold The similarity threshold passed to `clonalCluster()` if
#' `call.related.clones = TRUE`. See `?clonalCluster` for details.
#' @param chain The chain to use for clustering when `call.related.clones = TRUE`.
#' Passed to `clonalCluster()`. Default is `"IGH"`.
#' @param sequence The sequence type (`"nt"` or `"aa"`) to use for clustering.
#' Passed to `clonalCluster()`. Default is `"nt"`.
#' @param dist.type The distance metric to use. Options: `"levenshtein"` (default),
#' `"hamming"`, `"damerau"`, `"nw"` (Needleman-Wunsch), or `"sw"` (Smith-Waterman).
#' @param dist.mat The substitution matrix to use for alignment-based metrics
#' (`"nw"` or `"sw"`). Options include `"BLOSUM62"`, `"PAM30"`, etc.
#' @param normalize Method for normalizing distances. Options: `"none"` (default),
#' `"maxlen"`, or `"length"`.
#' @param gap.open Penalty for opening a gap in alignment metrics (default: -10).
#' @param gap.extend Penalty for extending a gap in alignment metrics (default: -1).
#' @param use.V Logical. If `TRUE`, sequences must share the same V gene to be
#' clustered together.
#' @param use.J Logical. If `TRUE`, sequences must share the same J gene to be
#' clustered together.
#' @param cluster.method The clustering algorithm to use. Defaults to `"components"`,
#' which finds connected subgraphs.
#' @param remove.multi Logical. If `TRUE`, removes cells that have more than
#' one distinct heavy or light chain after processing.
#' @param filter.multi Logical. If `TRUE`, filters multi-chain cells to retain
#' only the most abundant IGH and IGL/IGK chains.
#' @param remove.na This will remove any chain without values.
#' @param filter.nonproductive Logical. If `TRUE`, removes non-productive contigs
#' from the analysis.
#' @param removeNA \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `remove.na` instead.
#' @param removeMulti \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `remove.multi` instead.
#' @param filterMulti \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `filter.multi` instead.
#' @param filterNonproductive \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `filter.nonproductive` instead.
#' @param dist_type \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `dist.type` instead.
#' @param dist_mat \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `dist.mat` instead.
#' @param gap_open \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `gap.open` instead.
#' @param gap_extend \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `gap.extend` instead.
#'
#' @export
#' @concept Loading_and_Processing_Contigs
#' @return A list of data frames, where each data frame represents a sample.
#' Each row corresponds to a unique cell barcode, with columns detailing the
#' BCR chains and the assigned clone ID.
combineBCR <- function(input.data,
                       samples = NULL,
                       ID = NULL,
                       chain = "IGH",
                       sequence = "nt",
                       dist.type = NULL,
                       dist.mat = NULL,
                       normalize = "length",
                       gap.open = NULL,
                       gap.extend = NULL,
                       call.related.clones = TRUE,
                       group.by = NULL,
                       threshold = 0.85,
                       cluster.method = "components",
                       use.V = TRUE,
                       use.J = TRUE,
                       remove.na = NULL,
                       remove.multi = NULL,
                       filter.multi = NULL,
                       filter.nonproductive = NULL,
                       # Deprecated arguments
                       removeNA = NULL,
                       removeMulti = NULL,
                       filterMulti = NULL,
                       filterNonproductive = NULL,
                       dist_type = NULL,
                       dist_mat = NULL,
                       gap_open = NULL,
                       gap_extend = NULL) {

  # Handle deprecated arguments
  remove.na <- .deprecate_arg(removeNA, remove.na, "removeNA", "remove.na",
                              "combineBCR", default = FALSE)
  remove.multi <- .deprecate_arg(removeMulti, remove.multi, "removeMulti", "remove.multi",
                                 "combineBCR", default = FALSE)
  filter.multi <- .deprecate_arg(filterMulti, filter.multi, "filterMulti", "filter.multi",
                                 "combineBCR", default = TRUE)
  filter.nonproductive <- .deprecate_arg(filterNonproductive, filter.nonproductive,
                                         "filterNonproductive", "filter.nonproductive",
                                         "combineBCR", default = TRUE)
  dist.type <- .deprecate_arg(dist_type, dist.type, "dist_type", "dist.type",
                              "combineBCR", default = "levenshtein")
  dist.mat <- .deprecate_arg(dist_mat, dist.mat, "dist_mat", "dist.mat",
                             "combineBCR", default = "BLOSUM80")
  gap.open <- .deprecate_arg(gap_open, gap.open, "gap_open", "gap.open",
                             "combineBCR", default = -10)
  gap.extend <- .deprecate_arg(gap_extend, gap.extend, "gap_extend", "gap.extend",
                               "combineBCR", default = -1)
  
  # Initial Contig Processing and Filtering
  processed_list <- input.data %>%
    .checkList() %>%
    .checkContigs() %>%
    unname() %>%
    purrr::imap(function(x, i) {
      x <- subset(x, chain %in% c("IGH", "IGK", "IGL"))
      if (!is.null(ID)) x$ID <- ID[i]
      if (filter.nonproductive && "productive" %in% colnames(x)) {
        x <- subset(x, tolower(productive) == "true")
      }
      if (filter.multi) {
        # Keep IGH / IGK / IGL info in save_chain
        x$save_chain <- x$chain
        # Collapse IGK and IGL chains
        x$chain <- ifelse(x$chain == "IGH", "IGH", "IGLC")
        x <- .filteringMulti(x)
        # Get back IGK / IGL distinction
        x$chain <- x$save_chain
        x$save_chain <- NULL
      }
      x
    }) %>%
    # Add sample/ID prefixes
    (function(x) {
      if (!is.null(samples)) {
        .modifyBarcodes(x, samples, ID)
      } else { 
        x
      }
    }) %>%
    # Reshape data to one row per barcode with columns for each chain
    lapply(function(x) {
      data2 <- data.frame(x)
      data2 <- .makeGenes(cellType = "B", data2)
      unique_df <- unique(data2$barcode)
      Con.df <- data.frame(matrix(NA, length(unique_df), 9))
      colnames(Con.df) <- c("barcode", heavy_lines, light_lines)
      Con.df$barcode <- unique_df
      Con.df <- .parseBCR(Con.df, unique_df, data2)
      Con.df <- .assignCT(cellType = "B", Con.df)
      if(!is.null(group.by)) { #retain group.by variable for clustering
        barcode_groups <- data2[!duplicated(data2$barcode), c("barcode", group.by)]
        Con.df[[group.by]] <- barcode_groups[[group.by]][match(Con.df$barcode, barcode_groups$barcode)]
      }
      Con.df %>% 
        mutate(length1 = nchar(cdr3_nt1)) %>%
        mutate(length2 = nchar(cdr3_nt2))
    })
  
  # Getting CTstrict based on clusters
  if (call.related.clones) {
    clusters <- clonalCluster(processed_list,
                              sequence = sequence,
                              chain = chain,
                              threshold = threshold,
                              group.by = group.by,
                              use.V = use.V,
                              use.J = use.J,
                              cluster.method = cluster.method,
                              dist.type = dist.type,
                              dist.mat = dist.mat,
                              normalize = normalize,
                              gap.open = gap.open,
                              gap.extend = gap.extend)
  }
  
  # Defining element names for the final output
  list_names <- if (!is.null(samples)) {
    if (is.null(ID)) samples else paste0(samples, "_", ID)
  } else {
    paste0("S", seq_along(processed_list))
  }
  
  final_list <- purrr::map2(processed_list, seq_along(processed_list), function(df, i) {
    # Assigning CTstrict
    if (call.related.clones) {
      # Get the cluster column from clonalCluster output
      cluster_col <- clusters[[i]][, ncol(clusters[[i]])]
      
      # ========== CTstrict FORMATTING LOGIC ==========
      seq_col <- ifelse(sequence == "aa", "cdr3_aa", "cdr3_nt")
      heavy_seq_col <- paste0(seq_col, "1")
      light_seq_col <- paste0(seq_col, "2")
      
      # Create unique identifiers for each chain (vgene.sequence format)
      heavy_unique <- ifelse(
        !is.na(df[, "vgene1"]) & !is.na(df[, heavy_seq_col]),
        paste0(df[, "vgene1"], ".", df[, heavy_seq_col]),
        "NA"
      )
      
      light_unique <- ifelse(
        !is.na(df[, "vgene2"]) & !is.na(df[, light_seq_col]),
        paste0(df[, "vgene2"], ".", df[, light_seq_col]),
        "NA"
      )
      
      if (chain == "both") {
        # Both chains in same network - use cluster ID for both parts
        heavy_part <- ifelse(is.na(cluster_col), heavy_unique, cluster_col)
        light_part <- ifelse(is.na(cluster_col), light_unique, cluster_col)
        
      } else if (chain == "IGH") {
        # Only heavy chain clustered
        heavy_part <- ifelse(is.na(cluster_col), heavy_unique, cluster_col)
        light_part <- light_unique  # Light chain always uses unique ID
        
      } else if (chain %in% c("IGL", "IGK", "Light")) {
        # Only light chain clustered
        heavy_part <- heavy_unique  # Heavy chain always uses unique ID
        light_part <- ifelse(is.na(cluster_col), light_unique, cluster_col)
        
      } else {
        # Fallback for any other chain specification
        heavy_part <- heavy_unique
        light_part <- light_unique
      }
      
      # Combine into CTstrict with underscore separator
      df[, "CTstrict"] <- paste0(heavy_part, "_", light_part)
      
    } else {
      df[, "CTstrict"] <- paste0(df[, "vgene1"], ".", df[, "cdr3_aa1"], "_",
                                 df[, "vgene2"], ".", df[, "cdr3_aa2"])
    }
    # Adding samples/ID if applicable
    if (!is.null(samples)) df$sample <- samples[i]
    if (!is.null(ID)) df$ID <- ID[i]
    
    # Cleaning up the "NA"
    df[df == "NA_NA" | df == "NA.NA_NA.NA" | df == "NA;NA_NA;NA" | df == "NA"] <- NA
    
    # Select, reorder, and filter final columns
    col_selection <- c("barcode", "sample", "ID",
                       heavy_lines[c(1, 2, 3)], light_lines[c(1, 2, 3)], CT_lines)
    col_selection <- col_selection[col_selection %in% names(df)] # Keep only existing cols
    df <- df[, col_selection]
    df <- df[!duplicated(df$barcode), ]
    df <- df[rowSums(is.na(df)) < (ncol(df) - 1), ] 
    return(df)
  })
  
  # Set the names of the final list
  names(final_list) <- list_names
  
  # Final Optional Filtering
  if (remove.na) final_list <- .removingNA(final_list)
  if (remove.multi) final_list <- .removingMulti(final_list)
  
  return(final_list)
}

# test script for percentKmer.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                     "P19B","P19L", "P20B", "P20L"))
                                    
test_that("percentKmer: output type is correct for aa", {
  
  # Test for ggplot output by default
  p <- percentKmer(combined, 
                   clone.call = "aa", 
                   motif.length = 3, 
                   min.depth = 1)
  expect_s3_class(p, "ggplot")
  
  # Test for matrix output when export.table = TRUE
  mat <- percentKmer(combined, 
                     clone.call = "aa", 
                     motif.length = 3, 
                     min.depth = 1, 
                     export.table = TRUE)
  expect_true(is.matrix(mat))
})

test_that("percentKmer: output type is correct for nt", {
  
  # Test for ggplot output by default
  p <- percentKmer(combined, 
                   clone.call = "nt", 
                   motif.length = 3, 
                   min.depth = 1)
  expect_s3_class(p, "ggplot")
  
  # Test for matrix output when export.table = TRUE
  mat <- percentKmer(combined, 
                     clone.call = "nt", 
                     motif.length = 3, 
                     min.depth = 1, 
                     export.table = TRUE)
  expect_true(is.matrix(mat))
})


test_that("percentKmer: matrix calculations are accurate", {
  
  mat <- percentKmer(combined, 
                     clone.call = "aa", 
                     motif.length = 3, 
                     min.depth = 1, 
                     top.motifs = NULL, 
                     export.table = TRUE)
  
  # Check dimensions
  expect_equal(nrow(mat), 8)
  expect_equal(ncol(mat), 5500)
  expect_equal(rownames(mat), c("P17B", "P17L", "P18B", "P18L", "P19B", 
                                "P19L", "P20B", "P20L"))
  
  # Check if all row sums equal 1 (for proportions)
  expect_true(all(abs(rowSums(mat) - 1) < 1e-9))
})


test_that("percentKmer: parameter handling works as expected", {
  
  mat_top1 <- percentKmer(combined, 
                          clone.call = "aa", 
                          motif.length = 3, 
                          min.depth = 1, 
                          top.motifs = 1,
                          export.table = TRUE)
  expect_true(ncol(mat_top1) == 1)
  expect_equal(colnames(mat_top1), "ASS")
  
  # Test `motif.length` parameter
  mat_len2 <- percentKmer(combined, 
                          clone.call = "aa", 
                          motif.length = 2, 
                          top.motifs = NULL, 
                          export.table = TRUE)
  expect_true(all(nchar(colnames(mat_len2)) == 2))
})


test_that("percentKmer: input validation and error handling", {
  
  # Test for error on invalid `clone.call`
  expect_error(
    percentKmer(combined, clone.call = "invalid_option"),
    "Please select either nucleotide (nt) or amino acid (aa) sequences for clone.call",
    fixed = TRUE
  )
})

# test script for subsetClones.R - testcases are NOT comprehensive!

combined <- combineTCR(contig_list,
                      samples = c("P17B", "P17L", "P18B", "P18L",
                                  "P19B", "P19L", "P20B", "P20L"))


test_that("Subsetting to a single variable works correctly", {
  subset_single <- subsetClones(combined, name = "sample", variables = "P17B")
  expect_type(subset_single, "list")
  expect_length(subset_single, 1)
  expect_equal(names(subset_single), "P17B")
  expect_identical(subset_single[[1]], combined[["P17B"]])
})

test_that("Subsetting to multiple variables works correctly", {
  vars_to_subset <- c("P17L", "P19B", "P20L")
  subset_multi <- subsetClones(combined, name = "sample", variables = vars_to_subset)
  expect_type(subset_multi, "list")
  expect_length(subset_multi, 3)
  expect_equal(names(subset_multi), vars_to_subset)
  expect_identical(subset_multi, combined[vars_to_subset])
})

test_that("Returns an empty list when no variables match", {
  subset_empty <- subsetClones(combined, name = "sample", variables = "nonexistent_sample")
  expect_type(subset_empty, "list")
  expect_length(subset_empty, 0)
  expect_equal(subset_empty, list())
})

test_that("Subsetting works with other added metadata columns", {
  # Add a new variable to the combined object for testing
  combined_with_type <- addVariable(combined, 
                                    variable.name = "Type", 
                                    variables = rep(c("B_cell", "L_cell"), 4))
  subset_type <- subsetClones(combined_with_type, name = "Type", variables = "B_cell")
  expect_length(subset_type, 4)
  expected_names <- c("P17B", "P18B", "P19B", "P20B")
  expect_equal(names(subset_type), expected_names)
})

test_that("Function fails gracefully with an invalid column name", {
  expect_error(subsetClones(combined, name = "invalid_column_name", variables = "any_value"))
})

# test script for eportClones.R - testcases are NOT comprehensive!

# Setup: Create a combined object for testing using internal package data
data("contig_list", package = "scRepertoire")
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))

test_that("exportClones handles invalid format", {
  expect_error(
    exportClones(combined, format = "invalid_format"),
    "'arg' should be one of"
  )
})

test_that("'paired' format export works correctly", {
  # Test returning a data frame
  paired_df <- exportClones(combined, format = "paired", write.file = FALSE)
  expect_s3_class(paired_df, "data.frame")
  expect_named(paired_df, c("chain1_aa", "chain1_nt", "chain1_genes",
                            "chain2_aa", "chain2_nt", "chain2_genes", "group"))
  expect_true(nrow(paired_df) > 0)
  
  # Test writing to file
  temp_dir <- tempdir()
  file_path <- file.path(temp_dir, "paired_test.csv")
  exportClones(combined, format = "paired", dir = temp_dir, file.name = "paired_test.csv")
  expect_true(file.exists(file_path))
  
  # Verify file content
  read_df <- read.csv(file_path, row.names = 1) # row.names are barcodes
  expect_s3_class(read_df, "data.frame")
  expect_equal(ncol(read_df), ncol(paired_df))
  unlink(file_path) # Clean up
})

test_that("'TCRMatch' format export works correctly", {
  tcr_match_df <- exportClones(combined, format = "TCRMatch", write.file = FALSE)
  expect_s3_class(tcr_match_df, "data.frame")
  expect_named(tcr_match_df, c("chain2_aa", "group", "clonalFrequency"))
  expect_true(nrow(tcr_match_df) > 0)
})

test_that("'tcrpheno' format export works correctly", {
  tcr_pheno_df <- exportClones(combined, format = "tcrpheno", write.file = FALSE)
  expect_s3_class(tcr_pheno_df, "data.frame")
  expect_named(tcr_pheno_df, c("cell", "TCRA_cdr3aa", "TCRA_vgene", "TCRA_jgene",
                               "TCRA_cdr3nt", "TCRB_cdr3aa", "TCRB_vgene",
                               "TCRB_jgene", "TCRB_cdr3nt"))
  expect_true(nrow(tcr_pheno_df) > 0)
})

test_that("'airr' format export works correctly", {
  airr_df <- exportClones(combined, format = "airr", write.file = FALSE)
  expect_s3_class(airr_df, "data.frame")
  expect_named(airr_df, c("cell_id", "locus", "v_call", "d_call", "j_call",
                          "c_call", "junction", "junction_aa"))
  expect_true(nrow(airr_df) > 0)
})

test_that("'immunarch' format export works correctly", {
  skip_if_not_installed("dplyr")
  
  # Test returning a list object
  immunarch_list <- exportClones(combined, format = "immunarch", write.file = FALSE)
  expect_type(immunarch_list, "list")
  expect_named(immunarch_list, c("data", "meta"))
  expect_s3_class(immunarch_list$meta, "data.frame")
  expect_type(immunarch_list$data, "list")
  expect_true(length(immunarch_list$data) > 0)
  expect_named(immunarch_list$data[[1]], c("Clones", "Proportion", "CDR3.nt",
                                           "CDR3.aa", "V.name", "D.name",
                                           "J.name", "C.name", "Barcode"))
  
  # Test writing to file (with the bug fix)
  temp_dir <- tempdir()
  file_path <- file.path(temp_dir, "immunarch_test.csv")
  exportClones(combined, format = "immunarch", dir = temp_dir, file.name = "immunarch_test.csv")
  expect_true(file.exists(file_path))
  
  # Verify file content
  read_df <- read.csv(file_path)
  expect_s3_class(read_df, "data.frame")
  expect_true(nrow(read_df) > 0)
  expect_true("Sample" %in% names(read_df))
  unlink(file_path) # Clean up
})

# test script for alluvialClones.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list, samples = c("P17B", "P17L", "P18B", "P18L", "P19B", "P19L", "P20B", "P20L"))
scRep_example <- combineExpression(combined, scRep_example)
scRep_example$Patient <- substring(scRep_example$orig.ident, 1, 3)
scRep_example$Type <- substring(scRep_example$orig.ident, 4, 4)

test_that("Input validation works correctly", {
  expect_error(alluvialClones(scRep_example, y.axes = NULL),
               "Make sure you have selected the variable\\(s\\) to visualize")
})

test_that("Output formats (table vs. plot) are correct", {
  # Test that export.table = TRUE returns a data.frame in lodes format
  table_output <- alluvialClones(scRep_example, y.axes = c("Patient", "Type"), export.table = TRUE)
  expect_s3_class(table_output, "data.frame")
  expect_true(all(c("x", "stratum", "alluvium") %in% names(table_output)))
  
  # Test that the default behavior returns a ggplot object
  plot_output <- alluvialClones(scRep_example, y.axes = c("Patient", "Type"))
  expect_s3_class(plot_output, "ggplot")
})

test_that("Aesthetic parameters (`color`, `alpha`, `facet`) are handled correctly", {
  # Test with color as a column
  plot_color <- alluvialClones(scRep_example, y.axes = c("Patient", "Type"), color = "Type")
  expect_equal(plot_color$labels$fill, "Type")
  
  # Test with color as a specific clone
  plot_color_clone <- alluvialClones(scRep_example, y.axes = c("Patient", "Type"), color = "cloneA")
  expect_equal(plot_color_clone$labels$fill, "clone(s)")
  
  # Test with facet
  plot_facet <- alluvialClones(scRep_example, y.axes = c("ident", "Type"), facet = "Patient")
  expect_s3_class(plot_facet$facet, "FacetWrap")
})

test_that("Lodes data frame is generated correctly", {
  lodes_df <- alluvialClones(scRep_example, y.axes = c("Patient", "Type"), export.table = TRUE)
  axis_labels <- unique(as.character(lodes_df$x))
  expect_equal(axis_labels, c("Patient", "Type"))
})

test_that("Plot structure has correct geoms", {
  # No color or alpha
  plot_plain <- alluvialClones(scRep_example, y.axes = c("Patient", "Type"))
  expect_true("GeomStratum" %in% sapply(plot_plain$layers, function(l) class(l$geom)[1]))
  expect_true("GeomAlluvium" %in% sapply(plot_plain$layers, function(l) class(l$geom)[1]))
  
  # With color
  plot_color <- alluvialClones(scRep_example, y.axes = c("Patient", "Type"), color = "Type")
  expect_true("GeomFlow" %in% sapply(plot_color$layers, function(l) class(l$geom)[1]))
  
  # All plots should have geom_text
  expect_true("GeomText" %in% sapply(plot_plain$layers, function(l) class(l$geom)[1]))
})

# test script for getCirclize.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list, 
                       samples = c("P17B", "P17L", "P18B", "P18L", 
                                   "P19B", "P19L", "P20B", "P20L"))
scRep_example <- combineExpression(combined, scRep_example)

test_that("Output has the correct format", {
  df <- getCirclize(scRep_example, group.by = "seurat_clusters")
  
  expect_s3_class(df, "data.frame")
  expect_equal(names(df), c("from", "to", "value"))
  expect_true(is.numeric(df$value))
})

test_that("`include.self = FALSE` correctly removes self-referential rows", {
  df <- getCirclize(scRep_example, group.by = "seurat_clusters", include.self = FALSE)
  self_referential_rows <- df[df$from == df$to, ]
  expect_equal(nrow(self_referential_rows), 0)
  expect_gt(nrow(df), 0)
})


test_that("Function handles different `clone.call` methods", {
  expect_silent(
    df_strict <- getCirclize(scRep_example, group.by = "seurat_clusters", clone.call = "strict")
  )
  df_aa <- getCirclize(scRep_example, group.by = "seurat_clusters", clone.call = "aa")
  expect_equal(df_strict$value, df_aa$value)
})

# test script for positionalEntropy.R - testcases are NOT comprehensive!

combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined, 
                        variable.name = "Type", 
                        variables = rep(c("B", "L"), 4))


test_that("Function returns a ggplot object by default", {
  p <- positionalEntropy(combined)
  expect_s3_class(p, "ggplot")
})

test_that("Function returns a data.frame when export.table is TRUE", {
  df_out <- positionalEntropy(combined, export.table = TRUE)
  expect_s3_class(df_out, "data.frame")
  expect_true(all(c("entropy", "Position", "group") %in% colnames(df_out)))
})

test_that("Grouping with 'group.by' works correctly", {
  df_grouped <- positionalEntropy(combined, group.by = "Type", export.table = TRUE)
  
  # Check for correct group names
  expect_equal(sort(unique(df_grouped$group)), c("B", "L"))
  
  # Check calculations for group B
  group_b_data <- subset(df_grouped, group == "B")
  expect_equal(subset(group_b_data, Position == 1)$entropy, 0) 
  expect_equal(subset(group_b_data, Position == 20)$entropy, 0) 
})

test_that("'method' parameter changes", {
  df_shannon <- positionalEntropy(combined, 
                                  method = "shannon", 
                                  export.table = TRUE)
  expect_s3_class(df_shannon, "data.frame")
  expect_true(all(c("entropy", "Position", "group") %in% colnames(df_shannon)))
  
  df_inv_simpson <- positionalEntropy(combined, 
                                      method = "inv.simpson", 
                                      export.table = TRUE)
  expect_s3_class(df_inv_simpson, "data.frame")
  expect_true(all(c("entropy", "Position", "group") %in% colnames(df_inv_simpson)))
  
  df_gini_simpson <- positionalEntropy(combined, 
                                       method = "gini.simpson", 
                                       export.table = TRUE)
  expect_s3_class(df_gini_simpson, "data.frame")
  expect_true(all(c("entropy", "Position", "group") %in% colnames(df_gini_simpson)))

  df_hill0 <- positionalEntropy(combined, 
                                method = "hill0", 
                                export.table = TRUE)
  expect_s3_class(df_hill0 , "data.frame")
  expect_true(all(c("entropy", "Position", "group") %in% colnames(df_hill0 )))
  
  df_hill1 <- positionalEntropy(combined, 
                                method = "hill1", 
                                export.table = TRUE)
  expect_s3_class(df_hill1, "data.frame")
  expect_true(all(c("entropy", "Position", "group") %in% colnames(df_hill1)))
  
  df_hill2 <- positionalEntropy(combined, 
                                method = "hill2", 
                                export.table = TRUE)
  expect_s3_class(df_hill2, "data.frame")
  expect_true(all(c("entropy", "Position", "group") %in% colnames(df_hill2)))

})

test_that("Ordering with 'order.by' works correctly", {
  order_vec <- c("L", "B")
  df_ordered <- positionalEntropy(combined, 
                                  group.by = "Type", 
                                  order.by = order_vec, 
                                  export.table = TRUE)
  
  expect_s3_class(df_ordered$group, "factor")
  expect_equal(levels(df_ordered$group), order_vec)
})

t

# test script for percentAA.R - testcases are NOT comprehensive!

combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined, 
                        variable.name = "Type", 
                        variables = rep(c("B", "L"), 4))

test_that("Function returns a ggplot object by default", {
  p <- percentAA(combined)
  expect_s3_class(p, "ggplot")
})

test_that("Function returns a data.frame when export.table is TRUE", {
  df_out <- percentAA(combined, export.table = TRUE)
  expect_s3_class(df_out, "data.frame")
  # Check for expected columns
  expected_cols <- c("AminoAcid", "Position", "Frequency", "group")
  expect_true(all(expected_cols %in% colnames(df_out)))
})

test_that("Grouping with 'group.by' works correctly", {
  p_grouped <- percentAA(combined, group.by = "Type")
  df_grouped <- percentAA(combined, group.by = "Type", export.table = TRUE)
  
  # Check data.frame for correct groups
  expect_equal(sort(unique(df_grouped$group)), c("B", "L"))
  
  # Check data integrity for group B
  group_b_data <- df_grouped[df_grouped$group == "B", ]
  pos1_C <- subset(group_b_data, Position == 1 & AminoAcid == "C")
  expect_equal(pos1_C$Frequency, 1) 
})


test_that("Ordering with 'order.by' works correctly", {
  # Groups are B, L. Default order is alphabetical. We want L, B.
  order_vec <- c("L", "B")
  df_ordered <- percentAA(combined, group.by = "Type", order.by = order_vec, export.table = TRUE)
  
  # Check if the 'group' column is a factor with the specified levels
  expect_s3_class(df_ordered$group, "factor")
  expect_equal(levels(df_ordered$group), order_vec)
})

# test script for clonalLength.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined,
                        variable.name = "Type",
                        variables = rep(c("B", "L"), 4))

test_that("Input validation for `clone.call` works", {
  error_message <- "Please make a selection of the type of CDR3 sequence to analyze by using `clone.call`"
  expect_error(clonalLength(combined, clone.call = "gene"), error_message, fixed = TRUE)
  expect_error(clonalLength(combined, clone.call = "strict"), error_message, fixed = TRUE)
})

test_that("Output formats (table vs. plot) are correct", {
  table_output <- clonalLength(combined, export.table = TRUE)
  expect_s3_class(table_output, "data.frame")
  expect_true(all(c("length", "values") %in% names(table_output)))
  plot_output <- clonalLength(combined)
  expect_s3_class(plot_output, "ggplot")
})

test_that("`scale` parameter correctly switches plot type", {
  # Test with scale = FALSE (default) for a bar plot
  plot_bar <- clonalLength(combined, scale = FALSE)
  expect_s3_class(plot_bar$layers[[1]]$geom, "GeomBar")
  expect_equal(plot_bar$labels$y, "Number of CDR3 (AA)")
  
  # Test with scale = TRUE for a density plot
  plot_density <- clonalLength(combined, scale = TRUE)
  expect_s3_class(plot_density$layers[[1]]$geom, "GeomDensity")
  expect_equal(plot_density$labels$y, "Density of CDR3 (AA)")
})

test_that("`group.by` parameter functions correctly", {
  # Test exporting a table grouped by "Type"
  table_grouped <- clonalLength(combined, group.by = "Type", export.table = TRUE)
  expect_true("Type" %in% names(table_grouped))
  expect_true("values" %in% names(table_grouped))
  expect_true(all(unique(table_grouped$Type) %in% c("B", "L")))
  
  # Test plotting grouped by "Type"
  plot_grouped <- clonalLength(combined, group.by = "Type")
  expect_s3_class(plot_grouped, "ggplot")
})

test_that("`clone.call` and `chain` parameters execute correctly", {
  # Test with clone.call = "nt"
  plot_nt <- clonalLength(combined, clone.call = "nt")
  expect_s3_class(plot_nt, "ggplot")
  expect_equal(plot_nt$labels$y, "Number of CDR3 (NT)")
  table_nt <- clonalLength(combined, clone.call = "nt", export.table = TRUE)
  table_aa <- clonalLength(combined, clone.call = "aa", export.table = TRUE)
  expect_gt(mean(table_nt$length), mean(table_aa$length))
  
  # Test single chain selection
  expect_silent(plot_tra <- clonalLength(combined, chain = "TRA"))
  expect_s3_class(plot_tra, "ggplot")
  table_tra <- clonalLength(combined, chain = "TRA", export.table = TRUE)
  expect_s3_class(table_tra, "data.frame")
  expect_lt(nrow(table_tra), nrow(table_aa))
})

# test script for combineExpression.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list, 
                       samples = c("P17B", "P17L", "P18B", "P18L", 
                                   "P19B", "P19L", "P20B", "P20L"))
scRep_example <- combineExpression(combined, scRep_example)
scRep_example$Patient <- substring(scRep_example$orig.ident, 1, 3)
scRep_example$Type <- substring(scRep_example$orig.ident, 4, 4)


test_that("Basic functionality with Seurat object and default parameters", {
  # Run the function with default settings
  sc <- combineExpression(combined, scRep_example)
  expect_s4_class(sc, "Seurat")
  expect_equal(ncol(sc), ncol(scRep_example))
  metadata_cols <- colnames(sc[[]])
  expected_cols <- c("CTstrict", "clonalFrequency", "clonalProportion", "cloneSize")
  expect_true(all(expected_cols %in% metadata_cols))
  expect_s3_class(sc$cloneSize, "factor")
})


test_that("`clone.call` parameter works correctly", {
  # Test with clone.call = "nt"
  sc_nt <- combineExpression(combined, scRep_example, clone.call = "nt")
  expect_true("CTnt" %in% colnames(sc_nt[[]]))
  
  # Test with clone.call = "aa"
  sc_aa <- combineExpression(combined, scRep_example, clone.call = "aa")
  expect_true("CTaa" %in% colnames(sc_aa[[]]))
  
  # Test with clone.call = "gene"
  sc_gene <- combineExpression(combined, scRep_example, clone.call = "gene")
  expect_true("CTgene" %in% colnames(sc_gene[[]]))
})


test_that("`proportion` and `cloneSize` parameters work as expected", {
  # Test with frequency calculation (proportion = FALSE)
  sc_freq <- combineExpression(combined, 
                               scRep_example, 
                               proportion = FALSE, 
                               clone.size = c(1,2,5,10))
  meta_freq <- sc_freq[[]]
  max_freq <- max(meta_freq$clonalFrequency, na.rm = TRUE)
  clone_id_max_freq <- meta_freq$CTstrict[which.max(meta_freq$clonalFrequency)]
  
  # Default cloneSize for frequency goes up to the max value automatically
  expect_true(grepl(paste0("<= ", max_freq, ")"), levels(meta_freq$cloneSize)[1]))
  
  # Test error when proportion = FALSE and cloneSize bins are < 1
  bad_cloneSize <- c(Rare = 0.1, Small = 0.5)
  expect_error(
    combineExpression(combined, scRep_example, proportion = FALSE, clone.size = bad_cloneSize),
    "Adjust the clone.size parameter - there are groupings < 1"
  )
})

test_that("`filter.na` parameter correctly subsets the object", {
  sc_filtered <- combineExpression(combined, scRep_example, filter.na = TRUE)
  expect_lt(ncol(sc_filtered), ncol(scRep_example))
  expect_false(any(is.na(sc_filtered$CTstrict)))
  non_na_barcodes <- rownames(scRep_example[[]])[!is.na(combineExpression(combined, scRep_example)$CTstrict)]
  expect_equal(ncol(sc_filtered), length(non_na_barcodes))
})

test_that("`chain` parameter filters correctly for calculations", {
  sc_tra <- combineExpression(combined, scRep_example, chain = "TRA")
  sc_trb <- combineExpression(combined, scRep_example, chain = "TRB")
  
  # The clonal proportions should be different when calculated on different chains
  expect_false(identical(sc_tra$clonalProportion, sc_trb$clonalProportion))
  sc_both <- combineExpression(combined, scRep_example, chain = "both")
  
  # Pick a cell to compare
  cell_barcode <- na.omit(sc_tra[[]])$barcode[1]
  
  # But the strict clonotype ID should be identical, preserving the paired info
  id_tra <- sc_tra[[c("CTstrict")]][cell_barcode, ]
  id_both <- sc_both[[c("CTstrict")]][cell_barcode, ]
  expect_equal(id_tra, id_both)
})

test_that("Function issues a warning for high barcode mismatch", {
  # Create a Seurat object with non-matching barcodes
  combined.mm <- lapply(combined, function(x) {
                x$barcode <- paste0("mismatch_", x$barcode )
                x
  })
  
  # Expect the specific warning message
  expect_error(
    combineExpression(combined.mm, scRep_example),
    regexp =  "< 1% of barcodes match"
  )
})

# test script for clonalAbundance.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined, 
                        variable.name = "Type", 
                        variables = rep(c("B", "L"), 4))

test_that("export.table = TRUE returns a data.frame", {
  table_output <- clonalAbundance(combined, export.table = TRUE)
  expect_s3_class(table_output, "data.frame")
  expect_true(all(c("Abundance", "values") %in% names(table_output)))
})

test_that("Default behavior returns a ggplot object", {
  plot_output <- clonalAbundance(combined)
  expect_s3_class(plot_output, "ggplot")
})

test_that("scale = TRUE produces a density plot", {
  plot_output <- clonalAbundance(combined, scale = TRUE)
  expect_s3_class(plot_output, "ggplot")
  expect_equal(plot_output$labels$y, "Density of Clones")
  expect_s3_class(plot_output$layers[[1]]$geom, "GeomDensity")
})

test_that("scale = FALSE produces a line plot", {
  plot_output <- clonalAbundance(combined, scale = FALSE)
  expect_s3_class(plot_output, "ggplot")
  expect_equal(plot_output$labels$y, "Number of Clones")
  expect_s3_class(plot_output$layers[[1]]$geom, "GeomLine")
})

test_that("group.by correctly groups the data", {
  table_output_grouped <- clonalAbundance(combined, group.by = "Type", export.table = TRUE)
  expect_s3_class(table_output_grouped, "data.frame")
  expect_true("Type" %in% names(table_output_grouped))
  expect_lt(length(unique(table_output_grouped$patient)), length(combined))
  
  plot_output_grouped <- clonalAbundance(combined, group.by = "Type")
  expect_s3_class(plot_output_grouped, "ggplot")
  expect_equal(plot_output_grouped$labels$colour, "Type")
})

test_that("clone.call handles different arguments", {
  # Test with clone.call = "gene"
  expect_s3_class(clonalAbundance(combined, clone.call = "gene", export.table = TRUE), "data.frame")
  expect_s3_class(clonalAbundance(combined, clone.call = "gene"), "ggplot")
  
  # Test with clone.call = "nt"
  expect_s3_class(clonalAbundance(combined, clone.call = "nt", export.table = TRUE), "data.frame")
  expect_s3_class(clonalAbundance(combined, clone.call = "nt"), "ggplot")
  
  # Test with clone.call = "aa"
  expect_s3_class(clonalAbundance(combined, clone.call = "aa", export.table = TRUE), "data.frame")
  expect_s3_class(clonalAbundance(combined, clone.call = "aa"), "ggplot")
})

test_that("chain selection works correctly", {
  table_TRA <- clonalAbundance(combined, chain = "TRA", export.table = TRUE)
  expect_s3_class(table_TRA, "data.frame")
  
  # Test with chain = "TRB"
  table_TRB <- clonalAbundance(combined, chain = "TRB", export.table = TRUE)
  expect_s3_class(table_TRB, "data.frame")
  
  # Test with chain = "both" (default)
  table_both <- clonalAbundance(combined, chain = "both", export.table = TRUE)
  expect_s3_class(table_both, "data.frame")
  
  expect_lte(nrow(table_TRA), nrow(table_both))
  expect_lte(nrow(table_TRB), nrow(table_both))
})

# test script for clonalDiversity.R - testcases are NOT comprehensive!

# Generate some data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))

combined <- addVariable(combined, 
            variable.name = "Type", 
            variables = rep(c("B", "L"), 4))

test_that("clonalDiversity: Input validation works", {
  expect_error(
    clonalDiversity(combined_input_list, metric = "invalid_metric"),
    "Invalid `metric`"
  )
  expect_error(
    clonalDiversity(combined_input_list, metric = c("shannon", "inv.simpson")),
    "`metric` must be a single string."
  )
})

test_that("clonalDiversity: skip.boots = TRUE calculates correct values", {
  # Test with a list input
  result <- clonalDiversity(
    combined,
    clone.call = "CTaa",
    metric = "shannon",
    skip.boots = TRUE,
    export.table = TRUE
  )
  
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 8)
  expect_equal(result$value[result$Group == "P17B"], 4.108008, tolerance = 1e-6)
  expect_equal(result$value[result$Group == "P17L"], 7.33391, tolerance = 1e-6)
  
  # Test with a data.frame input
  result_df <- clonalDiversity(
    combined,
    clone.call = "CTaa",
    metric = "inv.simpson",,
    skip.boots = TRUE,
    export.table = TRUE
  )
  
  expect_equal(result_df$value[result_df$Group == "P17B"], 8.029998, tolerance = 1e-6)
  expect_equal(result_df$value[result_df$Group == "P17L"], 627.9127, tolerance = 1e-6)
})


test_that("clonalDiversity: Bootstrapping returns correct structure", {
  set.seed(42) # for reproducibility
  n_boots <- 10
  
  # Test with mean of bootstraps (default)
  result <- clonalDiversity(
    combined,
    clone.call = "CTaa",
    metric = "shannon",
    n.boots = n_boots,
    skip.boots = FALSE,
    export.table = TRUE
  )
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 8)
  expect_true(all(c("Group", "value", "metric") %in% names(result)))
  # Value should be a single mean value per group
  expect_true(is.numeric(result$value))
  
  # Test with return.boots = TRUE
  result_all_boots <- clonalDiversity(
    combined,
    clone.call = "CTaa",
    metric = "shannon",
    n.boots = n_boots,
    return.boots = TRUE # This also sets export.table = TRUE
  )
  expect_s3_class(result_all_boots, "data.frame")
  # Should have n.boots rows for each group
  expect_equal(nrow(result_all_boots), 8 * n_boots)
  expect_equal(sum(result_all_boots$Group == "P17B"), n_boots)
  expect_equal(sum(result_all_boots$Group == "P17L"), n_boots)
})


test_that("clonalDiversity: export.table and plotting works", {
  # Returns a data.frame when export.table = TRUE
  result_table <- clonalDiversity(
    combined,
    export.table = TRUE
  )
  expect_s3_class(result_table, "data.frame")
  
  # Returns a ggplot object when export.table = FALSE (default)
  result_plot <- clonalDiversity(
    combined,
    export.table = FALSE
  )
  expect_s3_class(result_plot, "ggplot")
  
  # Test x.axis functionality returns a ggplot object
  result_plot_xaxis <- clonalDiversity(
    combined,
    x.axis = "Type",
    export.table = FALSE
  )
  expect_s3_class(result_plot_xaxis, "ggplot")
})

# test script for clonalCluster.R 

library(testthat)
library(igraph)
library(SingleCellExperiment) 
library(Matrix)

# --- Global Setup ---
# Using the user-provided setup for general structure tests
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))

set.seed(42)
combined <- lapply(combined, function(x) {
  x <- x[sample(nrow(x), nrow(x) * .25),]
  x
})

# --- Existing Tests (Preserved) ---

test_that("Basic functionality and default output structure", {
  clustered_list <- clonalCluster(combined[1:2])
  expect_length(clustered_list, length(combined[1:2]))
  expect_true(all(sapply(clustered_list, is.data.frame)))
  expect_true("TRB.Cluster" %in% names(clustered_list[[1]]))
  
  # Check for data integrity
  expect_equal(nrow(clustered_list[[1]]), nrow(combined[[1]]))
  expect_false(all(is.na(clustered_list[[1]]$TRB.Cluster)))
})

test_that("export.graph  = TRUE returns a valid igraph object", {
  graph_obj <- clonalCluster(combined[1:2], export.graph = TRUE)
  expect_s3_class(graph_obj, "igraph")
  expect_gt(vcount(graph_obj), 0)
  expect_gt(ecount(graph_obj), 0)
  expect_true("cluster" %in% igraph::vertex_attr_names(graph_obj))
  expect_true("weight" %in% igraph::edge_attr_names(graph_obj))
})

test_that("export.adj.matrix = TRUE returns a valid sparse matrix", {
  adj_matrix <- clonalCluster(combined[3:4], export.adj.matrix  = TRUE)
  all_barcodes <- unique(do.call(rbind, combined[3:4])[["barcode"]])
  num_barcodes <- length(all_barcodes)
  expect_s4_class(adj_matrix, "dgCMatrix")
  expect_equal(dim(adj_matrix), c(num_barcodes, num_barcodes))
  expect_equal(rownames(adj_matrix), all_barcodes)
})

test_that("chain parameter works correctly", {
  clustered_tra <- clonalCluster(combined[5:6], chain = "TRA")
  expect_true("TRA.Cluster" %in% names(clustered_tra[[1]]))
  expect_false("TRB.Cluster" %in% names(clustered_tra[[1]]))
  clustered_both <- clonalCluster(combined[5:6], chain = "both")
  expect_true("Multi.Cluster" %in% names(clustered_both[[1]]))
})

test_that("group.by parameter functions without error", {
  clustered_grouped <- clonalCluster(combined[1:2], 
                                     group.by = "sample")
  expect_true("TRB.Cluster" %in% names(clustered_grouped[[1]]))
  expect_false(all(is.na(clustered_grouped[[1]]$TRB.Cluster)))
})

test_that("Different `cluster.method` options work", {
  louvain_graph <- clonalCluster(combined[5:6], 
                                 cluster.method = "louvain", 
                                 export.graph  = TRUE)
  expect_s3_class(louvain_graph, "igraph")
  expect_true("cluster" %in% igraph::vertex_attr_names(louvain_graph))
})

test_that("Input validation and error handling", {
  expect_error(
    clonalCluster(combined, export.graph  = TRUE, export.adj.matrix = TRUE),
    "Please set only one of `export.graph` or `export.adj.matrix` to TRUE."
  )
  expect_error(
    clonalCluster(combined, cluster.method = "invalid_method"),
    "Unsupported clustering.method"
  )
})


test_that("Alignment (NW/SW) and Matrix selection", {
  # Create a small controlled dataset to ensure alignment logic runs
  toy_data <- combined[1]
  
  # Test Needleman-Wunsch with BLOSUM62
  res_nw <- clonalCluster(toy_data, 
                          dist.type = "nw", 
                          dist.mat = "BLOSUM62", 
                          threshold = 0.8) 
  expect_true("TRB.Cluster" %in% names(res_nw[[1]]))
  
  # Test Damerau (Transposition)
  res_dam <- clonalCluster(toy_data, dist.type = "damerau")
  expect_true("TRB.Cluster" %in% names(res_dam[[1]]))
})

test_that("Normalization parameters function correctly", {
  # normalize = "maxlen"
  res_max <- clonalCluster(combined[1], normalize = "maxlen", threshold = 0.1)
  expect_true("TRB.Cluster" %in% names(res_max[[1]]))
  
  # normalize = "length" (mean length)
  res_len <- clonalCluster(combined[1], normalize = "length", threshold = 0.1)
  expect_true("TRB.Cluster" %in% names(res_len[[1]]))
})

# test script for clonalRarefaction.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined, 
                        variable.name = "Type", 
                        variables = rep(c("B", "L"), 4))

combined <- combined[1:4] #subset to reduce size

test_that("Output formats (data.frame vs. plot) are correct", {
  table_output_list <- clonalRarefaction(combined, export.table = TRUE, n.boots = 2)
  expect_type(table_output_list, "list")
  expect_s3_class(table_output_list$data, "data.frame")
  
  # Test that the default behavior returns a ggplot object
  plot_output <- clonalRarefaction(combined, n.boots = 2)
  expect_s3_class(plot_output, "ggplot")
})

test_that("`plot.type` parameter correctly generates plots", {
  expect_silent(p1 <- clonalRarefaction(combined, plot.type = 1, n.boots = 2))
  expect_s3_class(p1, "ggplot")
  expect_silent(p2 <- clonalRarefaction(combined, plot.type = 2, n.boots = 2))
  expect_s3_class(p2, "ggplot")
  expect_silent(p3 <- clonalRarefaction(combined, plot.type = 3, n.boots = 2))
  expect_s3_class(p3, "ggplot")
})

test_that("`hill.numbers` parameter is passed to iNEXT correctly", {
  df_q0 <- clonalRarefaction(combined, hill.numbers = 0, n.boots = 2, export.table = TRUE)$data
  expect_true(all(df_q0$Order.q == 0))
  df_q01 <- clonalRarefaction(combined, hill.numbers = c(0, 1), n.boots = 2, export.table = TRUE)$data
  expect_true(all(unique(df_q01$Order.q) %in% c(0, 1)))
})

test_that("`group.by` parameter correctly aggregates data", {
  # Group by the "Type" variable
  df_grouped <- clonalRarefaction(combined, group.by = "Type", n.boots = 2, export.table = TRUE)$data
  num_groups <- 2
  expect_equal(length(unique(df_grouped$Assemblage)), num_groups)
  expect_true(all(unique(df_grouped$Assemblage) %in% c("B", "L")))
})

test_that("`clone.call` and `chain` parameters execute correctly", {
  expect_silent(
    res_aa <- clonalRarefaction(combined, clone.call = "aa", n.boots = 2, export.table = TRUE)
  )
  expect_type(res_aa, "list")
  expect_silent(
    res_tra <- clonalRarefaction(combined, chain = "TRA", n.boots = 2, export.table = TRUE)
  )
  expect_type(res_tra, "list")
})

# test script for clonalProportion.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined,
                        variable.name = "Type",
                        variables = rep(c("B", "L"), 4))

test_that("Output formats (matrix vs. plot) are correct", {
  mat_output <- clonalProportion(combined, export.table = TRUE)
  expect_true(is.matrix(mat_output))
  plot_output <- clonalProportion(combined)
  expect_s3_class(plot_output, "ggplot")
})

test_that("The returned matrix has correct dimensions and properties", {
  mat_output <- clonalProportion(combined, export.table = TRUE)
  expect_equal(nrow(mat_output), length(combined)) # Rows should equal number of samples
  expect_equal(ncol(mat_output), 6)               # Default has 6 bins
  total_clones_per_sample <- sapply(combined, function(x) nrow(x[!is.na(x$CTstrict),]))
  expect_equal(rowSums(mat_output), total_clones_per_sample, tolerance = 1e-9)
  expected_colnames <- c("[1:10]", "[11:100]", "[101:1000]", "[1001:10000]", "[10001:30000]", "[30001:1e+05]")
  expect_equal(colnames(mat_output), expected_colnames)
})

test_that("Custom `clonal.split` parameter works as expected", {
  # Define custom bins
  custom_bins <- c(5, 50, 500)
  mat_custom <- clonalProportion(combined, clonal.split = custom_bins, export.table = TRUE)
  
  # Check dimensions with custom bins
  expect_equal(ncol(mat_custom), 3)
  
  # Check custom column names
  expected_custom_colnames <- c("[1:5]", "[6:50]", "[51:500]")
  expect_equal(colnames(mat_custom), expected_custom_colnames)
})

test_that("`group.by` parameter correctly aggregates the data", {
  mat_grouped <- clonalProportion(combined, group.by = "Type", export.table = TRUE)
  expect_equal(nrow(mat_grouped), 2)
  expect_true(all(rownames(mat_grouped) %in% c("B", "L")))
  grouped_list <- scRepertoire:::.groupList(combined, "Type")
  total_clones_per_group <- sapply(grouped_list, function(x) nrow(x[!is.na(x$CTstrict),]))
  expect_equal(rowSums(mat_grouped), total_clones_per_group, tolerance = 1e-9)
})

test_that("`clone.call` and `chain` parameters execute correctly", {
  mat_gene <- clonalProportion(combined, clone.call = "gene", export.table = TRUE)
  expect_true(is.matrix(mat_gene))
  expect_equal(nrow(mat_gene), length(combined))
  mat_tra <- clonalProportion(combined, chain = "TRA", export.table = TRUE)
  expect_true(is.matrix(mat_tra))
  expect_equal(nrow(mat_tra), length(combined))
})

test_that("Plotting output has the correct components", {
  # Generate the plot
  plot_output <- clonalProportion(combined)
  
  # Check for the correct geometry and position
  expect_s3_class(plot_output$layers[[1]]$geom, "GeomBar")
  expect_s3_class(plot_output$layers[[1]]$position, "PositionFill")
  
  # Check the labels
  expect_equal(plot_output$labels$y, "Occupied Repertoire Space")
  expect_equal(plot_output$labels$x, "Samples")
  expect_equal(plot_output$labels$fill, "Clonal Indices")
})

# test script for highlightClones.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list, samples = c("P17B", "P17L", "P18B", "P18L", "P19B", "P19L", "P20B", "P20L"))
scRep_example <- combineExpression(combined, scRep_example)

test_that("Function correctly highlights a single sequence", {
  # Highlight cloneB
  sc_highlighted <- highlightClones(scRep_example, clone.call = "aa", sequence = "CVVSDNTGGFKTIF_CASSVRRERANTGELFF")
  
  # Check that output is a Seurat object and has the new column
  expect_s4_class(sc_highlighted, "Seurat")
  expect_true("highlight" %in% names(sc_highlighted@meta.data))
  
  # Verify the contents of the 'highlight' column
  highlight_col <- sc_highlighted$highlight
  cloneB_indices <- which(scRep_example$CTaa == "CVVSDNTGGFKTIF_CASSVRRERANTGELFF")
  other_indices <- which(scRep_example$CTaa != "CVVSDNTGGFKTIF_CASSVRRERANTGELFF")
  
  expect_equal(unname(highlight_col[cloneB_indices]), rep("CVVSDNTGGFKTIF_CASSVRRERANTGELFF", 11))
})

# test script for addVariable.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))

test_that("Function correctly adds a new variable column", {
  variable_values <- rep(c("B_cell", "L_cell"), 4)
  
  combined_new <- addVariable(combined, 
                              variable.name = "Type", 
                              variables = variable_values)
  
  expect_type(combined_new, "list")
  expect_length(combined_new, length(combined))
  
  for (i in seq_along(combined_new)) {
    df <- combined_new[[i]]
    expect_true("Type" %in% names(df))
    expect_equal(unique(df$Type), variable_values[i])
  }
})

test_that("Function throws an error with mismatched lengths", {
  mismatched_variables <- c("A", "B", "C")
  error_message <- "Make sure the variables match the length of the contig list"
  expect_error(addVariable(combined, 
                           variable.name = "Mismatch", 
                           variables = mismatched_variables),
               error_message, fixed = TRUE)
})

test_that("Function can handle different data types", {
  numeric_vars <- 1:length(combined)
  combined_numeric <- addVariable(combined, 
                                  variable.name = "Batch", 
                                  variables = numeric_vars)
  expect_true("Batch" %in% names(combined_numeric[[1]]))
  expect_type(combined_numeric[[1]]$Batch, "integer")
  expect_equal(unique(combined_numeric[[3]]$Batch), 3)
  factor_vars <- factor(rep(c("Group1", "Group2"), 4))
  combined_factor <- addVariable(combined, 
                                 variable.name = "Group", 
                                 variables = factor_vars)
  
  expect_true("Group" %in% names(combined_factor[[1]]))
  expect_s3_class(combined_factor[[1]]$Group, "factor")
})

test_that("Function can overwrite an existing column", {
  combined_initial <- addVariable(combined, 
                                  variable.name = "Status", 
                                  variables = rep("Original", length(combined)))
  expect_equal(unique(combined_initial[[1]]$Status), "Original")
  combined_overwritten <- addVariable(combined_initial, 
                                      variable.name = "Status", 
                                      variables = rep("Updated", length(combined)))
  for (df in combined_overwritten) {
    expect_equal(unique(df$Status), "Updated")
  }
})

test_that("Function works with a single-element list", {
  single_element_list <- list(P17B = combined$P17B)
  single_variable <- "SingleSample"
  result <- addVariable(single_element_list, 
                        variable.name = "Experiment", 
                        variables = single_variable)
  
  expect_length(result, 1)
  expect_true("Experiment" %in% names(result$P17B))
  expect_equal(unique(result$P17B$Experiment), "SingleSample")
})

# test script for clonalBias.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list, samples = c("P17B", "P17L", "P18B", "P18L"))
scRep_example <- suppressMessages(combineExpression(combined, scRep_example, clone.call = "aa"))
scRep_example$Patient <- substring(scRep_example$orig.ident, 1, 3)
scRep_example$Type <- substring(scRep_example$orig.ident, 4, 4)

test_that("Output formats (table vs. plot) are correct", {
  table_output <- clonalBias(scRep_example, clone.call = "aa", split.by = "Patient", 
                             group.by = "seurat_clusters", n.boots = 2, min.expand = 2, 
                             export.table = TRUE)
  expect_s3_class(table_output, "data.frame")
  plot_output <- clonalBias(scRep_example, clone.call = "aa", split.by = "Patient", 
                            group.by = "seurat_clusters", n.boots = 2, min.expand = 2)
  expect_s3_class(plot_output, "ggplot")
})

test_that("Data frame output has correct structure", {
  table_output <- clonalBias(scRep_example, clone.call = "aa", split.by = "Patient", 
                             group.by = "seurat_clusters", n.boots = 2, min.expand = 2, 
                             export.table = TRUE)
  
  expected_cols <- c("Sample", "Clone", "ncells", "Top_state", "bias", "Z.score", "cloneSize")
  expect_true(all(expected_cols %in% names(table_output)))
  expect_true(is.numeric(table_output$ncells))
  expect_true(is.numeric(table_output$bias))
  expect_true(is.numeric(table_output$Z.score))
})

test_that("`min.expand` parameter filters correctly", {
  table_min2 <- clonalBias(scRep_example, clone.call = "aa", split.by = "Patient", 
                           group.by = "seurat_clusters", n.boots = 2, min.expand = 2, 
                           export.table = TRUE)
  expect_true(all(table_min2$ncells >= 2))
  table_min1 <- clonalBias(scRep_example, clone.call = "aa", split.by = "Patient", 
                           group.by = "seurat_clusters", n.boots = 2, min.expand = 1, 
                           export.table = TRUE)
  expect_gt(nrow(table_min1), nrow(table_min2))
})

test_that("Plot structure is correct", {
  plot_obj <- clonalBias(scRep_example, clone.call = "aa", split.by = "Patient", 
                         group.by = "seurat_clusters", n.boots = 2, min.expand = 2)
  expect_s3_class(plot_obj$layers[[1]]$geom, "GeomPoint")
  expect_s3_class(plot_obj$layers[[2]]$stat, "StatQuantile")
  expect_equal(plot_obj$labels$x, "Clonal Size")
  expect_equal(plot_obj$labels$y, "Clonal Bias")
  expect_equal(plot_obj$labels$fill, "Group")
  expect_equal(plot_obj$labels$size, "cloneSize")
})

# test script for clonalCompare.R - testcases are NOT comprehensive!


# Data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined, 
                        variable.name = "Type", 
                        variables = rep(c("B", "L"), 4))


test_that("export.table = TRUE returns a data.frame", {
  table_output <- clonalCompare(combined, top.clones = 5, export.table = TRUE)
  expect_s3_class(table_output, "data.frame")
  expect_true(all(c("clones", "Proportion", "Sample") %in% names(table_output)))
})

test_that("Default behavior returns a ggplot object (alluvial)", {
  plot_output <- clonalCompare(combined, top.clones = 5)
  expect_s3_class(plot_output, "ggplot")
  geoms <- sapply(plot_output$layers, function(l) class(l$geom)[1])
  expect_true("GeomStratum" %in% geoms)
  expect_true("GeomFlow" %in% geoms)
})

test_that("graph = 'area' produces an area plot", {
  plot_output <- clonalCompare(combined, top.clones = 5, graph = "area")
  expect_s3_class(plot_output, "ggplot")
  geoms <- sapply(plot_output$layers, function(l) class(l$geom)[1])
  expect_true("GeomArea" %in% geoms)
})

test_that("proportion = FALSE uses 'Count' and proportion = TRUE uses 'Proportion'", {
  # Test with proportion = FALSE (raw counts)
  table_count <- clonalCompare(combined, top.clones = 5, proportion = FALSE, export.table = TRUE)
  expect_true("Count" %in% names(table_count))
  expect_false("Proportion" %in% names(table_count))
  expect_true(all(table_count$Count >= 1))
  expect_type(table_count$Count, "integer")
  
  # Test with proportion = TRUE (relative proportions)
  table_prop <- clonalCompare(combined, top.clones = 5, proportion = TRUE, export.table = TRUE)
  expect_true("Proportion" %in% names(table_prop))
  expect_false("Count" %in% names(table_prop))
  expect_true(all(table_prop$Proportion > 0 & table_prop$Proportion <= 1))
})

test_that("Filtering by `samples`, `clones`, and `top.clones` works", {
  # Get a list of top clones to use for the `clones` parameter test
  all_clones <- clonalCompare(combined, export.table = TRUE, top.clones = 2)
  clones_to_filter <- unique(all_clones$clones)
  
  # Test filtering by specific clones
  table_clones <- clonalCompare(combined, clones = as.character(clones_to_filter), export.table = TRUE)
  expect_true(all(unique(table_clones$clones) %in% clones_to_filter))
  
  # Test filtering by specific samples
  samples_to_filter <- c("P17B", "P18L")
  table_samples <- clonalCompare(combined, top.clones = 5, samples = samples_to_filter, export.table = TRUE)
  expect_true(all(unique(table_samples$Sample) %in% samples_to_filter))
  
  # Test that `clones` takes precedence over `top.clones`
  table_precedence <- clonalCompare(combined, clones = as.character(clones_to_filter), top.clones = 10, export.table = TRUE)
  expect_true(all(unique(table_precedence$clones) %in% clones_to_filter))
  expect_length(unique(table_precedence$clones), length(clones_to_filter))
  
  # Test error when filtering results in no clones
  expect_error(clonalCompare(combined, clones = c("nonexistent_clone_1", "nonexistent_clone_2")),
               regexp = "Please reasses the filtering strategies here")
  
})

test_that("group.by correctly groups the data", {
  table_grouped <- clonalCompare(combined, top.clones = 3, group.by = "Type", export.table = TRUE)
  expect_s3_class(table_grouped, "data.frame")
  expect_true(all(unique(table_grouped$Sample) %in% c("B", "L")))
})

test_that("relabel.clones = TRUE renames clones numerically", {
  table_relabel <- clonalCompare(combined, top.clones = 4, relabel.clones = TRUE, export.table = TRUE)
  expect_true(all(grepl("^Clone: \\d+$", unique(table_relabel$clones))))
  expect_true("original.clones" %in% names(table_relabel))
})

test_that("highlight.clones modifies colors correctly", {
  # Get top clones to use for highlighting
  clones_to_highlight <- clonalCompare(combined, top.clones = 2, export.table = TRUE)$clones
  plot_highlight <- clonalCompare(combined, top.clones = 10, highlight.clones = clones_to_highlight)
  
  # Build the plot to inspect its components
  plot_build <- ggplot_build(plot_highlight)
  plot_data <- plot_build$data[[1]] # First layer data
  
  # Check the number of unique fill colors
  unique_colors <- unique(plot_data$fill)
  expect_length(unique_colors, length(unique(clones_to_highlight)) + 1)
  expect_true("grey" %in% unique_colors)
})

# test script for positionalProperty.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list, 
                       samples = c("P17B", "P17L", "P18B", "P18L", 
                                   "P19B","P19L", "P20B", "P20L"))

test_that("positionalProperty: Output structure is correct", {
  
  # Test with export.table = TRUE
  df_output <- positionalProperty(combined, 
                                  method = "kideraFactors", 
                                  export.table = TRUE)
  
  expect_s3_class(df_output, "data.frame")
  
  expected_cols <- c("property", "position", "mean", "sd", "n", "ci_lower", "ci_upper", "group")
  expect_true(all(expected_cols %in% names(df_output)))
  
  # Test that it returns a ggplot object by default
  plot_output <- positionalProperty(combined, 
                                    method = "atchleyFactors")
  
  expect_s3_class(plot_output, "ggplot")
})

test_that("positionalProperty: Input validation works", {
  
  # Error on invalid method
  expect_error(
    positionalProperty(combined, 
                       method = "InvalidMethod"),
    "Please select a compatible method"
  )
  
})

test_that("positionalProperty: Core calculations are plausible", {
  
  df_output <- positionalProperty(combined, 
                                  chain = "TRB",
                                  aa.length = 15, 
                                  export.table = TRUE)
  
  p17_b_subset <- df_output[df_output$group == "P17B", ]
  expect_true(all(p17_b_subset$n <= 2851)) 
})

test_that("positionalProperty: Parameters behave as expected", {
  
  # `aa.length` parameter
  df_len10 <- positionalProperty(combined, 
                                 aa.length = 10, 
                                 export.table = TRUE)
  expect_equal(max(df_len10$position), 10)
  
  df_len25 <- positionalProperty(combined, 
                                 aa.length = 25, 
                                 export.table = TRUE)
  expect_equal(max(df_len25$position), 25)
})

test_that("positionalProperty: ggplot object is correctly formed", {
  
  plot_output <- positionalProperty(combined, 
                                    method = "kideraFactors")
  
  # Check layers
  expect_true("GeomRibbon" %in% sapply(plot_output$layers, function(x) class(x$geom)[1]))
  expect_true("GeomLine" %in% unlist(sapply(plot_output$layers, function(x) class(x$geom))))
  
  # Check aesthetics
  expect_equal(rlang::as_name(plot_output$mapping$y), "mean")
  expect_equal(rlang::as_name(plot_output$mapping$group), "group")
})

# test script for combineContigs.R - testcases are NOT comprehensive!

test_that("combineTCR works with default parameters", {
  combined <- combineTCR(contig_list[1:2], samples = c("P17B", "P17L"))
  expect_type(combined, "list")
  expect_length(combined, 2)
  expect_s3_class(combined[[1]], "data.frame")
  # Check if barcodes are prefixed
  expect_true(startsWith(combined[[1]]$barcode[1], "P17B_"))
})

test_that("combineTCR `samples` and `ID` parameters work", {
  combined <- combineTCR(contig_list[1], samples = "S1", ID = "A")
  expect_equal(names(combined)[1], "S1_A")
  expect_true(startsWith(combined[[1]]$barcode[1], "S1_A_"))
})

test_that("combineTCR `filter.nonproductive = FALSE` keeps non-productive chains", {
  contig_mock <- contig_list[[1]]
  contig_mock$productive[1:50] <- "False"
  combined_filtered <- combineTCR(list(contig_mock), samples="S1")
  combined_unfiltered <- combineTCR(list(contig_mock), samples="S1", filter.nonproductive = FALSE)
  expect_lt(nrow(combined_filtered[[1]]), nrow(combined_unfiltered[[1]]))
})

test_that("combineTCR `remove.na` and `remove.multi` work", {
  contig_mock <- contig_list[[1]]
  combined_remove.na <- combineTCR(list(contig_mock), samples="S1", remove.na = TRUE)[[1]]
  expect_true(all(!grepl("NA_", combined_remove.na$CTaa)))
  expect_true(all(!grepl("_NA", combined_remove.na$CTnt)))
  
  combined_remove.multi <- combineTCR(list(contig_mock), samples="S1", remove.multi = TRUE)
  expect_true(all(!grepl(";", combined_remove.multi$CTaa)))
  expect_true(all(!grepl(";", combined_remove.multi$CTnt)))
})

# --- combineBCR testing -------------------------------------------------------

# test-combineBCR.R
# Comprehensive tests for combineBCR function

# Setup: Load test data
BCR_SOURCE <- read.csv("https://www.borch.dev/uploads/contigs/b_contigs.csv")
BCR_LIST <- list(P1 = BCR_SOURCE,
                 P2 = BCR_SOURCE)
BCR_LIST$P2$barcode <- paste0(BCR_LIST$P2$barcode, "_2")

# =============================================================================
# BASIC FUNCTIONALITY TESTS
# =============================================================================

test_that("Standard combineBCR functionality (Legacy & Basic)", {
  combined_bcr <- combineBCR(BCR_SOURCE, samples = "Patient1")
  
  expect_true(any(grepl("cluster", combined_bcr[[1]]$CTstrict)))
  expect_type(combined_bcr, "list")
  expect_length(combined_bcr, 1)
  expect_s3_class(combined_bcr[[1]], "data.frame")
  expect_true(startsWith(combined_bcr[[1]]$barcode[1], "Patient1_"))
  expect_true(all(c("cdr3_aa1", "cdr3_nt1", "CTgene", "CTnt") %in% colnames(combined_bcr[[1]])))
})

test_that("combineBCR with multiple samples", {
  combined_multi <- combineBCR(BCR_LIST, samples = c("P1", "P2"))
  
  expect_type(combined_multi, "list")
  expect_length(combined_multi, 2)
  expect_equal(names(combined_multi), c("P1", "P2"))
  expect_true(all(startsWith(combined_multi[["P1"]]$barcode, "P1_")))
  expect_true(all(startsWith(combined_multi[["P2"]]$barcode, "P2_")))
})

test_that("Output Structure and samples/ID handling", {
  # Test with Sample + ID
  combined_id <- combineBCR(BCR_LIST[1], 
                            samples = "P1", 
                            ID = "Timepoint1")
  
  first_barcode <- combined_id[[1]]$barcode[1]
  # Format should be: Sample_ID_Barcode
  expect_true(startsWith(first_barcode, "P1_Timepoint1_"))
  
  # Check column existence for specific BCR chains
  cols <- colnames(combined_id[[1]])
  expect_true(all(c("IGH", "IGLC") %in% cols))
})

test_that("combineBCR without samples parameter", {
  combined_no_samples <- combineBCR(BCR_LIST)
  
  expect_type(combined_no_samples, "list")
  expect_length(combined_no_samples, 2)
  # Should use default naming S1, S2, etc.
  expect_equal(names(combined_no_samples), c("S1", "S2"))
})

# =============================================================================
# CLUSTERING METHOD TESTS
# =============================================================================

test_that("combineBCR with Alignment Metrics", {
  # Test Needleman-Wunsch (Global Alignment)
  combined_nw <- combineBCR(BCR_LIST[1], 
                            samples = "Patient1",
                            dist.type = "nw", 
                            dist.mat = "BLOSUM62",
                            threshold = 0.85,
                            normalize = "length")
  
  expect_true("CTstrict" %in% colnames(combined_nw[[1]]))
  expect_true(any(grepl("^cluster", combined_nw[[1]]$CTstrict)))
  
  # Test Smith-Waterman (Local Alignment)
  combined_sw <- combineBCR(BCR_LIST[1], 
                            samples = "Patient1",
                            dist.type = "sw", 
                            dist.mat = "PAM30",
                            threshold = 2,
                            normalize = "none")
  
  expect_true("CTstrict" %in% colnames(combined_sw[[1]]))
})

test_that("combineBCR with different distance metrics", {
  # Levenshtein (default)
  combined_lev <- combineBCR(BCR_LIST[1], 
                             samples = "Patient1",
                             dist.type = "levenshtein")
  expect_true("CTstrict" %in% colnames(combined_lev[[1]]))
  
  # Damerau-Levenshtein
  combined_dam <- combineBCR(BCR_LIST[1], 
                             samples = "Patient1",
                             dist.type = "damerau")
  expect_true("CTstrict" %in% colnames(combined_dam[[1]]))
})

test_that("Clustering Logic: call.related.clones = FALSE", {
  combined_exact <- combineBCR(BCR_LIST[1], 
                               samples = "Patient1", 
                               call.related.clones = FALSE)
  
  sample_ct <- combined_exact[[1]]$CTstrict[1]
  expect_false(grepl("cluster", sample_ct))
  expect_true(grepl("_", sample_ct)) 
})

# =============================================================================
# CTstrict FORMAT TESTS
# =============================================================================

test_that("CTstrict format: Heavy_Light underscore structure", {
  combined_bcr <- combineBCR(BCR_SOURCE, samples = "Patient1")
  
  # All CTstrict values should have exactly one underscore separating heavy and light
  ctstrict_values <- na.omit(combined_bcr[[1]]$CTstrict)
  
  # Count underscores in each value
  underscore_counts <- sapply(ctstrict_values, function(x) {
    length(gregexpr("_", x, fixed = TRUE)[[1]])
  })
  
  # Each CTstrict should have at least one underscore (Heavy_Light format)
  # Note: sequences may contain underscores, so we check for at least one
  expect_true(all(underscore_counts >= 1))
  
  # Each CTstrict should be splittable into exactly 2 parts by the first underscore
  # This validates the Heavy_Light structure
  for (val in ctstrict_values[1:min(10, length(ctstrict_values))]) {
    parts <- strsplit(val, "_", fixed = TRUE)[[1]]
    expect_gte(length(parts), 2)
  }
})

test_that("CTstrict format: Clustered vs Singlet cells", {
  combined_bcr <- combineBCR(BCR_SOURCE, samples = "Patient1")
  
  ctstrict_values <- combined_bcr[[1]]$CTstrict
  
  # Clustered cells should have cluster.X prefix on the heavy chain side (chain="IGH" default)
  clustered_pattern <- "^cluster\\.\\d+_"
  clustered_cells <- grep(clustered_pattern, ctstrict_values, value = TRUE)
  expect_true(length(clustered_cells) > 0)
  
  # Singlets should NOT have cluster pattern - they use vgene.jgene.seq format
  non_cluster <- ctstrict_values[!grepl("^cluster\\.", ctstrict_values) & !is.na(ctstrict_values)]
  
  if (length(non_cluster) > 0) {
    # Singlets should contain gene names (e.g., IGHV, IGKV, IGLV) or NA
    has_gene_or_na <- sapply(non_cluster, function(x) {
      grepl("IG[HKL]", x) | grepl("^NA_", x) | grepl("_NA$", x)
    })
    expect_true(all(has_gene_or_na))
    
    # Should still have underscore separator
    expect_true(all(grepl("_", non_cluster)))
  }
})

test_that("CTstrict: Clustered cells share same cluster ID on both sides (chain='both')", {
  combined_bcr <- combineBCR(BCR_SOURCE, 
                             samples = "Patient1",
                             chain = "both")
  
  clustered <- grep("^cluster\\.\\d+_cluster\\.\\d+$", 
                    combined_bcr[[1]]$CTstrict, 
                    value = TRUE)
  
  if (length(clustered) > 0) {
    # For chain="both", both sides should have the SAME cluster ID
    for (ct in clustered) {
      parts <- strsplit(ct, "_")[[1]]
      expect_equal(parts[1], parts[2], 
                   info = paste("Cluster IDs should match for chain='both':", ct))
    }
  }
})

test_that("CTstrict format: Single chain clustering (IGH only)", {
  combined_igh <- combineBCR(BCR_SOURCE, 
                             samples = "Patient1",
                             chain = "Heavy")
  
  ctstrict_values <- na.omit(combined_igh[[1]]$CTstrict)
  
  # With chain = "IGH", clustered cells should have:
  # cluster.X on heavy side, unique ID on light side
  # Format: cluster.X_IGKV... or cluster.X_IGLV... or cluster.X_NA
  clustered_igh <- grep("^cluster\\.\\d+_", ctstrict_values, value = TRUE)
  
  if (length(clustered_igh) > 0) {
    # Light chain side should NOT be cluster (should be vgene.jgene.seq or NA)
    expect_false(any(grepl("_cluster\\.", clustered_igh)),
                 info = "Light chain side should not be cluster when chain='IGH'")
    
    # Light chain side should have light chain gene or NA
    light_side_valid <- sapply(clustered_igh, function(x) {
      light_part <- strsplit(x, "_")[[1]][2]
      grepl("^IG[KL]", light_part) | light_part == "NA"
    })
    expect_true(all(light_side_valid))
  }
})

test_that("CTstrict format: Single chain clustering (Light only)", {
  combined_light <- combineBCR(BCR_SOURCE, 
                               samples = "Patient1",
                               chain = "Light")
  
  ctstrict_values <- na.omit(combined_light[[1]]$CTstrict)
  
  # With chain = "Light", clustered cells should have:
  # unique ID on heavy side, cluster.X on light side
  # Format: IGHV..._cluster.X or NA_cluster.X
  clustered_light <- grep("_cluster\\.\\d+$", ctstrict_values, value = TRUE)
  
  if (length(clustered_light) > 0) {
    # Heavy chain side should NOT be cluster
    expect_false(any(grepl("^cluster\\.", clustered_light)),
                 info = "Heavy chain side should not be cluster when chain='Light'")
    
    # Heavy chain side should have heavy chain gene or NA
    heavy_side_valid <- sapply(clustered_light, function(x) {
      heavy_part <- strsplit(x, "_")[[1]][1]
      grepl("^IGHV", heavy_part) | heavy_part == "NA"
    })
    expect_true(all(heavy_side_valid))
  }
})

test_that("CTstrict format: Single chain clustering (IGL only)", {
  combined_igl <- combineBCR(BCR_SOURCE, 
                             samples = "Patient1",
                             chain = "IGL")
  
  ctstrict_values <- na.omit(combined_igl[[1]]$CTstrict)
  
  # Similar to Light chain test
  clustered_igl <- grep("_cluster\\.\\d+$", ctstrict_values, value = TRUE)
  
  if (length(clustered_igl) > 0) {
    # Heavy chain side should NOT be cluster
    expect_false(any(grepl("^cluster\\.", clustered_igl)))
  }
})

test_that("CTstrict format: Single chain clustering (IGK only)", {
  combined_igk <- combineBCR(BCR_SOURCE, 
                             samples = "Patient1",
                             chain = "IGK")
  
  ctstrict_values <- na.omit(combined_igk[[1]]$CTstrict)
  
  # Similar to Light chain test
  clustered_igk <- grep("_cluster\\.\\d+$", ctstrict_values, value = TRUE)
  
  if (length(clustered_igk) > 0) {
    # Heavy chain side should NOT be cluster
    expect_false(any(grepl("^cluster\\.", clustered_igk)))
  }
})

test_that("CTstrict format: Cells with missing heavy chain",
          {
            combined_bcr <- combineBCR(BCR_SOURCE, samples = "Patient1")
            
            df <- combined_bcr[[1]]
            
            # Find cells with only light chain (heavy chain is NA)
            light_only <- df[is.na(df$IGH) & !is.na(df$IGLC), ]
            
            if (nrow(light_only) > 0) {
              # CTstrict should have NA on the heavy side (or be fully NA)
              ctstrict_light_only <- light_only$CTstrict[!is.na(light_only$CTstrict)]
              
              if (length(ctstrict_light_only) > 0) {
                # Should start with NA_ or be cluster-based
                valid_format <- sapply(ctstrict_light_only, function(x) {
                  grepl("^NA_", x) | grepl("^cluster\\.", x)
                })
                expect_true(all(valid_format),
                            info = "Light-only cells should have 'NA_' prefix or be clustered")
              }
            }
          })

test_that("CTstrict format: Cells with missing light chain", {
  combined_bcr <- combineBCR(BCR_SOURCE, samples = "Patient1")
  
  df <- combined_bcr[[1]]
  
  # Find cells with only heavy chain (light chain is NA)
  heavy_only <- df[!is.na(df$IGH) & is.na(df$IGLC), ]
  
  if (nrow(heavy_only) > 0) {
    # CTstrict should have NA on the light side (or be fully NA)
    ctstrict_heavy_only <- heavy_only$CTstrict[!is.na(heavy_only$CTstrict)]
    
    if (length(ctstrict_heavy_only) > 0) {
      # Should end with _NA or be cluster-based
      valid_format <- sapply(ctstrict_heavy_only, function(x) {
        grepl("_NA$", x) | grepl("_cluster\\.", x)
      })
      expect_true(all(valid_format),
                  info = "Heavy-only cells should have '_NA' suffix or be clustered")
    }
  }
})

test_that("CTstrict format: Consistency with sequence parameter (aa vs nt)", {
  # Test with amino acid sequences
  combined_aa <- combineBCR(BCR_SOURCE, 
                            samples = "Patient1",
                            sequence = "aa",
                            threshold = 0.8)
  
  # Test with nucleotide sequences
  combined_nt <- combineBCR(BCR_SOURCE, 
                            samples = "Patient1",
                            sequence = "nt",
                            threshold = 0.8)
  
  # Get singlets from each (non-cluster entries)
  singlets_aa <- combined_aa[[1]]$CTstrict[!grepl("^cluster\\.", combined_aa[[1]]$CTstrict) & 
                                             !is.na(combined_aa[[1]]$CTstrict) &
                                             !grepl("^NA_NA$", combined_aa[[1]]$CTstrict)]
  
  singlets_nt <- combined_nt[[1]]$CTstrict[!grepl("^cluster\\.", combined_nt[[1]]$CTstrict) & 
                                             !is.na(combined_nt[[1]]$CTstrict) &
                                             !grepl("^NA_NA$", combined_nt[[1]]$CTstrict)]
  
  if (length(singlets_aa) > 0 && length(singlets_nt) > 0) {
    # NT sequences should be longer than AA sequences (roughly 3x for coding)
    avg_len_aa <- mean(nchar(singlets_aa))
    avg_len_nt <- mean(nchar(singlets_nt))
    
    expect_gt(avg_len_nt, avg_len_aa)
  }
})

# =============================================================================
# FILTERING PARAMETER TESTS
# =============================================================================

test_that("combineBCR remove.na parameter", {
  combined_with_na <- combineBCR(BCR_SOURCE, 
                                 samples = "Patient1",
                                 remove.na = FALSE)
  
  combined_no_na <- combineBCR(BCR_SOURCE, 
                               samples = "Patient1",
                               remove.na = TRUE)
  
  # remove.na=TRUE should have fewer or equal rows
  expect_lte(nrow(combined_no_na[[1]]), nrow(combined_with_na[[1]]))
})

test_that("combineBCR remove.multi parameter", {
  combined_with_multi <- combineBCR(BCR_SOURCE, 
                                    samples = "Patient1",
                                    remove.multi = FALSE)
  
  combined_no_multi <- combineBCR(BCR_SOURCE, 
                                  samples = "Patient1",
                                  remove.multi = TRUE)
  
  # remove.multi=TRUE should have fewer or equal rows
  expect_lte(nrow(combined_no_multi[[1]]), nrow(combined_with_multi[[1]]))
})

test_that("combineBCR filter.multi parameter", {
  combined_filter <- combineBCR(BCR_SOURCE, 
                                samples = "Patient1",
                                filter.multi = TRUE)
  
  combined_no_filter <- combineBCR(BCR_SOURCE, 
                                   samples = "Patient1",
                                   filter.multi = FALSE)
  
  # Both should produce valid output
  expect_s3_class(combined_filter[[1]], "data.frame")
  expect_s3_class(combined_no_filter[[1]], "data.frame")
})

test_that("combineBCR filter.nonproductive parameter", {
  combined_productive <- combineBCR(BCR_SOURCE, 
                                    samples = "Patient1",
                                    filter.nonproductive = TRUE)
  
  combined_all <- combineBCR(BCR_SOURCE, 
                             samples = "Patient1",
                             filter.nonproductive = FALSE)
  
  # filter.nonproductive=TRUE should have fewer or equal rows
  expect_lte(nrow(combined_productive[[1]]), nrow(combined_all[[1]]))
})

# =============================================================================
# V/J GENE USAGE TESTS
# =============================================================================

test_that("combineBCR use.V parameter affects clustering", {
  combined_use_v <- combineBCR(BCR_SOURCE, 
                               samples = "Patient1",
                               use.V = TRUE)
  
  combined_no_v <- combineBCR(BCR_SOURCE, 
                              samples = "Patient1",
                              use.V = FALSE)
  
  # Both should produce valid output with CTstrict
  expect_true("CTstrict" %in% colnames(combined_use_v[[1]]))
  expect_true("CTstrict" %in% colnames(combined_no_v[[1]]))
  
  # Clustering results may differ
  # When use.V=FALSE, more sequences may cluster together
  clustered_use_v <- sum(grepl("^cluster\\.", combined_use_v[[1]]$CTstrict), na.rm = TRUE)
  clustered_no_v <- sum(grepl("^cluster\\.", combined_no_v[[1]]$CTstrict), na.rm = TRUE)
  
  # Just verify both produce some clusters
  expect_true(clustered_use_v > 0 || clustered_no_v > 0)
})

test_that("combineBCR use.J parameter affects clustering", {
  combined_use_j <- combineBCR(BCR_SOURCE, 
                               samples = "Patient1",
                               use.J = TRUE)
  
  combined_no_j <- combineBCR(BCR_SOURCE, 
                              samples = "Patient1",
                              use.J = FALSE)
  
  # Both should produce valid output with CTstrict
  expect_true("CTstrict" %in% colnames(combined_use_j[[1]]))
  expect_true("CTstrict" %in% colnames(combined_no_j[[1]]))
})

test_that("combineBCR use.V and use.J together", {
  combined_both <- combineBCR(BCR_SOURCE, 
                              samples = "Patient1",
                              use.V = TRUE,
                              use.J = TRUE)
  
  combined_neither <- combineBCR(BCR_SOURCE, 
                                 samples = "Patient1",
                                 use.V = FALSE,
                                 use.J = FALSE)
  
  # When neither V nor J is used, more sequences may cluster
  clustered_both <- sum(grepl("^cluster\\.", combined_both[[1]]$CTstrict), na.rm = TRUE)
  clustered_neither <- sum(grepl("^cluster\\.", combined_neither[[1]]$CTstrict), na.rm = TRUE)
  
  # More permissive clustering (no V/J requirement) should produce >= clusters
  expect_gte(clustered_neither, clustered_both)
})

# =============================================================================
# GROUP.BY PARAMETER TESTS
# =============================================================================

test_that("combineBCR group.by parameter", {
  # Add a grouping column to the data
  bcr_with_group <- BCR_SOURCE
  bcr_with_group$patient_group <- sample(c("GroupA", "GroupB"), 
                                         nrow(bcr_with_group), 
                                         replace = TRUE)
  
  combined_grouped <- combineBCR(bcr_with_group, 
                                 samples = "Patient1",
                                 group.by = "patient_group")
  
  # Should produce valid output
  expect_s3_class(combined_grouped[[1]], "data.frame")
  expect_true("CTstrict" %in% colnames(combined_grouped[[1]]))
})

# =============================================================================
# EDGE CASES AND ERROR HANDLING
# =============================================================================


test_that("combineBCR CTstrict values are unique for different clones", {
  combined_bcr <- combineBCR(BCR_SOURCE, 
                             samples = "Patient1",
                             call.related.clones = FALSE)
  
  ctstrict_values <- combined_bcr[[1]]$CTstrict
  
  # Check that duplicated CTstrict values have matching chain information
  dup_ctstrict <- ctstrict_values[duplicated(ctstrict_values) & !is.na(ctstrict_values)]
  
  if (length(dup_ctstrict) > 0) {
    for (dup in unique(dup_ctstrict)) {
      matching_rows <- combined_bcr[[1]][combined_bcr[[1]]$CTstrict == dup & 
                                           !is.na(combined_bcr[[1]]$CTstrict), ]
      # All rows with same CTstrict should have same CDR3 sequences
      expect_equal(length(unique(matching_rows$cdr3_aa1)), 1)
      expect_equal(length(unique(matching_rows$cdr3_aa2)), 1)
    }
  }
})

test_that("combineBCR produces consistent output across runs", {
  set.seed(42)
  combined_run1 <- combineBCR(BCR_SOURCE, samples = "Patient1")
  
  set.seed(42)
  combined_run2 <- combineBCR(BCR_SOURCE, samples = "Patient1")
  
  # Results should be identical with same seed
  expect_equal(nrow(combined_run1[[1]]), nrow(combined_run2[[1]]))
  expect_equal(combined_run1[[1]]$barcode, combined_run2[[1]]$barcode)
})

# test script for clonalQuant.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))
combined <- addVariable(combined,
                        variable.name = "Type",
                        variables = rep(c("B", "L"), 4))

test_that("Input validation for `group.by` works", {
  error_message <- "Only one item in the group.by variable can be listed."
  expect_error(clonalQuant(combined, group.by = c("Type", "Patient")),
               error_message, fixed = TRUE)
})

test_that("Output formats (table vs. plot) are correct", {
  # Test that export.table = TRUE returns a data.frame
  table_output <- clonalQuant(combined, export.table = TRUE)
  expect_s3_class(table_output, "data.frame")
  
  # Test that the default behavior returns a ggplot object
  plot_output <- clonalQuant(combined)
  expect_s3_class(plot_output, "ggplot")
})

test_that("`scale` parameter correctly calculates values", {
  table_raw <- clonalQuant(combined, export.table = TRUE, scale = FALSE)
  expect_true(all(c("contigs", "values", "total") %in% names(table_raw)))
  expect_true(is.numeric(table_raw$contigs))
  expect_false("scaled" %in% names(table_raw))
  table_scaled <- clonalQuant(combined, export.table = TRUE, scale = TRUE)
  expect_true("scaled" %in% names(table_scaled))
  expect_equal(table_scaled$scaled, (table_scaled$contigs / table_scaled$total) * 100)
  expect_true(all(table_scaled$scaled >= 0 & table_scaled$scaled <= 100))
})

test_that("Data frame contents are correct", {
  table_output <- clonalQuant(combined, clone.call = "aa", export.table = TRUE)
  expect_equal(nrow(table_output), length(combined))
  first_sample_data <- combined[[1]]
  manual_unique_count <- length(na.omit(unique(first_sample_data$CTaa)))
  manual_total_count <- length(na.omit(first_sample_data$CTaa))
  expect_equal(table_output$contigs[1], manual_unique_count)
  expect_equal(table_output$total[1], manual_total_count)
})

test_that("`group.by` parameter functions correctly", {
  table_grouped <- clonalQuant(combined, group.by = "Type", export.table = TRUE)
  expect_equal(nrow(table_grouped), 2) # Should have 2 groups: "B" and "L"
  expect_true("Type" %in% names(table_grouped))
  grouped_list <- scRepertoire:::.groupList(combined, "Type")
  b_group_data <- grouped_list$B
  manual_unique_b <- length(na.omit(unique(b_group_data$CTstrict)))
  manual_total_b <- length(na.omit(b_group_data$CTstrict))
  b_row <- table_grouped[table_grouped$Type == "B",]
  expect_equal(b_row$contigs, manual_unique_b)
  expect_equal(b_row$total, manual_total_b)
})

test_that("Plotting output has the correct components and labels", {
  plot_raw <- clonalQuant(combined, scale = FALSE)
  expect_s3_class(plot_raw, "ggplot")
  expect_equal(plot_raw$labels$y, "Unique Clones")
  expect_equal(plot_raw$labels$fill, "Samples")
  plot_scaled <- clonalQuant(combined, scale = TRUE)
  expect_s3_class(plot_scaled, "ggplot")
  expect_equal(plot_scaled$labels$y, "Percent of Unique Clones")
  plot_grouped <- clonalQuant(combined, group.by = "Type")
  expect_s3_class(plot_grouped, "ggplot")
  expect_equal(plot_grouped$labels$fill, "Type")
})

# test script for clonalBin.R - comprehensive test cases

# Data setup
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))

# Add a variable for grouped tests
combined_with_type <- addVariable(combined,
                                  variable.name = "Type",
                                  variables = rep(c("B", "L"), 4))

# Basic functionality tests
test_that("clonalBin returns a list with the same structure as input", {
  result <- clonalBin(combined)
  expect_type(result, "list")
  expect_equal(length(result), length(combined))
  expect_equal(names(result), names(combined))
})

test_that("clonalBin adds clonalFrequency, clonalProportion, and cloneSize columns", {
  result <- clonalBin(combined)
  expected_cols <- c("clonalFrequency", "clonalProportion", "cloneSize")
  for (i in seq_along(result)) {
    expect_true(all(expected_cols %in% colnames(result[[i]])))
  }
})

test_that("cloneSize is a factor with correct levels", {
  result <- clonalBin(combined)
  for (i in seq_along(result)) {
    expect_s3_class(result[[i]]$cloneSize, "factor")
    # Levels should be in reverse order (hyperexpanded first, None last)
    levels_vec <- levels(result[[i]]$cloneSize)
    expect_true(grepl("Hyperexpanded", levels_vec[1]))
    expect_true(grepl("None", levels_vec[length(levels_vec)]))
  }
})

# clone.call parameter tests
test_that("clone.call parameter works with 'strict' (default)", {
  result <- clonalBin(combined, clone.call = "strict")
  expect_true(all(c("CTstrict", "cloneSize") %in% colnames(result[[1]])))
})

test_that("clone.call parameter works with 'nt'", {
  result <- clonalBin(combined, clone.call = "nt")
  expect_true("CTnt" %in% colnames(result[[1]]))
})

test_that("clone.call parameter works with 'aa'", {
  result <- clonalBin(combined, clone.call = "aa")
  expect_true("CTaa" %in% colnames(result[[1]]))
})

test_that("clone.call parameter works with 'gene'", {
  result <- clonalBin(combined, clone.call = "gene")
  expect_true("CTgene" %in% colnames(result[[1]]))
})

# Proportion vs Frequency tests
test_that("proportion = TRUE uses clonalProportion for binning", {
  result <- clonalBin(combined, proportion = TRUE)
  # Check that clonalProportion values are between 0 and 1
  for (i in seq_along(result)) {
    props <- result[[i]]$clonalProportion
    props <- props[!is.na(props)]
    expect_true(all(props >= 0 & props <= 1))
  }
})

test_that("proportion = FALSE uses clonalFrequency for binning", {
  result <- clonalBin(combined,
                      proportion = FALSE,
                      clone.size = c(Rare = 1, Small = 5, Medium = 20,
                                    Large = 100, Hyperexpanded = 500))
  # Check that clonalFrequency values are integers >= 1
  for (i in seq_along(result)) {
    freqs <- result[[i]]$clonalFrequency
    freqs <- freqs[!is.na(freqs)]
    expect_true(all(freqs >= 1))
    expect_true(all(freqs == floor(freqs)))
  }
})

test_that("Error when proportion = FALSE and clone.size < 1", {
  bad_clone_size <- c(Rare = 0.1, Small = 0.5)
  expect_error(
    clonalBin(combined, proportion = FALSE, clone.size = bad_clone_size),
    "Adjust the clone.size parameter - there are groupings < 1"
  )
})

test_that("Auto-adjustment of upper bin limit when max frequency exceeds it", {
  result <- clonalBin(combined,
                      proportion = FALSE,
                      clone.size = c(Single = 1, Small = 2))
  # Find max frequency across all elements
  max_freq <- max(sapply(result, function(x) max(x$clonalFrequency, na.rm = TRUE)))
  # The highest bin should include the max frequency
  highest_level <- levels(result[[1]]$cloneSize)[1]
  # Extract the upper bound from the level name
  upper_bound <- as.numeric(gsub(".*<= ([0-9.]+)\\)", "\\1", highest_level))
  expect_gte(upper_bound, max_freq)
})

# Custom clone.size tests
test_that("Custom clone.size bins are applied correctly", {
  custom_bins <- c(Low = 0.01, Medium = 0.1, High = 1)
  result <- clonalBin(combined, clone.size = custom_bins)
  levels_vec <- levels(result[[1]]$cloneSize)
  expect_true(any(grepl("Low", levels_vec)))
  expect_true(any(grepl("Medium", levels_vec)))
  expect_true(any(grepl("High", levels_vec)))
})

# group.by parameter tests
test_that("group.by = NULL calculates per list element", {
  result <- clonalBin(combined, group.by = NULL)
  # Each list element should have its own frequency/proportion calculations
  expect_equal(length(result), length(combined))
})

test_that("group.by parameter works with variable grouping", {
  result <- clonalBin(combined_with_type, group.by = "Type")
  expect_type(result, "list")
  # Verify calculations are grouped by Type
  for (i in seq_along(result)) {
    expect_true("clonalProportion" %in% colnames(result[[i]]))
    expect_true("clonalFrequency" %in% colnames(result[[i]]))
  }
})

# chain parameter tests
test_that("chain = 'TRA' filters to TRA chain only", {
  result <- clonalBin(combined, chain = "TRA")
  expect_type(result, "list")
  expect_true("cloneSize" %in% colnames(result[[1]]))
})
test_that("chain = 'TRB' filters to TRB chain only", {
  result <- clonalBin(combined, chain = "TRB")
  expect_type(result, "list")
  expect_true("cloneSize" %in% colnames(result[[1]]))
})

test_that("chain = 'both' uses paired chain data", {
  result <- clonalBin(combined, chain = "both")
  expect_type(result, "list")
  expect_true("cloneSize" %in% colnames(result[[1]]))
})

test_that("Different chains produce different frequency calculations", {
  result_tra <- clonalBin(combined, chain = "TRA")
  result_trb <- clonalBin(combined, chain = "TRB")
  result_both <- clonalBin(combined, chain = "both")

  # The frequencies should differ between chain selections
  freq_tra <- result_tra[[1]]$clonalFrequency
  freq_trb <- result_trb[[1]]$clonalFrequency
  freq_both <- result_both[[1]]$clonalFrequency

  # They should not all be identical
  expect_false(identical(freq_tra, freq_trb))
})

# Data integrity tests
test_that("clonalBin preserves original data columns", {
  result <- clonalBin(combined)
  original_cols <- colnames(combined[[1]])
  for (col in original_cols) {
    expect_true(col %in% colnames(result[[1]]))
  }
})

test_that("clonalBin handles NA cloneCall values", {
  # Create test data with some NA values
  combined_with_na <- combined
  combined_with_na[[1]]$CTstrict[1:5] <- NA
  result <- clonalBin(combined_with_na, clone.call = "strict")
  expect_type(result, "list")
  # NAs should result in NA cloneSize
  expect_true(any(is.na(result[[1]]$cloneSize)))
})

test_that("clonalBin works with single list element input", {
  single_element <- combined[1]
  result <- clonalBin(single_element)
  expect_type(result, "list")
  expect_equal(length(result), 1)
  expect_true("cloneSize" %in% colnames(result[[1]]))
})

# Bin assignment tests
test_that("All non-NA clones are assigned to a bin",
{
  result <- clonalBin(combined)
  for (i in seq_along(result)) {
    # Rows with non-NA cloneCall should have non-NA cloneSize
    non_na_clones <- !is.na(result[[i]]$CTstrict)
    clone_sizes <- result[[i]]$cloneSize[non_na_clones]
    # The only NAs in cloneSize should be from NAs in CTstrict
    expect_true(sum(is.na(clone_sizes)) < sum(non_na_clones))
  }
})

test_that("Bin labels contain correct range format", {
  result <- clonalBin(combined)
  levels_vec <- levels(result[[1]]$cloneSize)
  # Each level should follow the pattern "Name (lower < X <= upper)"
  for (level in levels_vec) {
    expect_true(grepl("\\(.*< X <=.*\\)", level))
  }
})

# Deprecated parameter tests
test_that("Deprecated cloneCall parameter still works with warning", {
  expect_warning(
    result <- clonalBin(combined, cloneCall = "gene"),
    regexp = "cloneCall.*deprecated"
  )
  expect_true("CTgene" %in% colnames(result[[1]]))
})

test_that("Deprecated cloneSize parameter still works with warning", {
  custom_bins <- c(Low = 0.05, High = 1)
  expect_warning(
    result <- clonalBin(combined, cloneSize = custom_bins),
    regexp = "cloneSize.*deprecated"
  )
  levels_vec <- levels(result[[1]]$cloneSize)
  expect_true(any(grepl("Low", levels_vec)))
})

# Edge case tests
test_that("clonalBin handles data with all unique clones", {
  # Create a subset where all clones are unique
  small_combined <- lapply(combined, function(x) {
    unique_clones <- !duplicated(x$CTstrict) & !is.na(x$CTstrict)
    x[unique_clones, ][1:min(10, sum(unique_clones)), ]
  })
  result <- clonalBin(small_combined)
  expect_type(result, "list")
  # All should be in the smallest bin
  for (i in seq_along(result)) {
    freqs <- result[[i]]$clonalFrequency[!is.na(result[[i]]$clonalFrequency)]
    expect_true(all(freqs == 1))
  }
})

test_that("clonalBin handles data with highly expanded clones", {
  result <- clonalBin(combined,
                      proportion = FALSE,
                      clone.size = c(Single = 1, Small = 2, Medium = 3))
  # Check that clones above the specified bins are captured
  max_freqs <- sapply(result, function(x) max(x$clonalFrequency, na.rm = TRUE))
  # If any clone has frequency > 3, bin should be auto-adjusted
  if (any(max_freqs > 3)) {
    highest_level <- levels(result[[1]]$cloneSize)[1]
    upper_bound <- as.numeric(gsub(".*<= ([0-9.]+)\\)", "\\1", highest_level))
    expect_gte(upper_bound, max(max_freqs))
  }
})

# Consistency tests
test_that("Running clonalBin twice produces same results", {
  result1 <- clonalBin(combined)
  result2 <- clonalBin(combined)
  expect_equal(result1, result2)
})

test_that("clonalProportion sums to approximately 1 within each list element", {
  result <- clonalBin(combined)
  for (i in seq_along(result)) {
    # Get unique clones with their proportions
    unique_clones <- result[[i]][!duplicated(result[[i]]$CTstrict) &
                                   !is.na(result[[i]]$CTstrict), ]
    prop_sum <- sum(unique_clones$clonalProportion, na.rm = TRUE)
    expect_equal(prop_sum, 1, tolerance = 0.01)
  }
})

# testcases are NOT comprehensive

getTestTcrList <- function() {
  combined <- combineTCR(contig_list,
                         samples = c("P17B", "P17L", "P18B", "P18L",
                                     "P19B", "P19L", "P20B", "P20L"))[1:3]
}

#' Generate a [combineBCR()] output example with matching sample names
#' as the output of [getTestTcrList()] above and no doublets.
getTestBcrListNoDoublets <- function() {
    BCR <- read.csv("https://www.borch.dev/uploads/contigs/b_contigs.csv")
    bcr <- combineBCR(BCR, 
                      samples = "Patient1")
    bcr <- list(bcr[[1]][1:10, ], bcr[[1]][20:30, ], bcr[[1]][100:110, ])
    names(bcr) <- names(getTestTcrList())
    purrr::imap(bcr, function(df, sampleName) {
        df$sample <- sampleName
        df
    })
}

getTestBcrListWithDoublets <- function(doubletsPerSample, seed = 42) {
    if (!is.null(seed)) set.seed(seed)
    purrr:::map2(
        getTestBcrListNoDoublets(), getTestTcrList(),
        makeRandomBcrBarcodesMatchTcr,
        n = doubletsPerSample
    )
}

makeRandomBcrBarcodesMatchTcr <- function(bcrDf, tcrDf, n) {
    sampleUniqueBarcodeDf(bcrDf, n) %>%
        dplyr::mutate(
            tcrBarcode = sampleUniqueBarcodeDf(tcrDf, n, asDf = FALSE)
        ) %>%
        dplyr::full_join(bcrDf, by = "barcode") %>%
        dplyr::mutate(
            barcode = ifelse(is.na(tcrBarcode), barcode, tcrBarcode)
        ) %>%
        dplyr::select(-tcrBarcode)
}

sampleUniqueBarcodeDf <- function(contigDf, n, asDf = TRUE) {
    contigDf %>%
        dplyr::select(barcode) %>%
        dplyr::distinct() %>%
        dplyr::slice_sample(n = n) %>%
        (if (asDf) identity else function(x) x$barcode)
}

test_that("getContigDoublets works for no doublets", {

    expected_no_doublet_output <- structure(
        list(
            contigType = structure(
                integer(0), levels = c("BCR", "TCR"), class = "factor"
            ),
            barcode = character(0), sample = character(0),
            TCR1 = character(0), cdr3_aa1 = character(0),
            cdr3_nt1 = character(0), TCR2 = character(0),
            cdr3_aa2 = character(0), cdr3_nt2 = character(0),
            CTgene = character(0), CTnt = character(0), CTaa = character(0),
            CTstrict = character(0), IGH = character(0), IGLC = character(0)
        ),
        row.names = integer(0),
        class = "data.frame"
    )

    expect_identical(
        getContigDoublets(getTestTcrList(), getTestBcrListNoDoublets()),
        expected_no_doublet_output
    )
})

test_that("getContigDoublets works for inputs with doublets", {

    NUM_UNIQUE_DOUBLETS_PER_SAMPLE <- 3
    tcr <- getTestTcrList()
    bcr <- getTestBcrListWithDoublets(NUM_UNIQUE_DOUBLETS_PER_SAMPLE)

    doubletDf <- getContigDoublets(tcr, bcr)

    expect_equal(
        nrow(doubletDf),
        NUM_UNIQUE_DOUBLETS_PER_SAMPLE * length(tcr) * 2
    )

    expect_identical(
        colnames(doubletDf),
        c("contigType", "barcode", "sample", "TCR1", "cdr3_aa1", "cdr3_nt1",
          "TCR2", "cdr3_aa2", "cdr3_nt2", "CTgene", "CTnt", "CTaa", "CTstrict",
          "IGH", "IGLC")
    )

    getBarcodeSampleForContigType <- function(contigType) {
        doubletDf %>%
            dplyr::filter(contigType == contigType) %>%
            dplyr::select(barcode, sample) %>%
            dplyr::arrange(barcode, sample)
    }

    expect_identical(
        getBarcodeSampleForContigType("BCR"),
        getBarcodeSampleForContigType("TCR")
    )

    makeCharNaDf <- function(dfColnames, nrow) {
        matrix(nrow = nrow, ncol = length(dfColnames)) %>%
            data.frame() %>%
            (function(df) {
                colnames(df) <- dfColnames
                df
            }) %>%
            dplyr::mutate(dplyr::across(dplyr::everything(), as.character))
    }

    expect_identical(
        doubletDf %>%
            dplyr::filter(contigType == "TCR") %>%
            dplyr::select(IGH, IGLC),
        makeCharNaDf(
            c("IGH", "IGLC"), NUM_UNIQUE_DOUBLETS_PER_SAMPLE * length(tcr)
        )
    )

    expect_identical(
        doubletDf %>%
            dplyr::filter(contigType == "BCR") %>%
            dplyr::select(TCR1, TCR2),
        makeCharNaDf(
            c("TCR1", "TCR2"),
            NUM_UNIQUE_DOUBLETS_PER_SAMPLE * length(bcr)
        )
    )

})

# test script for clonalNetwork.R - testcases are NOT comprehensive!
library(ggraph)

# Data to use
combined <- combineTCR(contig_list, samples = c("P17B", "P17L", "P18B", "P18L", "P19B", "P19L", "P20B", "P20L"))
scRep_example <- combineExpression(combined, scRep_example)
scRep_example$Patient <- substring(scRep_example$orig.ident, 1, 3)
scRep_example$Type <- substring(scRep_example$orig.ident, 4, 4)
 
test_that("Output formats (table, clones, plot) are correct", {
  # Test export.table
  table_output <- clonalNetwork(scRep_example, reduction = "umap", group.by = "seurat_clusters", export.table = TRUE)
  expect_s3_class(table_output, "data.frame")
  expect_true(all(c("to", "from", "weight") %in% names(table_output)))
  
  # Test exportClones
  clones_output <- clonalNetwork(scRep_example, reduction = "umap", group.by = "seurat_clusters", export.clones = TRUE)
  expect_s3_class(clones_output, "data.frame")
  expect_true(all(c("clone", "sum") %in% names(clones_output)))
  
  # Test plot output
  plot_output <- clonalNetwork(scRep_example, reduction = "umap", group.by = "seurat_clusters")
  expect_s3_class(plot_output, "ggplot")
})

test_that("Core edge weight calculation is correct", {
  edge_list <- clonalNetwork(scRep_example, reduction = "umap", group.by = "seurat_clusters", export.table = TRUE)
  
  edge_c3_c5 <- edge_list[edge_list$from == "3" & edge_list$to == "5", ]
  edge_c5_c3 <- edge_list[edge_list$from == "5" & edge_list$to == "3", ]
  
  expect_equal(edge_c3_c5$weight, 0.118, tolerance = 1e-2)
  expect_equal(edge_c5_c3$weight, 0.1, tolerance = 1e-2)
})

test_that("Filtering parameters work as expected", {
  edge_list_prop <- clonalNetwork(scRep_example, reduction = "umap", group.by = "seurat_clusters", 
                                  filter.proportion = 0.2, export.table = TRUE)
  expect_equal(nrow(edge_list_prop), 1)
  expect_equal(edge_list_prop$weight, 0.25)
  edge_list_unfiltered <- clonalNetwork(scRep_example, reduction = "umap", group.by = "seurat_clusters", export.table = TRUE)
  edge_list_half <- clonalNetwork(scRep_example, reduction = "umap", group.by = "seurat_clusters", 
                                  filter.graph = TRUE, export.table = TRUE)
  expect_lt(nrow(edge_list_half), nrow(edge_list_unfiltered))
})

test_that("exportClones provides correctly summarized data", {
  clones_df <- clonalNetwork(scRep_example, reduction = "umap", group.by = "seurat_clusters", export.clones = TRUE)
  expect_equal(nrow(clones_df), 341)
  expect_equal(clones_df$sum[1:5], c(11,3,3,3,2))
})

test_that("Plot structure has correct geoms", {
  plot_obj <- clonalNetwork(scRep_example, reduction = "umap", group.by = "seurat_clusters")
  geoms_in_plot <- sapply(plot_obj$layers, function(l) class(l$geom)[1])
  expect_true("GeomPoint" %in% geoms_in_plot)    
  expect_true("GeomEdgePath" %in% geoms_in_plot)  
})

#' Plot Relative Amino Acid Composition by Position
#'
#' This function the proportion of amino acids along the residues 
#' of the CDR3 amino acid sequence.
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' 
#' # Using percentAA()
#' percentAA(combined, 
#'           chain = "TRB", 
#'           aa.length = 20)
  
#' @param input.data The product of [combineTCR()], [combineBCR()], or
#'  [combineExpression()].
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains 
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param group.by A column header in the metadata or lists to group the analysis 
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed 
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements 
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups 
#' automatically.
#' @param aa.length The maximum length of the CDR3 amino acid sequence.
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any [hcl.pals][grDevices::hcl.pals].
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#' @importFrom immApex calculateFrequency
#' @importFrom stats reshape
#' @export
#' @concept Summarize_Repertoire
#' @return A ggplot object visualizing amino acid by proportion or a data.frame if
#'`exportTable = TRUE`.
percentAA <- function(input.data,
                      chain = "TRB",
                      group.by = NULL,
                      order.by = NULL,
                      aa.length = 20,
                      export.table = NULL,
                      palette = "inferno",
                      # Deprecated arguments
                      exportTable = NULL,
                      ...)  {

  # Handle deprecated arguments
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "percentAA", default = FALSE)

  sco <- .is.seurat.or.se.object(input.data)
  input.data <- .dataWrangle(input.data, group.by, "CTaa", chain)
  if(!is.null(group.by) & !sco) {
    input.data <- .groupList(input.data, group.by)
  }
  
  # Calculating proportion 
  lapply(seq_along(input.data), function(x) {
    strings <- .processStrings(input.data[[x]][["CTaa"]], aa.length)
    freqs <- calculateFrequency(strings)    
    freqs_df <- as.data.frame(freqs)
    freqs_df$AminoAcid <- rownames(freqs_df)
    long_df <- reshape(
      freqs_df,
      varying = list(colnames(freqs)), 
      v.names = "Frequency",           
      timevar = "Position",            
      times = colnames(freqs),         
      direction = "long",              
      idvar = "AminoAcid"              
    )
    rownames(long_df) <- NULL # Remove the messy row names
    long_df$Position <- as.integer(gsub("Pos.", "", long_df$Position))
    long_df$AminoAcid[long_df$AminoAcid == "."] <- NA
    long_df$group <- names(input.data)[x]
    long_df
  }) -> res.list
  
  mat_melt <- do.call(rbind, res.list)
  if(!is.null(order.by)) {
    mat_melt <- .orderingFunction(vector = order.by,
                                  group.by = "group", 
                                  mat_melt)
  }
  if (export.table) {
    return(mat_melt)
  }
  
  # Plotting the result
  plot <- ggplot(mat_melt, aes(x=as.factor(Position), y = .data[["Frequency"]], fill=.data[["AminoAcid"]])) +
    geom_bar(stat = "identity", position="fill", lwd= 0.25, color = "black") +
    scale_fill_manual(name = "Amino Acid", 
                      values = rev(.colorizer(palette,20))) +
    xlab("Amino Acid Residues") +
    ylab("Relative Percent") +
    .themeRepertoire(...) + 
    theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1))
  if(length(res.list) > 1) {
    plot <- plot + facet_grid(group~.)
  }
  return(plot)
}    
    

#' Export Clonal Data in Various Formats
#'
#' Exports clonal information (gene sequences, amino acids, nucleotides) from
#' scRepertoire objects into a file or a data frame. The output format can be
#' tailored for compatibility with different analysis workflows.
#'
#' @details
#' The `format` parameter determines the structure of the output:
#' \itemize{
#'   \item{`paired`}: Exports a data frame where each row represents a barcode,
#'     with paired chain information (amino acid, nucleotide, genes) in separate
#'     columns.
#'   \item{`airr`}: Exports a data frame that adheres to the Adaptive Immune
#'     Receptor Repertoire (AIRR) Community format, with each row representing
#'     a single receptor chain.
#'   \item{`TCRMatch`}: Exports a data frame specifically for the TCRMatch
#'     algorithm, containing the TRB chain amino acid sequence and clonal
#'     frequency.
#'   \item{`tcrpheno`}: Exports a data frame compatible with the `tcrpheno`
#'     pipeline, with TRA and TRB chains in separate columns.
#'   \item{`immunarch`}: Exports a list containing a data frame and metadata
#'     formatted for use with the `immunarch` package.
#' }
#'
#' @param input.data The product of `combineTCR()`, `combineBCR()`, or
#'   `combineExpression()`.
#' @param format The format for exporting clones.
#'   Options are: `paired`, `airr`, `TCRMatch`, `tcrpheno`, `immunarch`.
#' @param group.by The variable in the metadata to use for grouping. If `NULL`,
#'   data will be grouped by the sample names.
#' @param write.file If `TRUE` (default), saves the output to a CSV file. If
#'   `FALSE`, returns the data frame or list to the R environment.
#' @param dir The directory where the output file will be saved.
#'   Defaults to the current working directory.
#' @param file.name The name of the file to be saved.
#'
#' @return A data frame or list in the specified format, either returned to the
#'   R environment or saved as a CSV file.
#'
#' @author Jonathan Noonan, Nick Borcherding
#'
#' @importFrom utils write.csv
#' @export
#' @concept Loading_and_Processing_Contigs
#'
#' @examples
#' \dontrun{
#' #Making combined contig data
#' combined <- combineTCR(contig_list,
#'                        samples = c("P17B", "P17L", "P18B", "P18L",
#'                                    "P19B", "P19L", "P20B", "P20L"))
#'
#' # Export as a paired data frame and save to a file
#' exportClones(combined, format = "paired", file.name = "paired_clones.csv")
#'
#' # Return an AIRR-formatted data frame to the environment
#' airr_df <- exportClones(combined, format = "airr", write.file = FALSE)
#' }
exportClones <- function(input.data,
                         format = "paired",
                         group.by = NULL,
                         write.file = TRUE,
                         dir = NULL,
                         file.name = "clones.csv") {
  
  # Validate format parameter
  format <- match.arg(format, c("paired", "airr", "TCRMatch", "tcrpheno", "immunarch"))
  
  # Select the appropriate internal export function
  exportFunc <- switch(format,
                       "paired"     = .pairedExport,
                       "airr"       = .airrExport,
                       "TCRMatch"   = .tcrMatchExport,
                       "tcrpheno"   = .tcrPhenoExport,
                       "immunarch"  = .immunarchExport
  )
  
  # Generate the data matrix/list
  output_data <- exportFunc(input.data, group.by)
  
  # Replace string "NA" with actual NA values
  if (is.data.frame(output_data)) {
    output_data[output_data == "NA"] <- NA
  }
  
  if (!write.file) {
    return(output_data)
  }
  
  # Handle file writing
  if (is.null(dir)) {
    dir <- "."
  }
  filepath <- file.path(dir, file.name)
  
  # Immunarch format returns a list of data frames. To save as a single CSV,
  # we bind them together and add a 'Sample' identifier column.
  if (format == "immunarch") {
    bound_data <- dplyr::bind_rows(output_data$data, .id = "Sample")
    write.csv(bound_data, file = filepath, row.names = FALSE)
  } else {
    write.csv(output_data, file = filepath)
  }
}


#' Split String and Pad to a Fixed-Width Matrix
#' @keywords internal
._split_and_pad <- function(x, split, n_cols) {
  s <- strsplit(x, split)
  # Create a matrix by safely subsetting each list element up to n_cols
  mat <- t(sapply(s, `[`, seq_len(n_cols)))
  mat[mat == "NA"] <- NA
  return(mat)
}

#' Bind a List of Contig Data Frames and Add Grouping Variable
#' @keywords internal
._bind_contig_list <- function(contig_list) {
  list_names <- names(contig_list)
  if (is.null(list_names)) {
    list_names <- as.character(seq_along(contig_list))
  }
  # Efficiently add group names before binding
  contig_list <- mapply(function(df, name) transform(df, group = name),
                        contig_list, list_names, SIMPLIFY = FALSE)
  # Combine all data frames
  return(do.call(rbind, contig_list))
}

# remap a genes matrix (n x 4) to V/D/J/C with NA for missing D in 3-gene chains
.remap_VDJC <- function(genes_mat) {
  # Coerce to a 4-col character matrix, pad if needed
  if (is.null(genes_mat)) stop(".remap_VDJC: genes_mat is NULL")
  genes_mat <- as.matrix(genes_mat)
  storage.mode(genes_mat) <- "character"
  if (ncol(genes_mat) < 4) {
    pad <- matrix(NA_character_, nrow(genes_mat), 4 - ncol(genes_mat))
    genes_mat <- cbind(genes_mat, pad)
  }
  
  # Clean tokens: drop everything after first ';', trim, map "", "NA" -> NA
  clean_token <- function(x) {
    x <- sub(";.*", "", x, perl = TRUE)
    x <- trimws(x)
    x[x == "" | x == "NA"] <- NA_character_
    x
  }
  genes_mat[] <- clean_token(genes_mat)
  
  vcol <- genes_mat[, 1]
  
  # Detect 3-gene layouts (TRA/TRG/IGK/IGL); handle NAs
  layout3 <- !is.na(vcol) & grepl("^(TRAV|TRGV|IGKV|IGLV)", vcol, perl = TRUE)
  
  # Initialize as 4-gene (V,D,J,C)
  V <- vcol
  D <- genes_mat[, 2]
  J <- genes_mat[, 3]
  C <- genes_mat[, 4]
  
  # Overwrite 3-gene rows: V, (no D), J <- col2, C <- col3
  if (any(layout3, na.rm = TRUE)) {
    idx <- which(layout3)
    D[idx] <- NA_character_
    J[idx] <- genes_mat[idx, 2]
    C[idx] <- genes_mat[idx, 3]
  }
  
  # Ensure character vectors with row length
  stopifnot(length(V) == nrow(genes_mat),
            length(D) == nrow(genes_mat),
            length(J) == nrow(genes_mat),
            length(C) == nrow(genes_mat))
  
  list(V = V, D = D, J = J, C = C)
}



# Format-Specific Export Functions 
.tcrMatchExport <- function(input.data, group.by) {
  # Wrangle data to get TRB chain information
  input.data.list <- .dataWrangle(input.data, group.by, "CTgene", "TRB")
  
  # Process each list element to ensure proper chain format
  input.data.list <- lapply(input.data.list, function(df) {
    df <- .offTheChain(df, "TRB", "CTaa", check = FALSE)
    .offTheChain(df, "TRB", "CTnt", check = FALSE)
  })
  
  bound_data <- ._bind_contig_list(input.data.list)
  # Ensure only TRB chains are included
  bound_data <- bound_data[substr(bound_data$CTgene, 1, 3) == "TRB", ]
  
  # Format for TCRMatch: remove leading character from CDR3 sequence
  mat <- data.frame(
    chain2_aa = substring(bound_data[, "CTaa"], 2),
    group = bound_data[, "group"]
  )
  mat <- na.omit(mat)
  
  # Calculate clonal frequency within each group
  mat$clonalFrequency <- ave(rep(1, nrow(mat)), mat[, c("group", "chain2_aa")], FUN = sum)
  mat <- unique(mat)
  
  return(mat)
}

#' @noRd
.pairedExport <- function(input.data, group.by) {
  input.data.list <- .dataWrangle(input.data, group.by, "CTgene", "both")
  bound_data <- ._bind_contig_list(input.data.list)
  
  # Split concatenated strings into separate columns
  genes <- ._split_and_pad(bound_data[, "CTgene"], "_", 2)
  aa    <- ._split_and_pad(bound_data[, "CTaa"], "_", 2)
  nt    <- ._split_and_pad(bound_data[, "CTnt"], "_", 2)
  
  mat <- data.frame(
    row.names    = bound_data[, "barcode"],
    chain1_aa    = aa[, 1],
    chain1_nt    = nt[, 1],
    chain1_genes = genes[, 1],
    chain2_aa    = aa[, 2],
    chain2_nt    = nt[, 2],
    chain2_genes = genes[, 2],
    group        = bound_data[, "group"]
  )
  return(mat)
}

#' @noRd
.tcrPhenoExport <- function(input.data, group.by) {
  input.data.list <- .dataWrangle(input.data, NULL, "CTgene", "both")
  dat <- do.call(rbind, unname(input.data.list))
  
  # Split sequence and gene columns
  aa_split   <- ._split_and_pad(dat$CTaa, "_", 2)
  nt_split   <- ._split_and_pad(dat$CTnt, "_", 2)
  gene_split <- ._split_and_pad(dat$CTgene, "_", 2)
  
  # Further split gene segments
  tcr1_genes <- ._split_and_pad(gene_split[, 1], "[.]", 2)
  tcr2_genes <- ._split_and_pad(gene_split[, 2], "[.]", 3)[, c(1, 3)]
  
  contigs <- data.frame(
    cell        = rownames(dat),
    TCRA_cdr3aa = aa_split[, 1],
    TCRA_vgene  = tcr1_genes[, 1],
    TCRA_jgene  = tcr1_genes[, 2],
    TCRA_cdr3nt = nt_split[, 1],
    TCRB_cdr3aa = aa_split[, 2],
    TCRB_vgene  = tcr2_genes[, 1],
    TCRB_jgene  = tcr2_genes[, 2],
    TCRB_cdr3nt = nt_split[, 2]
  )
  
  # Remove multi-chain artifacts (keep first sequence)
  columns_to_clean <- c("TCRA_cdr3aa", "TCRA_cdr3nt", "TCRB_cdr3aa", "TCRB_cdr3nt")
  contigs[columns_to_clean] <- lapply(contigs[columns_to_clean], function(x) {
    gsub(";.*", "", x)
  })
  
  # Clean up empty or "NA" strings
  contigs[contigs == "" | contigs == "NA"] <- NA
  return(contigs)
}


#' @noRd
.airrExport <- function(input.data, group.by) {
  input.data.list <- .dataWrangle(input.data, group.by, "CTgene", "both")
  bound_data <- ._bind_contig_list(input.data.list)
  
  # Helper to clean multi-chain entries by taking the first one
  clean_multichain <- function(x) {
    gsub(";.*", "", x)
  }
  
  # Split combined fields for both chains at once
  genes <- ._split_and_pad(bound_data$CTgene, "_", 2)
  aa <- ._split_and_pad(clean_multichain(bound_data$CTaa), "_", 2)
  nt <- ._split_and_pad(clean_multichain(bound_data$CTnt), "_", 2)
  
  # Process chain 1 and chain 2 gene calls
  genes1 <- ._split_and_pad(genes[, 1], "[.]", 4)
  genes2 <- ._split_and_pad(genes[, 2], "[.]", 4)
  
  # Function to create a data frame for a single chain
  format_chain_df <- function(cell_ids, locus_prefix, genes_mat, nt_seq, aa_seq) {
    # Determine if chain is heavy/beta (has D gene)
    is_heavy_or_beta <- locus_prefix %in% c("IGH", "TRB", "TRD")
    is_heavy_or_beta <- names(table(is_heavy_or_beta))[1]
    
    df <- data.frame(
      locus = paste0(locus_prefix, "1"),
      v_call = genes_mat[, 1],
      d_call = if (is_heavy_or_beta) genes_mat[, 2] else NA,
      j_call = if (is_heavy_or_beta) genes_mat[, 3] else genes_mat[, 2],
      c_call = if (is_heavy_or_beta) genes_mat[, 4] else genes_mat[, 3],
      junction = nt_seq,
      junction_aa = aa_seq
    )
    return(df)
  }
  
  # Locus prefixes for chain 1 and 2
  locus1_prefix <- substr(genes[, 1], 1, 3)
  locus2_prefix <- substr(genes[, 2], 1, 3)
  
  # Create data frames for each chain
  df1 <- format_chain_df(bound_data$barcode, locus1_prefix, genes1, nt[, 1], aa[, 1])
  df2 <- format_chain_df(bound_data$barcode, locus2_prefix, genes2, nt[, 2], aa[, 2])
  
  # Add cell_id to both
  df1$cell_id <- bound_data$barcode
  df2$cell_id <- bound_data$barcode
  
  # Combine chain data frames
  mat <- rbind(df1, df2)
  
  # Reorder columns to AIRR standard
  airr_cols <- c("cell_id", "locus", "v_call", "d_call", "j_call", "c_call", "junction", "junction_aa")
  mat <- mat[, airr_cols]
  
  # Remove rows that are entirely NA except for cell_id
  mat <- mat[rowSums(is.na(mat[, -1])) < (ncol(mat) - 1), ]
  
  return(mat)
}


#' @noRd
.immunarchExport <- function(input.data, group.by) {
  df_list <- .dataWrangle(input.data, group.by, "CTgene", "both")
  meta <- data.frame(Sample = names(df_list))
  
  data_out <- lapply(df_list, function(x) {
    # Summarize by clonotype (CTstrict)
    result <- x %>%
      dplyr::group_by(.data[["CTstrict"]]) %>%
      dplyr::summarise(
        Clones  = dplyr::n(),
        barcode = paste(barcode, collapse = ";"),
        CTaa    = dplyr::first(.data[["CTaa"]]),
        CTnt    = dplyr::first(.data[["CTnt"]]),
        CTgene  = dplyr::first(.data[["CTgene"]]),
        .groups = "drop"
      ) %>%
      dplyr::mutate(Proportion = Clones / sum(Clones))
    
    # Determine chain type from first half
    first_genes <- ._split_and_pad(result$CTgene, "_", 2)[, 1]
    chain_type  <- substr(first_genes, 1, 3)
    
    # Impute Missing Chain info (fix condition)
    if (any(is.na(chain_type)) && length(unique(chain_type)) == 2) {
      tbl.store <- rev(sort(table(chain_type, useNA = "ifany")))
      if (length(tbl.store) == 2) {
        chain_type[is.na(chain_type)] <- names(tbl.store)[1]
      }
    }
    
    # Split sequence/gene info
    aa_split   <- ._split_and_pad(result$CTaa,   "_", 2)
    nt_split   <- ._split_and_pad(result$CTnt,   "_", 2)
    gene_split <- ._split_and_pad(result$CTgene, "_", 2)
    
    .is_layout3_vec <- function(v_gene_first_token) {
      grepl("^(TRAV|TRGV|IGKV|IGLV)", v_gene_first_token %||% "", perl = TRUE)
    }
    
    # Adaptive per-row ordering: alpha/light first
    first.gene.pos <- ifelse(chain_type %in% c("TRA","TRG","IGK","IGL"), 1L, 2L)
    na_pos <- is.na(first.gene.pos)
    if (any(na_pos)) {
      left_is3  <- .is_layout3_vec(gsub("\\..*$", "", gene_split[, 1]))
      right_is3 <- .is_layout3_vec(gsub("\\..*$", "", gene_split[, 2]))
      first.gene.pos[na_pos] <- ifelse(left_is3[na_pos], 1L, 2L)
    }
    second.gene.pos <- 3L - first.gene.pos
    
    n    <- nrow(result)
    ridx <- seq_len(n)
    
    # Per-row pick halves
    aa_first   <- as.character(aa_split[cbind(ridx, first.gene.pos)])
    aa_second  <- as.character(aa_split[cbind(ridx, second.gene.pos)])
    nt_first   <- as.character(nt_split[cbind(ridx, first.gene.pos)])
    nt_second  <- as.character(nt_split[cbind(ridx, second.gene.pos)])
    g_first_s  <- as.character(gene_split[cbind(ridx, first.gene.pos)])
    g_second_s <- as.character(gene_split[cbind(ridx, second.gene.pos)])
    
    # Chain presence flags (entire half missing?)
    present1 <- !(is.na(g_first_s)  | g_first_s  == "" | g_first_s  == "NA")
    present2 <- !(is.na(g_second_s) | g_second_s == "" | g_second_s == "NA")
    
    # Tokenize per selected half
    genes1 <- ._split_and_pad(g_first_s,  "[.]", 4)
    genes2 <- ._split_and_pad(g_second_s, "[.]", 4)
    
    # Remap to V/D/J/C (D=NA for 3-gene chains)
    g1 <- .remap_VDJC(genes1)
    g2 <- .remap_VDJC(genes2)
    
    # Display helpers
    pair_strict <- function(lhs, rhs) {
      lhs <- ifelse(is.na(lhs) | lhs == "", "NA", as.character(lhs))
      rhs <- ifelse(is.na(rhs) | rhs == "", "NA", as.character(rhs))
      paste(lhs, rhs, sep = ";")
    }
    
    # For genes: if chain absent -> "NA"
    show_gene <- function(g, present) {
      out <- ifelse(is.na(g) | g == "", "NA", g)
      out[!present] <- "NA"
      out
    }
    # For D gene: if chain present but D missing (3-gene) -> "None"; if chain absent -> "NA"
    show_D <- function(d, present) {
      out <- d
      out <- ifelse(is.na(out) | out == "", "None", out)  
      out[!present] <- "NA"                               
      out
    }
    
    # Build display vectors
    V1 <- show_gene(g1$V, present1); V2 <- show_gene(g2$V, present2)
    D1 <- show_D(  g1$D, present1); D2 <- show_D(  g2$D, present2)
    J1 <- show_gene(g1$J, present1); J2 <- show_gene(g2$J, present2)
    C1 <- show_gene(g1$C, present1); C2 <- show_gene(g2$C, present2)
    
    # CDR3: if chain absent or blank -> "NA"
    aa1 <- ifelse(present1 & !(is.na(aa_first)  | aa_first  == ""), aa_first,  "NA")
    aa2 <- ifelse(present2 & !(is.na(aa_second) | aa_second == ""), aa_second, "NA")
    nt1 <- ifelse(present1 & !(is.na(nt_first)  | nt_first  == ""), nt_first,  "NA")
    nt2 <- ifelse(present2 & !(is.na(nt_second) | nt_second == ""), nt_second, "NA")
    
    # Assemble; always "lhs;rhs"
    out <- data.frame(
      Clones      = result[["Clones"]],
      Proportion  = result[["Proportion"]],
      CDR3.nt     = pair_strict(nt1, nt2),
      CDR3.aa     = pair_strict(aa1, aa2),
      V.name      = pair_strict(V1, V2),
      D.name      = pair_strict(D1, D2),  # preserves "None" vs "NA"
      J.name      = pair_strict(J1, J2),
      C.name      = pair_strict(C1, C2),
      Barcode     = result[["barcode"]],
      stringsAsFactors = FALSE
    )
    out
  })
  
  names(data_out) <- names(df_list)
  list(data = data_out, meta = meta)
}

#' Adding Clonal Information to Single-Cell Object
#'
#' This function adds the immune receptor information to the Seurat or 
#' SCE object to the meta data. By default this function also calculates 
#' the frequencies and proportion of the clones by sequencing 
#' run (`group.by` = NULL). To change how the frequencies/proportions
#' are calculated, select a column header for the `group.by` variable. 
#' Importantly, before using [combineExpression()] ensure the 
#' barcodes of the single-cell object object match the barcodes in the output 
#' of the [combineTCR()] or [combineBCR()]. 
#'
#' @examples
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' 
#' # Getting a sample of a Seurat object
#' scRep_example <- get(data("scRep_example"))
#' 
#' # Using combineExpresion()
#' scRep_example <- combineExpression(combined, scRep_example)
#' 
#' @param input.data The product of [combineTCR()], [combineBCR()] or a list of
#' both c([combineTCR()], [combineBCR()]).
#' @param sc.data The Seurat or Single-Cell Experiment (SCE) object to attach
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`
#' (for both light chains), `both`.
#' @param group.by A column header in lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, will be based on the list element.
#' @param proportion Whether to proportion (`TRUE`) or total
#' frequency (`FALSE`) of the clone based on the group.by variable.
#' @param clone.size The bins for the grouping based on proportion or frequency.
#' If proportion is `FALSE` and the clone.sizes are not set high enough
#' based on frequency, the upper limit of clone.sizes will be automatically
#' updated.
#' @param filter.na Method to subset Seurat/SCE object of barcodes without
#' clone information
#' @param add.label This will add a label to the frequency header, allowing
#' the user to try multiple group.by variables or recalculate frequencies after
#' subsetting the data.
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param cloneSize \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.size` instead.
#' @param filterNA \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `filter.na` instead.
#' @param addLabel \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `add.label` instead.
#' @importFrom dplyr left_join all_of coalesce
#' @importFrom  rlang %||% sym :=
#' @importFrom SummarizedExperiment colData<- colData
#' @importFrom S4Vectors DataFrame
#' @export
#' @concept SC_Functions
#' @return Single-cell object with clone information added to meta data
#' information
#'
combineExpression <- function(input.data,
                              sc.data,
                              clone.call = NULL,
                              chain = "both",
                              group.by = NULL,
                              proportion = TRUE,
                              filter.na = NULL,
                              clone.size = NULL,
                              add.label = NULL,
                              # Deprecated arguments
                              cloneCall = NULL,
                              cloneSize = NULL,
                              filterNA = NULL,
                              addLabel = NULL) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "combineExpression", default = "strict")
  clone.size <- .deprecate_arg(cloneSize, clone.size, "cloneSize", "clone.size",
                               "combineExpression", default = c(Rare = 1e-4, Small = 0.001, Medium = 0.01, Large = 0.1, Hyperexpanded = 1))
  filter.na <- .deprecate_arg(filterNA, filter.na, "filterNA", "filter.na",
                              "combineExpression", default = FALSE)
  add.label <- .deprecate_arg(addLabel, add.label, "addLabel", "add.label",
                              "combineExpression", default = FALSE)

  clonalFrequency <- NULL
  call_time <- Sys.time()
    options( dplyr.summarise.inform = FALSE )
    if (!proportion && any(clone.size < 1)) {
        stop("Adjust the clone.size parameter - there are groupings < 1")
    }
    clone.size <- c(None = 0, clone.size)

    clone.call <- .theCall(input.data, clone.call)
    if (chain != "both") {
      #Retain the full clone information
      full.clone <- lapply(input.data, function(x) {
        x[, c("barcode", clone.call)]
      })
      full.clone <- bind_rows(full.clone)
      for(i in seq_along(input.data)) {
        input.data[[i]] <- .offTheChain(input.data[[i]], chain, clone.call, check = FALSE)
      }
    }
    input.data <- .checkList(input.data)

    # Remove any existing clonal columns to allow recalculation
    cols_to_remove <- c("clonalFrequency", "clonalProportion", "cloneSize")
    for (i in seq_along(input.data)) {
      existing_cols <- intersect(cols_to_remove, colnames(input.data[[i]]))
      if (length(existing_cols) > 0) {
        input.data[[i]] <- input.data[[i]][, !colnames(input.data[[i]]) %in% existing_cols, drop = FALSE]
      }
    }

    #Getting Summaries of clones from combineTCR() or combineBCR()
    Con.df <- NULL
    meta <- .grabMeta(sc.data)
    cell.names <- rownames(meta)

    conDfColnamesNoCloneSize <- unique(c(
        "barcode", CT_lines, clone.call, "clonalProportion", "clonalFrequency"
    ))

    # Computes the clonalProportion and clonalFrequency for each clone
    if (is.null(group.by) || group.by == "none") {

        for (i in seq_along(input.data)) {

            data <- data.frame(input.data[[i]], stringsAsFactors = FALSE)
            data2 <- unique(data[,c("barcode", clone.call)])
            #This ensures all calculations are based on the cells in the SCO
            data2 <- na.omit(data2[data2[,"barcode"] %in% cell.names,])
            data2 <- data2 %>%
                        group_by(data2[,clone.call]) %>%
                        summarise(clonalProportion = dplyr::n()/nrow(data2),
                                  clonalFrequency = dplyr::n())
            colnames(data2)[1] <- clone.call
            data <- merge(data, data2, by = clone.call, all = TRUE)
            data <- data[, conDfColnamesNoCloneSize]
            Con.df <- rbind.data.frame(Con.df, data)
        }

    } else {
        data <- data.frame(bind_rows(input.data), stringsAsFactors = FALSE)
        data2 <- na.omit(unique(data[,c("barcode", clone.call, group.by)]))
        #This ensures all calculations are based on the cells in the SCO
        data2 <- data2[data2[,"barcode"] %in% cell.names, ]
        data2 <- data2 %>%
          group_by(.data[[clone.call]], .data[[group.by]]) %>%
          summarise(clonalFrequency = n(), .groups = "drop") %>%
          group_by(.data[[group.by]]) %>%
          mutate(clonalProportion = clonalFrequency / sum(clonalFrequency))

        colnames(data2)[c(1,2)] <- c(clone.call, group.by)
        data <- merge(data, data2, by = c(clone.call, group.by), all = TRUE)
        Con.df <- data[, conDfColnamesNoCloneSize]
    }

    #Use shared helper to assign cloneSize bins
    bin_result <- .assignCloneSizeBins(Con.df, clone.size, proportion)
    Con.df <- bin_result$df
    clone.size <- bin_result$clone.size

    #Formating the meta data to add and removing any duplicate barcodes
    PreMeta <- unique(Con.df[, c(conDfColnamesNoCloneSize, "cloneSize")])
    dup <- PreMeta$barcode[which(duplicated(PreMeta$barcode))]
    PreMeta <- PreMeta[!PreMeta$barcode %in% dup,]

    #Re-adding full clones
    if (chain != "both") {
      clone_sym <- sym(clone.call)
      PreMeta <- PreMeta %>%
        left_join(full.clone, by = "barcode", suffix = c("", ".from_full_clones")) %>%
        mutate(!!clone_sym := coalesce(!!sym(paste0(clone.call, ".from_full_clones")), !!clone_sym)) %>%
        dplyr::select(-all_of(paste0(clone.call, ".from_full_clones")))
    }
    barcodes <- PreMeta$barcode
    PreMeta <- PreMeta[,-1]
    rownames(PreMeta) <- barcodes
    if (!is.null(group.by) && group.by != "none" && add.label) {
      location <- which(colnames(PreMeta) %in% c("clonalProportion",
                          "clonalFrequency"))
      colnames(PreMeta)[location] <- paste0(c("clonalProportion",
                                            "clonalFrequency"), group.by)
    }
    
    if (.is.seurat.object(sc.data)) { 
        if (length(which(rownames(PreMeta) %in% 
                         rownames(sc.data[[]])))/length(rownames(sc.data[[]])) < 0.01) {
          getHighBarcodeMismatchError()
        }
        col.name <- names(PreMeta) %||% colnames(PreMeta)
        sc.data[[col.name]] <- PreMeta
    } else {
      rownames <- rownames(colData(sc.data))
      if (length(which(rownames(PreMeta) %in% 
                       rownames))/length(rownames) < 0.01) {
        getHighBarcodeMismatchError() }
      
      combined_col_names <- unique(c(colnames(colData(sc.data)), colnames(PreMeta)))
      full_data <- merge(colData(sc.data), PreMeta[rownames, , drop = FALSE], by = "row.names", all.x = TRUE)
      # at this point, the rows in full_data are shuffled. match back with the original colData
      full_data <- full_data[match(rownames, full_data[,1]), ]
      rownames(full_data) <- full_data[, 1]
      full_data  <- full_data[, -1]
      colData(sc.data) <- DataFrame(full_data[, combined_col_names])  
    }
    if (filter.na) {
      sc.data <- .filteringNA(sc.data)
    }
    sc.data$cloneSize <- factor(sc.data$cloneSize, levels = rev(names(clone.size)))
    
    if(.is.seurat.object(sc.data)) {
        sc.data@commands[["combineExpression"]] <- .makeScrepSeurat(
              call_time, sc.data@active.assay)
    }
    return(sc.data)
}

getHighBarcodeMismatchError <- function() {
  stop(
    paste0(
      "< 1% of barcodes match.\n",
      "Ensure the barcodes in the single-cell object match the barcodes from scRepertoire.\n",
      "For help, see: https://www.borch.dev/uploads/screpertoire/articles/faq"
    ),
    call. = FALSE 
  )
}

#' Deconvolute Contig Information from Multiplexed Experiments
#'
#' This function reprocess and forms a list of contigs for downstream analysis 
#' in scRepertoire, [createHTOContigList()] take the filtered contig 
#' annotation output and the single-cell RNA object to create the list. 
#' If using an integrated single-cell object, it is recommended to split the 
#' object by sequencing run and remove extra prefixes and suffixes on the 
#' barcode before using [createHTOContigList()]. Alternatively, 
#' the variable `multi.run` can be used to separate a list of contigs
#' by a meta data variable. This may have issues with the repeated barcodes.
#' 
#' @examples
#' \dontrun{
#' filtered.contig <- read.csv(".../Sample/outs/filtered_contig_annotations.csv")
#' 
#' contig.list <- createHTOContigList(contig = filtered.contig, 
#'                                    sc.data = Seurat.Obj, 
#'                                    group.by = "HTO_maxID")
#' }
#' 
#' @param contig The filtered contig annotation file from multiplexed experiment
#' @param sc.data The Seurat or Single-Cell Experiment object.
#' @param group.by One or more meta data headers to create the contig 
#' list based on. If more than one header listed, the function combines 
#' them into a single variable.
#' @param multi.run If using integrated single-cell object, the meta data 
#' variable that indicates the sequencing run.
#' @export
#' @concept Loading_and_Processing_Contigs
#' @return Returns a list of contigs as input for [combineBCR()] 
#' or [combineTCR()]


createHTOContigList <- function(contig, 
                                sc.data, 
                                group.by = NULL, 
                                multi.run = NULL){
  contig.list <- NULL
  .checkSingleObject(sc.data)
  meta <- .grabMeta(sc.data)
  if (length(group.by) > 1) {
    meta["group.by"] <- apply(meta[ , group.by] , 1 , paste , collapse = "." )
  } else {
    meta[,"group.by"] <- meta[ , group.by]
  }
  unique.groups <- unique(meta$group.by)
  
  if (!is.null(multi.run) & inherits(x=contig, what ="list")) {
    tmp.list <- list()
    runs <- unique(meta[,multi.run])
    for (x in seq_along(contig)) {
      cont.tmp <- contig[[x]]
      cont.tmp$barcode <- paste0(cont.tmp$barcode, "_", x)
      meta.tmp <- meta[meta[,multi.run] == runs[x],] #Subset meta in batch
      rownames(meta.tmp) <- substring(rownames(meta.tmp), nchar(rownames(meta.tmp)[1])-17, nchar(rownames(meta.tmp)[1]))
      cont.tmp <- cont.tmp[cont.tmp$barcode %in% row.names(meta.tmp) ] #subset contig in new meta
      for (y in seq_along(unique.groups)) {
        sub.con <- cont.tmp[cont.tmp$barcode %in% rownames(subset(meta.tmp, group.by == unique.groups[y])),]
        tmp.list[[y]] <- sub.con
      }
      if(x == 1) {
        contig.list  <- tmp.list
      } else {
        contig.list <- c(contig.list, tmp.list)
      }
    }
  } else {
    cont.tmp <- contig[contig$barcode %in% rownames(meta), ]
    for (i in seq_along(unique.groups)) {
      sub.con <- cont.tmp[cont.tmp$barcode %in% rownames(subset(meta, group.by == unique.groups[i])),]
      contig.list[[i]] <- sub.con
    }
    names(contig.list) <- unique.groups
  }
  contig.list
}

#' Generate Data Frame to Plot Chord Diagram
#'
#' This function will take the meta data from the product of
#' [combineExpression()] and generate a relational data frame to
#' be used for a chord diagram. Each chord will represent the number of
#' clones unique and shared across the multiple `group.by` variable.
#' If using the downstream circlize R package, please read and cite the
#' following [manuscript](https://pubmed.ncbi.nlm.nih.gov/24930139/).
#' If looking for more advanced ways for circular visualizations, there
#' is a great [cookbook](https://jokergoo.github.io/circlize_book/book/)
#' for the circlize package.
#'
#' @examples
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list,
#'                         samples = c("P17B", "P17L", "P18B", "P18L",
#'                                     "P19B","P19L", "P20B", "P20L"))
#'
#' # Getting a sample of a Seurat object
#' scRep_example <- get(data("scRep_example"))
#' scRep_example <- combineExpression(combined,
#'                                    scRep_example)
#'
#' # Getting data frame output for Circlize
#' circles <- getCirclize(scRep_example,
#'                        group.by = "seurat_clusters")
#'
#' # Multi-level grouping for hierarchical chord diagrams
#' scRep_example$Patient <- substring(scRep_example$orig.ident, 1, 3)
#' circles <- getCirclize(scRep_example,
#'                        group.by = c("Patient", "seurat_clusters"))
#'
#' # Get rich output with sector metadata
#' result <- getCirclize(scRep_example,
#'                       group.by = "seurat_clusters",
#'                       include.metadata = TRUE)
#'
#' @param sc.data The single-cell object after [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param group.by A column header (or vector of column headers for hierarchical
#' grouping) in the metadata to group the analysis by (e.g., "sample", "treatment").
#' If `NULL`, data will be analyzed by active identity. When multiple columns are
#' provided, they are combined with "_" separator for multi-level annotations.
#' @param method The method for calculating link values: `"unique"` (default) counts
#' unique shared clones, `"abundance"` sums clone frequencies, `"jaccard"` calculates
#' Jaccard similarity, `"overlap"` calculates overlap coefficient.
#' @param proportion Calculate the relationship by unique clones (`FALSE`, default)
#' or normalized by proportion (`TRUE`).
#' @param symmetric If `TRUE` (default), returns symmetric relationships. If `FALSE`,
#' returns directional flow showing proportion of source's clones found in destination.
#' @param include.self Include counting the clones within a single group.by comparison.
#' @param include.metadata If `TRUE`, returns a list with links data frame and
#' sector-level metadata including cell counts, clone counts, and expansion metrics.
#' @param min.shared Minimum number of shared clones to include a link (default 0).
#' @param top.links Keep only the top N links by value. If `NULL` (default), keep all.
#' @param filter.sectors Character vector of sectors to include. If `NULL`, include all.
#' @param palette Colors to use for sector color suggestions - input any
#' [hcl.pals][grDevices::hcl.pals].
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#'
#' @export
#' @concept SC_Functions
#' @return A data frame of shared clones between groups formatted for
#' [chordDiagram][circlize::chordDiagram]. If `include.metadata = TRUE`, returns
#' a list with `links` (the edge data frame), `sectors` (sector-level statistics),
#' and `colors` (suggested colors for each sector).
#' @author Dillon Corvino, Nick Borcherding
getCirclize <- function(sc.data,
                        clone.call = NULL,
                        group.by = NULL,
                        method = c("unique", "abundance", "jaccard", "overlap"),
                        proportion = FALSE,
                        symmetric = TRUE,
                        include.self = TRUE,
                        include.metadata = FALSE,
                        min.shared = 0,
                        top.links = NULL,
                        filter.sectors = NULL,
                        palette = "inferno",
                        # Deprecated arguments
                        cloneCall = NULL) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "getCirclize", default = "strict")

  method <- match.arg(method)

  meta <- .grabMeta(sc.data)
  clone.call <- .theCall(meta, clone.call)

  if(is.null(group.by)) {
    group.by <- "ident"
  }


  # Handle multi-level grouping
  if(length(group.by) > 1) {
    # Create compound grouping variable
    meta$`.compound_group` <- apply(meta[, group.by, drop = FALSE], 1,
                                    function(x) paste(x, collapse = "_"))
    group.var <- ".compound_group"
    hierarchy.info <- meta[, c(group.by, ".compound_group"), drop = FALSE]
    hierarchy.info <- unique(hierarchy.info)
  } else {
    group.var <- group.by
    hierarchy.info <- NULL
  }

  # Filter sectors if specified
  if(!is.null(filter.sectors)) {
    meta <- meta[meta[[group.var]] %in% filter.sectors, ]
    if(nrow(meta) == 0) {
      stop("No cells remaining after filtering by filter.sectors")
    }
  }

  unique.sectors <- unique(meta[[group.var]])

  # Build group pairs based on symmetry
  if(symmetric) {
    group_pairs <- expand.grid(group1 = unique.sectors, group2 = unique.sectors,
                               stringsAsFactors = FALSE)
    group_pairs <- unique(t(apply(group_pairs, 1, function(x) sort(x))))
    group_pairs <- as.data.frame(group_pairs, stringsAsFactors = FALSE)
    colnames(group_pairs) <- c("from", "to")
  } else {
    # Asymmetric: all ordered pairs
    group_pairs <- expand.grid(from = unique.sectors, to = unique.sectors,
                               stringsAsFactors = FALSE)
  }

  if(!include.self) {
    group_pairs <- group_pairs[group_pairs[,1] != group_pairs[,2], ]
  }

  # Count clones across all identities
  clone.table <- .cloneCounter(meta, group.var, clone.call)
  clone.table[[clone.call]] <- as.character(clone.table[[clone.call]])

  group_pairs$value <- NA

  for(i in seq_len(nrow(group_pairs))) {
    pair1 <- group_pairs[i, 1]
    pair2 <- group_pairs[i, 2]

    clone1.data <- clone.table[clone.table[,1] == pair1 & clone.table[["n"]] > 0, ]
    clone2.data <- clone.table[clone.table[,1] == pair2 & clone.table[["n"]] > 0, ]

    clone1 <- clone1.data[[clone.call]]
    clone2 <- clone2.data[[clone.call]]

    common <- intersect(clone1, clone2)

    # Calculate value based on method
    if(method == "unique") {
      value <- length(common)

      if(symmetric && pair1 == pair2) {
        # For self-links in symmetric mode, subtract shared clones
        tmp <- clone.table[clone.table[[clone.call]] %in% common &
                           clone.table[,1] != pair1, ]
        shared.clones <- unique(tmp[[clone.call]])
        value <- value - length(shared.clones)
      }

    } else if(method == "abundance") {
      # Sum of minimum frequencies for shared clones
      if(length(common) == 0) {
        value <- 0
      } else {
        freq1 <- clone1.data[clone1.data[[clone.call]] %in% common, "n"]
        names(freq1) <- clone1.data[clone1.data[[clone.call]] %in% common, clone.call]
        freq2 <- clone2.data[clone2.data[[clone.call]] %in% common, "n"]
        names(freq2) <- clone2.data[clone2.data[[clone.call]] %in% common, clone.call]
        value <- sum(pmin(freq1[common], freq2[common]))
      }

    } else if(method == "jaccard") {
      union_clones <- union(clone1, clone2)
      if(length(union_clones) == 0) {
        value <- 0
      } else {
        value <- length(common) / length(union_clones)
      }

    } else if(method == "overlap") {
      min_size <- min(length(clone1), length(clone2))
      if(min_size == 0) {
        value <- 0
      } else {
        value <- length(common) / min_size
      }
    }

    # Apply proportion normalization if requested (for unique method)
    if(proportion && method == "unique") {
      denominator <- length(unique(clone.table[clone.table[,1] == pair2, clone.call]))
      if(denominator > 0) {
        value <- value / denominator
      } else {
        value <- 0
      }
    }

    # For asymmetric mode, calculate directional proportion
    if(!symmetric && pair1 != pair2) {
      n_source <- length(clone1)
      if(n_source > 0) {
        value <- length(common) / n_source
      } else {
        value <- 0
      }
    }

    group_pairs$value[i] <- value
  }

  # Filter by minimum shared value
  if(min.shared > 0) {
    group_pairs <- group_pairs[group_pairs$value >= min.shared, ]
  }

  # Keep only top links
  if(!is.null(top.links) && nrow(group_pairs) > top.links) {
    group_pairs <- group_pairs[order(group_pairs$value, decreasing = TRUE), ]
    group_pairs <- group_pairs[seq_len(top.links), ]
  }

  # If not including metadata, return simple data frame
 if(!include.metadata) {
    return(group_pairs)
  }

  # Build sector metadata
  sector.stats <- lapply(unique.sectors, function(s) {
    sector.meta <- meta[meta[[group.var]] == s, ]
    sector.clones <- clone.table[clone.table[,1] == s, ]

    n.cells <- nrow(sector.meta)
    n.clones <- nrow(sector.clones[sector.clones$n > 0, ])

    # Count shared clones (appearing in other sectors)
    sector.clone.ids <- sector.clones[[clone.call]]
    other.clones <- clone.table[clone.table[,1] != s, clone.call]
    n.shared <- length(intersect(sector.clone.ids, other.clones))

    # Simple expansion metric: 1 - (unique clones / total cells)
    expansion <- ifelse(n.cells > 0, 1 - (n.clones / n.cells), 0)

    data.frame(
      sector = s,
      n.cells = n.cells,
      n.clones = n.clones,
      n.shared = n.shared,
      expansion = expansion,
      stringsAsFactors = FALSE
    )
  })
  sector.stats <- do.call(rbind, sector.stats)

  # Add hierarchy info if multi-level grouping
  if(!is.null(hierarchy.info)) {
    sector.stats <- merge(sector.stats, hierarchy.info,
                          by.x = "sector", by.y = ".compound_group",
                          all.x = TRUE)
  }

  # Generate colors
  n.sectors <- length(unique.sectors)
  colors <- .colorizer(palette, n.sectors)
  names(colors) <- unique.sectors

  result <- list(
    links = group_pairs,
    sectors = sector.stats,
    colors = colors
  )

  return(result)
}

#' Bin Clones by Frequency or Proportion
#'
#' This function adds a clonal grouping variable (`cloneSize`) to the output
#' of [combineTCR()], [combineBCR()], or [combineExpression()]. It calculates
#' the clonal frequency and proportion, then bins clones into categories based
#' on customizable thresholds. This is useful for categorizing clones prior to
#' downstream analysis or visualization.
#'
#' @examples
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list,
#'                         samples = c("P17B", "P17L", "P18B", "P18L",
#'                                     "P19B","P19L", "P20B", "P20L"))
#'
#' # Adding clonal bins with default settings (proportion-based)
#' combined <- clonalBin(combined)
#'
#' # Adding clonal bins based on frequency
#' combined <- clonalBin(combined,
#'                       proportion = FALSE,
#'                       clone.size = c(Rare = 1, Small = 5, Medium = 20,
#'                                      Large = 100, Hyperexpanded = 500))
#'
#' # Using a custom grouping variable
#' combined <- addVariable(combined,
#'                         variable.name = "Type",
#'                         variables = rep(c("B", "L"), 4))
#' combined <- clonalBin(combined, group.by = "Type")
#'
#' @param input.data The product of [combineTCR()], [combineBCR()], or
#' [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`
#' (for both light chains), `both`.
#' @param group.by A column header in the metadata to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed
#' by list element.
#' @param proportion Whether to use proportion (`TRUE`) or total
#' frequency (`FALSE`) of the clone for binning.
#' @param clone.size The bins for the grouping based on proportion or frequency.
#' If proportion is `FALSE` and the clone.size values are not set high enough
#' based on frequency, the upper limit of clone.size will be automatically
#' updated.
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param cloneSize \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.size` instead.
#' @importFrom dplyr bind_rows group_by summarise mutate n
#' @importFrom rlang .data
#' @importFrom stats na.omit
#' @export
#' @concept Clonal_Analysis
#' @author Nick Borcherding
#' @return A list of data frames with clonal frequency, clonal proportion,
#' and cloneSize columns added.
#'
clonalBin <- function(input.data,
                      clone.call = NULL,
                      chain = "both",
                      group.by = NULL,
                      proportion = TRUE,
                      clone.size = NULL,
                      # Deprecated arguments
                      cloneCall = NULL,
                      cloneSize = NULL) {

  # Avoid R CMD check note for dplyr column references
  clonalFrequency <- NULL

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "clonalBin", default = "strict")
  clone.size <- .deprecate_arg(cloneSize, clone.size, "cloneSize", "clone.size",
                               "clonalBin",
                               default = c(Rare = 1e-4, Small = 0.001,
                                          Medium = 0.01, Large = 0.1,
                                          Hyperexpanded = 1))

  # Suppress summarise messages
  options(dplyr.summarise.inform = FALSE)

 # Validate clone.size for frequency-based binning
  if (!proportion && any(clone.size < 1)) {
    stop("Adjust the clone.size parameter - there are groupings < 1")
  }

  # Add the baseline bin
  clone.size <- c(None = 0, clone.size)

  # Resolve clone.call to actual column name
  clone.call <- .theCall(input.data, clone.call)

  # Handle chain filtering if needed
  if (chain != "both") {
    for (i in seq_along(input.data)) {
      input.data[[i]] <- .offTheChain(input.data[[i]], chain, clone.call, check = FALSE)
    }
  }

  # Ensure input.data is a list
  input.data <- .checkList(input.data)

  # Remove any existing clonal columns to allow recalculation
  cols_to_remove <- c("clonalFrequency", "clonalProportion", "cloneSize")
  for (i in seq_along(input.data)) {
    existing_cols <- intersect(cols_to_remove, colnames(input.data[[i]]))
    if (length(existing_cols) > 0) {
      input.data[[i]] <- input.data[[i]][, !colnames(input.data[[i]]) %in% existing_cols, drop = FALSE]
    }
  }

  # Initialize output list
  output.data <- list()

  # Compute clonal frequency and proportion
  if (is.null(group.by) || group.by == "none") {
    # Calculate per list element
    for (i in seq_along(input.data)) {
      data <- data.frame(input.data[[i]], stringsAsFactors = FALSE)
      data2 <- unique(data[, c("barcode", clone.call)])
      data2 <- na.omit(data2)

      # Calculate frequency and proportion
      data2 <- data2 %>%
        group_by(data2[, clone.call]) %>%
        summarise(clonalProportion = dplyr::n() / nrow(data2),
                  clonalFrequency = dplyr::n())
      colnames(data2)[1] <- clone.call

      # Merge back to original data
      data <- merge(data, data2, by = clone.call, all = TRUE)
      output.data[[i]] <- data
    }
    names(output.data) <- names(input.data)

  } else {
    # Calculate across all elements grouped by specified variable
    data <- data.frame(bind_rows(input.data), stringsAsFactors = FALSE)
    data2 <- na.omit(unique(data[, c("barcode", clone.call, group.by)]))

    data2 <- data2 %>%
      group_by(.data[[clone.call]], .data[[group.by]]) %>%
      summarise(clonalFrequency = n(), .groups = "drop") %>%
      group_by(.data[[group.by]]) %>%
      mutate(clonalProportion = clonalFrequency / sum(clonalFrequency))

    colnames(data2)[c(1, 2)] <- c(clone.call, group.by)
    data <- merge(data, data2, by = c(clone.call, group.by), all = TRUE)

    # Split back into list by original list element names
    if ("element.names" %in% colnames(data)) {
      output.data <- split(data, data$element.names)
    } else {
      # If no element.names, re-split based on input names
      output.data <- list(data)
      names(output.data) <- "combined"
    }
  }

  # Combine all data for bin assignment to get consistent bin thresholds
  Con.df <- bind_rows(output.data)

  # Use shared helper to assign bins (returns both df and formatted clone.size)
  bin_result <- .assignCloneSizeBins(Con.df, clone.size, proportion)
  clone.size <- bin_result$clone.size

  # Apply binning to each element in output.data
  for (i in seq_along(output.data)) {
    element_result <- .assignCloneSizeBins(output.data[[i]], clone.size, proportion)
    output.data[[i]] <- element_result$df

    # Convert to factor with reversed levels (largest first)
    output.data[[i]]$cloneSize <- factor(
      output.data[[i]]$cloneSize,
      levels = rev(names(clone.size))
    )
  }

  return(output.data)
}

#' Summarizes and Visualizes Gene Usage
#'
#' This function quantifies and visualizes the usage of V, D, or J genes,
#' or gene pairings across T or B cell clones. 
#' 
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#'                                     
#' # Visualize single gene (TRBV) usage as a heatmap, grouped by sample
#' percentGeneUsage(combined,
#'                  genes = "TRBV",
#'                  group.by = "sample",
#'                  plot.type = "heatmap",
#'                  summary.fun = "percent")
#'
#' # Visualize single gene (TRBV) usage as a barplot, grouped by sample
#' percentGeneUsage(combined,
#'                  genes = "TRBV",
#'                  group.by = "sample",
#'                  plot.type = "barplot",
#'                  summary.fun = "count")
#'
#' # Visualize paired gene (TRBV-TRBJ) usage as a heatmap
#' percentGeneUsage(combined[1:2],
#'                  genes = c("TRBV", "TRBJ"),
#'                  group.by = "sample",
#'                  plot.type = "heatmap",
#'                  summary.fun = "proportion")
#'
#' # Export the raw data table for single gene usage
#' trbv_usage_table <- percentGeneUsage(combined,
#'                                      genes = "TRBV",
#'                                      group.by = "sample",
#'                                      export.table = TRUE,
#'                                      summary.fun = "count")
#'
#' # Export the raw data table for paired gene usage
#' trbv_trbj_usage_table <- percentGeneUsage(combined,
#'                                           genes = c("TRBV", "TRBJ"),
#'                                           group.by = "sample",
#'                                           export.table = TRUE,
#'                                           summary.fun = "percent")
#' 
#'
#' @param input.data The product of [combineTCR()], [combineBCR()],
#'or [combineExpression()].
#' @param genes A character vector specifying the gene loci to analyze.
#' Can be a single gene e.g., "TRBV" or "IGHJ" or a pair for genes analysis 
#' (e.g., c("TRBV", "TRAV"), or "TRBV", "TRBJ").
#' @param group.by A column header in the metadata or lists to group the analysis 
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed as 
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements 
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups 
#' automatically.
#' @param summary.fun Character string choosing the summary statistic -
#' `"percent"` (default), `"proportion"`, or `"count"`.
#' @param plot.type The type of plot to return: `"heatmap"` (default for paired loci,
#' also available for single loci), or `"barplot"` (for single loci).
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' [hcl.pals][grDevices::hcl.pals].
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @importFrom immApex calculateGeneUsage getIR
#' @importFrom graphics plot
#' @importFrom grDevices hcl.pals
#' @importFrom utils stack
#' @importFrom dplyr group_by summarise ungroup mutate 
#'
#' @export
#'
#' @concept Summarize_Repertoire
#' @concept Visualizing_Clones
#'
#' @return A ggplot object displaying a heatmap or bar plot of gene usage.
#' If `exportTable = TRUE`, a matrix or data frame of the raw data is returned.
percentGeneUsage <- function(input.data,
                             chain = "TRB",
                             genes = "TRBV",
                             group.by = NULL,
                             order.by = NULL,
                             summary.fun = c("percent", "proportion", "count"),
                             plot.type = "heatmap",
                             export.table = NULL,
                             palette = "inferno",
                             # Deprecated arguments
                             exportTable = NULL,
                             ...) {

  # Handle deprecated arguments
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "percentGeneUsage", default = FALSE)

  sco <- .is.seurat.or.se.object(input.data)
  
  summary.fun <- match.arg(summary.fun)
  if (!is.character(genes) || !(length(genes) %in% 1:2)) {
    stop("Parameter 'genes' must be a character vector of length 1 or 2, specifying gene loci.")
  }
  
  chains_to_extract <- unique(substr(genes, 1,3))
  if (length(chains_to_extract) > 2) {
    stop("Analysis limited to a maximum of two chains. Please refine 'genes' input.")
  }
  
  gene_segment_map <- c("V" = "v", "D" = "d", "J" = "j", "C" = "c")
  column.headers <- character(length(genes))
  for (i in seq_along(genes)) {
    gene_type_char <- substr(genes[i], 4, 4)
    if (gene_type_char %in% names(gene_segment_map)) {
      column.headers[i] <- gene_segment_map[gene_type_char]
    } else {
      stop(paste0("Unsupported gene segment format: '", genes[i], "'. Expected format like 'TRBV', 'IGHJ', etc."))
    }
  }
  
  ir_data_list <- lapply(chains_to_extract, function(ch) {
    dfs <- immApex::getIR(input.data, 
                          chains = .chainConverter(ch), 
                          sequence.type = "aa", 
                          group.by = group.by)
    if(inherits(dfs, "list")) {
      dfs <- do.call(rbind, dfs) 
    } 
    return(dfs)
  })
  
  # Handling NULL group.by
  if(is.null(group.by)) {
    if(sco) {
      group.by <- "ident"
      groupings <- .bound.input.return(input.data)[,group.by, drop = FALSE]
    } else {
      group.by <- "element.names" 
      groupings <- .bound.input.return(input.data)[,c("barcode", group.by), drop = FALSE]
      rownames(groupings) <- groupings[["barcode"]]
      groupings <- groupings[,-which(colnames(groupings) == "barcode"), drop = FALSE]
      
    }
    ir_data_list <- lapply(ir_data_list, function(x) {
      merge(x, groupings, by.x = "barcode", by.y = 0)
    })
  }
  gene.data <- ir_data_list[[1]]

  
  # If second chain is being used
  if (length(chains_to_extract) == 2 && length(genes) == 2) {
    second_chain_data <- ir_data_list[[2]]
    common_cols <- intersect(names(gene.data), names(second_chain_data))
    cols_to_rename <- setdiff(names(second_chain_data), c(group.by, "barcode"))
    colnames(second_chain_data)[match(cols_to_rename, colnames(second_chain_data))] <- paste0(cols_to_rename, ".2")
    gene.data <- merge(gene.data, second_chain_data, by = c("barcode", group.by), all = TRUE)
    column.headers[2] <- paste0(column.headers[2], ".2")
  }
  
  # Final separation before calculation
  grouped_data <- split(gene.data, gene.data[[group.by]])
  calculated_usage <- lapply(grouped_data, function(x) {
      df_for_calc <- x
      immApex::calculateGeneUsage(
        input.data = df_for_calc,
        loci = column.headers,
        summary = summary.fun
      )
    })
  
  mat_melt_list <- lapply(names(calculated_usage), function(group_name) {
    item <- calculated_usage[[group_name]]
    if (is.atomic(item) && !is.null(names(item)) && length(dim(item)) < 2) {
      # Single locus result (vector)
      data.frame(
        Var1 = names(item),
        Weight = as.vector(unname(item)),
        Group = group_name,
        stringsAsFactors = FALSE
      )
    } else if (is.matrix(item)) {
      # Paired loci result 
      df_melted <- as.data.frame(as.table(item), stringsAsFactors = FALSE)
      names(df_melted) <- c("Var1", "Var2", "Weight")
      data.frame(
        Var1 = df_melted$Var1,
        Var2 = df_melted$Var2,
        Weight = df_melted$Weight,
        Group = group_name,
        stringsAsFactors = FALSE
      )
    }
  })
  
  mat_melt <- do.call(rbind, mat_melt_list)
  if (export.table) {
    if (length(genes) == 1) {
      # Single gene usage: rows are genes, columns are groups
      output_matrix <- tapply(mat_melt$Weight, list(mat_melt$Var1, mat_melt$Group), sum)
    } else {
      # Paired gene usage: rows are gene pairs, columns are groups
      row_variable <- paste(mat_melt$Var1, mat_melt$Var2, sep = "_")
      output_matrix <- tapply(mat_melt$Weight, list(row_variable, mat_melt$Group), sum)
    }
    output_matrix[is.na(output_matrix)] <- 0
    return(output_matrix)
  }
  
  col.lab <- .toCapitilize(summary.fun)
  
  if (!is.null(order.by)) {
      mat_melt <- .orderingFunction(vector = order.by, group.by = "Group", mat_melt)
  }
  
  if (length(genes) == 1) {
    if (plot.type == "barplot") {
      plot <- ggplot(mat_melt, aes(x = .data[["Var1"]], y = .data[["Weight"]])) +
        geom_bar(stat = "identity", color = "black", lwd = 0.25) +
        .themeRepertoire(...) + 
        labs(y = col.lab) + 
        theme(axis.title = element_blank(),,
              axis.ticks.x = element_blank(),
              axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1, size=rel(0.8)))
      
      if (length(unique(mat_melt$Group)) > 1) {
        plot <- plot + facet_grid(.data[["Group"]] ~ .)
      }
      
    } else { 
      plot <- ggplot(mat_melt, aes(y = .data[["Var1"]], x = .data[["Group"]], fill = .data[["Weight"]])) +
        geom_tile(lwd = 0.1, color = "black") +
        scale_fill_gradientn(colors = .colorizer(palette, 21)) +
        labs(fill = col.lab) + 
        .themeRepertoire(...) + 
        theme(axis.title = element_blank())
    }
  } else { # Paired genes (heatmap only)
    plot <- ggplot(mat_melt, aes(y = .data[["Var1"]], x = .data[["Var2"]], fill = round(.data[["Weight"]], 2))) +
      geom_tile(lwd = 0.1, color = "black") +
      scale_fill_gradientn(colors = .colorizer(palette, 21)) +
      labs(fill = col.lab) + 
      .themeRepertoire(...) + 
      theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust = 1),
            axis.title = element_blank())
    if (length(unique(mat_melt$Group)) > 1) {
      plot <- plot + facet_wrap(~Group)
    }
  }
  return(plot)
}

#' @rdname percentGeneUsage
#' @param x.axis Gene segments to separate the x-axis, such as `TRAV`, 
#' `TRBD`, `IGKJ`.
#' @param y.axis Variable to separate the y-axis, can be both categorical 
#' or other gene gene segments, such as `TRAV`, `TRBD`, `IGKJ`.
#' @param plot The type of plot to return - heatmap or barplot. 
#' @examples
#' # Visualize paired gene (TRAV-TRAJ) usage as a heatmap
#' vizGenes(combined[1:2],
#'          x.axis = "TRAV",
#'          y.axis = "TRAJ",
#'          group.by = "sample",
#'          summary.fun = "count")
#'
#' # Visualize cross-chain gene pairing (TRBV-TRAV)
#' vizGenes(combined[1:2],
#'          x.axis = "TRBV",
#'          y.axis = "TRAV",
#'          group.by = "sample",
#'          summary.fun = "percent")
#' 
#' @export
vizGenes <- function(input.data,
                     x.axis = "TRBV",
                     y.axis = NULL,
                     group.by = NULL,
                     plot = "heatmap",
                     order.by = NULL,
                     summary.fun = c("percent", "proportion", "count"),
                     export.table = NULL,
                     palette = "inferno",
                     # Deprecated arguments
                     exportTable = NULL) {

  # Handle deprecated arguments
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "vizGenes", default = FALSE)

  summary.fun <- match.arg(summary.fun)
  genes_param <- x.axis
  plot_type_param <- plot
  group_by_param <- group.by
  order_by_param <- order.by 
  
  # Handle y.axis for paired gene analysis
  if (!is.null(y.axis) && grepl("TRA|TRB|TRG|TRD|IGH|IGL|IGK", y.axis)) {
    genes_param <- c(x.axis, y.axis)
    plot_type_param <- "heatmap" 
  } else if (!is.null(y.axis)) {
    # If y.axis is a categorical variable, use it as group.by
    group_by_param <- y.axis
  } else if(is.null(y.axis)) {
    y.axis <- group_by_param
  }
  
  percentGeneUsage(
    input.data = input.data,
    genes = genes_param,
    group.by = group_by_param,
    order.by = order_by_param,
    summary.fun = summary.fun,
    plot.type = plot_type_param,
    export.table = export.table,
    palette = palette
  )
}

#' @rdname percentGeneUsage
#' @param chain The TCR/BCR chain to use. Accepted values: `TRA`, `TRB`, `TRG`, 
#' `TRD`, `IGH`, `IGL` (for both light chains) 
#' @param gene `Vgene`, `Dgene` or `Jgene`
#' @examples
#' 
#' # Quantify and visualize TRA V-gene usage as a heatmap
#' percentGenes(combined,
#'              chain = "TRA",
#'              gene = "Vgene",
#'              group.by = "sample",
#'              summary.fun = "percent")
#'
#' # Quantify TRA J-gene usage and export the table
#' traj_usage_table <- percentGenes(combined,
#'                                  chain = "TRA",
#'                                  gene = "Jgene",
#'                                  group.by = "sample",
#'                                  export.table = TRUE,
#'                                  summary.fun = "count")
#' 
#' @export
percentGenes <- function(input.data,
                         chain = "TRB",
                         gene = "Vgene",
                         group.by = NULL,
                         order.by = NULL,
                         export.table = NULL,
                         summary.fun = c("percent", "proportion", "count"),
                         palette = "inferno",
                         # Deprecated arguments
                         exportTable = NULL) {

  # Handle deprecated arguments
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "percentGenes", default = FALSE)

  summary.fun <- match.arg(summary.fun)
  
  
  gene_segment_prefix <- switch(chain,
                                "TRA" = "TRA", "TRB" = "TRB", "TRG" = "TRG", "TRD" = "TRD",
                                "IGH" = "IGH", "IGL" = "IGL", "IGK" = "IGK",
                                stop("Invalid chain argument.")
  )
  
  gene_type_suffix <- switch(tolower(substr(gene,1,1)),
                             "v" = "V", "d" = "D", "j" = "J",
                             stop("Invalid gene argument. Please use 'V', 'D', or 'J'.")
  )
  
  # Construct the 'genes' argument for percentGeneUsage
  genes_param <- paste0(gene_segment_prefix, gene_type_suffix)
  
  percentGeneUsage(
    input.data = input.data,
    genes = genes_param,
    group.by = group.by,
    order.by = order.by,
    summary.fun = summary.fun,
    plot.type = "heatmap",
    export.table = export.table,
    palette = palette
  )
}

#' @rdname percentGeneUsage
#' @param chain The TCR/BCR chain to use. Accepted values: `TRA`, `TRB`, `TRG`,
#' `TRD`, `IGH`, `IGL` (for both light chains)
#' @examples

#' # Quantify and visualize TRB V-J gene pairings as a heatmap
#' percentVJ(combined[1:2],
#'           chain = "TRB",
#'           group.by = "sample",
#'           summary.fun = "percent")
#'
#' # 2. Quantify TRA V-J gene pairings and export the table
#' trav_traj_table <- percentVJ(combined,
#'                              chain = "TRA",
#'                              group.by = "sample",
#'                              export.table = TRUE,
#'                              summary.fun = "proportion")
#' @export
percentVJ <- function(input.data,
                      chain = "TRB",
                      group.by = NULL,
                      order.by = NULL,
                      summary.fun = c("percent", "proportion", "count"),
                      export.table = NULL,
                      palette = "inferno",
                      # Deprecated arguments
                      exportTable = NULL) {

  # Handle deprecated arguments
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "percentVJ", default = FALSE)

  summary.fun <- match.arg(summary.fun)
  genes_param <- switch(chain,
                        "TRA" = c("TRAV", "TRAJ"),
                        "TRB" = c("TRBV", "TRBJ"),
                        "TRG" = c("TRGV", "TRGJ"),
                        "TRD" = c("TRDV", "TRDJ"), 
                        "IGH" = c("IGHV", "IGHJ"),
                        "IGL" = c("IGLV", "IGLJ"),
                        "IGK" = c("IGKV", "IGKJ"),
                        stop("Invalid chain argument.")
  )
  
  percentGeneUsage(
    input.data = input.data,
    genes = genes_param,
    group.by = group.by,
    order.by = order.by,
    summary.fun = summary.fun,
    plot.type = "heatmap",
    export.table = export.table,
    palette = palette
  )
}

# test script for percentGeneUsage.R - testcases are NOT comprehensive!

# Generate some data to use
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))

combined <- addVariable(combined, 
            variable.name = "Type", 
            variables = rep(c("B", "L"), 4))

test_that("percentGeneUsage returns a ggplot object for single gene heatmap", {
  p <- percentGeneUsage(combined, 
                        genes = "TRBV", 
                        group.by = "sample", 
                        plot.type = "heatmap")
  expect_s3_class(p, "ggplot")
  expect_true("GeomTile" %in% class(p$layers[[1]]$geom))
})

test_that("percentGeneUsage returns a ggplot object for single gene barplot", {
  p <- percentGeneUsage(combined, 
                        genes = "TRBV", 
                        group.by = "sample", 
                        plot.type = "barplot")
  expect_s3_class(p, "ggplot")
  expect_true("GeomBar" %in% class(p$layers[[1]]$geom))
})

test_that("percentGeneUsage returns a ggplot object for paired gene heatmap", {
  p <- percentGeneUsage(combined, 
                        genes = c("TRBV", "TRBJ"), 
                        group.by = "sample", 
                        plot.type = "heatmap")
  expect_s3_class(p, "ggplot")
  expect_true("GeomTile" %in% class(p$layers[[1]]$geom))
})

test_that("percentGeneUsage returns a matrix when export.table = TRUE (single gene)", {
  mat <- percentGeneUsage(combined, 
                          genes = "TRBV", 
                          group.by = "sample", 
                          export.table = TRUE)
  expect_type(mat, "double")
  expect_true(is.matrix(mat))
  expect_equal(colnames(mat), c("P17B", "P17L", "P18B", "P18L", "P19B", "P19L", "P20B", "P20L"))
  expect_true(any(rownames(mat) %in% c("TRBV7-2", "TRBV6-5")))
})

test_that("percentGeneUsage returns a matrix when export.table = TRUE (paired genes)", {
  mat <- percentGeneUsage(combined, 
                          genes = c("TRBV", "TRBJ"), 
                          group.by = "sample", 
                          export.table = TRUE)
  expect_type(mat, "double")
  expect_true(is.matrix(mat))
  expect_equal(colnames(mat), c("P17B", "P17L", "P18B", "P18L", "P19B", "P19L", "P20B", "P20L"))
})

test_that("percentGeneUsage handles different summary.fun (percent, proportion, count)", {
  mat_percent <- percentGeneUsage(combined, 
                                  genes = "TRBV",
                                  group.by = "sample",
                                  summary.fun = "percent", 
                                  export.table = TRUE)
  mat_proportion <- percentGeneUsage(combined, 
                                     genes = "TRBV", 
                                     group.by = "sample",
                                     summary.fun = "proportion", 
                                     export.table = TRUE)
  mat_count <- percentGeneUsage(combined, 
                                genes = "TRBV", 
                                group.by = "sample",
                                summary.fun = "count", 
                                export.table = TRUE)
  
  # Check sums for percentages and proportions
  expect_equal(colSums(mat_percent)[1], c(P17B = 100), tolerance = 1e-6)
  expect_equal(colSums(mat_proportion)[1], c(P17B = 1), tolerance = 1e-6)
  
  expect_equal(mat_proportion * 100, mat_percent, tolerance = 1e-6)
  expect_equal(mat_count["TRBV7-2", "P17B"], 44)
  expect_equal(mat_count["TRBV6-5", "P17B"], 98)
})



test_that("percentGeneUsage handles order.by = custom vector", {
  custom_order <- c("P17L", "P17B")
  p <- percentGeneUsage(combined[1:2], 
                        genes = "TRBV", 
                        group.by = "sample", 
                        order.by = custom_order)
  expect_s3_class(p, "ggplot")
  expect_equal(levels(p$data$Group), custom_order)
})

test_that("percentGeneUsage throws error for invalid genes parameter", {
  expect_error(percentGeneUsage(combined, genes = "InvalidGene"), 
               "Unsupported gene segment format")
  expect_error(percentGeneUsage(combined, genes = c("TRBV", "TRBJ", "TRBC")), 
               "Parameter 'genes' must be a character vector of length 1 or 2")
})

test_that("percentGeneUsage throws error for invalid group.by column", {
  expect_error(percentGeneUsage(combined, genes = "TRBV", 
                                group.by = "non_existent_col"), 
               regexp = "column 'non_existent_col' not found")
})

# --- Alias Function Tests ---

test_that("vizGenes calls percentGeneUsage correctly for single gene heatmap", {
  p <- vizGenes(combined, 
                x.axis = "TRBV", 
                group.by = "sample", 
                plot = "heatmap", 
                summary.fun = "proportion")
  expect_s3_class(p, "ggplot")
  expect_true("GeomTile" %in% class(p$layers[[1]]$geom))
  expect_equal(p$labels$fill, "Proportion")
})

test_that("vizGenes calls percentGeneUsage correctly for single gene barplot", {
  p <- vizGenes(combined, 
                x.axis = "TRBV", 
                group.by = "sample", 
                plot = "barplot")
  expect_s3_class(p, "ggplot")
  expect_true("GeomBar" %in% class(p$layers[[1]]$geom))
})

test_that("vizGenes handles y.axis as categorical variable (maps to group.by)", {
  p <- vizGenes(combined, 
                x.axis = "TRBV", 
                y.axis = "Type", 
                plot = "heatmap")
  expect_s3_class(p, "ggplot")
})

test_that("vizGenes handles y.axis as a gene (maps to paired genes)", {
  p <- vizGenes(combined, 
                x.axis = "TRBV", 
                y.axis = "TRAV", 
                group.by = "sample",
                plot = "heatmap")
  expect_s3_class(p, "ggplot")
  expect_true("GeomTile" %in% class(p$layers[[1]]$geom))
  # Check if it's a paired gene plot (Var1 and Var2 in data)
  expect_true("Var1" %in% colnames(p$data))
  expect_true("Var2" %in% colnames(p$data))
})

test_that("percentGenes calls percentGeneUsage correctly", {
  p <- percentGenes(combined, 
                    chain = "TRB", 
                    gene = "Vgene", 
                    group.by = "sample")
  expect_s3_class(p, "ggplot")
  expect_true("GeomTile" %in% class(p$layers[[1]]$geom))
  # Check if it's a single gene plot (only Var1 in data, and Group)
  expect_true("Var1" %in% colnames(p$data))
  expect_false("Var2" %in% colnames(p$data))
  expect_equal(p$labels$fill, "Percent")
})

test_that("percentGenes handles different summary.fun", {
  p_count <- percentGenes(combined, 
                          chain = "TRB", 
                          gene = "Vgene", 
                          group.by = "sample",
                          summary.fun = "count")
  expect_equal(p_count$labels$fill, "Count")
})

test_that("percentVJ calls percentGeneUsage correctly", {
  p <- percentVJ(combined, chain = "TRB", group.by = "sample")
  expect_s3_class(p, "ggplot")
  expect_true("GeomTile" %in% class(p$layers[[1]]$geom))
  # Check if it's a paired gene plot (Var1 and Var2 in data)
  expect_true("Var1" %in% colnames(p$data))
  expect_true("Var2" %in% colnames(p$data))
  expect_equal(p$labels$fill, "Percent")
})

test_that("percentVJ handles different chains correctly", {
  # For TRA, it should use TRAV and TRAJ
  p_tra <- percentVJ(combined, chain = "TRA", group.by = "sample")
  expect_s3_class(p_tra, "ggplot")
  expect_true(all(grepl("TRAV", levels(p_tra$data$Var1))))
  expect_true(all(grepl("TRAJ", levels(p_tra$data$Var2))))
})

#' Visualize Clonal Relationships as a Chord Diagram
#'
#' This function creates a chord diagram visualization of shared clones between
#' groups using the circlize package. It provides a convenient wrapper around
#' [getCirclize()] that handles the circlize plotting code automatically.
#'
#' @details
#' This function requires the circlize package to be installed. If circlize is

#' not available, the function will return the data that would be used for plotting
#' and provide instructions for manual plotting.
#'
#' The chord diagram shows relationships between groups (sectors) where the width
#' of each chord represents the number or proportion of shared clones between
#' the connected groups.
#'
#' @examples
#' \dontrun{
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list,
#'                         samples = c("P17B", "P17L", "P18B", "P18L",
#'                                     "P19B","P19L", "P20B", "P20L"))
#'
#' # Getting a sample of a Seurat object
#' scRep_example <- get(data("scRep_example"))
#' scRep_example <- combineExpression(combined,
#'                                    scRep_example)
#'
#' # Simple chord diagram
#' vizCirclize(scRep_example, group.by = "seurat_clusters")
#'
#' # Directional chord diagram with arrows
#' vizCirclize(scRep_example,
#'             group.by = "seurat_clusters",
#'             directional = TRUE)
#'
#' # Multi-level grouping
#' scRep_example$Patient <- substring(scRep_example$orig.ident, 1, 3)
#' vizCirclize(scRep_example,
#'             group.by = c("Patient", "seurat_clusters"))
#' }
#'
#' @param sc.data The single-cell object after [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param group.by A column header (or vector of column headers for hierarchical
#' grouping) in the metadata to group the analysis by.
#' @param method The method for calculating link values: `"unique"` (default) counts
#' unique shared clones, `"jaccard"` calculates Jaccard similarity,
#' `"overlap"` calculates overlap coefficient.
#' @param proportion Calculate the relationship by unique clones (`FALSE`, default)
#' or normalized by proportion (`TRUE`).
#' @param directional If `TRUE`, show directional arrows on chords. Default is `FALSE`.
#' @param self.link How to handle self-links. `1` = show as loops, `2` = show as
#' parallel lines. Default is `1`.
#' @param include.self Include self-links (clones within a single group). Default is `TRUE`.
#' @param transparency Transparency of the chord links (0-1). Default is `0.5`.
#' @param link.visible If `FALSE`, hide the chord links and show only sectors.
#' @param annotate.sectors If `TRUE` (default), display sector names.
#' @param min.shared Minimum number of shared clones to include a link (default 0).
#' @param palette Colors to use for sectors - input any [hcl.pals][grDevices::hcl.pals].
#' @param sector.colors Named vector of colors for specific sectors. Overrides palette.
#' @param export.table If `TRUE`, returns the data instead of plotting.
#'
#' @export
#' @concept SC_Functions
#' @return Invisibly returns the circlize data list. If circlize is not installed
#' or `export.table = TRUE`, returns the data that would be used for plotting.
#' @seealso [getCirclize()] for generating the underlying data
#' @author Nick Borcherding
vizCirclize <- function(sc.data,
                        clone.call = NULL,
                        group.by = NULL,
                        method = c("unique", "jaccard", "overlap"),
                        proportion = FALSE,
                        directional = FALSE,
                        self.link = 1,
                        include.self = TRUE,
                        transparency = 0.5,
                        link.visible = TRUE,
                        annotate.sectors = TRUE,
                        min.shared = 0,
                        palette = "inferno",
                        sector.colors = NULL,
                        export.table = FALSE) {

  method <- match.arg(method)

  # Check for circlize package
  if(!requireNamespace("circlize", quietly = TRUE)) {
    message("The circlize package is required for visualization.")
    message("Install it with: install.packages('circlize')")
    message("Returning data for manual plotting instead.")
    export.table <- TRUE
  }

  # Get the circlize data
  circ.data <- getCirclize(
    sc.data = sc.data,
    clone.call = clone.call,
    group.by = group.by,
    method = method,
    proportion = proportion,
    symmetric = !directional,
    include.self = include.self,
    include.metadata = TRUE,
    min.shared = min.shared,
    palette = palette
  )

  # Apply custom sector colors if provided
  if(!is.null(sector.colors)) {
    for(s in names(sector.colors)) {
      if(s %in% names(circ.data$colors)) {
        circ.data$colors[s] <- sector.colors[s]
      }
    }
  }

  if(export.table) {
    return(circ.data)
  }

  # Check if there are links to plot
  if(nrow(circ.data$links) == 0) {
    warning("No links to plot after filtering. Try reducing min.shared or check your data.")
    return(invisible(circ.data))
  }

  # Clear any existing circlize plots
  circlize::circos.clear()

  # Set up circlize parameters
  circlize::circos.par(
    gap.after = 2,
    track.margin = c(0.01, 0.01)
  )

  # Determine annotation track setting
  annotation_track <- if(annotate.sectors) "name" else NULL

  # Create the chord diagram
  if(directional) {
    circlize::chordDiagram(
      circ.data$links,
      grid.col = circ.data$colors,
      transparency = transparency,
      self.link = self.link,
      link.visible = link.visible,
      directional = 1,
      direction.type = "arrows",
      link.arr.type = "big.arrow",
      annotationTrack = annotation_track
    )
  } else {
    circlize::chordDiagram(
      circ.data$links,
      grid.col = circ.data$colors,
      transparency = transparency,
      self.link = self.link,
      link.visible = link.visible,
      annotationTrack = annotation_track
    )
  }

  # Clear circos parameters at the end
  on.exit(circlize::circos.clear(), add = TRUE)

  # Return data invisibly
  invisible(circ.data)
}

#-------------------------------------------
#-------------Background Functions-----------
#-------------------------------------------
# utility functions use camelCase

#-------------------------------------------
#---------Argument Deprecation Helper-------
#-------------------------------------------
# Helper function for soft-deprecating function arguments
# This allows old argument names to still work while warning users
# to switch to the new names
#
# @param old_arg The value passed to the deprecated argument (or NULL)
# @param new_arg The value passed to the new argument (or NULL)
# @param old_name Character string of the deprecated argument name
# @param new_name Character string of the new argument name
# @param func_name Character string of the function name
# @param default The default value if neither argument is provided
# @param version The version in which the argument was deprecated (default "2.10.0")
# @return The value to use (from old_arg if provided, else new_arg, else default)
# @keywords internal

.deprecate_arg <- function(old_arg, new_arg, old_name, new_name,
                           func_name, default = NULL, version = "2.10.0") {
  # Check if old argument was explicitly provided (not NULL/missing)
  old_provided <- !is.null(old_arg)
  new_provided <- !is.null(new_arg)
  
  if (old_provided && new_provided) {
    # Both provided - warn and use new
    lifecycle::deprecate_warn(
      version,
      paste0(func_name, "(", old_name, ")"),
      paste0(func_name, "(", new_name, ")"),
      details = paste0("Both `", old_name, "` and `", new_name,
                       "` were provided. Using `", new_name, "`.")
    )
    return(new_arg)
  } else if (old_provided) {
    # Only old provided - warn and use old
    lifecycle::deprecate_soft(
      version,
      paste0(func_name, "(", old_name, ")"),
      paste0(func_name, "(", new_name, ")")
    )
    return(old_arg)
  } else if (new_provided) {
    # Only new provided - use it
    return(new_arg)
  } else {
    # Neither provided - use default
    return(default)
  }
} 

#-------------------------------------------
#---------Clone Size Binning Helper--------
#-------------------------------------------
# Internal helper function to assign cloneSize bins to a data frame
# Used by both clonalBin() and combineExpression() to ensure DRY principles
#
# @param df Data frame containing clonalProportion and/or clonalFrequency columns
# @param clone.size Named numeric vector of bin thresholds (should already include None = 0)
# @param proportion Logical; if TRUE use clonalProportion, if FALSE use clonalFrequency
# @param format.names Logical; if TRUE, format bin names with ranges. Set to FALSE
#   when clone.size names are already formatted (e.g., when calling on subsets)
# @return A list with two elements:
#   - df: The data frame with cloneSize column added
#   - clone.size: The clone.size vector with formatted names (for factor level creation)
# @keywords internal
.assignCloneSizeBins <- function(df, clone.size, proportion, format.names = TRUE) {

  # Check if names are already formatted (contain " < X <= ")
  names_already_formatted <- any(grepl(" < X <= ", names(clone.size)))

  # Determine which column to use for binning
  cloneRatioColname <- ifelse(proportion, "clonalProportion", "clonalFrequency")

 # Verify the required column exists
  if (!cloneRatioColname %in% colnames(df)) {
    stop(paste0("Column '", cloneRatioColname, "' not found in data frame. ",
                "Ensure clonal frequency/proportion has been calculated."))
  }

  # Auto-adjust upper bin limit if using frequency and max exceeds threshold
  # Only do this if names aren't already formatted (first call)
  if (!names_already_formatted && !proportion) {
    max_freq <- max(na.omit(df[, "clonalFrequency"]))
    if (max_freq > clone.size[length(clone.size)]) {
      clone.size[length(clone.size)] <- max_freq
    }
  }

  # Format bin names with ranges (only if not already formatted)
  if (format.names && !names_already_formatted) {
    for (x in seq_along(clone.size)) {
      names(clone.size)[x] <- paste0(names(clone.size[x]), ' (', clone.size[x - 1],
                                     ' < X <= ', clone.size[x], ')')
    }
  }

  # Initialize cloneSize column
  df$cloneSize <- NA

  # Assign cloneSize bins
  for (i in 2:length(clone.size)) {
    df$cloneSize <- ifelse(
      df[, cloneRatioColname] > clone.size[i - 1] &
        df[, cloneRatioColname] <= clone.size[i],
      names(clone.size[i]),
      df$cloneSize
    )
  }

  return(list(df = df, clone.size = clone.size))
}

.themeRepertoire <- function(base_size = 12,
                             base_family = "sans",
                             grid_lines = "Y",
                             axis_lines = FALSE,
                             legend_position = "right") {
  
  t <- ggplot2::theme_bw(base_size = base_size, base_family = base_family)
  t <- t %+replace%
    ggplot2::theme(
      # Plot titles and caption
      plot.title = ggplot2::element_text(
        size = rel(1.2), hjust = 0, face = "bold",
        margin = ggplot2::margin(b = base_size / 2)
      ),
      plot.subtitle = ggplot2::element_text(
        size = rel(1.0), hjust = 0,
        margin = ggplot2::margin(b = base_size)
      ),
      plot.caption = ggplot2::element_text(
        size = rel(0.8), hjust = 1, color = "grey50",
        margin = ggplot2::margin(t = base_size / 2)
      ),
      
      # Backgrounds and borders
      plot.background = ggplot2::element_rect(fill = "white", color = NA),
      panel.background = ggplot2::element_rect(fill = "white", color = NA),
      panel.border = ggplot2::element_rect(fill = NA, color = "black", linewidth = 0.75),
      
      # Remove all grid lines by default; they will be added back conditionally
      panel.grid.major = ggplot2::element_blank(),
      panel.grid.minor = ggplot2::element_blank(),
      
      # Axis text, titles, and ticks
      axis.title = ggplot2::element_text(size = rel(1.0)),
      axis.text = ggplot2::element_text(size = rel(0.9), color = "black"),
      axis.ticks = ggplot2::element_line(color = "black", linewidth = 0.5),
      
      # Legend customization
      legend.background = ggplot2::element_blank(),
      legend.key = ggplot2::element_blank(),
      legend.title = ggplot2::element_text(size = rel(0.9), face = "bold"),
      legend.text = ggplot2::element_text(size = rel(0.85)),
      legend.position = legend_position,
      
      # Facet (strip) customization
      strip.background = ggplot2::element_rect(fill = "grey90", color = "black", linewidth = 0.75),
      strip.text = ggplot2::element_text(
        size = rel(1.0), face = "bold", color = "black",
        margin = ggplot2::margin(t = base_size / 4, b = base_size / 4)
      )
    )
  
  # Conditionally add major grid lines based on the 'grid_lines' parameter
  grid_lines <- toupper(grid_lines)
  if (grid_lines %in% c("Y", "XY")) {
    t <- t + ggplot2::theme(panel.grid.major.y = ggplot2::element_line(color = "grey85", linewidth = 0.5))
  }
  if (grid_lines %in% c("X", "XY")) {
    t <- t + ggplot2::theme(panel.grid.major.x = ggplot2::element_line(color = "grey85", linewidth = 0.5))
  }
  
  # Conditionally add axis lines
  if (axis_lines) {
    t <- t + ggplot2::theme(axis.line = ggplot2::element_line(color = "black", linewidth = 0.5))
  }
  
  return(t)
}

.processStrings <- function(seq, aa.length) {
  strings <- unlist(strsplit(seq, ";"))
  strings <- strings[!is.na(strings)]
  strings <- strings[strings != "NA"]
  strings <- strings[nchar(strings) <= aa.length]
  return(strings)
}

# function needed for makeScrepSeurat
#' @keywords internal
.seuratExtractField <- function(string, field = 1, delim = "_") {
  fields <- as.numeric(
    x = unlist(x = strsplit(x = as.character(x = field), split = ","))
  )
  if (length(x = fields) == 1) {
    return(strsplit(x = string, split = delim)[[1]][field])
  }
  return(paste(
    strsplit(x = string, split = delim)[[1]][fields], collapse = delim
  ))
}

# seurat's command adding but if a param is a dataframe or list of dataframes,
# completely omits them to save memory. 
#' @keywords internal
.makeScrepSeurat <- function(call_time, assay) {
  
  if (as.character(x = sys.calls()[[1]])[1] == "do.call") {
    call_string <- deparse(expr = sys.calls()[[1]])
    command_name <- as.character(x = sys.calls()[[1]])[2]
  } else {
    command_name <- as.character(
      x = deparse(expr = sys.calls()[[sys.nframe() - 1]])
    )
    command_name <- gsub(
      pattern = "\\.Seurat",
      replacement = "",
      x = command_name
    )
    call_string <- command_name
    command_name <- .seuratExtractField(
      string = command_name,
      field = 1,
      delim = "\\("
    )
  }
  
  argnames <- names(x = formals(fun = sys.function(which = sys.parent(n = 1))))
  argnames <- grep(
    pattern = "object",
    x = argnames,
    invert = TRUE,
    value = TRUE
  )
  argnames <- grep(
    pattern = "anchorset",
    x = argnames,
    invert = TRUE,
    value = TRUE
  )
  argnames <- grep(
    pattern = "\\.\\.\\.",
    x = argnames,
    invert = TRUE,
    value = TRUE
  )
  
  params <- list()
  p.env <- parent.frame(n = 1)
  argnames <- intersect(x = argnames, y = ls(name = p.env))
  for (arg in argnames) {
    param_value <- get(x = arg, envir = p.env)
    if (.is.seurat.object(param_value) || .is.df.or.list.of.df(param_value)) {
      next
    }
    params[[arg]] <- param_value
  }
  
  command_name <- sub(
    pattern = "[\\.]+$",
    replacement = "",
    x = command_name,
    perl = TRUE
  )
  command_name <- sub(
    pattern = "\\.\\.", replacement = "\\.", x = command_name, perl = TRUE
  )
  
  # return the command object
  methods::new(
    Class = 'SeuratCommand',
    name = command_name,
    params = params,
    time.stamp = call_time,
    call.string = call_string,
    assay.used = assay
  )
}


.alphanumericalSort <- function(x, ignore.case = TRUE) {
  if (!is.character(x)) {
    message("Input 'x' is not a character vector. Attempting to convert.")
    x <- as.character(x)
  }
  unique_elements <- unique(x)
  alpha_part <- gsub("[0-9]+", "", unique_elements, perl = TRUE)
  if (ignore.case) {
    alpha_part <- tolower(alpha_part)
  }
  numeric_part <- suppressWarnings(as.numeric(gsub("[^0-9]", "", unique_elements)))
  sorted_levels <- unique_elements[
    order(alpha_part, numeric_part)
  ]
  
  return(sorted_levels)
}


.toCapitilize <- function(name) {
  gsub("([\\w])([\\w]+)", "\\U\\1\\L\\2", name, perl = TRUE)
}

.orderingFunction <- function(vector,
                              group.by,
                              data.frame) {
  if (length(vector) == 1 && vector == "alphanumeric") {
    unique_elements <- as.character(unique(data.frame[, group.by]))
    sorted_levels <- unique_elements[
      order(gsub("[0-9]", "", unique_elements),
            as.numeric(gsub("[^0-9]", "", unique_elements)))
    ]
    # Apply the sorted levels to the factor
    data.frame[, group.by] <- factor(data.frame[, group.by], levels = sorted_levels)
  } else {
    data.frame[, group.by] <- factor(data.frame[, group.by], levels = vector)
  }
  
  return(data.frame)
}

# Get position of chain
.chainPositionParser <- function(chain) {
  chain1 <- toupper(chain) #to just make it easier
  if (chain1 %in% c("TRA", "TRG", "IGH", "HEAVY")) {
    x <- 1
  } else if (chain1 %in% c("TRB", "TRD", "IGL", "IGK", "LIGHT")) {
    x <- 2
  } else {
    # It's good practice to use stop() for fatal errors
    stop("'", chain, "' is not a valid entry for the `chain` argument.
         Please use 'TRA', 'TRG', 'IGH', 'TRB', 'TRD', 'IGL', 'IGK', 'Heavy', or 'Light'.")
  }
  return(x)
}

#Use to shuffle between chains
#' @keywords internal
#' @author Ye-Lin Son, Nick Borcherding
.offTheChain <- function(dat, chain, cloneCall, check = TRUE) {
  # Get position of chain
  x <- .chainPositionParser(chain)
  

  if (check) {
    if(chain == "Light") {
      chain.pattern <- "IGL|IGK"
    } else if (chain == "Heavy") {
      chain.pattern <- "IGH"
    } else {
      chain.pattern <- chain
    }
    
    split_genes <- strsplit(dat[, "CTgene"], "_")
    gene_elements <- sapply(split_genes, `[`, x)
    
    chain.check <- substr(gene_elements, 1, 3)
    
    # The following lines for handling NAs remain the same
    chain.check[chain.check == "NA"] <- NA
    chain.check[is.na(chain.check)] <- NA #
    chain.check[chain.check == "Non"] <- NA
    
    any.alt.chains <- which(!is.na(chain.check) & !grepl(chain.pattern, chain.check))
    
    if (length(any.alt.chains) > 0) {
      dat <- dat[-any.alt.chains, ]
      split_clones <- strsplit(dat[, cloneCall], "_")
    } else {
      split_clones <- strsplit(dat[, cloneCall], "_")
    }
  } else {
    split_clones <- strsplit(dat[, cloneCall], "_")
  }
  
  # Extract the relevant part of the cloneCall
  dat[, cloneCall] <- sapply(split_clones, `[`, x)
  
  # Handle NA-like values.
  dat[, cloneCall][dat[, cloneCall] == "NA"] <- NA
  dat[, cloneCall][is.na(dat[, cloneCall])] <- NA
  
  return(dat)
}

#' @importFrom stats na.omit ave
.cloneCounter <- function(meta,
                           group.by,
                           cloneCall) {
  
  meta_filtered <- na.omit(meta[, c(group.by, cloneCall)])
  clone.table <- as.data.frame(table(meta_filtered), responseName = "n")
  clone.table <- clone.table[order(clone.table$n, decreasing = TRUE), ]
  clone.table$group.sum <- ave(clone.table$n, clone.table[, group.by], FUN = sum)
  clone.table$clone.sum <- ave(clone.table$n, clone.table[, cloneCall], FUN = sum)
  rownames(clone.table) <- NULL
  return(clone.table)
}

#Pulling a color palette for visualizations
#' @importFrom grDevices hcl.colors
#' @keywords internal
.colorizer <- function(palette = "inferno", 
                        n= NULL) {
  colors <- hcl.colors(n=n, palette = palette, fixup = TRUE)
  return(colors)
}



#Remove list elements that contain all NA values
#' @keywords internal
.checkBlanks <- function(df, cloneCall) {
  count <- NULL
  for (i in seq_along(df)) {
    # First, check if there are no rows
    if (nrow(df[[i]]) == 0) {
      count <- c(i, count)
    } 
    # If there are rows, then proceed with blank checks
    else if (length(df[[i]][,cloneCall]) == length(which(is.na(df[[i]][,cloneCall]))) |
             length(which(!is.na(df[[i]][,cloneCall]))) == 0) {
      count <- c(i, count)
    } else {
      next()
    }
  }
  if (!is.null(count)) {
    df <- df[-count]
  }
  return(df)
}

#reshuffling df
#' @keywords internal
.groupList <- function(df, group.by) {
  if(inherits(df, "list")) {
    df <- do.call(rbind, df)
  }
    df <- split(df, df[,group.by])
    return(df)
}

# Ensure df is in list format
#' @keywords internal
.checkList <- function(df) {
  df <- tryCatch(
      {
          if (!inherits(df, "list")) {
              df <- list(df)
          }
          df
      },
      error = function(e) {
          stop(
              "Please ensure that the input consists of at least one data frame"
          )
      }
    )
    df
}

#' @keywords internal
.checkContigs <- function(df) {
    df <- lapply(seq_len(length(df)), function(x) {
        df[[x]] <- if(!is.data.frame(df[[x]])) as.data.frame(df[[x]]) else df[[x]]
        df[[x]][df[[x]] == ""] <- NA
        df[[x]]
    })
    df
}

#' @keywords internal
.bound.input.return <- function(df) {
  if (.is.seurat.or.se.object(df)) {
    return(.grabMeta(df))
  } 
  bind_rows(df, .id = "element.names")
}

#' @keywords internal
.list.input.return <- function(df, split.by) {
    if (.is.seurat.or.se.object(df)) {
        if(is.null(split.by)){
            split.by <- "ident"
        }
        df <- .expression2List(df, split.by)
    } 
    df
}

#Get UMAP or other coordinates
#' @importFrom SingleCellExperiment reducedDim reducedDimNames
#' @keywords internal
.getCoord <- function(sc, reduction) { 
  if (is.null(reduction)) {
    reduction <- "pca"
  }
  if (.is.seurat.object(sc)) {
    if (!reduction %in% names(sc@reductions)) {
      stop(paste("Reduction", reduction, "not found in Seurat object."))
    }
    coord <- sc@reductions[[reduction]]@cell.embeddings
  } else if (.is.se.object(sc)) {
    if (!reduction %in% reducedDimNames(sc)) {
      stop(paste("Reduction", reduction, "not found in SingleCellExperiment object."))
    }
    coord <- reducedDim(sc, reduction)
  }
  return(coord)
}

#This is to check the single-cell expression object
#' @keywords internal
.checkSingleObject <- function(sc) {
    if (!.is.seurat.or.se.object(sc)){
        stop("Object indicated is not of class 'Seurat' or 
            'SummarizedExperiment', make sure you are using
            the correct data.") 
    }
}

#This is to grab the meta data from a seurat or SCE object
#' @importFrom SingleCellExperiment colData 
#' @importFrom SeuratObject Idents
#' @importFrom methods slot
#' @keywords internal
.grabMeta <- function(sc) {
  messString <- c("Meta data contains an 'ident' column and will likely result
                  in errors downstream.")
    if (.is.seurat.object(sc)) {
        meta <- data.frame(sc[[]], slot(sc, "active.ident"))
        if("ident" %in% colnames(meta)) {
          message(messString)
        }
        colnames(meta)[length(meta)] <- "ident"
    } else if (.is.se.object(sc)){
        meta <- data.frame(colData(sc))
        if("ident" %in% colnames(meta)) {
          message(messString)
        }
        rownames(meta) <- sc@colData@rownames
        clu <- which(colnames(meta) == "label") # as set by colLabels()
        colnames(meta)[clu] <- "ident"
    } else {
        stop("Object indicated is not of class 'Seurat' or 
            'SummarizedExperiment', make sure you are using
            the correct data.")
    }
    return(meta)
}

#This is to add the sample and ID prefixes for combineBCR()/TCR()
#' @keywords internal
.modifyBarcodes <- function(df, samples, ID) {
    out <- NULL
    for (x in seq_along(df)) {
        data <- df[[x]] 
        if (!is.null(ID)){
          data$barcode <- paste0(samples[x], "_", ID[x], "_", data$barcode)
        } else {
          data$barcode <- paste0(samples[x], "_", data$barcode)
        }
        out[[x]] <- data }
    return(out)
}

#Removing barcodes with NA recovered
#' @keywords internal
#' @importFrom stats na.omit
.removingNA <- function(final) {
    for(i in seq_along(final)) {
        final[[i]] <- na.omit(final[[i]])}
    return(final)
}

#Removing barcodes with > 2 clones recovered
#' @keywords internal
.removingMulti <- function(final){
    for(i in seq_along(final)) {
        final[[i]] <- filter(final[[i]], !grepl(";",CTnt))}
    return(final)
}

#Removing extra clones in barcodes with > 2 productive contigs
#' @importFrom stats ave
#' @keywords internal
.filteringMulti <- function(x) {
  x <- as.data.frame(x)
  x <- x[order(x$reads, decreasing = TRUE), ]
  x <- x[!duplicated(x[, c("barcode", "chain")]), ]
  contig_counts <- as.data.frame(table(x$barcode))
  table <- subset(contig_counts, Freq > 2)
  if (nrow(table) > 0) {
    barcodes <- as.character(unique(table$Var1))
    multichain <- NULL
    multi_subset <- subset(x, barcode %in% barcodes)
    multi_subset <- multi_subset[order(multi_subset$barcode, multi_subset$reads, decreasing = TRUE), ]
    multi_subset$rank <- ave(multi_subset$reads, multi_subset$barcode, FUN = seq_along)
    multichain <- subset(multi_subset, rank <= 2)
    multichain$rank <- NULL
    x <- subset(x, !barcode %in% barcodes)
    x <- rbind(x, multichain)
  }
  x <- x[order(x$barcode, x$chain), ]
  rownames(x) <- NULL 
  return(x)
}


#Filtering NA contigs out of single-cell expression object
#' @importFrom SingleCellExperiment colData
#' @keywords internal
.filteringNA <- function(sc) {
  meta <- .grabMeta(sc)
  evalNA <- data.frame(indicator = meta[, "cloneSize"], 
                       row.names = rownames(meta))
  evalNA$indicator <- ifelse(is.na(evalNA$indicator), 0, 1)
  if (.is.se.object(sc)) {
    colData(sc)[["evalNA"]] <- evalNA
    return(sc[, !is.na(sc$cloneSize)])
  } else {
    pos <- which(evalNA[, "indicator"] != 0)
    sc <- subset(sc, cells = rownames(evalNA)[pos])
    return(sc)
  }
}

#Organizing list of contigs for visualization
#' @keywords internal
#' @author Justin Reimertz
.parseContigs <- function(df, i, names, cloneCall) {
	data <- df[[i]] %>% 
		# Count the abundance of each clone per specified cloneCall category and
		# cell type
		group_by("{cloneCall}" := df[[i]][[cloneCall]]) %>%
		# Add sample ids to df
		mutate(values = names[i], Abundance = n()) %>%
		select(all_of(cloneCall), all_of("values"), all_of("Abundance")) %>%
		ungroup() 
	return(data)
}

# This is to help sort the type of clone data to use
#' @keywords internal
.theCall <- function(df, x, check.df = TRUE, silent = FALSE) {
    x <- .convertClonecall(x, silent)
    if(check.df) {
      if(inherits(df, "list") & !any(colnames(df[[1]]) %in% x)) {
        stop("Check the clonal variable (cloneCall) being used in the function, it does not appear in the data provided.")
      } else if (inherits(df, "data.frame") & !any(colnames(df) %in% x)) {
        stop("Check the clonal variable (cloneCall) being used in the function, it does not appear in the data provided.")
      }
    }
    return(x)
}

.chainConverter <- function(chain) {
  if(chain == "IGH") {
    chain <- "Heavy"
  } else if (chain %in% c("IGL", "IGK")) {
    chain <- "Light"
  }
  return(chain)
}

# helper for .theCall # Qile: on second thought - converting to x to lowercase may be a bad idea...
.convertClonecall <- function(x, silent) {

  clonecall_dictionary <- list(
    "gene" = "CTgene",
		"genes" = "CTgene",
		"ctgene" = "CTgene",
		"ctstrict" = "CTstrict",
		"nt" = "CTnt",
		"nucleotide" = "CTnt",
		"nucleotides" = "CTnt",
		"ctnt" = "CTnt",
		"aa" = "CTaa",
		"amino" = "CTaa",
		"ctaa" = "CTaa",
		"gene+nt" = "CTstrict",
		"strict" = "CTstrict",
		"ctstrict" = "CTstrict"
	)

	if (!is.null(clonecall_dictionary[[tolower(x)]])) {
		return(clonecall_dictionary[[tolower(x)]])
	} else {
	  if(!silent) {
		  message("A custom variable ", x, " will be used to call clones")
	  }
		return(x)
	}
}


# Assigning positions for TCR contig data
# Used to be .parseTCR(Con.df, unique_df, data2) in v1
# but now also constructs Con.df and runs the parseTCR algorithm on it, all in Rcpp
#' @author Gloria Kraus, Nick Bormann, Nicky de Vrij, Nick Borcherding, Qile Yang
#' @keywords internal
.constructConDfAndParseTCR <- function(data2) {
  rcppConstructConDfAndParseTCR(
    dplyr::arrange(data2, chain, cdr3_nt),
    uniqueData2Barcodes = unique(data2$barcode)
  )
}

# Assigning positions for BCR contig data
# Now assumes lambda over kappa in the context of only 2 light chains
#' @author Gloria Kraus, Nick Bormann, Nick Borcherding, Qile Yang
#' @keywords internal
.parseBCR <- function (Con.df, unique_df, data2) {
  barcodeIndex <- rcppConstructBarcodeIndex(unique_df, data2$barcode)
  for (y in seq_along(unique_df)) {
    location.i <- barcodeIndex[[y]]
    
    for (z in seq_along(location.i)) {
      where.chain <- data2[location.i[z],"chain"]
      
      if (where.chain == "IGH") {
        if(is.na(Con.df[y,"IGH"])) {
          Con.df[y,heavy_lines] <- data2[location.i[z],h_lines]
        } else {
          Con.df[y,heavy_lines] <- paste(Con.df[y, heavy_lines],
                                         data2[location.i[z],h_lines],sep=";") 
        }
      } else if (where.chain == "IGK") {
        if(is.na(Con.df[y,"IGLC"])) {
          Con.df[y,light_lines] <- data2[location.i[z],k_lines]
        } else {
          Con.df[y,light_lines] <- paste(Con.df[y, light_lines],
                                         data2[location.i[z],k_lines],sep=";") 
        }
      }else if (where.chain == "IGL") {
        if(is.na(Con.df[y,"IGLC"])) {
          Con.df[y,light_lines] <- data2[location.i[z],l_lines]
        } else {
          Con.df[y,light_lines] <- paste(Con.df[y, light_lines],
                                         data2[location.i[z],l_lines],sep=";") 
        }
      }
    }
    
  }
  return(Con.df)
}



# Producing a data frame to visualize for lengthContig()
#' @keywords internal
#' @importFrom stats na.omit
.lengthDF <- function(df, cloneCall, chain, group) {
  Con.df <- NULL
  names <- names(df)
  
  if (identical(chain, "both")) {
    for (i in seq_along(df)) {
      clones <- gsub("_NA", "", df[[i]][, cloneCall])
      clones <- gsub("NA_", "", clones)
      length <- nchar(gsub("_", "", clones))
      val <- df[[i]][, cloneCall]
      
      if (!is.null(group)) {
        cols <- df[[i]][, group]
        data <- na.omit(data.frame(length, val, cols, names[i]))
        colnames(data) <- c("length", "CT", group, "values")
        Con.df <- rbind(Con.df, data) 
      } else {
        data <- na.omit(data.frame(length, val, names[i]))
        colnames(data) <- c("length", "CT", "values")
        Con.df <- rbind(Con.df, data)
      }
    }
  } else {
    for (x in seq_along(df)) {
      strings <- df[[x]][, cloneCall]
      val1 <- strings
      split_strings <- strsplit(val1, ";")
      val1 <- sapply(split_strings, `[`, 1)
      chain1 <- nchar(val1)
      
      if (!is.null(group)) {
        cols1 <- df[[x]][, group]
        data1 <- data.frame(chain1, val1, names[x], chain, cols1)
        colnames(data1) <- c("length", "CT", "values", "chain", group)
      } else { 
        data1 <- data.frame(chain1, val1, names[x], chain)
        colnames(data1) <- c("length", "CT", "values", "chain")
      }
      
      data <- na.omit(data1)
      data <- data[data$CT != "NA" & data$CT != "" & !is.na(data$CT), ]
      Con.df <- rbind(Con.df, data)
    }
  }
  return(Con.df)
}
# General combination of nucleotide, aa, and gene sequences for T/B cells
#' @keywords internal
.assignCT <- function(cellType, Con.df) {
    if (cellType == "T") {
        Con.df$CTgene <- paste(Con.df$TCR1, Con.df$TCR2, sep="_")
        Con.df$CTnt <- paste(Con.df$cdr3_nt1, Con.df$cdr3_nt2, sep="_")
        Con.df$CTaa <- paste(Con.df$cdr3_aa1, Con.df$cdr3_aa2, sep="_")
        Con.df$CTstrict <- paste0(Con.df$TCR1, ";", Con.df$cdr3_nt1, "_",
            Con.df$TCR2, ";", Con.df$cdr3_nt2)
    } else { # assume cellType = B
        Con.df$CTgene <- paste(Con.df$IGH, Con.df$IGLC, sep="_")
        Con.df$CTnt <- paste(Con.df$cdr3_nt1, Con.df$cdr3_nt2, sep="_")
        Con.df$CTaa <- paste(Con.df$cdr3_aa1, Con.df$cdr3_aa2, sep="_")
    }
    return(Con.df)
}


# Sorting the V/D/J/C gene sequences for T and B cells
#' @keywords internal
.makeGenes <- function(cellType, data2, chain1, chain2) {

  replace_na_with_empty <- function(x) {
    x[is.na(x)] <- ""
    return(x)
  }
  
  if (cellType == "T") {
    # Filter for T-cell chains
    data2 <- data2[data2$chain %in% c("TRA", "TRB", "TRG", "TRD"), ]
    
    # Create TCR1 column for TRA/TRG chains
    is_tcr1_chain <- data2$chain %in% c("TRA", "TRG")
    data2$TCR1 <- ifelse(
      is_tcr1_chain,
      paste(
        replace_na_with_empty(data2$v_gene),
        replace_na_with_empty(data2$j_gene),
        replace_na_with_empty(data2$c_gene),
        sep = "."
      ),
      NA
    )
    
    # Create TCR2 column for TRB/TRD chains
    is_tcr2_chain <- data2$chain %in% c("TRB", "TRD")
    data2$TCR2 <- ifelse(
      is_tcr2_chain,
      paste(
        replace_na_with_empty(data2$v_gene),
        replace_na_with_empty(data2$d_gene),
        replace_na_with_empty(data2$j_gene),
        replace_na_with_empty(data2$c_gene),
        sep = "."
      ),
      NA
    )
    
  } else if (cellType == "B") {
    # Filter for B-cell chains
    data2 <- data2[data2$chain %in% c("IGH", "IGK", "IGL"), ]
    
    data2$IGHct <- NA
    data2$IGKct <- NA
    data2$IGLct <- NA
    
    # Populate IGHct for IGH chains
    is_heavy <- data2$chain == "IGH" & !is.na(data2$chain)
    data2$IGHct[is_heavy] <- paste(
      replace_na_with_empty(data2$v_gene[is_heavy]),
      replace_na_with_empty(data2$d_gene[is_heavy]),
      replace_na_with_empty(data2$j_gene[is_heavy]),
      replace_na_with_empty(data2$c_gene[is_heavy]),
      sep = "."
    )
    
    # Populate IGKct for IGK chains
    is_kappa <- data2$chain == "IGK" & !is.na(data2$chain)
    data2$IGKct[is_kappa] <- paste(
      replace_na_with_empty(data2$v_gene[is_kappa]),
      replace_na_with_empty(data2$j_gene[is_kappa]),
      replace_na_with_empty(data2$c_gene[is_kappa]),
      sep = "."
    )
    
    # Populate IGLct for IGL chains
    is_lambda <- data2$chain == "IGL" & !is.na(data2$chain)
    data2$IGLct[is_lambda] <- paste(
      replace_na_with_empty(data2$v_gene[is_lambda]),
      replace_na_with_empty(data2$j_gene[is_lambda]),
      replace_na_with_empty(data2$c_gene[is_lambda]),
      sep = "."
    )
  }
  return(data2)
}




# Pulling meta data 
#' @keywords internal
.expression2List <- function(sc, split.by) {
  .checkSingleObject(sc)
  meta <- .grabMeta(sc)
  if(is.null(split.by)){
    split.by <- "ident"
  }
  unique_elements <- as.character(unique(meta[, split.by]))
  sorted_unique <- unique_elements[
    order(gsub("[0-9]", "", unique_elements),
          as.numeric(gsub("[^0-9]", "", unique_elements)))
  ]
  df <- list()
  
  for (i in seq_along(sorted_unique)) {
    current_level <- sorted_unique[i]
    
    # Filter metadata for the current level
    temp_df <- meta[meta[, split.by] == current_level & !is.na(meta[, split.by]), ]
    
    # Filter out rows where cloneSize is NA, checking if the column exists first
    if ("cloneSize" %in% colnames(temp_df)) {
      temp_df <- temp_df[!is.na(temp_df$cloneSize), ]
    }
    
    df[[i]] <- temp_df
  }
  
  names(df) <- sorted_unique
  return(df)
}

# Making lists for single-cell object, check blanks and apply chain filter
#' @keywords internal
.dataWrangle <- function(df, split.by, cloneCall, chain) {
  df <- .list.input.return(df, split.by)
  df <- .checkBlanks(df, cloneCall)
  for (i in seq_along(df)) {
    if (chain != "both") {
      df[[i]] <- .offTheChain(df[[i]], chain, cloneCall)
    }
  }
  return(df)
}


#-------------------------------------------
#-----------------Type Checks---------------
#-------------------------------------------
# Separate from utility functions with dot designations
.is.seurat.object <- function(obj) inherits(obj, "Seurat")

.is.se.object <- function(obj) inherits(obj, "SummarizedExperiment")

.is.seurat.or.se.object <- function(obj) {
  .is.seurat.object(obj) || .is.se.object(obj)
}

.is.named.numeric <- function(obj) {
  is.numeric(obj) && !is.null(names(obj))
}

.is.df.or.list.of.df <- function(x) {
  if (is.data.frame(x)) {
    return(TRUE)
  } else if (is.list(x)) {
    if (length(x) == 0) {
      return(FALSE)
    }
    return(all(sapply(x, is.data.frame)))
  } else {
    return(FALSE)
  }
}

# test script for utils.R - testcases are NOT comprehensive!

# Mock data frame simulating contig information
mock_df_contigs <- data.frame(
  barcode = rep(c("bc1", "bc2", "bc3", "bc4"), each = 2),
  chain = rep(c("TRA", "TRB"), 4),
  reads = c(100, 90, 200, 180, 50, 40, 300, 290),
  CTgene = c("TRAV1_TRBV1", "TRAV1_TRBV1", "TRAV2_TRBV2", "TRAV2_TRBV2", 
             "TRAV3_TRBV3", "TRAV3_TRBV3", "TRAV4_TRBV4;TRAV5_TRBV5", "TRAV4_TRBV4"),
  CTnt = c("AAA_CCC", "AAA_CCC", "GGG_TTT", "GGG_TTT", 
           "CCC_AAA", "CCC_AAA", "GGG_TTT;AAA_CCC", "GGG_TTT"),
  CTaa = c("K_C", "K_C", "G_F", "G_F", "P_K", "P_K", "G_F;K_C", "G_F"),
  CTstrict = c("TRAV1;AAA_TRBV1;CCC", "TRAV1;AAA_TRBV1;CCC", "TRAV2;GGG_TRBV2;TTT", "TRAV2;GGG_TRBV2;TTT",
               "TRAV3;CCC_TRBV3;AAA", "TRAV3;CCC_TRBV3;AAA", "TRAV4;GGG;TRAV5_TRBV4;TTT", "TRAV4;GGG_TRBV4;TTT"),
  sample = rep(c("s1", "s2"), each = 4),
  cluster = rep(c("c1", "c1", "c2", "c2"), 2)
)

# Mock metadata for clone counting
mock_meta <- data.frame(
  barcode = paste0("cell", 1:10),
  group = rep(c("group1", "group2"), each = 5),
  CTnt = c("AAA", "AAA", "GGG", "CCC", NA, "AAA", "GGG", "GGG", "TTT", "TTT"),
  clone.size = c(3, 3, 2, 1, NA, 3, 2, 2, 2, 2)
)

# Mock list of data frames
mock_df_list <- list(
  sample1 = data.frame(
    barcode = c("s1_bc1", "s1_bc2"),
    CTnt = c("AAA", "GGG"),
    CTaa = c("K", "G")
  ),
  sample2 = data.frame(
    barcode = c("s2_bc1"),
    CTnt = c("TTT"),
    CTaa = c("F")
  ),
  sample3_all_na = data.frame(
    barcode = c("s3_bc1", "s3_bc2"),
    CTnt = c(NA, NA),
    CTaa = c(NA, NA)
  ),
  sample4_empty = data.frame()
)

# Mock Seurat/SCE objects - simplified S4 objects for testing
setClass("mock_seurat", representation(meta.data = "data.frame", active.ident = "factor"))
setClass("mock_sce", representation(colData = "DataFrame"))



test_that(".toCapitilize works correctly", {
  expect_equal(.toCapitilize("hello"), "Hello")
  expect_equal(.toCapitilize("WORLD"), "World")
  expect_equal(.toCapitilize("miXeD"), "Mixed")
  expect_equal(.toCapitilize(""), "")
  expect_equal(.toCapitilize(c("one", "two")), c("One", "Two"))
})

test_that(".orderingFunction works correctly", {
  df <- data.frame(group = c("g2", "g10", "g1"))
  
  # Test alphanumeric ordering
  df_alpha <- .orderingFunction(df, group.by = "group", vector = "alphanumeric")
  expect_equal(levels(df_alpha$group), c("g1", "g2", "g10"))
  
  # Test custom vector ordering
  custom_order <- c("g10", "g1", "g2")
  df_custom <- .orderingFunction(df, group.by = "group", vector = custom_order)
  expect_equal(levels(df_custom$group), custom_order)
})

test_that(".chainPositionParser works correctly", {
  expect_equal(.chainPositionParser("TRA"), 1)
  expect_equal(.chainPositionParser("trb"), 2)
  expect_equal(.chainPositionParser("IGH"), 1)
  expect_equal(.chainPositionParser("IGL"), 2)
  
  # Test error on invalid input
  expect_error(.chainPositionParser("INVALID"), 
               "'INVALID' is not a valid entry for the `chain` argument.")
})

test_that(".offTheChain works correctly", {
  df <- data.frame(
    barcode = c("bc1", "bc2", "bc3"),
    CTgene = c("TRA.V1_TRB.V1", "IGH.V2_IGL.V2", "TRA.V3_NA"),
    CTaa = c("K_C", "G_F", "P_NA")
  )
  
  # Test TRA chain extraction
  df_tra <- .offTheChain(df, chain = "TRA", cloneCall = "CTaa", check = FALSE)
  expect_equal(df_tra$CTaa, c("K", "G", "P"))
  
  # Test TRB chain extraction
  df_trb <- .offTheChain(df, chain = "TRB", cloneCall = "CTaa", check = FALSE)
  expect_equal(df_trb$CTaa, c("C", "F", NA))
})

test_that(".cloneCounter works correctly", {
  counts <- .cloneCounter(mock_meta, group.by = "group", cloneCall = "CTnt")
  
  # Check dimensions
  expect_equal(nrow(counts), 8)
  expect_equal(ncol(counts), 5)
  
  # Check calculations for a specific clone
  aaa_group1 <- subset(counts, group == "group1" & CTnt == "AAA")
  expect_equal(aaa_group1$n, 2)
  expect_equal(aaa_group1$group.sum, 4)
  expect_equal(aaa_group1$clone.sum, 3)
})

test_that(".colorizer works correctly", {
  # Test default palette
  colors <- .colorizer(n = 5)
  expect_type(colors, "character")
  expect_length(colors, 5)
  
  # Test different palette
  colors_viridis <- .colorizer(palette = "viridis", n = 10)
  expect_type(colors_viridis, "character")
  expect_length(colors_viridis, 10)
  expect_false(identical(colors, colors_viridis[1:5]))
})

test_that(".checkBlanks works correctly", {
  filtered_list <- .checkBlanks(mock_df_list, cloneCall = "CTnt")
  
  # Expecting sample3 (all NA) and sample4 (empty) to be removed
  expect_length(filtered_list, 2)
  expect_equal(names(filtered_list), c("sample1", "sample2"))
})

test_that(".groupList works correctly", {
  df_bound <- bind_rows(mock_df_list[1:2], .id = "sample")
  df_grouped <- .groupList(df_bound, group.by = "sample")
  
  expect_type(df_grouped, "list")
  expect_length(df_grouped, 2)
  expect_equal(names(df_grouped), c("sample1", "sample2"))
  expect_equal(nrow(df_grouped$sample1), 2)
})

test_that(".checkList works correctly", {
  # Test with a data frame
  df <- data.frame(a = 1)
  expect_type(.checkList(df), "list")
  expect_length(.checkList(df), 1)
  
  # Test with a list
  ls <- list(df)
  expect_identical(.checkList(ls), ls)
})

test_that(".checkContigs works correctly", {
  df_list <- list(data.frame(a = c(1, ""), b = c("", 2)))
  checked_list <- .checkContigs(df_list)
  
  expect_true(is.na(checked_list[[1]][2, "a"]))
  expect_true(is.na(checked_list[[1]][1, "b"]))
})


test_that(".modifyBarcodes works correctly", {
  df <- list(data.frame(barcode = "bc1"), data.frame(barcode = "bc2"))
  
  # Test with samples only
  modified <- .modifyBarcodes(df, samples = c("s1", "s2"), ID = NULL)
  expect_equal(modified[[1]]$barcode, "s1_bc1")
  expect_equal(modified[[2]]$barcode, "s2_bc2")
  
  # Test with samples and ID
  modified_id <- .modifyBarcodes(df, samples = c("s1", "s2"), ID = c("id1", "id2"))
  expect_equal(modified_id[[1]]$barcode, "s1_id1_bc1")
  expect_equal(modified_id[[2]]$barcode, "s2_id2_bc2")
})

test_that(".removingNA works correctly", {
  df <- list(data.frame(a = c(1, NA), b = c(3, 4)))
  cleaned <- .removingNA(df)
  expect_equal(nrow(cleaned[[1]]), 1)
  expect_equal(cleaned[[1]]$a, 1)
})

test_that(".removingMulti works correctly", {
  df <- list(data.frame(CTnt = c("A;B", "C", "D;E")))
  cleaned <- .removingMulti(df) 
  expect_equal(nrow(cleaned[[1]]), 1)
  expect_equal(cleaned[[1]]$CTnt, "C")
})

test_that(".filteringMulti works correctly", {
  df <- data.frame(
    barcode = c("bc1", "bc1", "bc1", "bc2", "bc2"),
    chain = c("TRA", "TRB", "TRG", "TRA", "TRB"),
    reads = c(100, 90, 80, 200, 180)
  )
  filtered <- .filteringMulti(df)
  
  expect_equal(nrow(filtered), 4)
  expect_false("TRG" %in% subset(filtered, barcode == "bc1")$chain)
  expect_equal(subset(filtered, barcode == "bc1" & chain == "TRA")$reads, 100)
})

test_that(".theCall and .convertClonecall work together", {
  

  # Test dictionary conversion
  expect_equal(.theCall(mock_df_contigs, "gene"), "CTgene")
  expect_equal(.theCall(mock_df_contigs, "nt"), "CTnt")
  expect_equal(.theCall(mock_df_contigs, "amino"), "CTaa")
  expect_equal(.theCall(mock_df_contigs, "strict"), "CTstrict")
  
  # Test custom variable with warning
  expect_message(.theCall(mock_df_contigs, "cluster"), "A custom variable cluster will be used to call clones")
})

test_that(".assignCT works correctly", {
  con_df_t <- data.frame(TCR1 = "TRAV1", TCR2 = "TRBV1", cdr3_nt1 = "AAA", cdr3_nt2 = "CCC", cdr3_aa1 = "K", cdr3_aa2 = "C")
  con_df_b <- data.frame(IGH = "IGHV1", IGLC = "IGLC1", cdr3_nt1 = "GGG", cdr3_nt2 = "TTT", cdr3_aa1 = "G", cdr3_aa2 = "F")
  
  # Test for T cells
  t_assigned <- .assignCT("T", con_df_t)
  expect_equal(t_assigned$CTgene, "TRAV1_TRBV1")
  expect_equal(t_assigned$CTnt, "AAA_CCC")
  expect_equal(t_assigned$CTaa, "K_C")
  expect_equal(t_assigned$CTstrict, "TRAV1;AAA_TRBV1;CCC")
  
  # Test for B cells
  b_assigned <- .assignCT("B", con_df_b)
  expect_equal(b_assigned$CTgene, "IGHV1_IGLC1")
  expect_equal(b_assigned$CTnt, "GGG_TTT")
  expect_equal(b_assigned$CTaa, "G_F")
})

test_that(".makeGenes works correctly", {
  data_t <- data.frame(chain = c("TRA", "TRB"), v_gene = c("vA", "vB"), j_gene = c("jA", "jB"), c_gene = c("cA", "cB"), d_gene = c(NA, "dB"))
  data_b <- data.frame(chain = c("IGH", "IGK"), v_gene = c("vH", "vK"), j_gene = c("jH", "jK"), c_gene = c("cH", "cK"), d_gene = c("dH", NA))
  
  # Test T-cell gene creation
  t_genes <- .makeGenes("T", data_t)
  expect_equal(t_genes$TCR1[1], "vA.jA.cA")
  expect_equal(t_genes$TCR2[2], "vB.dB.jB.cB")
  
  # Test B-cell gene creation
  b_genes <- .makeGenes("B", data_b)
  expect_equal(b_genes$IGHct[1], "vH.dH.jH.cH")
  expect_equal(b_genes$IGKct[2], "vK.jK.cK")
})

test_that(".lengthDF works correctly", {
  df_list <- list(sample1 = data.frame(CTnt = c("AAA_CCC", "GGG_TTT"), group = c("A", "B")))
  
  # Test with chain = "both"
  ldf_both <- .lengthDF(df_list, cloneCall = "CTnt", chain = "both", group = "group")
  expect_equal(names(ldf_both), c("length", "CT", "group", "values"))
  expect_equal(ldf_both$length, c(6, 6))
  
  # Test with specific chain
  ldf_chain <- .lengthDF(df_list, cloneCall = "CTnt", chain = "TRA", group = "group")
  expect_equal(names(ldf_chain), c("length", "CT", "values", "chain", "group"))
  expect_equal(ldf_chain$length, c(7, 7)) # Length of first part of the string
})

#' Calculate Clonal Bias Towards a Cluster or Compartment
#'
#' The metric seeks to quantify how individual clones are skewed towards
#' a specific cellular compartment or cluster. A clone bias of `1` -
#' indicates that a clone is composed of cells from a single
#' compartment or cluster, while a clone bias of `0` - matches the
#' background subtype distribution. Please read and cite the following
#' [manuscript](https://pubmed.ncbi.nlm.nih.gov/35829695/)
#' if using [clonalBias()].
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list,
#'                         samples = c("P17B", "P17L", "P18B", "P18L",
#'                                     "P19B","P19L", "P20B", "P20L"))
#'
#' # Getting a sample of a Seurat object
#' scRep_example <- get(data("scRep_example"))
#'
#' # Using combineExpresion()
#' scRep_example <- combineExpression(combined, scRep_example)
#' scRep_example$Patient <- substring(scRep_example$orig.ident,1,3)
#'
#' # Using clonalBias()
#' clonalBias(scRep_example,
#'               clone.call = "aa",
#'               split.by = "Patient",
#'               group.by = "seurat_clusters",
#'               n.boots = 5,
#'               min.expand = 2)
#'
#'
#' @param sc.data The single-cell object after [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param group.by A column header in the metadata that bias will be based on.
#' @param split.by The variable to use for calculating the baseline frequencies.
#' For example, "Type" for lung vs peripheral blood comparison
#' @param n.boots number of bootstraps to downsample.
#' @param min.expand clone frequency cut off for the purpose of comparison.
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals].
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @importFrom quantreg rqss
#' @importFrom stats sd
#' @export
#' @concept SC_Functions
#' @return ggplot scatter plot with clone bias
clonalBias <- function(sc.data,
                       clone.call = NULL,
                       split.by = NULL,
                       group.by = NULL,
                       n.boots = 20,
                       min.expand = 10,
                       export.table = NULL,
                       palette = "inferno",
                       # Deprecated arguments
                       cloneCall = NULL,
                       exportTable = NULL,
                       ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "clonalBias", default = "strict")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "clonalBias", default = FALSE)

  .checkSingleObject(sc.data)
  clone.call <- .theCall(.grabMeta(sc.data), clone.call)
  # Calculating bias
  bias <- .get_clono_bias(sc.data,
                         split.by = split.by,
                         group.by = group.by,
                         clone.call = clone.call,
                         min.expand = min.expand)
  df_shuffle.list <- list()
  # Bootstrapping
  for (ii in seq_len(n.boots)) {
    df_shuffle.list[[ii]] <- .get_clono_bias(sc.data,
                                             split.by   = split.by,
                                             group.by   = group.by,
                                             clone.call = clone.call,
                                             min.expand = min.expand,
                                             do.shuffle = TRUE,
                                             seed       = ii)
  }
  df_shuffle <- Reduce(rbind, df_shuffle.list)
  
  # Summarizing boot straps
  stat.summary <- df_shuffle %>%
    group_by(ncells) %>%
    summarise(mean = mean(bias), 
              std = sd(bias))
  
  corrected_p <- 1-(0.05/nrow(bias))
  bias$Top_state <- factor(bias$Top_state, 
                           .alphanumericalSort(unique(bias$Top_state)))
  
  # Calculating Bias Z-score
  bias$Z.score <- NA
  for(i in seq_len(nrow(bias))) {
    stat.pos <- bias[i,]$ncells
    row.pull <- stat.summary[stat.summary$ncells == stat.pos,]
    z.score <- (bias[i,]$bias - row.pull$mean)/row.pull$std
    bias$Z.score[i] <- z.score
  }
  
  # Attaching the cloneSize of original combineExpression()
  meta <- .grabMeta(sc.data)
  meta <- meta[meta[,clone.call] %in% bias[,"Clone"],]
  meta <- unique(meta[,c(clone.call, split.by, "cloneSize")])

  bias$cloneSize <- NA
  for(i in seq_len(nrow(bias))) {
    split <- bias[,1][i]
    clone <- bias[,3][i]
    bias$cloneSize[i] <- as.vector(meta[which(meta[,clone.call] == clone & meta[,split.by] == split),"cloneSize"])[1]
  }

  bias$cloneSize <- factor(bias$cloneSize, rev(levels(meta[, "cloneSize"])))

  if (export.table) {
    return(bias)
  }
  bias$dotSize <- as.numeric(bias$cloneSize)
  
  #else, return the plot 
  ggplot(bias, aes(x=.data[["ncells"]],y=bias)) + 
    geom_point(aes(fill=.data[["Top_state"]], size = .data[["dotSize"]]), 
               shape = 21, 
               stroke = 0.25) + 
    stat_quantile(data=df_shuffle, 
                  quantiles = c(corrected_p), 
                  method = "rqss", 
                  lambda = 3, 
                  color = "black", 
                  lty = 2) + 
    #This is ridiculous way to get around the internal ggplot "style"-based warnings
    scale_size_continuous(breaks = as.vector(unique(bias$dotSize)), 
                          labels = as.vector(unique(bias$cloneSize))) + 
    scale_fill_manual(values = .colorizer(palette,  length(unique(bias[,"Top_state"])))) + 
    .themeRepertoire(...) + 
    labs(size = "cloneSize", fill = "Group") + 
    xlab("Clonal Size") +
    ylab("Clonal Bias")
}

#Background summary of clones
.get_clono_bg <- function(df,
                         split.by = split.by,
                         group.by = group.by,
                         clone.call = clone.call,
                         min.expand = 10) {
  df <- .dataWrangle(df, split.by, clone.call, "both")

  bg <- list()
  for (s in seq_along(df)) {

    clones <-  table(df[[s]][,clone.call])
    clones <- clones[clones>=min.expand]

    if (length(clones)>0) {

      expanded <- df[[s]][which(df[[s]][,clone.call] %in% names(clones)), group.by]
      bg[[s]] <- table(expanded) / sum(table(expanded))
    }
  }
  names(bg) <- names(df)[seq_len(length(bg))]
  return(bg)
}

#Clone Bias
.get_clono_bias <- function(df,
                           split.by = NULL,
                           group.by = NULL,
                           clone.call = clone.call,
                           min.expand = 10,
                           do.shuffle = FALSE,
                           seed = 42) {

  dat <- data.frame(Sample=character(),
                    Clone_i=character(),
                    Clone=character(),
                    ncells=integer(),
                    Top_state=double(),
                    freq=double(),
                    freq_diff=double(),
                    bias=double())

  bg <- .get_clono_bg(df,
                     split.by = split.by,
                     min.expand = min.expand,
                     group.by = group.by,
                     clone.call = clone.call)
  if (!is.null(split.by)) {
    df <- .list.input.return(df, split.by)
  } else {
    if (inherits(x=df, what ="Seurat") | inherits(x=df, what ="SummarizedExperiment")) {
      df <- list("Object" = .grabMeta(df))
    }
  }
  clone.call <- .theCall(df, clone.call)
  df <- .checkBlanks(df, clone.call)

  for (s in names(bg)) {

    #All cells have the same type. Cannot calculate bias for this sample
    if (length(bg[[s]]) > 1) {

      clones <-  table(df[[s]][,clone.call])
      clones <- clones[clones>=min.expand]
      subtypes <- names(bg[[s]])

      if (length(clones)>0) {
        clones <- sort(clones, decreasing = TRUE)
        expanded <- df[[s]][which(df[[s]][,clone.call] %in% names(clones)), c(group.by, clone.call)]

        if (do.shuffle) {  #reshuffle annotation column
          expanded[[group.by]] <- sample(expanded[[group.by]])
        }

        for (i in seq_along(clones)) {
          this <- names(clones)[i]
          this.s <- paste0(this, "_", s)

          ncells <- clones[i]
          sub <- expanded[which(expanded[, clone.call] == this), group.by]
          sub <- factor(sub, levels=subtypes)

          comp <- table(sub) / sum(table(sub))
          diff <- comp - bg[[s]]
          diff_norm <- round((comp - bg[[s]])/(1-bg[[s]]), 3)
          top_state <- names(which.max(diff_norm))[1]

          new_row <- c(s, this.s, this, as.integer(ncells), top_state,
                       as.double(comp[top_state]),
                       as.double(diff[top_state]),
                       as.double(diff_norm[top_state]))
          dat[nrow(dat) + 1,] <- new_row
        }
      }
    }
  }
  dat$ncells <- as.numeric(dat$ncells)
  dat$freq <- as.double(dat$freq)
  dat$freq_diff <- as.double(dat$freq_diff)
  dat$bias <- as.double(dat$bias)
  return(dat)
}

#' Plot Positional Physicochemical Property Analysis
#'
#' This function analyzes the physicochemical properties of amino acids at each
#' position along the CDR3 sequence. It calculates the mean property value and
#' the 95% confidence interval for each position across one or more groups,
#' visualizing the results as a line plot with a confidence ribbon.
#' 
#' @details
#' The function uses one of several established physicochemical property scales
#' to convert amino acid sequences into numerical vectors. More information for
#' the individual methods can be found at the following citations:
#'
#' **atchleyFactors:** [citation](https://pubmed.ncbi.nlm.nih.gov/15851683/)
#'
#' **crucianiProperties:** [citation](https://analyticalsciencejournals.onlinelibrary.wiley.com/doi/abs/10.1002/cem.856)
#'
#' **FASGAI:** [citation](https://pubmed.ncbi.nlm.nih.gov/18318694/)
#'
#' **kideraFactors:** [citation](https://link.springer.com/article/10.1007/BF01025492)
#'
#' **MSWHIM:** [citation](https://pubs.acs.org/doi/10.1021/ci980211b)
#'
#' **ProtFP:** [citation](https://pubmed.ncbi.nlm.nih.gov/24059694/)
#'
#' **stScales:** [citation](https://pubmed.ncbi.nlm.nih.gov/19373543/)
#'
#' **tScales:** [citation](https://www.sciencedirect.com/science/article/abs/pii/S0022286006006314)
#'
#' **VHSE:** [citation](https://pubmed.ncbi.nlm.nih.gov/15895431/)
#'
#' **zScales:** [citation](https://pubmed.ncbi.nlm.nih.gov/9651153/)
#' 
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' 
#' # Using positionalProperty()
#' positionalProperty(combined, 
#'                    chain = "TRB",
#'                    method = "atchleyFactors", 
#'                    aa.length = 20)
#'                    
#' @param input.data The product of [combineTCR()], [combineBCR()], or 
#' [combineExpression()]
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains 
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param group.by A column header in the metadata or lists to group the analysis 
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed 
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements 
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups 
#' automatically.
#' @param aa.length The maximum length of the CDR3 amino acid sequence. 
#' @param method Character string (one of the supported names)
#' Defaults to `"atchleyFactors"`, but includes: `"crucianiProperties"`,
#' `"FASGAI"`, `"kideraFactors"`, `"MSWHIM"`, `"ProtFP"`, `"stScales"`,
#' `"tScales"`, `"VHSE"`, `"zScales"`
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any [hcl.pals][grDevices::hcl.pals]
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#' @importFrom stats qt sd
#' @importFrom utils getFromNamespace
#' @export
#' @concept Summarize_Repertoire
#' @return A ggplot object displaying property by amino acid position.
#' If `exportTable = TRUE`, a matrix of the raw data is returned.
#' @author Florian Bach, Nick Borcherding

positionalProperty <- function(input.data,
                               chain = "TRB",
                               group.by = NULL,
                               order.by = NULL,
                               aa.length = 20,
                               method = "atchleyFactors",
                               export.table = NULL,
                               palette = "inferno",
                               # Deprecated arguments
                               exportTable = NULL,
                               ...)  {

  # Handle deprecated arguments
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "positionalProperty", default = FALSE)

  factors <- c("atchleyFactors", "crucianiProperties", "FASGAI", "kideraFactors",
  "MSWHIM", "ProtFP", "stScales", "tScales", "VHSE", "zScales")
  if (!method %in% c(factors)) {
    stop("Please select a compatible method: ", paste0(factors, collapse = ", "))
  }
  
  amino.acids <- c("A", "R", "N", "D", "C", "Q", "E", "G", "H", "I", "L", "K", 
                   "M", "F", "P", "S", "T", "W", "Y", "V")
  
  sco <- .is.seurat.or.se.object(input.data)
  input.data <- .dataWrangle(input.data, 
                             group.by, 
                             .theCall(input.data, "CTaa", 
                                      check.df = FALSE, silent = TRUE), 
                             chain)
  cloneCall <- .theCall(input.data, "CTaa")
  
  .aa.property.matrix <- getFromNamespace(".aa.property.matrix", "immApex")
  .padded.strings <- getFromNamespace(".padded.strings", "immApex")
  
  if(!is.null(group.by) & !sco) {
    input.data <- .groupList(input.data, group.by)
  }
  
  # Load the selected property matrix
  S <- .aa.property.matrix(method)
  S <- S[, amino.acids, drop = FALSE]
  
  # Main calculation loop over groups
  results_list <- lapply(names(input.data), function(current_group_name) {
    x <- input.data[[current_group_name]]
    input.sequences <- x[["CTaa"]]
    
    # Clean sequence vector
    if (any(grepl(";", input.sequences))) {
      input.sequences <- unlist(strsplit(input.sequences, ";"))
    }
    input.sequences <- na.omit(input.sequences)
    input.sequences <- input.sequences[nchar(input.sequences) > 0]
    
    if (length(input.sequences) == 0) return(NULL)
    
    n_seq <- length(input.sequences)
    k <- nrow(S)
    
    # Vectorized profile creation
    padded_seqs <- .padded.strings(input.sequences, 
                                   max.length =  aa.length, 
                                   pad = ".")
    char_matrix <- do.call(rbind, strsplit(padded_seqs, ""))
    index_matrix <- matrix(match(char_matrix, colnames(S)), nrow = n_seq)
    profile_array <- array(NA_real_, dim = c(k, aa.length, n_seq), dimnames = list(rownames(S), 1:aa.length, NULL))
    
    for (i in 1:k) {
      prop_values_matrix <- matrix(S[i, index_matrix], nrow = n_seq)
      profile_array[i, , ] <- t(prop_values_matrix)
    }
    
    # Summarization
    mean_vals <- apply(profile_array, c(1, 2), mean, na.rm = TRUE)
    sd_vals   <- apply(profile_array, c(1, 2), sd, na.rm = TRUE)
    n_obs     <- apply(profile_array, c(1, 2), function(val) sum(!is.na(val)))
    
    # 95% CI Calculation
    conf.level <- 0.95
    t_critical <- qt(1 - (1 - conf.level) / 2, df = pmax(1, n_obs - 1))
    margin_of_error <- t_critical * (sd_vals / sqrt(n_obs))
    
    # Construct summary data.frame
    summary_df <- as.data.frame.table(mean_vals, responseName = "mean")
    names(summary_df) <- c("property", "position", "mean")
    
    summary_df$sd <- as.vector(sd_vals)
    summary_df$n <- as.vector(n_obs)
    summary_df$ci_lower <- as.vector(mean_vals - margin_of_error)
    summary_df$ci_upper <- as.vector(mean_vals + margin_of_error)
    summary_df$group <- current_group_name
    
    return(summary_df)
  })
  
  # Combine results and handle ordering
  mat <- do.call(rbind, results_list[!sapply(results_list, is.null)])
  if (is.null(mat) || nrow(mat) == 0) {
    message("No sequences found to plot after filtering.")
    return(NULL)
  }
  #Removing any CI for calculations based on < 2 residues
  cols_to_modify <- c("sd", "ci_lower", "ci_upper")
  mat[mat[["n"]] <= 2, cols_to_modify] <- 0
  
  if (!is.null(order.by)) {
    mat <- .orderingFunction(order.by, "group", mat)
  }
  mat$position <- as.integer(mat$position)
  
  if (export.table) {
    return(mat)
  }
  
  colors <- .colorizer(palette, length(unique(mat[["group"]])))
  # Visualization
  plot <- ggplot(mat, aes(x = as.factor(position), y = mean, group = group, color = group)) +
    geom_ribbon(aes(ymin = ci_lower, ymax = ci_upper, fill = group), alpha = 0.3, linetype = 0) +
    geom_line() +
    scale_color_manual(name = "Group", values = colors) +
    scale_fill_manual(name = "Group", values = colors) +
    labs(x = "Amino Acid Position", y = "Mean Property Value") +
    facet_grid(property~., scales = "free_y") +
    guides(fill = "none", 
           color = guide_legend(override.aes = list(fill = NA))) + 
    .themeRepertoire(...) + 
    theme(
      strip.background = element_rect(fill="white", color = "grey"),
      strip.text = element_text(face = "bold"),
      legend.key = element_rect(fill = NA)
    )
  
  return(plot)
}

#' Visualize Clonal Network in Dimensional Reductions
#'
#' This function generates a network based on clonal
#' proportions of an indicated identity and then superimposes
#' the network onto a single-cell object dimensional reduction
#' plot.
#'
#' @examples

#' \dontrun{
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list,
#'                         samples = c("P17B", "P17L", "P18B", "P18L",
#'                                     "P19B","P19L", "P20B", "P20L"))
#'
#' # Getting a sample of a Seurat object
#' scRep_example  <- get(data("scRep_example"))
#'
#' # Using combineExpresion()
#' scRep_example  <- combineExpression(combined, scRep_example)
#'
#' # Using clonalNetwork()
#' clonalNetwork(scRep_example,
#'               reduction = "umap",
#'               group.by = "seurat_clusters")
#' }
#'
#' @param sc.data The single-cell object after [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param reduction The name of the dimensional reduction of the
#' single-cell object.
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). This will be the nodes overlaid onto the graph.
#' @param filter.clones Use to select the top n clones (e.g., ``filter.clones`**`
#' = 2000) or n of clones based on the minimum number of all the comparators
#' (e.g., `filter.clone`` = "min").
#' @param filter.identity Display the network for a specific level of the
#' indicated identity.
#' @param filter.proportion Remove clones from the network below a specific
#'  proportion.
#' @param filter.graph Remove the reciprocal edges from the half of the graph,
#' allowing for cleaner visualization.
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param export.clones Exports a table of clones that are shared
#' across multiple identity groups and ordered by the total number
#' of clone copies.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals].
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param exportClones \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.clones` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @importFrom igraph graph_from_data_frame V `V<-`
#' @importFrom dplyr summarize_all count across all_of desc
#' @importFrom tidygraph as_tbl_graph activate
#' @importFrom ggraph ggraph geom_edge_bend  geom_node_point scale_edge_colour_gradientn circle guide_edge_colourbar
#' @importFrom stats setNames
#' @export
#' @concept SC_Functions
#' @return ggplot object
#'
clonalNetwork <- function(sc.data,
                          clone.call = NULL,
                          chain = "both",
                          reduction = "umap",
                          group.by = "ident",
                          filter.clones = NULL,
                          filter.identity = NULL,
                          filter.proportion = NULL,
                          filter.graph = FALSE,
                          export.clones = NULL,
                          export.table = NULL,
                          palette = "inferno",
                          # Deprecated arguments
                          cloneCall = NULL,
                          exportClones = NULL,
                          exportTable = NULL,
                          ...) {

    # Handle deprecated arguments
    clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                                 "clonalNetwork", default = "strict")
    export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                   "clonalNetwork", default = FALSE)
    export.clones <- .deprecate_arg(exportClones, export.clones, "exportClones", "export.clones",
                                    "clonalNetwork", default = FALSE)

    to <- from <- weight <- y <- NULL
    meta <- .grabMeta(sc.data)
    clone.call <- .theCall(meta, clone.call)
    coord <- data.frame(.getCoord(sc.data, reduction), group.by = meta[,group.by])
    min <- c()
    meta <- .grabMeta(sc.data)
    if (!is.null(filter.clones))  {
      if(filter.clones == "min") {
        meta <- .grabMeta(sc.data)
        id.meta <- split(meta, meta[,group.by])
        for (x in seq_along(id.meta)) {
            min.tmp <- length(which(!is.na(unique(id.meta[[x]][,clone.call]))))
            min <- c(min.tmp, min)
        }
        #Filtering clones based on the minimum value
        min_val <- min(min)
        table <- .cloneCounter(meta, group.by, clone.call)
        cut <- which.min(abs(table$clone.sum - min_val))
        clones.to.filter <- table[,1][seq_len(cut)]
      } else if (is.numeric(filter.clones)) {
          #Filtering based on a numeric value
          table <- meta %>%
            dplyr::count(meta[, clone.call]) %>%
            na.omit() %>%
            arrange(desc(n)) %>%
            mutate(cumSum = cumsum(n))
          cut <- which.min(abs(table$cumSum - filter.clones))
          clones.to.filter <- table[seq_len(cut),1]
      }
      meta <- meta[meta[,clone.call] %in% clones.to.filter,]
    } 
    clones.duplicated <- na.omit(unique(meta[which(duplicated(meta[,clone.call])),clone.call]))
    
    if(export.clones) {
      #Summarizing all the clones by group.by
      table <- .cloneCounter(meta, group.by, clone.call)[,seq_len(3)]
      #Identifying the clones across the group by
      clones.across.identities <- names(which(table(table[,2]) > 1))
      if(length(clones.across.identities) < 1) {
        stop("No shared clones across group.by variables for the current parameters selected")
      }
      #Getting the clones to output
      table <- table[table[,2] %in% clones.across.identities,]
      colnames(table) <- c("id", "clone", "n")
      dupl.clones <- table %>%
                      group_by(clone) %>%
                      summarise(sum = sum(n))%>%
                      arrange(desc(sum)) 
      return(dupl.clones)
    }
    
    id <- as.vector(meta[,group.by])
    id.names <- id
    names(id.names) <- row.names(meta)
    unique.id <- unique(id)
    
    #Placing the Position of the nodes
    id.positions <- data.frame(coord)
    colnames(id.positions)[seq_len(2)] <- c("x", "y")
    centers <- id.positions %>%
      group_by(group.by) %>%
      summarise(across(c(x, y), mean, .names = "{col}")) %>%
      as.data.frame()
    
    row.names(centers) <- centers$group.by
    centers$group.by <- NULL
    
    #Unique clones per group.by
    clone.number <- meta %>%
      select(all_of(c(clone.call, group.by))) %>%
      group_by(meta[,group.by]) %>%
      na.omit() %>%
      unique() %>%
      summarise(n = dplyr::n()) %>%
      {setNames(.$n, .$`meta[, group.by]`)} 
    
    #Total clones per group.by
    total.number <- meta %>%
      select(all_of(c(clone.call, group.by))) %>%
      group_by(meta[,group.by]) %>%
      na.omit() %>%
      summarise(n = dplyr::n()) %>%
      {setNames(.$n, .$`meta[, group.by]`)} 
    
    edge.list <- NULL
    for (i in seq_along(clones.duplicated)) {
       pos <- which(meta[,clone.call] == clones.duplicated[i])
       num <- table(meta[pos,group.by])
       num <- num[num > 0]
       if(length(num) == 1) {
         next()
       }
       grid <- expand.grid(names(num),names(num))
       grid <- grid[grid[,1] != grid[,2],]
       for (x in seq_len(nrow(grid))) {
          summary <- c(to = as.vector(grid[x,1]), 
                       from = as.vector(grid[x,2]), 
                       weight = num[grid[x,2]]/total.number[as.vector(grid[x,1])])
          edge.list <- rbind(edge.list, summary)
       }
    }
    if(is.null(edge.list)) {
      stop("No shared clones between the indicated identity")
    }
    
    edge.list <- data.frame(edge.list)
    colnames(edge.list)[3] <-"weight"
    #Filtering based on identity
    if (!is.null(filter.identity)) { 
        col1 <- which(edge.list[,1] == filter.identity)
        col2 <- which(edge.list[,2] == filter.identity)
        edge.list <- edge.list[c(col1,col2),]
    }
    #Remove reciprocals 
    if (filter.graph) {
        unique.id <- .alphanumericalSort(unique.id)
        edge.list <- edge.list[edge.list[,1] %in% unique.id[seq_len(length(unique.id)/2)],]
    }
    #Removing any clones below proportion threshold
    if(!is.null(filter.proportion)) {
        edge.list <-edge.list[edge.list[,3] > filter.proportion,]
    }
    edge.list1 <- edge.list %>%
        group_by(to, from) %>%
        summarise(weight = sum(as.numeric(weight)))
    graph <- graph_from_data_frame(edge.list1)
    clone.number <- clone.number[match(names(V(graph)), names(clone.number))]
    V(graph)$size <- unname(clone.number)
    centers <- centers[rownames(centers) %in% names(V(graph)),]

    if (export.table) {
      return(edge.list1)
    }
    #Warning from this is produced by geom_edge_bend and there is nothng that can be done
    #until the authors of the package update it.
    plot <- ggraph(graph, layout = centers[match(names(V(graph)), rownames(centers)),]) + 
                  geom_point(data = coord, aes(x = coord[,1], 
                                               y = coord[,2], 
                                               color = group.by)) + 
                  geom_edge_bend(aes(edge_color = as.numeric(weight)), 
                                 alpha = 0.7, 
                                 width = 1,
                                 arrow = arrow(length = unit(4, 'mm')), 
                                 end_cap = circle(3, 'mm'), 
                                 angle_calc = "across", 
                                 check_overlap = TRUE, 
                                 strength = 0.75) + 
                  geom_node_point(aes(size = size)) + 
                  ylab(paste0(reduction, "_2")) + 
                  xlab(paste0(reduction, "_1")) + 
                  guides(color = "none") + 
                  scale_edge_colour_gradientn(colors  = .colorizer(palette,13), trans = "log10") + 
                  labs(size = "Unique Clones", 
                       edge_color = "Relative Proportion of \nClones of Starting Node") + 
                  .themeRepertoire(...) + 
                  theme(
                      panel.background = element_blank(),
                      panel.border = element_blank(), 
                      panel.grid.major = element_blank(),
                      panel.grid.minor = element_blank(), 
                      axis.line = element_line(colour = "black"), 
                      legend.key=element_blank()
                  )
    return(plot)
}

#' Plot the Distribution of Sequence Lengths
#'
#' This function displays either the nucleotide `nt` or amino
#' acid `aa` sequence length. The sequence length visualized
#' can be selected using the chains parameter, either the combined clone
#' (both chains) or across all single chains. Visualization can either
#' be a histogram or if `scale` = TRUE, the output will
#' be a density plot. Multiple sequencing runs can be group together
#' using the group.by parameter.
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list,
#'                         samples = c("P17B", "P17L", "P18B", "P18L",
#'                                     "P19B","P19L", "P20B", "P20L"))
#' clonalLength(combined, clone.call = "aa", chain = "both")
#'
#' @param input.data The product of [combineTCR()],
#' [combineBCR()], or [combineExpression()]
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `nt` (CDR3 nucleotide sequence) or `aa` (CDR3 amino acid sequence)
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups
#' automatically.
#' @param scale Converts the graphs into density plots in order to show
#' relative distributions.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals]
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @export
#' @concept Visualizing_Clones
#' @return A ggplot object visualizing the distributions by length, or a data.frame if
#' `export.table = TRUE`.
clonalLength <- function(input.data,
                         clone.call = NULL,
                         chain = "both",
                         group.by = NULL,
                         order.by = NULL,
                         scale = FALSE,
                         export.table = NULL,
                         palette = "inferno",
                         # Deprecated arguments
                         cloneCall = NULL,
                         exportTable = NULL,
                         ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "clonalLength", default = "aa")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "clonalLength", default = FALSE)

  input.data <- .dataWrangle(input.data,
                             group.by,
                             .theCall(input.data, clone.call,
                                      check.df = FALSE, silent = TRUE),
                             chain)
  clone.call <- .theCall(input.data, clone.call)

  #Sorting out graphing parameters
  xlab <- "Length"
  if(clone.call == "CTnt") {
    ylab <- "CDR3 (NT)"
  } else if (clone.call == "CTaa") {
    ylab <- "CDR3 (AA)"
  } else {
    stop("Please make a selection of the type of CDR3 sequence to analyze by using `clone.call`")
  }
  
  #Calculating Length
  Con.df <- NULL
  Con.df <- .lengthDF(input.data, clone.call, chain, group.by)

  names <- names(input.data)

  #Skip plotting if want to export table
  if (export.table == TRUE) {
    return(Con.df)
  }
  
  if(!is.null(order.by)) {
    if (!is.null(group.by)) { 
      Con.df <- .orderingFunction(vector = order.by,
                                  group.by = group.by, 
                                  data.frame = Con.df)
    } else {
      Con.df <- .orderingFunction(vector = order.by,
                                  group.by = "values", 
                                  data.frame = Con.df)
    }
  }
  
  #Plotting
  if (!is.null(group.by)) { 
    fill <- group.by
    col <- length(unique(Con.df[,group.by]))
    if (scale == TRUE) { 
      yplus <- "Percent of "
      plot <- ggplot(Con.df, aes(fill=Con.df[,group.by],
                                 x = length,
                                 y = (after_stat(count))/sum(after_stat(count))*100)) + 
                     geom_density(aes(y=after_stat(scaled)),
                                  alpha=.5, lwd=.25, color="black")
    } else { 
      yplus <- "Number of "
      plot <- ggplot(Con.df,aes(x = as.factor(length),
                                fill=Con.df[,group.by]))+
                     geom_bar(position = position_dodge2(preserve = "single"), 
                              color="black", lwd=0.25, width=0.9)  +
                      scale_x_discrete(breaks = round(seq(min(Con.df$length), 
                                          max(Con.df$length), by = 5),10)) }
  } else if (is.null(group.by)){ 
      fill <- "Samples"
      col <- length(unique(Con.df$values))
      if (scale == TRUE) { 
        yplus <- "Density of "
        plot <- ggplot(Con.df, aes(x = length, 
                                   y = (after_stat(count))/sum(after_stat(count))*100, 
                                   fill=values)) + 
                        geom_density(aes(y=after_stat(scaled)), 
                                     alpha=0.5, lwd=0.25, color="black")
      }  else { 
        yplus <- "Number of "
        plot <- ggplot(Con.df, aes(as.factor(length), fill=values)) +
                        geom_bar(position = position_dodge2(preserve = "single"), 
                                 color="black", lwd=0.25) +
                        scale_x_discrete(breaks = round(seq(min(Con.df$length), 
                                          max(Con.df$length), by = 5),10))} }
  plot <- plot + 
          scale_fill_manual(values = .colorizer(palette,col)) +
          labs(fill = fill) + 
          ylab(paste(yplus, ylab, sep="")) +
          xlab(xlab) + 
          .themeRepertoire(...)
  
  return(plot)
}

#' Compare Clonal Abundance Across Variables
#'
#' This function visualizes the relative abundance of specific clones across
#' different samples or groups. It is useful for tracking how the proportions
#' of top clones change between conditions. The output can be an alluvial plot
#' to trace clonal dynamics or an area plot to show compositional changes.
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list,
#'                        samples = c("P17B", "P17L", "P18B", "P18L",
#'                                    "P19B","P19L", "P20B", "P20L"))
#'
#' # Using clonalCompares()
#' clonalCompare(combined,
#'               top.clones = 5,
#'               samples = c("P17B", "P17L"),
#'               clone.call = "aa")
#'
#' @param input.data The product of \code{\link{combineTCR}},
#' \code{\link{combineBCR}}, or \code{\link{combineExpression}}.
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param samples The specific samples to isolate for visualization.
#' @param clones The specific clonal sequences of interest
#' @param top.clones The top number of clonal sequences per group.
#' (e.g., top.clones = 5)
#' @param highlight.clones Clonal sequences to highlight, if present,
#' all other clones returned will be grey
#' @param relabel.clones Simplify the legend of the graph by returning
#' clones that are numerically indexed
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups
#' automatically.
#' @param graph The type of plot to generate. Accepted values are `alluvial`
#' (default) or `area`
#' @param proportion If `TRUE` (default), the y-axis will represent the
#' proportional abundance of clones. If `FALSE`, the y-axis will represent
#' raw clone counts.`
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' \link[grDevices]{hcl.pals}
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @export
#' @importFrom dplyr slice_max
#' @concept Visualizing_Clones
#' @return A ggplot object visualizing proportions of clones by groupings, or a
#' data.frame if `export.table = TRUE`.
clonalCompare <- function(input.data,
                          clone.call = NULL,
                          chain = "both",
                          samples = NULL,
                          clones = NULL,
                          top.clones = NULL,
                          highlight.clones = NULL,
                          relabel.clones = FALSE,
                          group.by = NULL,
                          order.by = NULL,
                          graph = "alluvial",
                          proportion = TRUE,
                          export.table = NULL,
                          palette = "inferno",
                          # Deprecated arguments
                          cloneCall = NULL,
                          exportTable = NULL,
                          ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "clonalCompare", default = "strict")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "clonalCompare", default = FALSE)

  #Tie goes to indicated clones over top clones
  if(!is.null(top.clones) && !is.null(clones)) {
    top.clones <- NULL
  }
  input.data <- .dataWrangle(input.data,
                              group.by,
                              .theCall(input.data, clone.call,
                                       check.df = FALSE, silent = TRUE),
                              chain)
  clone.call <- .theCall(input.data, clone.call)

  sco <- .is.seurat.or.se.object(input.data)
  if(!is.null(group.by) && !sco) {
    input.data <- .groupList(input.data, group.by)
  }

  compareColname <- ifelse(proportion, "Proportion", "Count")
  normalizer <- ifelse(proportion, sum, length)

  Con.df <- input.data %>%
    purrr::imap(function(df, columnNames) {
      tbl <- as.data.frame(table(df[, clone.call]))
      if (proportion) {
        tbl[, 2] <- tbl[, 2] / normalizer(tbl[, 2])
      }
      colnames(tbl) <- c("clones", compareColname)
      tbl$Sample <- columnNames
      tbl
    }) %>%
    dplyr::bind_rows()

  #Filtering steps
  if (!is.null(samples)) {
    Con.df <- Con.df[Con.df$Sample %in% samples,]
  }
  if (!is.null(clones)) {
    Con.df <- Con.df[Con.df$clones %in% clones,]
  } else if (!is.null(top.clones)) {
    top <- Con.df %>%
      group_by(Sample) %>%
      slice_max(
        n = top.clones,
        order_by = !!sym(compareColname),
        with_ties = FALSE
      )
    Con.df <- Con.df[Con.df$clones %in% top$clones,]
  }
  if (nrow(Con.df) < length(unique(Con.df$Sample)) || nrow(Con.df) == 0) {
    stop("Please reasses the filtering strategies here, there are not
            enough clones to examine.")
  }
  #Clones relabeling
  clones.returned <- as.vector(unique(Con.df[order(Con.df[, compareColname], decreasing = TRUE),"clones"]))
  if (relabel.clones) {
    new.clones <- paste0("Clone: ", seq_len(length(clones.returned)))
    names(new.clones) <- clones.returned
    #Isolated new clone names for highlight purposes
    if(!is.null(highlight.clones)) {
      highlight.clones <- unname(new.clones[which(names(new.clones) %in% highlight.clones)])
    }
    Con.df[,"original.clones"] <- Con.df[, "clones"]
    Con.df[,"clones"] <- new.clones[as.vector(Con.df[, "clones"])]
    Con.df[,"clones"] <- factor(Con.df[, "clones"],
                                levels = .alphanumericalSort(unique(Con.df[, "clones"])))
    clones.returned <- as.vector(unique(Con.df[, "clones"]))
  }

  if(!is.null(order.by)) {
    Con.df <- .orderingFunction(vector = order.by,
                                group.by = "Sample",
                                data.frame = Con.df)
  }

  if (export.table) {
    return(Con.df)
  }

  #Plotting Functions
  plot <- ggplot(Con.df, aes(x = .data[["Sample"]],
                             fill = clones,
                             group = clones,
                             stratum = clones,
                             alluvium = clones,
                             y = !!sym(compareColname),
                             label = clones)) +
    .themeRepertoire(...) +
    theme(axis.title.x = element_blank(),
          legend.text=element_text(size=rel(0.5)),
          legend.key.size = unit(0.5,"line"))
  if (graph == "alluvial") {
    plot <- plot +  geom_stratum() + geom_flow(stat = "alluvium")
  } else if (graph == "area") {
    plot <- plot +
      geom_area(aes(group = clones), color = "black")
  }

  #Highlighting specific clones
  if (!is.null(highlight.clones)) {
    clone.colors <- rep("grey", length(clones.returned))
    pos <- which(clones.returned %in% highlight.clones)
    clone.colors[pos] <- .colorizer(palette, length(pos))
    names(clone.colors) <- clones.returned
    plot <- plot + scale_fill_manual(values = clone.colors)
  } else {
    plot <- plot + scale_fill_manual(values = .colorizer(palette, length(unique(Con.df[,"clones"]))))
  }
  return(plot)
}

#' Examining the Diversity of Amino Acids by Position
#'
#' This function the diversity amino acids along the residues of the CDR3 
#' amino acid sequence. Please see [clonalDiversity()] for more information 
#' on the underlying methods for diversity/entropy calculations. Positions 
#' without variance will have a value reported as 0 for the purposes of comparison.
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' 
#' # Using positionalEntropy()
#' positionalEntropy(combined, 
#'                   chain = "TRB", 
#'                   aa.length = 20)
#'                   
#' @param input.data The product of [combineTCR()], 
#' [combineBCR()], or [combineExpression()]
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains 
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param group.by A column header in the metadata or lists to group the analysis 
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed as 
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements 
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups 
#' automatically.
#' @param aa.length The maximum length of the CDR3 amino acid sequence. 
#' @param method The method to calculate the entropy/diversity -
#' `"shannon"`, `"inv.simpson"`, `"gini.simpson"`, `"norm.entropy"`,
#' `"pielou"`, `"hill0"`, `"hill1"`, `"hill2"`
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals]
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#' 
#' @export
#' @importFrom immApex calculateEntropy
#' @concept Summarize_Repertoire
#' @return A ggplot object displaying entropy or diversity by amino acid position.
#' If `exportTable = TRUE`, a matrix of the raw data is returned.
positionalEntropy <- function(input.data,
                              chain = "TRB",
                              group.by = NULL,
                              order.by = NULL,
                              aa.length = 20,
                              method = "norm.entropy",
                              export.table = NULL,
                              palette = "inferno",
                              # Deprecated arguments
                              exportTable = NULL,
                              ...)  {

  # Handle deprecated arguments
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "positionalEntropy", default = FALSE)

  sco <- .is.seurat.or.se.object(input.data)
  input.data <- .dataWrangle(input.data, 
                              group.by, 
                              .theCall(input.data, "CTaa", 
                                       check.df = FALSE, silent = TRUE), 
                              chain)
  cloneCall <- .theCall(input.data, "CTaa")
  
  if(!is.null(group.by) & !sco) {
    input.data <- .groupList(input.data, group.by)
  }
  
  #Selecting Diversity Function
  lapply(seq_along(input.data), function(x){
    strings <- .processStrings(input.data[[x]][["CTaa"]], aa.length)
    entropy <- calculateEntropy(strings, method = method)
    entropy_df <- as.data.frame(entropy)
    entropy_df$Position <- 1:nrow(entropy_df)
    entropy_df$group <- names(input.data)[x]
    rownames(entropy_df) <- NULL # Remove the messy row names
    entropy_df
  }) -> group.results
  
  mat_melt <- do.call(rbind, group.results)
  
  if(!is.null(order.by)) {
    mat_melt <- .orderingFunction(vector = order.by,
                                  group.by = "group", 
                                  mat_melt)
  }
    
  plot <- ggplot(mat_melt, aes(x=as.factor(Position), 
                               y = .data[["entropy"]], 
                               group= .data[["group"]], 
                               color = group)) +
          geom_line(stat = "identity") +
          geom_point() + 
          scale_color_manual(name = "Groups", 
                            values = rev(.colorizer(palette,length(input.data)))) +
          xlab("Amino Acid Residues") +
          ylab("Relative Diversity") +
          .themeRepertoire(...) + 
          theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1))
    if (export.table == TRUE) {
      return(mat_melt)
    }
    return(plot)
}

  

#' Highlighting Specific Clones
#'
#' Use a specific clonal sequence to highlight on top of the dimensional 
#' reduction in single-cell object.
#'
#' @examples
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' 
#' # Getting a sample of a Seurat object
#' scRep_example  <- get(data("scRep_example"))
#' 
#' # Using combineExpresion()
#' scRep_example  <- combineExpression(combined, 
#'                                     scRep_example)
#' 
#' # Using highlightClones()
#' scRep_example   <- highlightClones(scRep_example,
#'                                    clone.call= "aa",
#'                                    sequence = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF"))
#' 
#' @param sc.data The single-cell object to attach after
#' [combineExpression()]
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param sequence The specific sequence or sequence to highlight
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @importFrom S4Vectors DataFrame
#' @export
#' @concept SC_Functions
#' @return Single-cell object object with new meta data column
#' for indicated clones
highlightClones <- function(sc.data,
                            clone.call = NULL,
                            sequence = NULL,
                            # Deprecated arguments
                            cloneCall = NULL){

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "highlightClones", default = "strict")

  if (!.is.seurat.or.se.object(sc.data)) {
    stop("Please select a single-cell object")
  }

  clone.call <- .theCall(.grabMeta(sc.data), clone.call)
  meta <- .grabMeta(sc.data)
  meta$highlight <- NA
  for(i in seq_along(sequence)) {
    meta$highlight <-  ifelse(meta[,clone.call] == sequence[i],
                              sequence[i], meta$highlight)
  }
  
  meta <- meta[,-(which(colnames(meta) == "ident"))]
  
  if(.is.se.object(sc.data)) {
    colData(sc.data) <- DataFrame(meta)
  } else {
    col.name <- names(meta) %||% colnames(meta)
    sc.data[[col.name]] <- meta
  }
  return(sc.data)
}

#' **DEPRECATED** Take the meta data in seurat/SCE and place it into a list
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' Allows users to perform more fundamental measures of clonotype analysis
#' using the meta data from the seurat or SCE object. For Seurat objects the
#' active identity is automatically added as "cluster". Remaining grouping
#' parameters or SCE or Seurat objects must appear in the meta data.
#'
#' This function is deprecated as of version 2 due to the confusion it caused
#' to many users. Users are encouraged to remain with the abstraction barrier
#' of combined single cell objects and the outputs of [combineTCR()] and
#' [combineBCR()] for all functions.
#'
#' We discourage the use of this function, but if you have to use it, set the
#' `force` argument to `TRUE`.
#'
#' @param sc output of [combineExpression()].
#' @param ... previously the `group` or `split.by` argument, indicating the
#' column header to group the new list by. This should strictly be one argument
#' and is an ellipsis for backwards compatibility. Everything after the first
#' argument is ignored.
#' @param force logical. If not `TRUE` (default), a deprecation error will
#' be thrown. Otherwise the function will run but not guaranteed to be stable.
#'
#' @export
#' @return list derived from the meta data of single-cell object with
#' elements divided by the group parameter
#' @keywords internal
#'
expression2List <- function(sc, ..., force = FALSE) {

    if (!force) {
        lifecycle::deprecate_stop(
            "2.0.0", "expression2List()",
            details = "If you MUST use it, see `?expression2List`"
        )
    }

    lifecycle::deprecate_soft("2.0.0", "expression2List()")
    .expression2List(sc, list(...)[[1]])
}

#' Plot the Relative Abundance of Clones
#'
#' Displays the number of clones at specific frequencies by sample
#' or group. Visualization can either be a line graph (
#' `scale` = FALSE) using calculated numbers or density
#' plot (`scale` = TRUE). Multiple sequencing runs can
#' be group together using the group parameter. If a matrix
#' output for the data is preferred, set
#' `export.table` = TRUE.
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list,
#'                         samples = c("P17B", "P17L", "P18B", "P18L",
#'                                     "P19B","P19L", "P20B", "P20L"))
#'
#' # Using clonalAbundance()
#' clonalAbundance(combined,
#'                 clone.call = "gene",
#'                 scale = FALSE)
#'
#' @param input.data The product of [combineTCR()],
#' [combineBCR()], or [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups
#' automatically.
#' @param scale Converts the graphs into density plots in order to show
#' relative distributions.
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals].
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @importFrom ggplot2 ggplot
#' @importFrom dplyr distinct
#' @export
#' @concept Visualizing_Clones
#' @author Nick Borcherding, Justin Reimertz
#' @return A ggplot object visualizing clonal abundance by group, or a
#' data.frame if `export.table = TRUE`.
clonalAbundance <- function(input.data,
                            clone.call = NULL,
                            chain = "both",
                            scale = FALSE,
                            group.by = NULL,
                            order.by = NULL,
                            export.table = NULL,
                            palette = "inferno",
                            # Deprecated arguments
                            cloneCall = NULL,
                            exportTable = NULL,
                            ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "clonalAbundance", default = "strict")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "clonalAbundance", default = FALSE)
  Con.df <- NULL
  xlab <- "Abundance"
  input.data <- .dataWrangle(input.data,
                             group.by,
                             .theCall(input.data, clone.call,
                                      check.df = FALSE, silent = TRUE),
                             chain)
  clone.call <- .theCall(input.data, clone.call)
  
  names <- names(input.data)
  if (!is.null(group.by)) {
  	Con.df <- bind_rows(lapply(seq_along(input.data), function(x) {
  		input.data %>%
  			.parseContigs(x, names, clone.call) %>%
  			mutate("{group.by}" := input.data[[x]][[group.by]])
  	})) %>% 
  		dplyr::distinct() %>%
  	  as.data.frame()
  	col <- length(unique(Con.df[[group.by]]))
    fill <- group.by
    if(!is.null(order.by)) {
        Con.df <- .orderingFunction(vector = order.by,
                                     group.by = group.by, 
                                     data.frame =  Con.df)
    }
    if (scale == TRUE) { 
        ylab <- "Density of Clones"
        plot <- ggplot(Con.df, aes(x=Abundance, 
                                   fill=Con.df[,group.by])) +
                      geom_density(aes(y=after_stat(scaled)), 
                                   alpha=0.5, 
                                   lwd=0.25, 
                                   color="black", 
                                   bw=0.5)  +
                      scale_fill_manual(values = .colorizer(palette,col)) +
                      labs(fill = fill)
    } else { 
        ylab <- "Number of Clones"
        plot <- ggplot(Con.df, aes(x=Abundance, group.by = values, 
                               color = Con.df[,group.by])) +
                        geom_line(stat="count") +
                        scale_color_manual(values = .colorizer(palette,col)) +
                        labs(color = fill)
    }
  } else {
    for (i in seq_along(input.data)) {
      data1 <- .parseContigs(input.data, i, names, clone.call) 
      Con.df<- rbind.data.frame(Con.df, data1) %>%
                  dplyr::distinct() %>%
                  as.data.frame()
    }
    Con.df <- data.frame(Con.df)
    if(!is.null(order.by)) {
      Con.df <- .orderingFunction(vector = order.by,
                                  group.by = "values", 
                                  data.frame = Con.df)
    }
    
    col <- length(unique(Con.df$values))
    fill <- "Samples"
    if (scale == TRUE) { 
      ylab <- "Density of Clones"
      plot <- ggplot(Con.df, aes(Abundance, fill=values)) +
                      geom_density(aes(y=after_stat(scaled)), 
                                   alpha=0.5, 
                                   lwd=0.25, 
                                   color="black", 
                                   bw=0.5) +
                      scale_fill_manual(values = .colorizer(palette,col)) +
                      labs(fill = fill)
    } else { 
      ylab <- "Number of Clones"
      plot <- ggplot(Con.df, aes(x=Abundance, group = values, 
                               color = values)) +
                      geom_line(stat="count") +
                      scale_color_manual(values = .colorizer(palette,col)) +
                      labs(color = fill)
  } }
  if (export.table == TRUE) { 
    return(Con.df) 
  }
  plot <- plot + scale_x_log10() + ylab(ylab) + xlab(xlab) +
    .themeRepertoire(...)
  return(plot) 
}

#' Plot Number or Proportions of Clones
#'
#' This function quantifies unique clones. The unique clones 
#' can be either reported as a raw output or scaled to the total number of 
#' clones recovered using the scale parameter. 
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' 
#' # Using clonalQuant()
#' clonalQuant(combined,
#'             clone.call="strict",
#'             scale = TRUE)
#'
#' @param input.data The product of [combineTCR()],
#' [combineBCR()], or [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups
#' automatically.
#' @param scale Converts the graphs into percentage of unique clones
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals]
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#' 
#' @export
#' @concept Visualizing_Clones
#' @return A ggplot object visualizing the total or relative number of clones 
#' or a data.frame if `exportTable = TRUE`.
clonalQuant <- function(input.data,
                        clone.call = NULL,
                        chain = "both",
                        scale=FALSE,
                        group.by = NULL,
                        order.by = NULL,
                        export.table = NULL,
                        palette = "inferno",
                        # Deprecated arguments
                        cloneCall = NULL,
                        exportTable = NULL,
                        ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "clonalQuant", default = "strict")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "clonalQuant", default = FALSE)

  if (length(group.by) > 1) {
    stop("Only one item in the group.by variable can be listed.")
  }
  input.data <- .dataWrangle(input.data,
                             group.by,
                             .theCall(input.data, clone.call,
                                      check.df = FALSE, silent = TRUE),
                             chain)
  clone.call <- .theCall(input.data, clone.call)
  
  sco <- .is.seurat.or.se.object(input.data)
  if(!is.null(group.by) & !sco) {
    input.data <- .groupList(input.data, group.by)
  }

  mat.names <- c("contigs","values", "total", group.by)
  #Set up mat to store and selecting graph parameters
  if (!is.null(group.by)) {
    x <- group.by
    labs <- group.by
  } else {
    x <- "values"
    labs <- "Samples"
    col <- length(unique(names(input.data)))
  }
  mat <- data.frame(matrix(NA, length(input.data), length(mat.names)))
  colnames(mat) <- mat.names
  for (i in seq_along(input.data)) {
      mat[i,1] <- length(na.omit(unique(input.data[[i]][,clone.call])))
      mat[i,2] <- names(input.data)[i]
      mat[i,3] <- length(na.omit(input.data[[i]][,clone.call]))
      if (!is.null(group.by)) {
        location <- which(colnames(input.data[[i]]) == group.by)
        mat[i,4] <- as.vector(input.data[[i]][1,location])
      }
  }
  if (scale) {
      y <- "scaled"
      mat$scaled <- mat$contigs/mat$total*100
      ylab <- "Percent of Unique Clones"
   } else {
      y <- "contigs"
      ylab <- "Unique Clones"
   }

  if (export.table) {
    if (length(input.data) > 1) {
      return(mat)
    }
    # if a single sample, remove the "values" column if NA
    if (is.na(mat[[2]])) {
      mat[[2]] <- NULL
    }
    return(mat)
  }
  
  if(!is.null(group.by)) {
    col <- length(unique(mat[,group.by]))
  }
  mat[,x] = factor(mat[,x], levels = names(input.data))
  
  if(!is.null(order.by)) {
    mat <- .orderingFunction(vector = order.by,
                             group.by = "values", 
                             data.frame = mat)
  }
  
  #Plotting
  plot <- ggplot(data = mat, 
                 aes(x=mat[,x], y=mat[,y], fill=mat[,x])) +
            stat_summary(geom = "errorbar", 
                         fun.data = mean_se, 
                         position = "dodge", 
                         width=.5) + 
              labs(fill = labs) +
              ylab(ylab) +
              stat_summary(fun=mean, geom="bar", color="black", lwd=0.25)+
              .themeRepertoire(...) + 
              xlab("Samples") + 
              scale_fill_manual(values = .colorizer(palette, col))
  
  # if it is a single run, remove x axis labels if sample name missing
  if ((length(input.data) == 1) && identical(names(input.data), NA_character_)) {
    plot <- plot +
      ggplot2::theme(
        axis.title.x = element_blank(),
        axis.text.x = element_blank(),
        axis.ticks.x = element_blank()
      )
  }
  return(plot)
}

## ----echo=FALSE, results="hide", message=FALSE, warning=FALSE-----------------
knitr::opts_chunk$set(error=FALSE, message=FALSE, warning=FALSE, tidy = FALSE)
library(BiocStyle)
library(scater)
library(Seurat)
library(igraph)

## -----------------------------------------------------------------------------
suppressMessages(library(scRepertoire))

## ----eval=FALSE---------------------------------------------------------------
# S1 <- read.csv(".../Sample1/outs/filtered_contig_annotations.csv")
# S2 <- read.csv(".../Sample2/outs/filtered_contig_annotations.csv")
# S3 <- read.csv(".../Sample3/outs/filtered_contig_annotations.csv")
# S4 <- read.csv(".../Sample4/outs/filtered_contig_annotations.csv")
# 
# contig_list <- list(S1, S2, S3, S4)

## ----eval=FALSE, tidy = FALSE-------------------------------------------------
# # Directory example
# contig.output <- c("~/Documents/MyExperiment")
# contig.list <- loadContigs(input = contig.output,
#                            format = "TRUST4")

## ----eval = FALSE, tidy = FALSE-----------------------------------------------
# # List of data frames example
# S1 <- read.csv("~/Documents/MyExperiment/Sample1/outs/barcode_results.csv")
# S2 <- read.csv("~/Documents/MyExperiment/Sample2/outs/barcode_results.csv")
# S3 <- read.csv("~/Documents/MyExperiment/Sample3/outs/barcode_results.csv")
# S4 <- read.csv("~/Documents/MyExperiment/Sample4/outs/barcode_results.csv")
# 
# contig.list <- list(S1, S2, S3, S4)
# contig.list <- loadContigs(input = contig.list,
#                            format = "WAT3R")

## ----eval = F, tidy = FALSE---------------------------------------------------
# contigs <- read.csv(".../outs/filtered_contig_annotations.csv")
# 
# contig.list <- createHTOContigList(contigs,
#                                    Seurat.Obj,
#                                    group.by = "HTO_maxID")

## ----tidy = FALSE-------------------------------------------------------------
data("contig_list") #the data built into scRepertoire

head(contig_list[[1]])

## ----tidy = FALSE-------------------------------------------------------------
combined.TCR <- combineTCR(contig_list, 
                           samples = c("P17B", "P17L", "P18B", "P18L", 
                                            "P19B","P19L", "P20B", "P20L"),
                           removeNA = FALSE, 
                           removeMulti = FALSE, 
                           filterMulti = FALSE)

head(combined.TCR[[1]])

## ----tidy = FALSE-------------------------------------------------------------
# Load example BCR contig data
BCR.contigs <- read.csv("https://www.borch.dev/uploads/contigs/b_contigs.csv")

## ----tidy = FALSE-------------------------------------------------------------

# Combine using the default similarity clustering
combined.BCR.clustered <- combineBCR(BCR.contigs, 
                                     samples = "Patient1", 
                                     threshold = 0.85)

# The CTstrict column contains cluster IDs (e.g., "cluster.1")
head(combined.BCR.clustered[[1]][, c("barcode", "CTstrict", "IGH", "cdr3_aa1")])

## -----------------------------------------------------------------------------
combined.BCR.aligned <- combineBCR(BCR.contigs, 
                                   samples = "Patient1",
                                   sequence = "aa",        
                                   dist_type = "nw",      
                                   dist_mat = "BLOSUM80",  
                                   threshold = 0.85)

head(combined.BCR.aligned[[1]][, c("barcode", "CTstrict", "IGH", "cdr3_aa1")])

## -----------------------------------------------------------------------------
cleaned.BCR <- combineBCR(BCR.contigs,
                          samples = "Patient1",
                          filterNonproductive = TRUE,
                          filterMulti = TRUE)

head(cleaned.BCR[[1]])

## ----tidy = FALSE-------------------------------------------------------------
combined.TCR <- addVariable(combined.TCR, 
                            variable.name = "Type", 
                            variables = rep(c("B", "L"), 4))

head(combined.TCR[[1]])

## ----tidy = FALSE-------------------------------------------------------------
subset1 <- subsetClones(combined.TCR, 
                        name = "sample", 
                        variables = c("P18L", "P18B"))

head(subset1[[1]][,1:4])

## -----------------------------------------------------------------------------
subset2 <- combined.TCR[c(3,4)]
head(subset2[[1]][,1:4])

## ----eval = FALSE, tidy = FALSE-----------------------------------------------
# exportClones(combined,
#              write.file = TRUE,
#              dir = "~/Documents/MyExperiment/Sample1/"
#              file.name = "clones.csv")

## -----------------------------------------------------------------------------
immunarch <- exportClones(combined.TCR, 
                          format = "immunarch", 
                          write.file = FALSE)
head(immunarch[[1]][[1]])

## -----------------------------------------------------------------------------
combined.TCR <- clonalBin(combined.TCR, clone.call = "strict")

# Check the cloneSize column
head(combined.TCR[[1]][, c("barcode", "CTstrict", "clonalProportion", "clonalFrequency", "cloneSize")])

# View the distribution of clone sizes
table(combined.TCR[[1]]$cloneSize)

## -----------------------------------------------------------------------------
combined.TCR.freq <- combineTCR(contig_list,
                                samples = c("P17B", "P17L", "P18B", "P18L",
                                            "P19B","P19L", "P20B", "P20L"))

combined.TCR.freq <- clonalBin(combined.TCR.freq,
                               clone.call = "strict",
                               proportion = FALSE,
                               clone.size = c(Rare = 1, Small = 5, Medium = 20,
                                              Large = 100, Hyperexpanded = 500))

# Check the frequency-based binning
head(combined.TCR.freq[[1]][, c("barcode", "clonalFrequency", "cloneSize")])

## -----------------------------------------------------------------------------
combined.TCR.grouped <- combineTCR(contig_list,
                                   samples = c("P17B", "P17L", "P18B", "P18L",
                                               "P19B","P19L", "P20B", "P20L"))

combined.TCR.grouped <- addVariable(combined.TCR.grouped,
                                    variable.name = "Type",
                                    variables = rep(c("B", "L"), 4))

combined.TCR.grouped <- clonalBin(combined.TCR.grouped,
                                  clone.call = "strict",
                                  group.by = "Type")

# Check grouped binning
head(combined.TCR.grouped[[1]][, c("barcode", "Type", "clonalProportion", "cloneSize")])

## ----eval = FALSE, tidy = FALSE-----------------------------------------------
# combined <- annotateInvariant(combined,
#                               type = "MAIT",
#                               species = "human")
# 
# combined <- annotateInvariant(combined,
#                               type = "iNKT",
#                               species = "human")

## ----tidy = FALSE-------------------------------------------------------------
clonalQuant(combined.TCR, 
            cloneCall="strict", 
            chain = "both", 
            scale = TRUE)

## -----------------------------------------------------------------------------
clonalQuant(combined.TCR, 
            cloneCall="gene", 
            group.by = "Type", 
            scale = TRUE)

## ----tidy = FALSE-------------------------------------------------------------
clonalAbundance(combined.TCR, 
                cloneCall = "gene", 
                scale = FALSE)

## -----------------------------------------------------------------------------
clonalAbundance(combined.TCR, 
                cloneCall = "gene", 
                scale = TRUE)

## ----tidy = FALSE-------------------------------------------------------------
clonalLength(combined.TCR, 
             cloneCall="aa", 
             chain = "both") 

## ----tidty = FALSE------------------------------------------------------------
clonalLength(combined.TCR, 
             cloneCall="aa", 
             chain = "TRA", 
             scale = TRUE) 

## ----tidy = FALSE-------------------------------------------------------------
clonalCompare(combined.TCR, 
                  top.clones = 10, 
                  samples = c("P17B", "P17L"), 
                  cloneCall="aa", 
                  graph = "alluvial")

## ----tidy = FALSE-------------------------------------------------------------
clonalCompare(combined.TCR, 
              top.clones = 10,
              highlight.clones = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF", "NA_CASSVRRERANTGELFF"),
              relabel.clones = TRUE,
              samples = c("P17B", "P17L"), 
              cloneCall="aa", 
              graph = "alluvial")

## -----------------------------------------------------------------------------
clonalCompare(combined.TCR, 
              clones = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF", "NA_CASSVRRERANTGELFF"),
              relabel.clones = TRUE,
              samples = c("P17B", "P17L"), 
              cloneCall="aa", 
              graph = "alluvial")

## ----tidy = FALSE-------------------------------------------------------------
clonalScatter(combined.TCR, 
              cloneCall ="gene", 
              x.axis = "P18B", 
              y.axis = "P18L",
              dot.size = "total",
              graph = "proportion")

## ----tidy = FALSE-------------------------------------------------------------
clonalHomeostasis(combined.TCR, 
                  cloneCall = "gene")


## ----tidy = FALSE-------------------------------------------------------------
clonalHomeostasis(combined.TCR, 
                  cloneCall = "gene",
                  cloneSize = c(Rare = 0.001, Small = 0.01, Medium = 0.1, 
                                Large = 0.3, Hyperexpanded = 1))

## ----tidy = FALSE-------------------------------------------------------------
combined.TCR <- addVariable(combined.TCR, 
                            variable.name = "Type", 
                            variables = rep(c("B", "L"), 4))

## ----tidy = FALSE-------------------------------------------------------------
clonalHomeostasis(combined.TCR, 
                  group.by = "Type",
                  cloneCall = "gene")

## ----tidy = FALSE-------------------------------------------------------------
clonalProportion(combined.TCR, 
                 cloneCall = "gene") 

## -----------------------------------------------------------------------------
clonalProportion(combined.TCR, 
                 cloneCall = "nt",
                 clonalSplit = c(1, 5, 10, 100, 1000, 10000)) 

## ----tidy = FALSE-------------------------------------------------------------
vizGenes(combined.TCR,
         x.axis = "TRBV",
         y.axis = NULL, # No specific y-axis variable, will group all samples
         plot = "barplot",
         summary.fun = "proportion") 

## ----tidy = FALSE-------------------------------------------------------------
# Peripheral Blood Samples
vizGenes(combined.TCR[c("P17B", "P18B", "P19B", "P20B")],
         x.axis = "TRBV",
         y.axis = "TRBJ",
         plot = "heatmap",
         summary.fun = "percent") # Display percentages

# Lung Samples
vizGenes(combined.TCR[c("P17L", "P18L", "P19L", "P20L")],
         x.axis = "TRBV",
         y.axis = "TRBJ",
         plot = "heatmap",
         summary.fun = "percent") # Display percentages

## ----tidy = FALSE-------------------------------------------------------------
vizGenes(combined.TCR[c("P17B", "P17L")],
         x.axis = "TRBV",
         y.axis = "TRAV",
         plot = "heatmap",
         summary.fun = "count") 

## -----------------------------------------------------------------------------
percentGenes(combined.TCR,
             chain = "TRB",
             gene = "Vgene",
             summary.fun = "percent")

## -----------------------------------------------------------------------------
df.genes <- percentGenes(combined.TCR,
                         chain = "TRB",
                         gene = "Vgene",
                         exportTable = TRUE,
                         summary.fun = "proportion") 

# Performing PCA on the gene usage matrix
pc <- prcomp(t(df.genes) )

# Getting data frame to plot from
df_plot <- as.data.frame(cbind(pc$x[,1:2], colnames(df.genes)))
colnames(df_plot) <- c("PC1", "PC2", "Sample")
df_plot$PC1 <- as.numeric(df_plot$PC1)
df_plot$PC2 <- as.numeric(df_plot$PC2)

ggplot(df_plot, aes(x = PC1, y = PC2)) +
  geom_point(aes(fill = Sample), shape = 21, size = 5) +
  guides(fill=guide_legend(title="Samples")) +
  scale_fill_manual(values = hcl.colors(nrow(df_plot), "inferno")) +
  theme_classic() +
  labs(title = "PCA of TRBV Gene Usage")

## -----------------------------------------------------------------------------
percentVJ(combined.TCR[1:2],
          chain = "TRB",
          summary.fun = "percent")

## -----------------------------------------------------------------------------
df.vj <- percentVJ(combined.TCR,
                   chain = "TRB",
                   exportTable = TRUE,
                   summary.fun = "proportion") # Export proportions for PCA

# Performing PCA on the V-J pairing matrix
pc.vj <- prcomp(t(df.vj))

# Getting data frame to plot from
df_plot_vj <- as.data.frame(cbind(pc.vj$x[,1:2], colnames(df.vj)))
colnames(df_plot_vj) <- c("PC1", "PC2", "Sample")
df_plot_vj$PC1 <- as.numeric(df_plot_vj$PC1)
df_plot_vj$PC2 <- as.numeric(df_plot_vj$PC2)

# Plotting the PCA results
ggplot(df_plot_vj, aes(x = PC1, y = PC2)) +
  geom_point(aes(fill = Sample), shape = 21, size = 5) +
  guides(fill=guide_legend(title="Samples")) +
  scale_fill_manual(values = hcl.colors(nrow(df_plot_vj), "inferno")) +
  theme_classic() +
  labs(title = "PCA of TRBV-TRBJ Gene Pairings")

## ----message = FALSE, tidy = FALSE--------------------------------------------
percentAA(combined.TCR, 
          chain = "TRB", 
          aa.length = 20)

## ----tidy = FALSE-------------------------------------------------------------
positionalEntropy(combined.TCR, 
                  chain = "TRB", 
                  aa.length = 20)

## ----tidy = FALSE-------------------------------------------------------------
positionalProperty(combined.TCR[c(1,2)], 
                  chain = "TRB", 
                  aa.length = 20, 
                  method = "atchleyFactors") + 
  scale_color_manual(values = hcl.colors(5, "inferno")[c(2,4)])

## ----tidy = FALSE-------------------------------------------------------------
percentKmer(combined.TCR, 
            cloneCall = "aa",
            chain = "TRB", 
            motif.length = 3, 
            top.motifs = 25)

## -----------------------------------------------------------------------------
percentKmer(combined.TCR, 
            cloneCall = "nt",
            chain = "TRB", 
            motif.length = 3, 
            top.motifs = 25)

## ----tidy = FALSE-------------------------------------------------------------
clonalDiversity(combined.TCR, 
                cloneCall = "gene")

## ----tidy = FALSE-------------------------------------------------------------
combined.TCR <- addVariable(combined.TCR, 
                            variable.name = "Patient", 
                             variables = c("P17", "P17", "P18", "P18", 
"P19","P19", "P20", "P20"))

## ----tidy = FALSE-------------------------------------------------------------
clonalDiversity(combined.TCR, 
                cloneCall = "gene", 
                group.by = "Patient", 
                metric = "inv.simpson")

## ----tidy = FALSE-------------------------------------------------------------
clonalDiversity(combined.TCR, 
                cloneCall = "gene", 
                x.axis = "Patient", 
                metric = "inv.simpson")

## ----message=FALSE, tidy = FALSE----------------------------------------------
clonalRarefaction(combined.TCR,
                  plot.type = 1,
                  hill.numbers = 0,
                  n.boots = 2)

clonalRarefaction(combined.TCR,
                  plot.type = 2,
                  hill.numbers = 0,
                  n.boots = 2)

clonalRarefaction(combined.TCR,
                  plot.type = 3,
                  hill.numbers = 0,
                  n.boots = 2)

## ----tidy = FALSE-------------------------------------------------------------
clonalRarefaction(combined.TCR,
                  plot.type = 1,
                  hill.numbers = 1,
                  n.boots = 2)

clonalRarefaction(combined.TCR,
                  plot.type = 2,
                  hill.numbers = 1,
                  n.boots = 2)

clonalRarefaction(combined.TCR,
                  plot.type = 3,
                  hill.numbers = 1,
                  n.boots = 2)

## ----tidy = FALSE-------------------------------------------------------------
clonalSizeDistribution(combined.TCR, 
                       cloneCall = "aa", 
                       method= "ward.D2")

## ----tidy = FALSE-------------------------------------------------------------
clonalOverlap(combined.TCR, 
              cloneCall = "strict", 
              method = "morisita")

## ----tidy = FALSE-------------------------------------------------------------
clonalOverlap(combined.TCR, 
              cloneCall = "strict", 
              method = "raw")

## ----tidy = FALSE-------------------------------------------------------------
scRep_example <- get(data("scRep_example"))

#Making a Single-Cell Experiment object
sce <- Seurat::as.SingleCellExperiment(scRep_example)

#Adding patient information
scRep_example$Patient <- substr(scRep_example$orig.ident, 1,3)

#Adding type information
scRep_example$Type <- substr(scRep_example$orig.ident, 4,4)

## ----tidy = FALSE-------------------------------------------------------------
# Check the first 10 variable features before removal
VariableFeatures(scRep_example)[1:10]

## ----tidy = FALSE-------------------------------------------------------------
# Remove TCR VDJ genes
scRep_example <- quietTCRgenes(scRep_example)

# Check the first 10 variable features after removal
VariableFeatures(scRep_example)[1:10]

## ----tidy = FALSE-------------------------------------------------------------
sce <- combineExpression(combined.TCR, 
                         sce, 
                         cloneCall="gene", 
                         group.by = "sample", 
                         proportion = TRUE)

#Define color palette 
colorblind_vector <- hcl.colors(n=7, palette = "inferno", fixup = TRUE)

plotUMAP(sce, colour_by = "cloneSize") +
    scale_color_manual(values=rev(colorblind_vector[c(1,3,5,7)]))

## ----tidy = FALSE-------------------------------------------------------------
scRep_example <- combineExpression(combined.TCR, 
                                   scRep_example, 
                                   cloneCall="gene", 
                                   group.by = "sample", 
                                   proportion = FALSE, 
                                   cloneSize=c(Single=1, Small=5, Medium=20, Large=100, Hyperexpanded=500))

# Bug in Seurat 5.3.1 no longer handles NA for DimPlot
# Will Update in Future Commits

# Seurat::DimPlot(scRep_example, group.by = "cloneSize") +
#    scale_color_manual(values=rev(colorblind_vector[c(1,3,4,5,7)]))

## ----eval=FALSE, tidy = FALSE-------------------------------------------------
# #This is an example of the process, which will not be evaluated during knit
# TCR <- combineTCR(...)
# BCR <- combineBCR(...)
# list.receptors <- c(TCR, BCR)
# 
# 
# seurat <- combineExpression(list.receptors,
#                             seurat,
#                             cloneCall="gene",
#                             proportion = TRUE)

## ----tidy = FALSE-------------------------------------------------------------
clonalOverlay(scRep_example, 
              reduction = "umap", 
              cutpoint = 1, 
              bins = 10, 
              facet.by = "Patient") + 
              guides(color = "none")

## ----tidy = FALSE-------------------------------------------------------------
#ggraph needs to be loaded due to issues with ggplot
library(ggraph)

#No Identity filter
clonalNetwork(scRep_example, 
              reduction = "umap", 
              group.by = "seurat_clusters",
              filter.clones = NULL,
              filter.identity = NULL,
              cloneCall = "aa")

## ----tidy = FALSE-------------------------------------------------------------
#Examining Cluster 3 only
clonalNetwork(scRep_example, 
              reduction = "umap", 
              group.by = "seurat_clusters",
              filter.identity = 3,
              cloneCall = "aa")

## ----tidy = FALSE-------------------------------------------------------------
shared.clones <- clonalNetwork(scRep_example, 
                               reduction = "umap", 
                               group.by = "seurat_clusters",
                               cloneCall = "aa", 
                               exportClones = TRUE)
head(shared.clones)

## ----tidy = FALSE-------------------------------------------------------------
scRep_example <- highlightClones(scRep_example, 
                    cloneCall= "aa", 
                    sequence = c("CAERGSGGSYIPTF_CASSDPSGRQGPRWDTQYF", 
                                 "CARKVRDSSYKLIF_CASSDSGYNEQFF"))

# Bug in Seurat 5.3.1 no longer handles NA for DimPlot
# Will Update in Future Commits

# Seurat::DimPlot(scRep_example, group.by = "highlight") + 
#   guides(color=guide_legend(nrow=3,byrow=TRUE)) + 
#   ggplot2::theme(plot.title = element_blank(), 
#                 legend.position = "bottom")

## ----tidy = FALSE-------------------------------------------------------------
clonalOccupy(scRep_example, 
              x.axis = "seurat_clusters")

## -----------------------------------------------------------------------------
clonalOccupy(scRep_example, 
             x.axis = "ident", 
             proportion = TRUE, 
             label = FALSE)

## ----tidy = FALSE-------------------------------------------------------------
alluvialClones(scRep_example,
               clone.call = "aa",
               y.axes = c("Patient", "ident", "Type"),
               color = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF", "NA_CASSVRRERANTGELFF")) +
    scale_fill_manual(values = c("grey", colorblind_vector[3]))

## ----tidy = FALSE-------------------------------------------------------------
alluvialClones(scRep_example,
               clone.call = "gene",
               y.axes = c("Patient", "ident", "Type"),
               color = "ident")

## ----tidy = FALSE-------------------------------------------------------------
alluvialClones(scRep_example,
               clone.call = "aa",
               y.axes = c("Patient", "ident"),
               top.clones = 25,
               color = "ident")

## ----tidy = FALSE-------------------------------------------------------------
alluvialClones(scRep_example,
               clone.call = "aa",
               y.axes = c("Patient", "ident", "Type"),
               highlight.clones = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF"),
               highlight.color = "red")

## ----tidy = FALSE-------------------------------------------------------------
alluvialClones(scRep_example,
               clone.call = "gene",
               y.axes = c("Patient", "ident"),
               color = "ident",
               stratum.width = 0.3,
               flow.alpha = 0.7,
               label.size = 3)

## ----tidy = FALSE-------------------------------------------------------------
library(circlize)
library(scales)

vizCirclize(scRep_example,
            group.by = "seurat_clusters",
            clone.call = "aa")

## ----tidy = FALSE-------------------------------------------------------------
circles <- getCirclize(scRep_example,
                       group.by = "seurat_clusters")

#Just assigning the normal colors to each cluster
grid.cols <- hue_pal()(length(unique(scRep_example$seurat_clusters)))
names(grid.cols) <- unique(scRep_example$seurat_clusters)

#Graphing the chord diagram
chordDiagram(circles, self.link = 1, grid.col = grid.cols)

## ----tidy = FALSE-------------------------------------------------------------
vizCirclize(scRep_example,
            group.by = c("Patient", "Type"),
            clone.call = "aa",
            min.shared = 1)

## ----tidy = FALSE-------------------------------------------------------------
vizCirclize(scRep_example,
            group.by = "seurat_clusters",
            clone.call = "aa",
            directional = TRUE)

## ----tidy = FALSE-------------------------------------------------------------
# Jaccard similarity
circles.jaccard <- getCirclize(scRep_example,
                               group.by = "seurat_clusters",
                               method = "jaccard",
                               include.metadata = TRUE)

# The include.metadata option returns sector statistics
head(circles.jaccard$sector.stats)

## -----------------------------------------------------------------------------
subset <- subset(scRep_example, Type == "L")

circles <- getCirclize(subset, group.by = "ident", proportion = TRUE)

grid.cols <- scales::hue_pal()(length(unique(subset@active.ident)))
names(grid.cols) <- levels(subset@active.ident)

chordDiagram(circles,
             self.link = 1,
             grid.col = grid.cols,
             directional = 1,
             direction.type = "arrows",
             link.arr.type = "big.arrow")

## -----------------------------------------------------------------------------
# Calculate and plot all three STARTRAC indices
StartracDiversity(scRep_example, 
                  type = "Type", 
                  group.by = "Patient")

## -----------------------------------------------------------------------------
# Calculate and plot only the clonal expansion index
StartracDiversity(scRep_example, 
                  type = "Type", 
                  group.by = "Patient",
                  index = "expa")

## -----------------------------------------------------------------------------
# # Calculate pairwise migration between tissues
StartracDiversity(scRep_example, 
                  type = "Type", 
                  group.by = "Patient",
                  index = "migr",
                  pairwise = "Type")

## ----message = FALSE, tidy = FALSE--------------------------------------------
clonalBias(scRep_example, 
           cloneCall = "aa", 
           split.by = "Patient", 
           group.by = "seurat_clusters",
           n.boots = 10, 
           min.expand =5)

## ----tidy = FALSE-------------------------------------------------------------
# Run clustering on the first two samples for the TRA chain
sub_combined <- clonalCluster(combined.TCR[c(1,2)], 
                              chain = "TRA", 
                              sequence = "aa", 
                              threshold = 0.85)

# View the new cluster column
head(sub_combined[[1]][, c("barcode", "TCR1", "TRA.Cluster")])

## ----tidy = FALSE-------------------------------------------------------------
#Adding patient information
scRep_example$Patient <- substr(scRep_example$orig.ident, 1,3)

#Adding type information
scRep_example$Type <- substr(scRep_example$orig.ident, 4,4)

## -----------------------------------------------------------------------------
# Run clustering, but group calculations by "Patient"
scRep_example <- clonalCluster(scRep_example, 
                               chain = "TRA", 
                               sequence = "aa", 
                               threshold = 0.85, 
                               group.by = "Patient")

#Define color palette 
num_clusters <- length(unique(na.omit(scRep_example$TRA.Cluster)))
cluster_colors <- hcl.colors(n = num_clusters, palette = "inferno")

# Bug in Seurat 5.3.1 no longer handles NA for DimPlot
# Will Update in Future Commits

# DimPlot(scRep_example, group.by = "TRA.Cluster") +
#  scale_color_manual(values = cluster_colors, na.value = "grey") + 
#  NoLegend()

## ----tidy = FALSE-------------------------------------------------------------
set.seed(42)
#Clustering Patient 19 samples
igraph.object <- clonalCluster(combined.TCR[c(5,6)],
                               chain = "TRB",
                               sequence = "aa",
                               group.by = "sample",
                               threshold = 0.85, 
                               exportGraph = TRUE)

#Setting color scheme
col_legend <- factor(igraph::V(igraph.object)$group)
col_samples <- hcl.colors(2,"inferno")[as.numeric(col_legend)]
color.legend <- factor(unique(igraph::V(igraph.object)$group))
sample.vertices <- V(igraph.object)[sample(length(igraph.object), 500)]

subgraph.object <- induced_subgraph(igraph.object, vids = sample.vertices)
V(subgraph.object)$degrees <- igraph::degree(subgraph.object)
edge_alpha_color <- adjustcolor("gray", alpha.f = 0.3)

#Plotting
plot(subgraph.object,
     layout = layout_nicely(subgraph.object),
     vertex.label = NA,
     vertex.size = sqrt(igraph::V(subgraph.object)$degrees), 
     vertex.color = col_samples[sample.vertices],
     vertex.frame.color = "white", 
     edge.color = edge_alpha_color,
     edge.arrow.size = 0.05,
     edge.curved = 0.05, 
     margin = -0.1)
legend("topleft", 
       legend = levels(color.legend), 
       pch = 16, 
       col = unique(col_samples), 
       bty = "n")

## -----------------------------------------------------------------------------
# Generate the sparse matrix
adj.matrix <- clonalCluster(combined.TCR[c(1,2)],
                            chain = "TRB",
                            exportAdjMatrix = TRUE)

# View the dimensions and a snippet of the matrix
dim(adj.matrix)
print(adj.matrix[1:10, 1:10])

## -----------------------------------------------------------------------------
# Cluster using both TRB and TRA chains simultaneously
clustered_both <- clonalCluster(combined.TCR[c(1,2)], 
                                chain = "both")

# View the new "Multi.Cluster" column
head(clustered_both[[1]][, c("barcode", "TCR1", "TCR2", "Multi.Cluster")])

## -----------------------------------------------------------------------------
# Cluster using the walktrap algorithm
graph_walktrap <- clonalCluster(combined.TCR[c(1,2)],
                                cluster.method = "walktrap",
                                exportGraph = TRUE)

# Compare the number of clusters found
length(unique(V(graph_walktrap)$cluster))

## -----------------------------------------------------------------------------
sessionInfo()

#' Remove TCR and BCR genes from variable gene results
#'
#' Most single-cell workflows use highly-expressed and highly-variable
#' genes for the initial calculation of PCA and subsequent dimensional
#' reduction. This function will remove the TCR and/or BCR genes from the
#' variable features in a Seurat object or from a vector (potentially
#' generated by the Bioconductor scran workflow).
#'
#' @examples
#' example <- quietVDJgenes(scRep_example)
#' scRep <- quietTCRgenes(scRep_example)
#' ibex_example <- quietBCRgenes(scRep_example)
#'
#' @param input.data Single-cell object in Seurat format or vector of variable genes
#' to use in reduction
#' @param assay The Seurat assay slot to use to remove immune receptor genes
#' from, NULL value will default to the default assay
#' @param ... Reserved for future arguments
#'
#' @export
#' @return Seurat object or vector list with TCR genes removed.
quietVDJgenes <- function(input.data, ...) {
    quietTCRgenes(quietBCRgenes(input.data, ...), ...)
}

#' @rdname quietVDJgenes
#' @export
#' @author Nicky de Vrij, Nikolaj Pagh, Nick Borcherding, Qile Yang
quietTCRgenes <- function(input.data, ...) {
    UseMethod("quietTCRgenes")
}

#' @rdname quietVDJgenes
#' @method quietTCRgenes default
#' @export
quietTCRgenes.default <- function(input.data, ...) {
    input.data[!shouldQuietTcrGene(input.data)]
}

shouldQuietTcrGene <- function(genes) {
    grepl("^TR[ABDG][VDJ][^D]", genes, ignore.case = TRUE)
}

#' @rdname quietVDJgenes
#' @method quietTCRgenes Seurat
#' @export
quietTCRgenes.Seurat <- function(input.data, assay = NULL, ...) {
    if (is.null(assay)) {
        assay <- SeuratObject::DefaultAssay(input.data)
    }
    SeuratObject::VariableFeatures(input.data, assay = assay) <-
        quietTCRgenes.default(SeuratObject::VariableFeatures(input.data, assay = assay))
    input.data
}

#' @rdname quietVDJgenes
#' @export
quietBCRgenes <- function(input.data, ...) {
    UseMethod("quietBCRgenes")
}

#' @rdname quietVDJgenes
#' @method quietBCRgenes default
#' @export
quietBCRgenes.default <- function(input.data, ...) {
    input.data[!shouldQuietBcrGene(input.data)]
}

shouldQuietBcrGene <- function(genes) {
    grepl("^IG[HLK][VDJCAGM]", genes, ignore.case = TRUE) &
    !genes %in% c(getHumanIgPseudoGenes(), getMouseIgPseudoGenes()) |
    toupper(genes) == "JCHAIN"
}

#' @rdname quietVDJgenes
#' @method quietBCRgenes Seurat
#' @export
quietBCRgenes.Seurat <- function(input.data, assay = NULL, ...) {
    if (is.null(assay)) {
        assay <- SeuratObject::DefaultAssay(input.data)
    }
    SeuratObject::VariableFeatures(input.data, assay = assay) <-
        quietBCRgenes.default(SeuratObject::VariableFeatures(input.data, assay = assay))
    input.data
}

#' Get Human Immunoglobulin pseudogenes
#'
#' This function returns a character vector of human immunoglobulin
#' pseudogenes. These are also the genes that are removed from the
#' variable gene list in [quietVDJgenes()].
#'
#' @return Character vector of human immunoglobulin pseudogenes.
#' @keywords internal
getHumanIgPseudoGenes <- function() {
    unique(c(
        "IGHJ1P", "IGHJ2P", "IGHJ3P", "IGLC4", "IGLC5", "IGHEP1", "IGHEP2",
        "IGHV1-12","IGHV1-14", "IGHV1-17", "IGHV1-67", "IGHV1-68",
        "IGHV2-10", "IGHV3-6", "IGHV3-19", "IGHV3-22", "IGHV3-25",
        "IGHV3-29", "IGHV3-32", "IGHV3-36", "IGHV3-37", "IGHV3-41",
        "IGHV3-42", "IGHV3-47", "IGHV3-50", "IGHV3-52", "IGHV3-54",
        "IGHV3-57", "IGHV3-60", "IGHV3-62", "IGHV3-63", "IGHV3-65",
        "IGHV3-71", "IGHV3-75", "IGHV3-76", "IGHV3-79", "IGHV4-55",
        "IGHV4-80", "IGHV5-78", "IGHV7-27", "IGHV7-40", "IGHV7-56",
        "IGHVIII-44", "IGHVIII-82", "IGKV1-22", "IGKV1-32", "IGKV1-35",
        "IGKV1D-22", "IGKV1D-27", "IGKV1D-32", "IGKV1D-35", "IGKVOR-2",
        "IGKVOR-3", "IGKVOR-4", "IGKV2-4", "IGKV2-10", "IGKV2-14", "IGKV2-18",
        "IGKV2-19", "IGKV2-23", "IGKV2-26", "IGKV2-36", "IGKV2-38",
        "IGKV2D-10", "IGKV2D-14", "IGKV2D-18", "IGKV2D-19", "IGKV2D-23",
        "IGKV2D-36", "IGKV2D-38", "IGKV3-25", "IGKV3-31", "IGKV3-34",
        "IGKV7-3", "IGLCOR22-1", "IGLCOR22-2", "IGLJCOR18", "IGLV1-41",
        "IGLV1-62", "IGLV2-5", "IGLV2-28", "IGLV2-34", "IGLV3-2",
        "IGLV3-4", "IGLV3-6", "IGLV3-7", "IGLV3-13", "IGLV3-15",
        "IGLV3-17", "IGLV3-24", "IGLV3-26", "IGLV3-29", "IGLV3-30",
        "IGLV3-31", "IGLV7-35", "IGLV10-67", "IGLVI-20", "IGLVI-38",
        "IGLVI-42", "IGLVI-56", "IGLVI-63", "IGLVI-68", "IGLVI-70",
        "IGLVIV-53", "IGLVIV-59", "IGLVIV-64", "IGLVIV-65", "IGLVV-58",
        "IGLVV-66", "IGHV1OR15-2", "IGHV1OR15-3", "IGHV1OR15-4", "IGHV1OR15-6",
        "IGHV1OR16-1", "IGHV1OR16-2", "IGHV1OR16-3", "IGHV1OR16-4", "IGHV3-30-2",
        "IGHV3-33-2", "IGHV3-69-1", "IGHV3OR15-7", "IGHV3OR16-6", "IGHV3OR16-7",
        "IGHV3OR16-11", "IGHV3OR16-14", "IGHV3OR16-15", "IGHV3OR16-16", "IGHV7-34-1",
        "IGHVII-1-1", "IGHVII-15-1", "IGHVII-20-1", "IGHVII-22-1", "IGHVII-26-2",
        "IGHVII-28-1", "IGHVII-30-1", "IGHVII-30-21", "IGHVII-31-1", "IGHVII-33-1",
        "IGHVII-40-1", "IGHVII-43-1", "IGHVII-44-2", "IGHVII-46-1", "IGHVII-49-1",
        "IGHVII-51-2", "IGHVII-53-1", "IGHVII-60-1", "IGHVII-62-1", "IGHVII-65-1",
        "IGHVII-67-1", "IGHVII-74-1", "IGHVII-78-1", "IGHVIII-2-1", "IGHVIII-5-1",
        "IGHVIII-5-2", "IGHVIII-11-1", "IGHVIII-13-1", "IGHVIII-16-1", "IGHVIII-22-2",
        "IGHVIII-25-1", "IGHVIII-26-1", "IGHVIII-38-1", "IGHVIII-47-1", "IGHVIII-67-2",
        "IGHVIII-67-3", "IGHVIII-67-4", "IGHVIII-76-1", "IGHVIV-44-1", "IGKV1OR1-1",
        "IGKV1OR2-1", "IGKV1OR2-2", "IGKV1OR2-3", "IGKV1OR2-6", "IGKV1OR2-9",
        "IGKV1OR2-11", "IGKV1OR2-118", "IGKV1OR9-1", "IGKV1OR9-2", "IGKV1OR10-1",
        "IGKV1OR15-118", "IGKV1OR22-1", "IGKV1OR22-5", "IGKV1ORY-1", "IGKV2OR2-1",
        "IGKV2OR2-2", "IGKV2OR2-4", "IGKV2OR2-7", "IGKV2OR2-7D", "IGKV2OR2-8",
        "IGKV2OR2-10", "IGKV2OR22-3", "IGKV2OR22-4", "IGKV3OR2-5", "IGKV3OR22-2",
        "IGKV8OR8-1", "IGLVIV-66-1", "IGLVIVOR22-1", "IGLVIVOR22-2", "IGLVVI-22-1",
        "IGLVVI-25-1", "IGLVVII-41-1"
    ))
}

getMouseIgPseudoGenes <- function() {
  unique(c(
    "Ighv1-1", "Ighv1-2", "Ighv1-3", "Ighv1-6", "Ighv1-8",
    "Ighv1-10", "Ighv1-13", "Ighv1-14", "Ighv1-17", "Ighv1-17-1",
    "Ighv1-18-1", "Ighv1-18-3", "Ighv1-18-5", "Ighv1-18-6", "Ighv1-18-8",
    "Ighv1-18-9", "Ighv1-18-10", "Ighv1-18-12", "Ighv1-18-15", "Ighv1-18-16",
    "Ighv1-18-18", "Ighv1-18-20", "Ighv1-18-21", "Ighv1-18-24", "Ighv1-18-25",
    "Ighv1-18-27", "Ighv1-18-29", "Ighv1-18-30", "Ighv1-18-31", "Ighv1-18-33",
    "Ighv1-18-34", "Ighv1-18-35", "Ighv1-18-38", "Ighv1-18-42", "Ighv1-18-43",
    "Ighv1-18-44", "Ighv1-18-46", "Ighv1-18-47", "Ighv1-18-48", "Ighv1-18-49",
    "Ighv1-18-51", "Ighv1-18-52", "Ighv1-18-54", "Ighv1-18-57", "Ighv1-18-59",
    "Ighv1-18-61", "Ighv1-18-61d", "Ighv1-18-63", "Ighv1-18-67", "Ighv1-18-70",
    "Ighv1-18-71", "Ighv1-18-72", "Ighv1-18-76", "Ighv1-18-77", "Ighv1-18-79",
    "Ighv1-18-83", "Ighv1-18-85", "Ighv1-18-87", "Ighv1-18-88", "Ighv1-18-90",
    "Ighv1-18-91", "Ighv1-18-93", "Ighv1-18-94", "Ighv1-18-95", "Ighv1-18-96",
    "Ighv1-18-97", "Ighv1-19-1", "Ighv1-21", "Ighv1-21-1", "Ighv1-25",
    "Ighv1-27", "Ighv1-28", "Ighv1-29", "Ighv1-30", "Ighv1-32",
    "Ighv1-33", "Ighv1-34", "Ighv1-35", "Ighv1-38", "Ighv1-40",
    "Ighv1-41", "Ighv1-42", "Ighv1-43", "Ighv1-44", "Ighv1-45",
    "Ighv1-46", "Ighv1-47-4", "Ighv1-47-12", "Ighv1-47-15", "Ighv1-47-16",
    "Ighv1-47-19", "Ighv1-47-21", "Ighv1-47-25", "Ighv1-47-38", "Ighv1-47-42",
    "Ighv1-47-49", "Ighv1-47-50", "Ighv1-47-54", "Ighv1-47-55", "Ighv1-47-57",
    "Ighv1-48", "Ighv1-51", "Ighv1-57", "Ighv1-58", "Ighv1-60",
    "Ighv1-62", "Ighv1-65", "Ighv1-67", "Ighv1-68", "Ighv1-70",
    "Ighv1-71-7", "Ighv1-71-13", "Ighv1-71-14", "Ighv1-73", "Ighv1-79",
    "Ighv1-83", "Ighv1-86", "Ighv1s9", "Ighv1s10", "Ighv1s11",
    "Ighv1s15", "Ighv1s16", "Ighv1s17", "Ighv1s18", "Ighv1s19",
    "Ighv1s28", "Ighv1s30", "Ighv1s51", "Ighv1s57", "Ighv1s65",
    "Ighv1s74", "Ighv1s84", "Ighv1s88", "Ighv1s101", "Ighv1s110",
    "Ighv1s112", "Ighv1s124", "Ighv2-1", "Ighv2-2-1", "Ighv2-7",
    "Ighv2-8", "Ighv3-2", "Ighv3-3", "Ighv3-3-1", "Ighv3-7",
    "Ighv4-2", "Ighv5-1", "Ighv5-3", "Ighv5-5", "Ighv5-5-1",
    "Ighv5-5-3", "Ighv5-7", "Ighv5-7-1", "Ighv5-7-2", "Ighv5-7-3",
    "Ighv5-7-4", "Ighv5-7-5", "Ighv5-7-6", "Ighv5-8", "Ighv5-8-1",
    "Ighv5-8-2", "Ighv5-8-3", "Ighv5-8-4", "Ighv5-10", "Ighv5-10-1",
    "Ighv5-10-2", "Ighv5-11", "Ighv5-11-1", "Ighv5-11-2", "Ighv5-11-3",
    "Ighv5-11d", "Ighv5-12-3", "Ighv5-13", "Ighv5-13-1", "Ighv5-18",
    "Ighv5-19", "Ighv5-21", "Ighv6-1", "Ighv6-1-1", "Ighv6-2",
    "Ighv6-2d", "Ighv7-2", "Ighv7-4", "Ighv8-1", "Ighv8-3",
    "Ighv8-4-1", "Ighv8-4-3", "Ighv8-4-5", "Ighv8-4-9", "Ighv8-4-10",
    "Ighv8-4-11", "Ighv8-4-20", "Ighv8-7", "Ighv8-8-1", "Ighv8-9",
    "Ighv8-10", "Ighv8-13", "Ighv8-13-1", "Ighv8-13-2", "Ighv8-13-5",
    "Ighv8-13-6", "Ighv8-13-7", "Ighv8-13-9", "Ighv8-14", "Ighv8-15",
    "Ighv8-16", "Ighv8-17", "Ighv8-18", "Ighv8s1", "Ighv8s6",
    "Ighv9-2-3", "Ighv10-1-1", "Ighv10-1-3", "Ighv10-1-5", "Ighv10-1-7",
    "Ighv10-2", "Ighv10-4", "Ighv10s5", "Ighv11-2-1", "Ighv12-1",
    "Ighv12-1-2", "Ighv12-1-3", "Ighv12-1-6", "Ighv12-1-7", "Ighv12-1-8",
    "Ighv12-2", "Ighv12s2", "Ighv13-1", "Ighv14-2", "Ighv15-1",
    "Ighv15-1-1", "Ighv16-1-1"
  ))
}

# test script for quietVDJgenes.R - testcases are NOT comprehensive!

# Sample gene vectors for testing
tcr_genes <- c("TRAV1-1", "TRBV2", "TRGV9", "TRDV3", "TRBJ1-1", "TRDJ2")
bcr_genes <- c("IGHV3-23", "IGKV1-5", "IGLV3-1", "IGHJ4", "IGLC2", "JCHAIN")
other_genes <- c("CD3D", "GZMB", "MS4A1", "CD19")
pseudogenes <- c("IGHV1-12", "IGKV1-22") # From getHumanIgPseudoGenes()
all_genes <- c(tcr_genes, bcr_genes, other_genes, pseudogenes)


test_that("shouldQuietTcrGene correctly identifies TCR genes", {
  expected <- c(TRUE, TRUE, TRUE, TRUE, TRUE, TRUE)
  expect_equal(shouldQuietTcrGene(tcr_genes), expected)
  
  expect_false(any(shouldQuietTcrGene(other_genes)))
  expect_false(any(shouldQuietTcrGene(bcr_genes)))
})

test_that("shouldQuietBcrGene correctly identifies BCR genes", {
  # All BCR genes except pseudogenes, plus JCHAIN, should be quieted
  bcr_genes_to_quiet <- c("IGHV3-23", "IGKV1-5", "IGLV3-1", "IGHJ4", "IGLC2", "JCHAIN")
  expect_true(all(shouldQuietBcrGene(bcr_genes_to_quiet)))
  expect_false(any(shouldQuietBcrGene(pseudogenes)))
  expect_false(any(shouldQuietBcrGene(other_genes)))
})

test_that("getHumanIgPseudoGenes returns expected character vector", {
  pseudos <- getHumanIgPseudoGenes()
  expect_type(pseudos, "character")
  expect_gt(length(pseudos), 0)
  expect_true("IGHV1-12" %in% pseudos)
  expect_true("IGLV1-41" %in% pseudos)
})

test_that("quietTCRgenes.default removes only TCR genes", {
  cleaned_genes <- quietTCRgenes.default(all_genes)
  expect_length(cleaned_genes, length(bcr_genes) + length(other_genes) + length(pseudogenes))
  expect_false(any(tcr_genes %in% cleaned_genes))
  expect_true(all(other_genes %in% cleaned_genes))
  expect_true(all(bcr_genes %in% cleaned_genes))
})

test_that("quietBCRgenes.default removes only BCR genes and JCHAIN", {
  cleaned_genes <- quietBCRgenes.default(all_genes)
  expected_genes <- c(tcr_genes, other_genes, pseudogenes)
  expect_equal(sort(cleaned_genes), sort(expected_genes))
  expect_false("JCHAIN" %in% cleaned_genes)
})

test_that("quietVDJgenes works as a wrapper on character vectors", {
  cleaned_genes <- quietVDJgenes(all_genes)
  expected_genes <- c(other_genes, pseudogenes)
  expect_equal(sort(cleaned_genes), sort(expected_genes))
  expect_length(cleaned_genes, length(other_genes) + length(pseudogenes))
})

test_that("Default methods handle empty input", {
  expect_equal(quietTCRgenes.default(character(0)), character(0))
  expect_equal(quietBCRgenes.default(character(0)), character(0))
  expect_equal(quietVDJgenes(character(0)), character(0))
})

#' Adding Variables After combineTCR() or combineBCR()
#'
#' This function adds variables to the product of [combineTCR()], 
#' or [combineBCR()] to be used in later visualizations. 
#' For each element, the function will add a column (labeled by 
#' `variable.name`) with the variable. The length of the 
#' `variables` parameter needs to match the length of the 
#' combined object.
#'
#' @examples
#' combined <- combineTCR(contig_list, 
#'                        samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' combined <- addVariable(combined, 
#'                         variable.name = "Type", 
#'                         variables = rep(c("B", "L"), 4))
#'
#' @param input.data The product of [combineTCR()] or [combineBCR()].
#' @param variable.name A character string that defines the new variable to add.
#' @param variables A character vector defining the desired column value for 
#' each list element. Must match the length of the product of [combineTCR()] or 
#' [combineBCR()].
#' @export
#' @concept Loading_and_Processing_Contigs
#' @return input.data list with the variable column added to each element.
addVariable <- function(input.data, 
                        variable.name = NULL, 
                        variables =  NULL) {
  if (length(input.data) != length(variables)) {
    stop("Make sure the variables match the length of the contig list")
  }
  for (i in seq_along(input.data)) {
    input.data[[i]][,variable.name] <- variables[i]
  }
  return(input.data)
}

#' A List of Eight Single-cell TCR Sequencing Runs.
#' 
#' @description A list of 8 `filtered_contig_annotations.csv` files
#' outputted from 10X Cell Ranger. More information on the 
#' data can be found in the following
#'  [manuscript](https://pubmed.ncbi.nlm.nih.gov/33622974/).
#'  
#' @docType data
#' @concept Data
#' @name contig_list
#' 
NULL

#' A Seurat Object of 500 Single T cells,
#' 
#' @description The object is compatible with `contig_list` and the TCR
#' sequencing data can be added with `combineExpression`.  The data is 
#' from 4 patients with acute respiratory distress, with samples taken
#' from both the lung and peripheral blood. More information on the 
#' data can be found in the following
#'  [manuscript](https://pubmed.ncbi.nlm.nih.gov/33622974/).
#' 
#' @name scRep_example
#' @concept Data
#' @docType data
#'
NULL

#' Get Contig Doublets
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' This function identifies potential doublets by finding common barcodes
#' between TCR and BCR outputs. It extracts unique barcodes from each list
#' of dataframes, finds the intersection of the barcodes, and joins the
#' resulting data.
#'
#' @param tcrOutput Output of [combineTCR()]. A list of data.frames containing
#' TCR contig information, each dataframe must have a `barcode` column.
#' @param bcrOutput Output of [combineBCR()]. A list of data.frames containing
#' BCR contig information, each dataframe must have a `barcode` column.
#'
#' @return
#' A dataframe of barcodes that exist in both the TCR and BCR data, with
#' columns from both sets of data. There will be an additional column
#' `contigType` of type factor with levels 'TCR' and 'BCR' indicating the
#' origin of the contig - this will be the new first column.
#'
#' If there are no doublets, the returned
#' data.frame will have the same colnames but no rows.
#'
#' @export
getContigDoublets <- function(tcrOutput, bcrOutput) {

    doubletBarcodes <- getContigDoubletBarcodes(tcrOutput, bcrOutput)

    if (length(doubletBarcodes) == 0) {
        autoFullJoin(tcrOutput[[1]][0, ], bcrOutput[[1]][0, ]) %>%
            dplyr::mutate(
                contigType = factor(character(0), levels = c("BCR", "TCR")),
                .before = 1
            ) %>%
            return()
    }

    list(TCR = tcrOutput, BCR = bcrOutput) %>%
        lapplyOnAll(function(df) {
            df[df$barcode %in% doubletBarcodes, ]
        }) %>%
        purrr::imap(function(x, type) {
            dplyr::bind_rows(x) %>%
                dplyr::mutate(contigType = type)
        }) %>%
        purrr::reduce(autoFullJoin) %>%
        dplyr::mutate(
            contigType = factor(contigType, levels = c("BCR", "TCR"))
        ) %>%
        dplyr::relocate(contigType)
}

getContigDoubletBarcodes <- function(tcrOutput, bcrOutput) {
    intersect(
        dplyr::bind_rows(tcrOutput)$barcode,
        dplyr::bind_rows(bcrOutput)$barcode
    )
}

lapplyOnAll <- function(listOfLists, fun) {
    lapply(listOfLists, function(x) lapply(x, fun))
}

autoFullJoin <- function(df1, df2) {
    suppressMessages(dplyr::full_join(df1, df2))
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

rcppConstructBarcodeIndex <- function(conDfBarcodes, data2Barcodes) {
    .Call(`_scRepertoire_rcppConstructBarcodeIndex`, conDfBarcodes, data2Barcodes)
}

rcppConstructConDfAndParseTCR <- function(data2, uniqueData2Barcodes) {
    .Call(`_scRepertoire_rcppConstructConDfAndParseTCR`, data2, uniqueData2Barcodes)
}

.onLoad <- function (libname, pkgname)
{
    utils::globalVariables ("barcode")
    utils::globalVariables ("cdr3")
    utils::globalVariables ("cdr3_nt")
    utils::globalVariables ("cdr3_nt1")
    utils::globalVariables ("cdr3_nt2")
    utils::globalVariables ("CTaa")
    utils::globalVariables ("CTnt")
    utils::globalVariables ("productive")
    utils::globalVariables ("Clones")
    utils::globalVariables ("group")
    utils::globalVariables ("contigType")
    utils::globalVariables ("include")
    utils::globalVariables (".") 
    utils::globalVariables ("Abundance") 
    utils::globalVariables ("Freq") 
    utils::globalVariables ("Sample") 
    utils::globalVariables ("ci_lower") 
    utils::globalVariables ("ci_upper") 
    utils::globalVariables ("clone")
    utils::globalVariables ("cloneSize") 
    utils::globalVariables ("ncells")
    utils::globalVariables ("position") 
    utils::globalVariables ("prop") 
    utils::globalVariables ("scaled") 
    utils::globalVariables ("size") 
    utils::globalVariables ("total") 
    utils::globalVariables ("value") 
    utils::globalVariables ("values")
    utils::globalVariables ("cluster")
    invisible ()
}

#' Cluster clones by sequence similarity
#'
#' This function clusters TCRs or BCRs based on the edit distance or alignment
#' score of their CDR3 sequences. It can operate on either nucleotide (`nt`)
#' or amino acid (`aa`) sequences and can optionally enforce that clones share
#' the same V and/or J genes. The output can be the input object with an added
#' metadata column for cluster IDs, a sparse adjacency matrix, or an `igraph`
#' graph object representing the cluster network.
#'
#' @details
#' The clustering process is as follows:
#' 1.  The function retrieves the relevant chain data from the input object.
#' 2.  It calculates the distance between all sequences within each group
#'     (or across the entire dataset if `group.by` is `NULL`).
#' 3.  An edge list is constructed, connecting sequences that meet the similarity
#'     `threshold`.
#' 4.  The `threshold` parameter behaves differently based on its value:
#'     - **`threshold` < 1 (e.g., 0.85):** Interpreted as a *normalized*
#'       distance. A higher value means greater similarity is required.
#'     - **`threshold` >= 1 (e.g., 2):** Interpreted as a maximum *raw* edit
#'       distance. A lower value means greater similarity is required.
#' 5.  **Distance Metrics:**
#'     - **Levenshtein/Hamming/Damerau:** Standard edit distance calculations.
#'     - **Alignment (NW/SW):** If `dist.type` is "nw" (Needleman-Wunsch) or
#'       "sw" (Smith-Waterman), alignment scores are calculated using the
#'       specified substitution matrix (`dist.mat`). These scores are converted
#'       to a distance-like metric for clustering.
#' 6.  An `igraph` graph is built from the edge list.
#' 7.  A clustering algorithm is run on the graph (default: connected components).
#'
#' @examples
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list,
#'                        samples = c("P17B", "P17L", "P18B", "P18L",
#'                                    "P19B","P19L", "P20B", "P20L"))
#'
#' # Standard Levenshtein clustering (85% similarity)
#' sub_combined <- clonalCluster(combined[c(1,2)],
#'                               chain = "TRA",
#'                               sequence = "aa",
#'                               threshold = 0.85)
#'
#' # Alignment-based clustering using BLOSUM80
#' sub_combined_nw <- clonalCluster(combined[c(1,2)],
#'                                  chain = "TRA",
#'                                  dist.type = "nw",
#'                                  dist.mat = "BLOSUM80",
#'                                  threshold = 0.85)
#'
#' # Export the graph object instead
#' graph_obj <- clonalCluster(combined[c(1,2)],
#'                            chain = "TRA",
#'                            export.graph = TRUE)
#'
#' @param input.data The product of [combineTCR()],
#' [combineBCR()] or [combineExpression()].
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param sequence Clustering based on either `aa` or `nt` sequences.
#' @param threshold The similarity threshold. If < 1, treated as normalized
#' similarity (higher is stricter). If >= 1, treated as raw edit distance
#' (lower is stricter).
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, clusters will be calculated across
#' all sequences.
#' @param use.V If `TRUE`, sequences must share the same V gene to be
#' clustered together.
#' @param use.J If `TRUE`, sequences must share the same J gene to be
#' clustered together.
#' @param dist.type The distance metric to use. Options: `"levenshtein"` (default),
#' `"hamming"`, `"damerau"` (allows transpositions), `"nw"` (Needleman-Wunsch),
#' or `"sw"` (Smith-Waterman).
#' @param dist.mat The substitution matrix to use for alignment-based metrics
#' (`"nw"` or `"sw"`). Options: `"BLOSUM45"`, `"BLOSUM50"`, `"BLOSUM62"`,
#' `"BLOSUM80"` (default), `"BLOSUM100"`, `"PAM30"`, `"PAM40"`, `"PAM70"`, `"PAM120"`,
#' `"PAM250"`, or `"identity"`.
#' @param normalize Method for normalizing distances. Options: `"none"`,
#' `"maxlen"` (divide by max sequence length), or `"length"` (default, divide
#' by mean sequence length). If `threshold < 1`, this controls how the
#' similarity is calculated.
#' @param gap.open Penalty for opening a gap in alignment metrics (default: -10).
#' @param gap.extend Penalty for extending a gap in alignment metrics (default: -1).
#' @param cluster.method The clustering algorithm to use. Defaults to `"components"`,
#' which finds connected subgraphs.
#' @param cluster.prefix A character prefix to add to the cluster names (e.g.,
#' "cluster.").
#' @param export.graph If `TRUE`, the function returns an `igraph`
#' object of the sequence network.
#' @param export.adj.matrix If `TRUE`, the function returns a sparse
#' adjacency matrix (`dgCMatrix`) of the network.
#' @param dist_type \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `dist.type` instead.
#' @param dist_mat \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `dist.mat` instead.
#' @param gap_open \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `gap.open` instead.
#' @param gap_extend \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `gap.extend` instead.
#' @param exportGraph \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.graph` instead.
#' @param exportAdjMatrix \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.adj.matrix` instead.
#' @importFrom igraph graph_from_edgelist E E<- V V<- as_data_frame
#' as_adjacency_matrix membership set_vertex_attr
#' @importFrom dplyr left_join ungroup
#' @importFrom rlang %||%
#' @importFrom SummarizedExperiment colData colData<-
#' @importFrom S4Vectors DataFrame
#' @importFrom Matrix sparseMatrix
#' @importFrom stats reshape
#'
#' @export
#' @concept Visualizing_Clones
#' @return
#' Depending on the export parameters, one of the following:
#' \itemize{
#'   \item An amended `input.data` object with a new metadata column containing cluster IDs (default).
#'   \item An `igraph` object if `export.graph = TRUE`.
#'   \item A sparse `dgCMatrix` object if `export.adj.matrix = TRUE`.
#' }

clonalCluster <- function(input.data,
                          chain = "TRB",
                          sequence = "aa",
                          threshold = 0.85,
                          group.by = NULL,
                          dist.type = NULL,
                          dist.mat = NULL,
                          normalize = "length",
                          gap.open = NULL,
                          gap.extend = NULL,
                          cluster.method = "components",
                          cluster.prefix = "cluster.",
                          use.V = TRUE,
                          use.J = FALSE,
                          export.adj.matrix = NULL,
                          export.graph = NULL,
                          # Deprecated arguments
                          dist_type = NULL,
                          dist_mat = NULL,
                          gap_open = NULL,
                          gap_extend = NULL,
                          exportAdjMatrix = NULL,
                          exportGraph = NULL) {

  # Handle deprecated arguments
  dist.type <- .deprecate_arg(dist_type, dist.type, "dist_type", "dist.type",
                              "clonalCluster", default = "levenshtein")
  dist.mat <- .deprecate_arg(dist_mat, dist.mat, "dist_mat", "dist.mat",
                             "clonalCluster", default = "BLOSUM80")
  gap.open <- .deprecate_arg(gap_open, gap.open, "gap_open", "gap.open",
                             "clonalCluster", default = -10)
  gap.extend <- .deprecate_arg(gap_extend, gap.extend, "gap_extend", "gap.extend",
                               "clonalCluster", default = -1)
  export.graph <- .deprecate_arg(exportGraph, export.graph, "exportGraph", "export.graph",
                                 "clonalCluster", default = FALSE)
  export.adj.matrix <- .deprecate_arg(exportAdjMatrix, export.adj.matrix, "exportAdjMatrix", "export.adj.matrix",
                                      "clonalCluster", default = FALSE)
  
  if (export.graph && export.adj.matrix) {
    stop("Please set only one of `export.graph` or `export.adj.matrix` to TRUE.")
  }
  
  #Prepping any single-cell object
  is.list <- inherits(input.data, "list")
  if(!is.list && !.is.seurat.or.se.object(input.data)) {
    input.data <- .checkList(input.data)
  }
  if (chain == "both") {
    # Use a placeholder to grab both alpha/beta or heavy/light chains
    chains_to_get <- c("TRA", "TRB")
    chain_data <- lapply(chains_to_get, function(x) {
      getIR(input.data, chains = x, sequence.type = sequence, group.by = group.by)
    })
  } else {
    chain_data <- getIR(input.data, chains = .chainConverter(chain), sequence.type = sequence, group.by = group.by)
    chain_data <- list(chain_data)
  }
  
  # Flatten the list if nested and get all barcodes
  if (is.list(chain_data[[1]]) && !is.data.frame(chain_data[[1]])) {
    chain_data <- unlist(chain_data, recursive = FALSE)
  }
  all_barcodes <- unique(do.call(rbind, chain_data)[["barcode"]])
  
  # Apply the network function to each data frame and combine into one edge list
  result_list <- lapply(chain_data, function(y) {
    y <- y[!is.na(y[,1]),]
    .buildNetwork(y, use.V, use.J, threshold,
                  dist.type, dist.mat, normalize,
                  gap.open, gap.extend)
  })
  full_edge_list <- do.call(rbind, result_list)
  
  # if group.by is invoked
  if (!is.null(group.by) && !is.null(names(chain_data))) {
    row_counts <- sapply(result_list, NROW)
    full_edge_list$group <- rep(names(chain_data), times = row_counts)
  }
  
  # pseudocount for same sequences
  full_edge_list$dist[full_edge_list$dist == 0] <- 1e-6
  
  # Add filtering
  if(threshold >= 1) {
    full_edge_list <- full_edge_list[full_edge_list$dist >= 1 & full_edge_list$dist <= threshold,]
  } else {
    max_ned <- 1-threshold
    full_edge_list <- full_edge_list[full_edge_list$dist <= max_ned, ]
  }
  full_edge_list <- unique(full_edge_list) 
  
  if (nrow(full_edge_list) == 0) {
    warning("No clusters found, returning original data.")
    return(input.data)
  }
  
  # Create the graph object
  full_g <- igraph::graph_from_data_frame(full_edge_list, 
                                          directed = FALSE)
  igraph::E(full_g)$weight <- full_edge_list$dist
  if (!is.null(group.by)) {
    vertex_map_df <- data.frame(
      name = c(full_edge_list$from, full_edge_list$to),
      group = rep(full_edge_list[["group"]], 2))
    unique_vertex_map <- vertex_map_df[!duplicated(vertex_map_df$name), ]
    vertex_group_lookup <- setNames(unique_vertex_map$group, unique_vertex_map$name)
    graph_vertex_names <- igraph::V(full_g)$name
    igraph::V(full_g)$group <- vertex_group_lookup[graph_vertex_names]
  }
  
  # Perform clustering using the specified method
  clusters <- .clusterGraph(cluster.method, full_g)
  cluster_membership <- igraph::membership(clusters)
  
  # Rename clusters sequentially with the specified prefix
  unique_clusters <- unique(cluster_membership)
  cluster_names <- paste0(cluster.prefix, seq_along(unique_clusters))
  name_map <- setNames(cluster_names, unique_clusters)
  renamed_membership <- name_map[as.character(cluster_membership)]
  names(renamed_membership) <- names(cluster_membership)
  
  # Add vertex communities 
  igraph::V(full_g)$cluster <- renamed_membership
  
  # Returning Graph
  if(export.graph) {
    # Adding additional vertex information if graph being exported
    vertex_df <- igraph::as_data_frame(full_g, what = "vertices")
    colnames(vertex_df)[1] <- "barcode" 
    full_meta_long <- do.call(rbind, chain_data)
    
    if (nrow(full_meta_long) > 0) {
      meta_to_process <- unique(full_meta_long[, c("barcode", "cdr3_aa", "v", "j")])
      meta_indexed <- meta_to_process %>% 
        group_by(barcode) %>%
        mutate(chain_num = dplyr::row_number()) %>%
        ungroup() %>%
        as.data.frame()
      original_meta <- reshape(
        data = meta_indexed, 
        idvar = "barcode",
        timevar = "chain_num",
        direction = "wide",
        v.names = c("cdr3_aa", "v", "j"),
        sep = "" )
      # Loop to add more vertex info
      vertex_names <- V(full_g)$name
      match_indices <- match(vertex_names, original_meta$barcode)
      for (col in colnames(original_meta)) {
        if (col == "barcode") next 
        full_g <- set_vertex_attr(
          graph = full_g,
          name = col,
          value = original_meta[[col]][match_indices])
      }
    }
    return(full_g)
  }
  
  # Returning Adjacency Matrix
  if (export.adj.matrix) {
    adj_from_graph <- igraph::as_adjacency_matrix(full_g, 
                                                  sparse = TRUE,
                                                  attr = "weight")
    barcodes_in_graph <- rownames(adj_from_graph)
    
    adjacency_matrix <- Matrix::sparseMatrix(
      i = {}, j = {}, x = 0.0,
      dims = c(length(all_barcodes), length(all_barcodes)),
      dimnames = list(all_barcodes, all_barcodes)
    )
    adjacency_matrix[barcodes_in_graph, barcodes_in_graph] <- adj_from_graph
    return(adjacency_matrix)
  }
  
  # Attaching to input.data
  bound <- igraph::as_data_frame(full_g, what = "vertices")
  colnames(bound)[2] <- ifelse(chain == "both", "Multi.Cluster", paste0(chain, ".Cluster"))
  
  #Adding to potential single-cell object
  if(.is.seurat.or.se.object(input.data)) {
    PreMeta <- bound[,-1, drop = FALSE]
    if (.is.seurat.object(input.data)) { 
      col.name <- names(PreMeta) %||% colnames(PreMeta)
      input.data[[col.name]] <- PreMeta
    } else {
      combined_col_names <- unique(c(colnames(colData(sc.data)), colnames(PreMeta)))
      full_data <- merge(colData(sc.data), PreMeta[rownames, , drop = FALSE], by = "row.names", all.x = TRUE)
      rownames(full_data) <- full_data[, 1]
      full_data  <- full_data[, -1]
      colData(sc.data) <- DataFrame(full_data[, combined_col_names])
    }
  } else {
    colnames(bound)[1] <- "barcode"
    input.data <- lapply(input.data, function(df) {
      dplyr::left_join(df, bound, by = "barcode")
    })
  }
  return(input.data)
}  

# Handling clustering options
#' @importFrom igraph cluster_fast_greedy cluster_walktrap cluster_louvain 
#' cluster_leiden cluster_spinglass cluster_edge_betweenness components
.clusterGraph <- function(clustering.method, graph, ...) {
  if (!igraph::is_igraph(graph)) {
    stop("The 'graph' argument must be a valid 'igraph' object.", call. = FALSE)
  }
  
  if (!is.character(clustering.method) || length(clustering.method) != 1) {
    stop("The 'clustering.method' argument must be a single character string.", call. = FALSE)
  }
  result <- switch(
    clustering.method,
    "fast_greedy" = igraph::cluster_fast_greedy(graph, ...),
    "walktrap" = igraph::cluster_walktrap(graph, ...),
    "louvain" = igraph::cluster_louvain(graph, ...),
    "leiden" = igraph::cluster_leiden(graph, ...),
    "spinglass" = igraph::cluster_spinglass(graph, ...),
    "edge_betweenness" = igraph::cluster_edge_betweenness(graph, ...),
    "components" = igraph::components(graph, mode = "weak", ...),
    
    # Default case: Handle unsupported methods
    stop(
      paste0(
        "Unsupported clustering.method: '", clustering.method, "'. ",
        "Supported methods are: 'fast_greedy', 'walktrap', 'louvain', ",
        "'leiden', 'spinglass', 'edge_betweenness', and 'components'."
      ),
      call. = FALSE
    )
  )
  
  return(result)
}

#' @importFrom immApex buildNetwork
.buildNetwork <- function(df, use.V, use.J, threshold,
                          dist.type, dist.mat, normalize,
                          gap.open, gap.extend) {
  edge_list <- buildNetwork(df,
                            seq_col   = "cdr3_aa",
                            v_col     = "v",
                            j_col     = "j",
                            filter.v  = use.V,
                            filter.j  = use.J,
                            ids       = df[["barcode"]],
                            threshold = threshold,
                            dist_type = dist.type,
                            dist_mat  = dist.mat,
                            normalize = normalize,
                            gap_open  = gap.open,
                            gap_extend= gap.extend)

  return(edge_list)
}

#' Calculate Clonal Diversity
#'
#' This function calculates a specified diversity metric for samples or groups
#' within a dataset. To control for variations in library size, the function
#' can perform bootstrapping with downsampling. It resamples each group to the
#' size of the smallest group and calculates the diversity metric across
#' multiple iterations, returning the mean value.
#' 
#' @details
#' The function operates by first splitting the dataset by the specified `group.by`
#' variable.
#'
#' **Downsampling and Bootstrapping:**
#' To make a fair comparison between groups of different sizes, diversity metrics
#' often require normalization. This function implements this by downsampling.
#' 1.  It determines the number of clones in the smallest group.
#' 2.  For each group, it performs `n.boots` iterations (default = 100).
#' 3.  In each iteration, it randomly samples the clones (with replacement) down to
#'     the size of the smallest group.
#' 4.  It calculates the selected diversity metric on this downsampled set.
#' 5.  The final reported diversity value is the mean of the results from all
#'     bootstrap iterations.
#'
#' This process can be disabled by setting `skip.boots = TRUE`.
#' 
#' Available Diversity Metrics (metric):
#' The function uses a registry of metrics imported from the immApex package.
#' You can select one of the following:
#' \itemize{
#'   \item{\code{"shannon"}: Shannon's Entropy. See \code{\link[immApex]{shannon_entropy}}.}
#'   \item{\code{"inv.simpson"}: Inverse Simpson Index. See \code{\link[immApex]{inv_simpson}}.}
#'   \item{\code{"gini.simpson"}: Gini-Simpson Index. See \code{\link[immApex]{gini_simpson}}.}
#'   \item{\code{"norm.entropy"}: Normalized Shannon Entropy. See \code{\link[immApex]{norm_entropy}}.}
#'   \item{\code{"pielou"}: Pielou's Evenness (same as norm.entropy). See \code{\link[immApex]{pielou_evenness}}.}
#'   \item{\code{"ace"}: Abundance-based Coverage Estimator. See \code{\link[immApex]{ace_richness}}.}
#'   \item{\code{"chao1"}: Chao1 Richness Estimator. See \code{\link[immApex]{chao1_richness}}.}
#'   \item{\code{"gini"}: Gini Coefficient for inequality. See \code{\link[immApex]{gini_coef}}.}
#'   \item{\code{"d50"}: The number of top clones making up 50% of the library. See \code{\link[immApex]{d50_dom}}.}
#'   \item{\code{"hill0"}, \code{"hill1"}, \code{"hill2"}: Hill numbers of order 0, 1, and 2. See \code{\link[immApex]{hill_q}}.}
#' }
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list,
#'                        samples = c("P17B", "P17L", "P18B", "P18L",
#'                                    "P19B","P19L", "P20B", "P20L"))
#'
#' # Calculate Shannon diversity, grouped by sample
#' clonalDiversity(combined,
#'                 clone.call = "gene",
#'                 metric = "shannon")
#'
#' # Calculate Inverse Simpson without bootstrapping
#' clonalDiversity(combined,
#'                 clone.call = "aa",
#'                 metric = "inv.simpson",
#'                 skip.boots = TRUE)
#'
#' @param input.data The product of [combineTCR()],
#' [combineBCR()], or [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups
#' automatically.
#' @param metric The diversity metric to calculate. Must be a single string from
#' the list of available metrics (see Details).
#' @param x.axis An additional metadata variable to group samples along the x-axis.
#' @param return.boots If `TRUE`, returns all bootstrap values instead of the
#' mean. Automatically enables `export.table`.
#' @param skip.boots If `TRUE`, disables downsampling and bootstrapping. The
#' metric will be calculated on the full dataset for each group. Defaults to `FALSE`.
#' @param n.boots The number of bootstrap iterations to perform (default is 100).
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals].
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @import ggplot2
#' @export
#' @concept Visualizing_Clones
#' @return A ggplot object visualizing the diversity metric, or a data.frame if
#' `export.table = TRUE`.
#' @author Andrew Malone, Nick Borcherding, Nathan Vanderkraan
clonalDiversity <- function(input.data,
                            clone.call = NULL,
                            metric = "shannon",
                            chain = "both",
                            group.by = NULL,
                            order.by = NULL,
                            x.axis = NULL,
                            export.table = NULL,
                            palette = "inferno",
                            n.boots = 100,
                            return.boots = FALSE,
                            skip.boots = FALSE,
                            # Deprecated arguments
                            cloneCall = NULL,
                            exportTable = NULL,
                            ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "clonalDiversity", default = "strict")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "clonalDiversity", default = FALSE)

  #Argument and Data Validation
  if (length(metric) != 1 || !is.character(metric)) {
    stop("`metric` must be a single string.")
  }
  if (!metric %in% names(.div.registry)) {
    stop("Invalid `metric`. Please choose from: ", paste(names(.div.registry), collapse = ", "))
  }
  if(return.boots) {
    export.table <- TRUE
  }

  # Data Wrangling
  sco <- .is.seurat.or.se.object(input.data)
  clone.call <- .theCall(input.data, clone.call, check.df = FALSE)
  input.data <- .dataWrangle(input.data, group.by, clone.call, chain)
  
  if(!is.null(group.by) && !sco) {
    grouping <- c(group.by, x.axis)
    input.data <- .groupList(input.data, grouping)
  }
  
  # Core Calculation 
  div_func <- .div.registry[[metric]]
  min_size <- min(sapply(input.data, function(x) nrow(x)))
  
  # Efficiently process each group using lapply
  results_list <- lapply(names(input.data), function(group_name) {
    sub_data <- input.data[[group_name]]
    clones <- sub_data[[clone.call]]
    
    if (skip.boots) {
      diversity_scores <- div_func(table(clones))
    } else {
      # Use replicate for efficient bootstrapping
      diversity_scores <- replicate(n.boots, {
        resampled_clones <- sample(clones, size = min_size, replace = TRUE)
        div_func(table(resampled_clones))
      })
    }
    
    if (return.boots) {
      data.frame(
        group = group_name,
        value = diversity_scores,
        metric = metric
      )
    } else {
      data.frame(
        group = group_name,
        value = mean(diversity_scores, na.rm = TRUE),
        metric = metric
      )
    }
  })
  
  # Combine results into a single data frame 
  output_df <- do.call(rbind, results_list)
  colnames(output_df)[colnames(output_df) == "group"] <- group.by %||% "Group"
  
  # Prepare for plotting or export 
  if (!is.null(x.axis)) {
    x.variable <- lapply(input.data, function(x) {
      unique(x[,x.axis])[1]
    })
    x.variable <- do.call(rbind, x.variable)
    x.variable <- as.data.frame(x.variable)
    colnames(x.variable) <- x.axis
    output_df <- merge(output_df, x.variable, by.x = group.by %||% "Group", by.y = "row.names")
    output_df[,1] <- gsub("\\..*", "", output_df[,1]) #removing the x.axis grouping
  } else {
    x.axis <- "x.axis"
    output_df[,x.axis] <- 1
  }
  
  if (export.table) {
    return(output_df)
  }
  
  # Arranging order if using order.by
  if(!is.null(order.by)) {
    output_df <- .orderingFunction(vector = order.by,
                                   group.by = names(output_df)[1],
                                   output_df)
  }
  
  # Plotting 
  metric.name <- gsub("(^|[[:space:]])([[:alpha:]])", "\\1\\U\\2", metric, perl=TRUE)
  plot <- ggplot(output_df, aes(x = .data[[x.axis]], y = as.numeric(.data[["value"]]))) +
    geom_boxplot(outlier.alpha = 0) +
    geom_jitter(aes(fill = .data[[group.by %||% "Group"]]),
                size = 3,
                shape = 21,
                stroke = 0.25,
                color = "black") +
    labs(fill = "Group", y = paste(metric.name, "Index Score")) +
    scale_fill_manual(values = .colorizer(palette, length(unique(output_df[[group.by %||% "Group"]])))) +
    .themeRepertoire(...) + 
    theme(axis.title.x = element_blank())
  
  if (x.axis == "x.axis") {
    plot <- plot + theme(
      axis.text.x = element_blank(),
      axis.ticks.x = element_blank()
    )
  }
  
  return(plot)
}

# Helper for null default
`%||%` <- function(a, b) {
  if (is.null(a)) b else a
}

#' @importFrom immApex ace_richness chao1_richness gini_coef d50_dom 
#' shannon_entropy inv_simpson gini_simpson norm_entropy pielou_evenness hill_q
.div.registry <- list(
  ace          = ace_richness,
  chao1        = chao1_richness,
  gini         = gini_coef,
  d50          = d50_dom,
  shannon      = shannon_entropy,
  inv.simpson  = inv_simpson,
  gini.simpson = gini_simpson,
  norm.entropy = norm_entropy,
  pielou       = pielou_evenness,
  hill0        = hill_q(0),   # richness
  hill1        = hill_q(1),   # exp(H)
  hill2        = hill_q(2)    # 1/Simpson
)

#' Subset The Product of combineTCR() or combineBCR()
#'
#' This function allows for the subsetting of the product of 
#' [combineTCR()] or [combineBCR()] 
#' by the name of the individual list element. 
#'
#' @examples
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' subset <- subsetClones(combined, name = "sample", variables = c("P17B"))
#'
#' @param input.data The product of [combineTCR()] or 
#' [combineBCR()].
#' @param name The column header/name to use for subsetting.
#' @param variables The values to subset by, must be in the in the variable.

#' @export
#' @concept Loading_and_Processing_Contigs
#' @return list of contigs that have been filtered for the name parameter
subsetClones <- function(input.data, 
                         name, 
                         variables = NULL) {
  names2 <- NULL
  input.data2 <- list()
  for (i in seq_along(input.data)) {
    if (input.data[[i]][1,name] %in% variables) {
      input.data2 <- append(input.data2, list(input.data[[i]]))
      n2 <- names(input.data)[i] 
      names2 <- c(names2, n2)
    }
    else {
      next()
    }
  }
  names(input.data2) <- names2
  return(input.data2)
}

#' Analyze K-mer Motif Composition
#' 
#' @description This function calculates and visualizes the frequency of k-mer 
#' motifs for either nucleotide (nt) or amino acid (aa) sequences. It produces 
#' a heatmap showing the relative composition of the most variable motifs across
#'  samples or groups.
#'
#' @details The function first calculates k-mer frequencies for each sample/group.
#' By default, it then identifies the 30 most variable motifs based on the Median
#' Absolute Deviation (MAD) across all samples and displays their frequencies 
#' in a heatmap.
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' 
#' # Using percentKmer()
#' percentKmer(combined, 
#'             chain = "TRB", 
#'             motif.length = 3)
#' 
#' @param input.data The product of [combineTCR()], 
#' [combineBCR()], or [combineExpression()]
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `nt` (CDR3 nucleotide sequence) or `aa` (CDR3 amino acid sequence).
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed as
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups
#' automatically.
#' @param motif.length The length of the kmer to analyze
#' @param min.depth Minimum count a motif must reach to be retained in the
#' output (`>= 1`). **Default:** `3`.
#' @param top.motifs Return the n most variable motifs as a function of
#' median absolute deviation
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals]
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#' 
#' @importFrom stats mad
#' @importFrom immApex calculateMotif
#' 
#' @export
#' 
#' @concept Summarize_Repertoire
#' 
#' @return A ggplot object displaying a heatmap of motif percentages.
#' If `exportTable = TRUE`, a matrix of the raw data is returned.
percentKmer <- function(input.data,
                        chain = "TRB",
                        clone.call = NULL,
                        group.by = NULL,
                        order.by = NULL,
                        motif.length = 3,
                        min.depth = 3,
                        top.motifs = 30,
                        export.table = NULL,
                        palette = "inferno",
                        # Deprecated arguments
                        cloneCall = NULL,
                        exportTable = NULL,
                        ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "percentKmer", default = "aa")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "percentKmer", default = FALSE)

  if(!clone.call %in% c("aa", "nt")) {
    stop("Please select either nucleotide (nt) or amino acid (aa) sequences for clone.call")
  }
  motifs.to.save <- NULL
  sco <- .is.seurat.or.se.object(input.data)
  input.data <- .dataWrangle(input.data,
                             group.by,
                             .theCall(input.data, clone.call,
                                      check.df = FALSE, silent = TRUE),
                             chain)
  clone.call <- .theCall(input.data, clone.call)
  if(!is.null(group.by) && !sco) {
    input.data <- .groupList(input.data, group.by)
  }

  # Generating all motif counts
  motif.list <- lapply(input.data, function(x) {
      immApex::calculateMotif(x[[clone.call]],
                          motif.lengths = motif.length,
                          min.depth = min.depth)
  })
  
  # Collecting motifs into a matrix
  unique.motifs <- unique(unlist(lapply(motif.list, `[`, , 1)))
  if(any(grepl(";", unique.motifs))) { #remove seq deliminator
    unique.motifs <- unique.motifs[-grep(";", unique.motifs)] 
  }
  mat <- matrix(0, ncol = length(unique.motifs), nrow = length(input.data))
  colnames(mat) <- unique.motifs
  rownames(mat) <- names(input.data)
  
  # Populating matrix
  for (i in seq_along(motif.list)) {
    sample_motifs <- motif.list[[i]]
    col_indices <- match(sample_motifs[,1], colnames(mat))
    mat[i, na.omit(col_indices)] <- sample_motifs[!is.na(col_indices), 2]
    mat[i, ] <- mat[i, ] / sum(mat[i,])
  }
  
  # Filtering for top.motifs
  if(!is.null(top.motifs)) {
    mads <- apply(mat, 2, mad)
    motifs.to.save <- names(sort(mads, decreasing = TRUE))[seq_len(top.motifs)]
    mat <- mat[, colnames(mat) %in% motifs.to.save, drop = FALSE]
  }
  
  # Export table if asked for
  if (export.table) {
    return(mat)
  }

  # Getting mat into a ggplot-compliant form
  mat_melt <- expand.grid(Var1 = rownames(mat), Var2 = colnames(mat), KEEP.OUT.ATTRS = FALSE)
  mat_melt$value <- as.vector(mat)
  if (!is.null(motifs.to.save)) {
    mat_melt$Var2 <- factor(mat_melt$Var2, levels = rev(motifs.to.save))
  }
  
  # Ordering group is order.by is not NULL
  if(!is.null(order.by)) {
    mat_melt <- .orderingFunction(vector = order.by,
                                  group.by = "Var1", 
                                  mat_melt)
  }
  
  #Plotting
  plot <- ggplot(mat_melt, aes(x=.data[["Var2"]], y = .data[["Var1"]], fill=.data[["value"]])) +
            geom_tile(lwd = 0.1, color = "black") + 
            scale_fill_gradientn(name = "Proportion", colors = .colorizer(palette,21)) +
            .themeRepertoire(...) + 
            coord_flip() + 
            theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust = 1),
                  axis.title = element_blank())
  return(plot)
}

#Making lodes to function in alluvial plots
#' @importFrom ggalluvial to_lodes_form
.makingLodes <- function(meta2, color, alpha, facet, set.axes) {
  diffuse_elements <- c()
  if (!is.null(color)) {
    diffuse_elements <- c(diffuse_elements, color)
  }
  if (!is.null(alpha)) {
    diffuse_elements <- c(diffuse_elements, alpha)
  }
  if (!is.null(facet)) {
    diffuse_elements <- c(diffuse_elements, facet)
  }

  lodes <- to_lodes_form(meta2, key = "x", value = "stratum",
                         id = "alluvium", axes = set.axes,
                         diffuse = if(length(diffuse_elements) > 0) diffuse_elements else NULL)
  return(lodes)
}

#' Alluvial Plotting for Single-Cell Object
#'
#' View the proportional contribution of clones by Seurat or SCE object
#' meta data after [combineExpression()]. The visualization
#' is based on the ggalluvial package, which requires the aesthetics
#' to be part of the axes that are visualized. Therefore, alpha, facet,
#' and color should be part of the the axes you wish to view or will
#' add an additional stratum/column to the end of the graph.
#'
#' @examples
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list,
#'                         samples = c("P17B", "P17L", "P18B", "P18L",
#'                                     "P19B","P19L", "P20B", "P20L"))
#'
#' # Getting a sample of a Seurat object
#' scRep_example <- get(data("scRep_example"))
#'
#' # Using combineExpresion()
#' scRep_example <- combineExpression(combined, scRep_example)
#' scRep_example$Patient <- substring(scRep_example$orig.ident, 1,3)
#'
#' # Using alluvialClones()
#' alluvialClones(scRep_example,
#'                    clone.call = "gene",
#'                    y.axes = c("Patient", "ident"),
#'                    color = "ident")
#'
#' # Show only top 50 most frequent clones
#' alluvialClones(scRep_example,
#'                    clone.call = "aa",
#'                    y.axes = c("Patient", "ident"),
#'                    top.clones = 50)
#'
#' # Highlight specific clones
#' alluvialClones(scRep_example,
#'                    clone.call = "aa",
#'                    y.axes = c("Patient", "ident"),
#'                    highlight.clones = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF"))
#'
#' @param sc.data The product of [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param y.axes The columns that will separate the proportional visualizations.
#' @param color The column header or clone(s) to be highlighted.
#' @param facet The column label to separate.
#' @param alpha The column header to have gradated opacity.
#' @param top.clones Show only the top N clones by frequency. If `NULL` (default),
#' show all clones.
#' @param min.freq Minimum frequency threshold for displaying flows. Clones
#' appearing fewer than this many times are filtered out.
#' @param highlight.clones Character vector of specific clone sequences to highlight.
#' These clones will be colored distinctly while others are shown in gray.
#' @param highlight.color Color to use for highlighted clones (default: "red").
#' @param stratum.width Width of the stratum bars (default: 0.2).
#' @param flow.alpha Transparency of the flows (default: 0.5). Highlighted clones
#' use full opacity.
#' @param show.labels If `TRUE` (default), display stratum labels.
#' @param label.size Text size for stratum labels (default: 2).
#' @param order.strata Named list specifying the order of levels within each stratum.
#' Names should match column names in y.axes.
#' @param export.table If `TRUE`, returns a data frame of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals].
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @importFrom ggalluvial StatStratum geom_flow geom_stratum to_lodes_form geom_alluvium
#'
#' @export
#' @concept SC_Functions
#' @return A ggplot object visualizing categorical distribution of clones, or a
#' data.frame if `export.table = TRUE`.
alluvialClones <- function(sc.data,
                           clone.call = NULL,
                           chain = "both",
                           y.axes = NULL,
                           color = NULL,
                           facet = NULL,
                           alpha = NULL,
                           top.clones = NULL,
                           min.freq = 0,
                           highlight.clones = NULL,
                           highlight.color = "red",
                           stratum.width = 0.2,
                           flow.alpha = 0.5,
                           show.labels = TRUE,
                           label.size = 2,
                           order.strata = NULL,
                           export.table = NULL,
                           palette = "inferno",
                           # Deprecated arguments
                           cloneCall = NULL,
                           exportTable = NULL,
                           ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "alluvialClones", default = "strict")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "alluvialClones", default = FALSE)

  x <- alluvium <- stratum <- NULL
  .checkSingleObject(sc.data)
  clone.call <- .theCall(.grabMeta(sc.data), clone.call)
  if (length(y.axes) == 0) {
    stop("Make sure you have selected the variable(s) to visualize")
  }
  meta <- .grabMeta(sc.data)
  if (chain != "both") {
    meta <- .offTheChain(meta, chain, clone.call)
  }
  meta$barcodes <- rownames(meta)
  meta <- meta[!is.na(meta[,clone.call]),]

  # Filter by minimum frequency
  if(min.freq > 0) {
    clone.counts <- table(meta[[clone.call]])
    keep.clones <- names(clone.counts)[clone.counts >= min.freq]
    meta <- meta[meta[[clone.call]] %in% keep.clones, ]
  }

  # Filter to top clones by frequency
  if(!is.null(top.clones)) {
    clone.counts <- sort(table(meta[[clone.call]]), decreasing = TRUE)
    keep.clones <- names(clone.counts)[seq_len(min(top.clones, length(clone.counts)))]
    meta <- meta[meta[[clone.call]] %in% keep.clones, ]
  }

  # Handle highlight.clones - creates a special highlighting column
  highlight.mode <- FALSE
  if(!is.null(highlight.clones)) {
    highlight.mode <- TRUE
    meta$`.highlight` <- ifelse(meta[[clone.call]] %in% highlight.clones,
                                "Highlighted", "Other")
    # Ensure highlighted clones are plotted last (on top)
    meta <- meta[order(meta$`.highlight` == "Highlighted"), ]
  }

  # Handle color parameter - check if it's clone sequences or a column
  check <- colnames(meta) == color
  if (length(unique(check)) == 1 & unique(check)[1] == FALSE &
      !is.null(color)) {
    meta <- meta %>% mutate("clone(s)" = ifelse(meta[,clone.call] %in%
                                                      color, "Selected", "Other"))
    color <- "clone(s)"
  }

  # Prepping the data for calculating lodes
  y.axes.plot <- unique(c(y.axes, color, alpha, facet))
  set.axes <- seq_along(y.axes.plot)
  cols.to.keep <- c(y.axes.plot, clone.call, "barcodes")
  if(highlight.mode) {
    cols.to.keep <- c(cols.to.keep, ".highlight")
  }
  meta2 <- meta[, cols.to.keep]
  meta2 <- unique(na.omit(meta2[!duplicated(as.list(meta2))]))

  # Apply stratum ordering if specified
  if(!is.null(order.strata)) {
    for(axis.name in names(order.strata)) {
      if(axis.name %in% colnames(meta2)) {
        meta2[[axis.name]] <- factor(meta2[[axis.name]],
                                     levels = order.strata[[axis.name]])
      }
    }
  }

  lodes <- .makingLodes(meta2, color, alpha, facet, set.axes)

  # Filtering the lodes
  if(any(lodes[,clone.call] != "")) {
    lodes <- lodes[lodes[,clone.call] != "",]
  }
  if(any(is.na(lodes[,clone.call]))) {
    lodes <- lodes[!is.na(lodes[,clone.call]),]
  }

  # Add frequency and proportion info for export
  if(export.table) {
    # Calculate frequency per clone
    clone.freq <- as.data.frame(table(meta[[clone.call]]))
    colnames(clone.freq) <- c(clone.call, "freq")
    lodes <- merge(lodes, clone.freq, by = clone.call, all.x = TRUE)

    # Calculate proportion
    total.cells <- nrow(meta)
    lodes$prop <- lodes$freq / total.cells

    # Add rank
    lodes$rank <- match(lodes[[clone.call]],
                        names(sort(table(meta[[clone.call]]), decreasing = TRUE)))
    return(lodes)
  }

  # Plotting
  plot <- ggplot(data = lodes, aes(x = x,
                                   stratum = stratum,
                                   alluvium = alluvium,
                                   label = stratum)) +
                geom_stratum(width = stratum.width, color = "black")

  # Handle highlighting mode
  if(highlight.mode) {
    # Use highlight column for coloring
    plot <- plot + geom_flow(aes(fill = lodes[[".highlight"]],
                                 alpha = lodes[[".highlight"]]),
                             stat = "alluvium",
                             lode.guidance = "forward",
                             width = stratum.width) +
                   scale_fill_manual(values = c("Highlighted" = highlight.color,
                                               "Other" = "grey70"),
                                    name = "Clone") +
                   scale_alpha_manual(values = c("Highlighted" = 1,
                                                "Other" = flow.alpha),
                                     guide = "none")
  } else if (is.null(color) & is.null(alpha)) {
    plot <- plot + geom_alluvium(width = stratum.width, alpha = flow.alpha)
  } else if (!is.null(color) & is.null(alpha)) {
    plot <- plot + geom_flow(aes(fill = lodes[,color]),
                             stat = "alluvium",
                             lode.guidance = "forward",
                             width = stratum.width,
                             alpha = flow.alpha) +
                   labs(fill = color)
  } else if (is.null(color) & !is.null(alpha)) {
    plot <- plot + geom_flow(aes(alpha = lodes[,alpha]),
                             stat = "alluvium",
                             lode.guidance = "forward",
                             width = stratum.width) +
                   labs(alpha = alpha)
  } else {
    plot <- plot + geom_flow(aes(alpha = lodes[,alpha],
                                 fill = lodes[,color]),
                             stat = "alluvium",
                             lode.guidance = "forward",
                             width = stratum.width) +
      labs(fill = color, alpha = alpha)
  }

  # Add faceting
  if (length(facet) == 1 & length(facet) < 2) {
    plot <- plot +
              facet_wrap(.~lodes[,facet], scales="free_y")
  }

  # Add labels
  if(show.labels) {
    plot <- plot +
            geom_text(stat = ggalluvial::StatStratum,
                      infer.label = FALSE,
                      reverse = TRUE,
                      size = label.size)
  }

  # Apply colors if not in highlight mode
  if(!highlight.mode && !is.null(color)) {
    n.colors <- length(unique(lodes[,color]))
    plot <- plot +
            scale_fill_manual(values = .colorizer(palette, n.colors))
  }

  plot <- plot +
            scale_x_discrete(expand = c(0.025, 0.025)) +
            .themeRepertoire(...) +
            theme(axis.title.x = element_blank(),
                  axis.ticks.x = element_blank(),
                  line = element_blank())

  return(plot)
}

#' Calculate Startrac-based Diversity Indices 
#' 
#' @description This function utilizes the STARTRAC approach to calculate T cell 
#' diversity metrics based on the work of Zhang et al. (2018, Nature) 
#' [PMID: 30479382](https://pubmed.ncbi.nlm.nih.gov/30479382/). It can compute 
#' three distinct indices: clonal expansion (`expa`), cross-tissue migration 
#' (`migr`), and state transition (`tran`). 
#' 
#' @details
#' The function requires a `type` variable in the metadata, which specifies the
#' tissue origin or any other categorical variable for migration analysis.
#'
#' **Indices:**
#' \itemize{
#'   \item{\strong{expa (Clonal Expansion):}} Measures the extent of clonal 
#'         proliferation within a T cell cluster. It is calculated as 
#'         `1 - normalized Shannon entropy`. A higher value indicates greater 
#'         expansion of a few clones.
#'   \item{\strong{migr (Cross-Tissue Migration):}} Quantifies the movement of 
#'         clonal T cells across different tissues (as defined by the `type`
#'         parameter). It is based on the entropy of a clonotype's distribution 
#'         across tissues.
#'   \item{\strong{tran (State Transition):}} Measures the developmental 
#'         transition of clonal T cells between different functional clusters. 
#'         It is based on the entropy  of a clonotype's distribution across 
#'         clusters.
#' }
#'
#' **Pairwise Analysis:**
#' The `pairwise` parameter enables the calculation of migration or transition
#' between specific pairs of tissues or clusters, respectively.
#' \itemize{
#'   \item{For migration (`index = "migr"`), set `pairwise` to the `type` column 
#'         (e.g., `pairwise = "Type"`).}
#'   \item{For transition (`index = "tran"`), set `pairwise` to `"cluster"`.}
#' }
#' 
#' @examples
#' # Getting the combined contigs
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' 
#' # Getting a sample of a Seurat object
#' scRep_example  <- get(data("scRep_example"))
#' scRep_example  <- combineExpression(combined, scRep_example)
#' scRep_example$Patient <- substring(scRep_example$orig.ident,1,3)
#' scRep_example$Type <- substring(scRep_example$orig.ident,4,4) 
#' 
#' # Calculate a single index (expansion)
#' StartracDiversity(scRep_example, 
#'                   type = "Type", 
#'                   group.by = "Patient",
#'                   index = "expa")
#'                   
#' # Calculate pairwise transition 
#' StartracDiversity(scRep_example, 
#'                   type = "Type", 
#'                   group.by = "Patient",
#'                   index = "tran",
#'                   pairwise = "cluster") 
#'
#' @param sc.data The single-cell object after [combineExpression()].
#' For SCE objects, the cluster variable must be in the meta data under
#' "cluster".
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param index A character vector specifying which indices to calculate.
#' Options: "expa", "migr", "tran". Default is all three.
#' @param type The metadata variable that specifies tissue type for migration
#' analysis.
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed as
#' by list element or active identity in the case of single-cell objects.
#' @param pairwise The metadata column to be used for pairwise comparisons.
#' Set to the `type` variable for pairwise migration or "cluster" for
#' pairwise transition.
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals].
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#' @importFrom stats reshape
#' @export
#' @concept SC_Functions
#' @return A ggplot object visualizing STARTRAC diversity metrics or data.frame if
#'`exportTable = TRUE`.
#' @author Liangtao Zheng
StartracDiversity <- function(sc.data,
                              clone.call = NULL,
                              chain = "both",
                              index = c("expa", "migr", "tran"),
                              type = NULL,
                              group.by = NULL,
                              pairwise = NULL,
                              export.table = NULL,
                              palette = "inferno",
                              # Deprecated arguments
                              cloneCall = NULL,
                              exportTable = NULL,
                              ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "StartracDiversity", default = "strict")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "StartracDiversity", default = FALSE)

  if(!all(index %in% c("expa", "migr", "tran"))) {
    stop("Please select 'expa', 'migr', and/or 'tran' for index.")
  }
  if (!is.null(pairwise) && length(index) > 1) {
    stop("Pairwise analysis can only be performed for a single index ('migr' or 'tran').")
  }
  if (!is.null(pairwise) && !index %in% c("migr", "tran")) {
    stop("Pairwise analysis is only supported for 'migr' or 'tran' indices.")
  }
  
  
  # Prepare data
  df <- .grabMeta(sc.data)
  clone.call <- .theCall(df, clone.call)
  barcodes <- rownames(df)
  colnames(df)[ncol(df)] <- "cluster"
    
  if (is.null(group.by)) {
    if (!"orig.ident" %in% colnames(df)) {
      stop("Please select a group.by variable.")
    }
    group.by <- "orig.ident"
  }
  group.levels <- unique(df[,group.by])
  
  if (chain != "both") {
    df <- .offTheChain(df, chain, clone.call)
  }

  # Process clonotypes
  df <- df %>%
    group_by(across(all_of(c(group.by, clone.call)))) %>%
    dplyr::mutate(n = n()) %>%
    as.data.frame()

  rownames(df) <- barcodes
  remove.pos <- which(is.na(df[,clone.call]) | df[,clone.call] == "")
  if (length(remove.pos) > 0) {
    df <- df[-remove.pos,]
  }

  processed <- data.frame(
    Cell_Name = rownames(df),
    clone.id = df[,clone.call],
    patient = df[,group.by],
    cluster = df[,"cluster"],
    loc = df[,type],
    stringsAsFactors = FALSE
  )
  processed[processed == "NA"] <- NA
  processed <- na.omit(processed)
  
  # Calculate indices
  mat.list <- lapply(group.levels, function(level) {
    subset_data <- processed[processed$patient == level,]
    if (!is.null(pairwise)) {
      comparison_col <- if (index == "migr") {
        "loc"
      } else { # index == "tran"
        "cluster"
      }
      .calculatePairwiseIndices(subset_data, index, comparison_col)
    } else {
      .calculateIndices(subset_data, index)
    }
  })
  
  mat <- bind_rows(mat.list, .id = "group")
  if (nrow(mat) == 0) {
    warning("No data available for calculation. Returning NULL.")
    return(NULL)
  }
  
  if(!is.null(pairwise)) {
    mat$variable <- index[1]
    mat <- mat[!is.nan(mat$value),]
  }
  
  if (export.table) {
    return(mat)
  }
  # Plotting logic
  if (!is.null(pairwise)) {
    
    if (pairwise == "cluster") {
        num_colors <- length(unique(mat[["cluster"]]))
        mat$cluster2 <- sapply(strsplit(mat$comparison, " "), `[`, 1)
        mat$cluster2 <- factor(mat$cluster2, levels = .alphanumericalSort(mat$cluster2))
        mat$cluster <- factor(mat$cluster, levels = .alphanumericalSort(mat$cluster))
        plot <- ggplot(mat, aes(x = cluster, y = .data$value)) +
          geom_boxplot(aes(fill = cluster), outlier.alpha = 0, na.rm = TRUE) +
          labs(y = "Pairwise Index Score", x = "Cluster") + 
          facet_grid(cluster2 ~ ., scales = "free_y") 
    } else {
      col_name <- colnames(mat)[grepl("comparison", colnames(mat))]
      num_colors <- length(unique(mat[[col_name]]))
      plot <- ggplot(mat, aes(x = .data[[col_name]], y = .data$value)) +
        geom_boxplot(aes(fill = .data[[col_name]]), outlier.alpha = 0, na.rm = TRUE) +
        labs(y = "Pairwise Index Score", x = "Comparison")
    }
  } else {
    mat_melt <- reshape(mat,
                        varying = index,
                        v.names = "value",
                        timevar = "variable",
                        times = index,
                        direction = "long")
    values <- .alphanumericalSort(unique(mat_melt$cluster))
    mat_melt$cluster <- factor(mat_melt$cluster, levels = values)
    mat_melt$value <- as.numeric(mat_melt$value)
    num_colors <- length(unique(mat_melt$cluster))
    
    plot <- ggplot(mat_melt, aes(x = cluster, y = .data[["value"]])) +
      geom_boxplot(aes(fill = cluster), outlier.alpha = 0, na.rm = TRUE) +
      labs(y = "Index Score", 
           x = "Clusters") +
      theme(axis.title.x = element_blank())
  }
  
  plot <- plot + 
    .themeRepertoire(...) + 
    guides(fill = "none") +
    scale_fill_manual(values = .colorizer(palette, num_colors))
  
  if(length(index) > 1) {
    plot <- plot + facet_grid(variable ~ ., scales = "free_y") 
  } 
  
  return(plot)
}


# Helper function for standard index calculation
.calculateIndices <- function(processed, indices) {
  if (nrow(processed) == 0) return(NULL)
  clonotype.dist.cluster <- table(processed[,c("clone.id", "cluster")])
  clonotype.dist.loc <- table(processed[,c("clone.id", "loc")])
  
  # Return NULL if no clusters are found
  if (ncol(clonotype.dist.cluster) == 0) return(NULL)
  
  calIndex.matrix <- data.frame(cluster = colnames(clonotype.dist.cluster))
  
  if ("expa" %in% indices) {
    entropy_val <- .mcolEntropy(clonotype.dist.cluster)
    entropy_max <- log2(colSums(clonotype.dist.cluster > 0))
    expa <- 1 - (entropy_val / entropy_max)
    calIndex.matrix$expa <- expa
  }
  
  # Check if there are clonotypes to process for migr/tran
  if (nrow(clonotype.dist.cluster) > 0) {
    clonotype.data <- data.frame(clone.id = rownames(clonotype.dist.cluster))
    weights.mtx <- sweep(clonotype.dist.cluster, 2, colSums(clonotype.dist.cluster), "/")
    
    if ("migr" %in% indices && "loc" %in% colnames(processed) && nrow(clonotype.dist.loc) > 0) {
      # Ensure clonotypes for migration calculation exist in the cluster distribution
      shared_clones_migr <- intersect(rownames(clonotype.dist.cluster), rownames(clonotype.dist.loc))
      if(length(shared_clones_migr) > 0) {
        clonotype.data$migr <- .mrowEntropy(clonotype.dist.loc[shared_clones_migr,,drop=FALSE])
        migr_matrix <- t(weights.mtx[shared_clones_migr, , drop=FALSE]) %*% as.matrix(clonotype.data$migr)
        calIndex.matrix$migr <- migr_matrix[,1]
      } else {
        calIndex.matrix$migr <- NA
      }
    }
    
    if ("tran" %in% indices) {
      clonotype.data$tran <- .mrowEntropy(clonotype.dist.cluster)
      tran_matrix <- t(weights.mtx) %*% as.matrix(clonotype.data$tran)
      calIndex.matrix$tran <- tran_matrix[,1]
    }
  } else {
    # If no clonotypes, set indices to NA
    if ("migr" %in% indices) calIndex.matrix$migr <- NA
    if ("tran" %in% indices) calIndex.matrix$tran <- NA
  }
  
  for (col in names(calIndex.matrix)) {
    if (is.numeric(calIndex.matrix[[col]])) {
      calIndex.matrix[[col]][is.nan(calIndex.matrix[[col]]) | is.infinite(calIndex.matrix[[col]])] <- NA
    }
  }
  
  return(calIndex.matrix)
}

# Helper function for pairwise index calculation
#' @importFrom utils combn
.calculatePairwiseIndices <- function(processed, index, pairwise_col) {
  if (nrow(processed) < 2) return(NULL)
  
  unique_items <- unique(processed[[pairwise_col]])
  if (length(unique_items) < 2) return(NULL)
  
  pairs <- combn(unique_items, 2, simplify = FALSE)
  
  pairwise_results <- lapply(pairs, function(p) {
    pair_data <- processed[processed[[pairwise_col]] %in% p,]
    
    if (index == "migr") {
      dist_table <- table(pair_data[,c("clone.id", "loc")])
      clonotype_dist_cluster <- table(pair_data[,c("clone.id", "cluster")])
    } else { # tran
      dist_table <- table(pair_data[,c("clone.id", "cluster")])
      clonotype_dist_cluster <- dist_table
    }
    
    if(nrow(dist_table) == 0) return(NULL)
    
    clonotype_data <- data.frame(clone.id = rownames(dist_table),
                                 value = .mrowEntropy(dist_table))
    
    weights_mtx <- sweep(clonotype_dist_cluster, 2, colSums(clonotype_dist_cluster), "/")
    
    # Ensure clone.ids match for matrix multiplication
    shared_clones <- intersect(rownames(weights_mtx), clonotype_data$clone.id)
    if (length(shared_clones) == 0) return(NULL)
    
    weights_mtx_filtered <- weights_mtx[shared_clones,, drop=FALSE]
    clonotype_data_filtered <- clonotype_data[clonotype_data$clone.id %in% shared_clones,]
    
    result_matrix <- t(weights_mtx_filtered) %*% as.matrix(clonotype_data_filtered$value)
    
    res <- data.frame(cluster = rownames(result_matrix), value = result_matrix[,1])
    res$comparison <- paste(p, collapse = " vs ")
    return(res)
  })
  
  bind_rows(pairwise_results)
}

# Entropy of each row of the input matrix
.mrowEntropy <- function(x) {
  freqs <- sweep(x, 1, rowSums(x), "/")
  H <- -rowSums(ifelse(freqs > 0, freqs * log2(freqs), 0))
  return(H)
}

# Entropy of each column of the input matrix
.mcolEntropy <- function(x) {
  freqs <- sweep(x, 2, colSums(x), "/")
  H <- -colSums(ifelse(freqs > 0, freqs * log2(freqs), 0))
  return(H)
}

# test script for StartracDiversity.R - testcases are NOT comprehensive!

# Data to use
combined <- combineTCR(contig_list, samples = c("P17B", "P17L", "P18B", "P18L", "P19B", "P19L", "P20B", "P20L"))
scRep_example <- combineExpression(combined, scRep_example)
scRep_example$Patient <- substring(scRep_example$orig.ident, 1, 3)
scRep_example$Type <- substring(scRep_example$orig.ident, 4, 4)

test_that("Input validation and error handling work correctly", {
  expect_error(
    StartracDiversity(scRep_example, index = c("migr", "tran"), pairwise = "Type"),
    "Pairwise analysis can only be performed for a single index"
  )
  
  expect_error(
    StartracDiversity(scRep_example, index = "expa", pairwise = "Type"),
    "Pairwise analysis is only supported for 'migr' or 'tran' indices."
  )
})


test_that("Output format and structure are correct", {
  # Returns a ggplot object by default
  plot_output <- StartracDiversity(scRep_example, 
                                   type = "Type", 
                                   group.by = "Patient")
  expect_s3_class(plot_output, "ggplot")
  
  # Returns a data.frame when export.table = TRUE
  table_output <- StartracDiversity(scRep_example, 
                                    type = "Type", 
                                    group.by = "Patient", 
                                    export.table = TRUE)
  expect_s3_class(table_output, "data.frame")
  
  # Check standard output columns
  expect_true(all(c("group", "cluster", "expa", "migr", "tran") %in% names(table_output)))
})


test_that("Helper functions for entropy calculate correctly", {
  # Test matrix
  mat <- matrix(c(1, 1, 0, 4, 2, 0, 3, 0, 0), nrow = 3, byrow = TRUE)
  
  expected_row_entropy <- c(1, 0.9182958, 0)
  expect_equal(.mrowEntropy(mat), expected_row_entropy, tolerance = 1e-6)

  expected_col_entropy <- c(1.405639, 0.9182958, NA)
  expect_equal(.mcolEntropy(mat), expected_col_entropy, tolerance = 1e-2)
})


test_that("Standard index calculations are mathematically correct", {
  results <- StartracDiversity(scRep_example, 
                               type = "Type", 
                               group.by = "Patient", 
                               export.table = TRUE)
  

  expect_equal(results$expa[results$cluster == "1"][1], 0, tolerance = 1e-4)
  expect_equal(results$expa[results$cluster == "2"][1], 0, tolerance = 1e-4)
  expect_equal(results$expa[results$cluster == "3"][1], 0, tolerance = 1e-4)
  
  expect_equal(results$migr[results$cluster == "1"][1], 0, tolerance = 1e-4)
  expect_equal(results$migr[results$cluster == "2"][1], 0, tolerance = 1e-4)
  expect_equal(results$migr[results$cluster == "3"][1], 0, tolerance = 1e-4)
  
  expect_equal(results$tran[results$cluster == "1"][1], 0, tolerance = 1e-4)
  expect_equal(results$tran[results$cluster == "2"][1], 0, tolerance = 1e-4)
  expect_equal(results$tran[results$cluster == "3"][1], 0.119958, tolerance = 1e-4)
})

test_that("Pairwise calculations are correct", {
  # Test 1: Pairwise migration (index = "migr", pairwise = "Type")
  pairwise_migr_results <- StartracDiversity(scRep_example,
                                             type = "Type",
                                             group.by = "Patient",
                                             index = "migr",
                                             pairwise = "Type",
                                             export.table = TRUE)
  expect_s3_class(pairwise_migr_results, "data.frame")
  expect_true(all(c("group", "cluster", "value", "comparison") %in% names(pairwise_migr_results)))
  expect_true(all(grepl("\\w vs \\w", pairwise_migr_results$comparison)))
  expect_equal(unique(pairwise_migr_results$comparison), "B vs L")
  
  
  # Test 2: Pairwise transition (index = "tran", pairwise = "cluster")
  pairwise_tran_results <- StartracDiversity(scRep_example,
                                             type = "Type",
                                             group.by = "Patient",
                                             index = "tran",
                                             pairwise = "cluster",
                                             export.table = TRUE)
  expect_s3_class(pairwise_tran_results, "data.frame")
  expect_true(all(c("group", "cluster", "value", "comparison") %in% names(pairwise_tran_results)))
  expect_true(all(grepl("\\w vs \\w", pairwise_tran_results$comparison)))
  expect_true(is.numeric(pairwise_tran_results$value))
})

#' Calculate rarefaction based on the abundance of clones
#' 
#' This functions uses the Hill numbers of order q: species richness (`q = 0`), 
#' Shannon diversity (`q = 1`), the exponential of Shannon entropy and Simpson 
#' diversity (`q = 2`, the inverse of Simpson concentration) to compute diversity 
#' estimates for rarefaction and extrapolation. The function relies on the
#' [iNEXT::iNEXT()] R package. Please read and cite the 
#' [manuscript](https://besjournals.onlinelibrary.wiley.com/doi/10.1111/2041-210X.12613) 
#' if using this function. The input into the iNEXT calculation is abundance, 
#' incidence-based calculations are not supported.
#' 
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#'                                     
#' # Using clonalRarefaction()
#' clonalRarefaction(combined[c(1,2)], clone.call = "gene", n.boots = 3)
#'
#'
#' @param input.data The product of [combineTCR()], [combineBCR()], or
#' [combineExpression()].
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed
#' by list element or active identity in the case of single-cell objects.
#' @param plot.type sample-size-based rarefaction/extrapolation curve
#' (`type = 1`); sample completeness curve (`type = 2`);
#' coverage-based rarefaction/extrapolation curve (`type = 3`).
#' @param hill.numbers The Hill numbers to be plotted out
#' (0 - species richness, 1 - Shannon, 2 - Simpson)
#' @param n.boots The number of bootstrap replicates used to derive confidence
#' intervals for the diversity estimates. More replicates can provide a more
#' reliable measure of statistical variability.
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals].
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @importFrom iNEXT iNEXT ggiNEXT
#' @export
#' @concept Visualizing_Clones
clonalRarefaction <- function(input.data,
                              clone.call = NULL,
                              chain = "both",
                              group.by = NULL,
                              plot.type = 1,
                              hill.numbers = 0,
                              n.boots = 20,
                              export.table = NULL,
                              palette = "inferno",
                              # Deprecated arguments
                              cloneCall = NULL,
                              exportTable = NULL,
                              ...) {

  # Handle deprecated arguments
  clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                               "clonalRarefaction", default = "strict")
  export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                 "clonalRarefaction", default = FALSE)

  input.data <- .dataWrangle(input.data,
                             group.by,
                             .theCall(input.data, clone.call,
                                      check.df = FALSE, silent = TRUE),
                             chain)
  clone.call <- .theCall(input.data, clone.call)
  sco <- .is.seurat.or.se.object(input.data)

  if(!is.null(group.by) & !sco) {
    input.data <- .groupList(input.data, group.by)
  }

  mat.list <- lapply(input.data, function(x) {
                  table(x[,clone.call])
  })
  col <- length(input.data)

  # Compute diversity estimates along with bootstrap-derived confidence intervals.
  # The iNEXT function uses 'n.boots' replicates to estimate variability.
  mat <- iNEXT(mat.list, q=hill.numbers, datatype="abundance",nboot = n.boots)
  plot <- suppressMessages(ggiNEXT(mat, type=plot.type) +
            scale_shape_manual(values = rep(16,col)) +
            scale_fill_manual(values = c(.colorizer(palette,col))) +
            scale_color_manual(values = c(.colorizer(palette,col)))  +
            .themeRepertoire(...))

  if (export.table) {
    return(plot["data"])
  } else {
    return(plot)
  }
  
}

#' Plot the Clonal Space Occupied by Specific Clones
#'
#' This function calculates the relative clonal space occupied by the 
#' clones. The grouping of these clones is based on the parameter 
#' `clonalSplit`, at default, `clonalSplit` will group the clones 
#' into bins of 1:10, 11:100, 101:1001, etc. To adjust the clones 
#' selected, change the numbers in the variable split. If a matrix output 
#' for the data is preferred, set `exportTable` = TRUE.
#'
#' @examples
#' # Making combined contig data
#' combined <- combineTCR(contig_list, 
#'                         samples = c("P17B", "P17L", "P18B", "P18L", 
#'                                     "P19B","P19L", "P20B", "P20L"))
#' # Using clonalProportion()
#' clonalProportion(combined, clone.call = "gene")
#'
#' @param input.data The product of [combineTCR()],
#' [combineBCR()], or [combineExpression()].
#' @param clonal.split The cut points for the specific clones, default = c(10,
#' 100, 1000, 10000, 30000, 1e+05)
#' @param clone.call Defines the clonal sequence grouping. Accepted values
#' are: `gene` (VDJC genes), `nt` (CDR3 nucleotide sequence), `aa` (CDR3 amino
#' acid sequence), or `strict` (VDJC + nt). A custom column header can also be used.
#' @param chain The TCR/BCR chain to use. Use `both` to include both chains
#' (e.g., TRA/TRB). Accepted values: `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`,
#' `IGK`, `Light` (for both light chains), or `both` (for TRA/B and Heavy/Light).
#' @param group.by A column header in the metadata or lists to group the analysis
#' by (e.g., "sample", "treatment"). If `NULL`, data will be analyzed
#' by list element or active identity in the case of single-cell objects.
#' @param order.by A character vector defining the desired order of elements
#' of the `group.by` variable. Alternatively, use `alphanumeric` to sort groups
#' automatically.
#' @param export.table If `TRUE`, returns a data frame or matrix of the results
#' instead of a plot.
#' @param palette Colors to use in visualization - input any
#' [hcl.pals][grDevices::hcl.pals]
#' @param clonalSplit \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clonal.split` instead.
#' @param cloneCall \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `clone.call` instead.
#' @param exportTable \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use `export.table` instead.
#' @param ... Additional arguments passed to the ggplot theme
#'
#' @export
#' @concept Visualizing_Clones
#' @return A ggplot object dividing space occupied by ranks of clones or a 
#' data.frame if `exportTable = TRUE`.
clonalProportion <- function(input.data,
                             clonal.split = NULL,
                             clone.call = NULL,
                             chain = "both",
                             group.by = NULL,
                             order.by = NULL,
                             export.table = NULL,
                             palette = "inferno",
                             # Deprecated arguments
                             clonalSplit = NULL,
                             cloneCall = NULL,
                             exportTable = NULL,
                             ...) {

    # Handle deprecated arguments
    clonal.split <- .deprecate_arg(clonalSplit, clonal.split, "clonalSplit", "clonal.split",
                                   "clonalProportion", default = c(10, 100, 1000, 10000, 30000, 100000))
    clone.call <- .deprecate_arg(cloneCall, clone.call, "cloneCall", "clone.call",
                                 "clonalProportion", default = "strict")
    export.table <- .deprecate_arg(exportTable, export.table, "exportTable", "export.table",
                                   "clonalProportion", default = FALSE)

    Con.df <- NULL
    input.data <- .dataWrangle(input.data,
                               group.by,
                               .theCall(input.data, clone.call,
                                        check.df = FALSE, silent = TRUE),
                               chain)
    clone.call <- .theCall(input.data, clone.call)
    sco <- .is.seurat.or.se.object(input.data)
    
    if(!is.null(group.by) & !sco) {
      input.data <- .groupList(input.data, group.by)
    }

    #Generating data matrix to store value
    mat <- matrix(0, length(input.data), length(clonal.split), dimnames = list(names(input.data),
            paste0('[', c(1, clonal.split[-length(clonal.split)] + 1), ':', clonal.split, ']')))

    #Assigning the clonal grouping
    input.data <- lapply(input.data, '[[', clone.call)
    input.data <- lapply(input.data, na.omit)
    input.data <- lapply(input.data, as.data.frame(table))
    for (i in seq_along(input.data)) {
        input.data[[i]] <- rev(sort(as.numeric(input.data[[i]][,2])))
    }
    cut <- c(1, clonal.split[-length(clonal.split)] + 1)
    for (i in seq_along(clonal.split)) {
        mat[,i] <- vapply(input.data, function (x)
                          sum(na.omit(x[cut[i]:clonal.split[i]])), FUN.VALUE = numeric(1))
    }
    if (export.table) {
        return(mat)
    }
    # Plotting
    mat_melt <- data.frame(
      Var1 = rep(rownames(mat), ncol(mat)),
      Var2 = rep(colnames(mat), each = nrow(mat)),
      value = as.vector(mat)
    )
    mat_melt[["Var2"]] <- factor(mat_melt[["Var2"]], 
                                 levels = colnames(mat))
    
    if(!is.null(order.by)) {
      mat_melt <- .orderingFunction(vector = order.by,
                                    group.by = "Var1", 
                                    data.frame = mat_melt)
    }
    col <- length(unique(mat_melt$Var2))
    plot <- ggplot(mat_melt, aes(x=as.factor(.data[["Var1"]]), y=.data[["value"]], fill=.data[["Var2"]])) +
        geom_bar(stat = "identity", position="fill", 
                    color = "black", lwd= 0.25) +
        scale_fill_manual(values = rev(.colorizer(palette,col))) +
        labs(x = "Samples",
             y = "Occupied Repertoire Space", 
             fill = "Clonal Indices") +
      .themeRepertoire(...)
    return(plot)
}

---
title: Starting work with scRepertoire.
author: 
  name: Nick Borcherding
  email: ncborch@gmail.com
  affiliation: Washington University in St. Louis, School of Medicine, St. Louis, MO, USA
date: 'Compiled: `r format(Sys.Date(), "%B %d, %Y")`'
output:
  BiocStyle::html_document:
    toc_float: true
package: scRepertoire
vignette: >
  %\VignetteIndexEntry{Using scRepertoire}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8} 
---

```{r, echo=FALSE, results="hide", message=FALSE, warning=FALSE}
knitr::opts_chunk$set(error=FALSE, message=FALSE, warning=FALSE, tidy = FALSE)
library(BiocStyle)
library(scater)
library(Seurat)
library(igraph)
```

# Introduction

scRepertoire is designed to take filter contig outputs from the 10x Genomics Cell Ranger pipeline, process that data to assign clonotype based on two TCR or Ig chains and analyze the clonotype dynamics. The latter can be separated into 1) clonotype-only analysis functions, such as unique clonotypes or clonal space quantification, and 2) interaction with mRNA expression data using Seurat or SingleCellExperiment packages.

## Loading Libraries

```{r}
suppressMessages(library(scRepertoire))
```

## Loading Data

### What data to load into scRepertoire?

`scRepertoire` primarily functions using the `filtered_contig_annotations.csv` output generated by 10x Genomics Cell Ranger. This file is typically found in the ./outs/ directory of your VDJ alignment folder. 

### Demonstrating Manual Data Loading (10x Genomics)

To prepare data for scRepertoire from 10x Genomics outputs, you would:

* Load the `filtered_contig_annotations.csv` file for each of your samples.
* Combine these loaded data frames into a single list in your R environment.

```{r, eval=FALSE}
S1 <- read.csv(".../Sample1/outs/filtered_contig_annotations.csv")
S2 <- read.csv(".../Sample2/outs/filtered_contig_annotations.csv")
S3 <- read.csv(".../Sample3/outs/filtered_contig_annotations.csv")
S4 <- read.csv(".../Sample4/outs/filtered_contig_annotations.csv")

contig_list <- list(S1, S2, S3, S4)
```

## Other alignment workflows

Beyond the default 10x Genomics Cell Ranger pipeline outputs, `scRepertoire` supports various other single-cell immune receptor sequencing formats through the `loadContigs()` function.

### Supported Formats and Expected File Names

* `10X`: "filtered_contig_annotations.csv"
* `AIRR`: "airr_rearrangement.tsv"
* `BD`: "Contigs_AIRR.tsv"
* `Dandelion`: "all_contig_dandelion.tsv"
* `Immcantation`: "_data.tsv" (or similar)
* `JSON`: ".json"
* `MiXCR`: "clones.tsv"
* `ParseBio`: "barcode_report.tsv"
* `TRUST4`: "barcode_report.tsv"
* `WAT3R`: "barcode_results.csv"

Key Parameter(s) for `loadContigs()`

* `input`: A directory path containing your contig files (the function will recursively search) or a list/data frame of pre-loaded contig data.
* `format`: A string specifying the data format (e.g., `10X`, `TRUST4`, `WAT3R`). If set to "auto", the function will attempt to automatically detect the format based on file names or data structure.

You can provide `loadContigs()` with a directory where your sequencing experiments are located, and it will recursively load and process the contig data based on the file names:

```{r, eval=FALSE, tidy = FALSE}
# Directory example
contig.output <- c("~/Documents/MyExperiment")
contig.list <- loadContigs(input = contig.output, 
                           format = "TRUST4")
```

Alternatively, `loadContigs()` can be given a list of pre-loaded data frames and process the contig data based on the specified format:

```{r eval = FALSE, tidy = FALSE}
# List of data frames example
S1 <- read.csv("~/Documents/MyExperiment/Sample1/outs/barcode_results.csv")
S2 <- read.csv("~/Documents/MyExperiment/Sample2/outs/barcode_results.csv")
S3 <- read.csv("~/Documents/MyExperiment/Sample3/outs/barcode_results.csv")
S4 <- read.csv("~/Documents/MyExperiment/Sample4/outs/barcode_results.csv")

contig.list <- list(S1, S2, S3, S4)
contig.list <- loadContigs(input = contig.list, 
                           format = "WAT3R")
```

## Multiplexed Experiment

It is now easy to create the contig list from a multiplexed experiment by first generating a single-cell RNA object (either Seurat or Single Cell Experiment), loading the filtered contig file, and then using `createHTOContigList()`. This function will return a list separated by the `group.by` variable(s).

### Important Considerations for `createHTOContigList()`

* This function depends on the match of barcodes between the single-cell object and contigs. If there is a prefix or different suffix added to the barcode, this will result in no contigs being recovered.
* It is currently recommended to perform this step before integration workflows, as integration commonly alters the barcodes.
* There is a multi.run variable that can be used on an integrated object. However, it assumes you have modified the barcodes with the Seurat pipeline (automatic addition of _# to the end), and your contig list is in the same order.

To create a contig list separated by HTO (Hash Tag Oligo) IDs from a single-cell object:

```{r, eval = F, tidy = FALSE}
contigs <- read.csv(".../outs/filtered_contig_annotations.csv")

contig.list <- createHTOContigList(contigs, 
                                   Seurat.Obj, 
                                   group.by = "HTO_maxID")
```

## Example Data in `scRepertoire`

`scRepertoire` includes a built-in example dataset to demonstrate the functionality of the R package. This dataset consists of T cells derived from four patients with acute respiratory distress with paired peripheral-blood (B) and bronchoalveolar lavage (L), effectively creating 8 distinct runs for T cell receptor (TCR) enrichment. More information on the data set can be found in the corresponding [manuscript](https://pubmed.ncbi.nlm.nih.gov/33622974/). 

The built-in example data is derived from the 10x Cell Ranger pipeline, so it is ready to go for downstream processing and analysis. 

To load and preview the example data built into `scRepertoire`:

```{r tidy = FALSE}
data("contig_list") #the data built into scRepertoire

head(contig_list[[1]])
```

***

# Combining Contigs into Clones

The `combineTCR()` function processes a list of TCR sequencing results, consolidating them to the level of individual cell barcodes. It handles potential issues with repeated barcodes by adding prefixes from `samples` and `ID` parameters. The output includes combined reads into clone calls by nucleotide sequence (`CTnt`), amino acid sequence (`CTaa`), VDJC gene sequence (`CTgene`), or a combination of nucleotide and gene sequence (`CTstrict`).

Key Parameter(s) for `combineTCR()`

* `input.data`: A list of filtered contig annotations (e.g., filtered_contig_annotations.csv from 10x Cell Ranger) or outputs from `loadContigs()`.
* `samples`: Labels for your samples (recommended).
* `ID`: Additional sample labels (optional).
* `removeNA`: If `TRUE`, removes any cell barcode with an NA value in at least one chain (default is `FALSE`).
* `removeMulti`: If `TRUE`, removes any cell barcode with more than two immune receptor chains (default is `FALSE`).
* `filterMulti`: If `TRUE`, isolates the top two expressed chains in cell barcodes with multiple chains (default is `FALSE`).
* `filterNonproductive`: If `TRUE`, removes non-productive chains if the variable exists in the contig data (default is `TRUE`).

To combine TCR contigs from `contig_list` and apply sample prefixes:

```{r tidy = FALSE}
combined.TCR <- combineTCR(contig_list, 
                           samples = c("P17B", "P17L", "P18B", "P18L", 
                                            "P19B","P19L", "P20B", "P20L"),
                           removeNA = FALSE, 
                           removeMulti = FALSE, 
                           filterMulti = FALSE)

head(combined.TCR[[1]])
```

`combineTCR()` is the essential first step for organizing raw TCR sequencing data into a structured format for `scRepertoire` analyses. It allows for flexible handling of single and paired chains, barcode disambiguation, and initial filtering, producing a list of data frames where each row represents a single cell and its associated TCR clonotypes.

## combineBCR

The ```combineBCR()``` function is the primary tool for processing raw B cell receptor contig data into a format ready for analysis. It is analogous to ```combineTCR()``` but includes specialized logic for handling the complexities of BCRs, such as somatic hypermutation. The function consolidates contigs into a single data frame per sample, with each row representing a unique cell.

By default `(call.related.clones = TRUE)`, ```combineBCR()``` groups B cells into clones based on the similarity of their CDR3 sequences.

### How `combineBCR()` Groups Related Clones

* Internally calling ```clonalCluster()``` to build a network of related sequences.
* Using the `threshold` parameter to define connections. The `threshold` is a normalized Levenshtein distance, where a value closer to 1.0 requires higher sequence similarity. The default of 0.85 is a good starting point.
* Assigning a cluster-based ID to the CTstrict column.

Additionally, the `group.by` argument allows you to constrain the clustering analysis to only occur within distinct categories in your metadata. For example, using `group.by = "sample"` ensures that sequences from different samples are never compared or clustered together, even if they are identical.

Key Parameter(s) for `combineBCR()`

* `call.related.clones`: If `TRUE` (default), uses `clonalCluster()` to identify related clones based on sequence similarity.
* `group.by`: The column header used to group clones for clustering (if `NULL`, clusters will be calculated across all samples).
* `threshold`: The similarity threshold for `clonalCluster()` (default: 0.85).
* `dist_type`: The metric for calculating difference: "levenshtein" (default), "hamming", "nw" (Needleman-Wunsch), or "sw" (Smith-Waterman).
* `dist_mat:` The substitution matrix used if dist_type is "nw" or "sw" (e.g., "BLOSUM62", "PAM30")
* `normalize`: How to handle the threshold: "none", "length", or "maxlen".
* `chain`: The chain to use for clustering when call.related.clones = TRUE (default: `both`).
* `sequence`: The sequence type (`nt` or `aa`) for clustering (default: `nt`).
* use.V, use.J: If `TRUE`, sequences must share the same V/J gene to be clustered (default: `TRUE`)
* `cluster.method`: The clustering algorithm to apply to the edit-distanc network (default: `components`).

First, load the example BCR contig data:

```{r tidy = FALSE}
# Load example BCR contig data
BCR.contigs <- read.csv("https://www.borch.dev/uploads/contigs/b_contigs.csv")
```

Then, combine BCR contigs using the default similarity clustering:

```{r tidy = FALSE}

# Combine using the default similarity clustering
combined.BCR.clustered <- combineBCR(BCR.contigs, 
                                     samples = "Patient1", 
                                     threshold = 0.85)

# The CTstrict column contains cluster IDs (e.g., "cluster.1")
head(combined.BCR.clustered[[1]][, c("barcode", "CTstrict", "IGH", "cdr3_aa1")])
```

### Advanced: Grouping by Alignment

For more biological accuracy—specifically when analyzing amino acid sequences—you can use alignment metrics. This allows the clustering to penalize conservative amino acid changes less than radical changes.

Here we use Needleman-Wunsch alignment with the BLOSUM80 substitution matrix:

```{r}
combined.BCR.aligned <- combineBCR(BCR.contigs, 
                                   samples = "Patient1",
                                   sequence = "aa",        
                                   dist_type = "nw",      
                                   dist_mat = "BLOSUM80",  
                                   threshold = 0.85)

head(combined.BCR.aligned[[1]][, c("barcode", "CTstrict", "IGH", "cdr3_aa1")])
```


### Filtering and Cleaning Data

`combineBCR()` includes several arguments to filter and clean the contig data during processing.
* `filterNonproductive = TRUE` (Default): Removes any contigs that are not classified as productive, ensuring that only functional receptor chains are included in the analysis.
* `filterMulti = TRUE` (Default): For cells with more than one heavy or light chain detected, this automatically selects the chain with the highest UMI count (read abundance) and discards the others. This helps resolve cellular multiplets or technical artifacts.

Here is an example applying these filters (though they are on by default):

```{r}
cleaned.BCR <- combineBCR(BCR.contigs,
                          samples = "Patient1",
                          filterNonproductive = TRUE,
                          filterMulti = TRUE)

head(cleaned.BCR[[1]])
```

`combineBCR()` is designed for processing B cell repertoire data, going beyond simple contig aggregation to incorporate advanced clustering based on CDR3 sequence similarity. This enables the identification of clonally related B cells, crucial for studying B cell development, affinity maturation, and humoral immune responses. Its filtering options further ensure the quality and interpretability of the processed data.

***

# Additional Processing

## addVariable: Adding Variables for Plotting

What if there are more variables to add than just sample and ID? We can add them by using the `addVariable()` function. For each element, the function will add a column (labeled by `variable.name`) with the `variable`. The length of the `variables` parameter needs to match the length of the combined object.

Key Parameter(s) for `addVariable()`

* `variable.name`: A character string that defines the new variable to add (e.g., "Type", "Treatment").
* `variables`: A character vector defining the desired column value for each list element. Its length must match the number of elements in the input.data list.

As an example, here we add the Type in which the samples were processed and sequenced to the `combined.TCR` object:

```{r tidy = FALSE}
combined.TCR <- addVariable(combined.TCR, 
                            variable.name = "Type", 
                            variables = rep(c("B", "L"), 4))

head(combined.TCR[[1]])
```

## subsetClones: Filter Out Clonal Information

Likewise, we can remove specific list elements after `combineTCR()` or `combineBCR()` using the `subsetClones()` function. In order to subset, we need to identify the column header we would like to use for subsetting (`name`) and the specific values to include (`variables`).

Key Parameter(s) for `subsetClones()`

* `name`: The column header/name in the metadata of input.data to use for subsetting (e.g., "sample", "Type").
* `variables`: A character vector of the specific values within the chosen name column to retain in the subsetted data.

Below, we isolate just the two sequencing results from "P18L" and "P18B" samples:

```{r, tidy = FALSE}
subset1 <- subsetClones(combined.TCR, 
                        name = "sample", 
                        variables = c("P18L", "P18B"))

head(subset1[[1]][,1:4])
```

Alternatively, we can also just select the list elements after `combineTCR()` or `combineBCR()`.

```{r}
subset2 <- combined.TCR[c(3,4)]
head(subset2[[1]][,1:4])
```

## exportClones: Save Clonal Data  

After assigning the clone by barcode, we can export the clonal information using `exportClones()` to save for later use or to integrate with other bioinformatics pipelines. This function supports various output formats tailored for different analytical needs.

Key Parameter(s) for `exportClones()`
* `format`: The desired output format for the clonal data.
  * `airr`: Exports data in an Adaptive Immune Receptor Repertoire (AIRR) Community-compliant format, with each row representing a single receptor chain.
   * `immunarch`: Exports a list containing a data frame and metadata formatted for use with the `immunarch` package. 
  * `paired`: Exports a data frame with paired chain information (amino acid, nucleotide, genes) per barcode. This is the default.
  * `TCRMatch`: Exports a data frame specifically for the TCRMatch algorithm, containing TRB chain amino acid sequence and clonal frequency.
  * `tcrpheno`: Exports a data frame compatible with the `tcrpheno` pipeline, with TRA and TRB chains in separate columns.
* `write.file`: If `TRUE` (default), saves the output to a CSV file. If `FALSE`, returns the data frame or list to the R environment.
* `dir`: The directory where the output file will be saved. Defaults to the current working directory.
* `file.name`: The name of the CSV file to be saved.

To export the combined clonotypes as a `paired` data frame and save it to a specified directory:

```{r, eval = FALSE, tidy = FALSE}
exportClones(combined, 
             write.file = TRUE,
             dir = "~/Documents/MyExperiment/Sample1/"
             file.name = "clones.csv")
```

To return an `immunarch`-formatted data frame directly to your R environment without saving a file:

```{r}
immunarch <- exportClones(combined.TCR, 
                          format = "immunarch", 
                          write.file = FALSE)
head(immunarch[[1]][[1]])
```

## clonalBin: Bin Clones by Frequency or Proportion

The `clonalBin()` function adds a clonal grouping variable (`cloneSize`) to the output of `combineTCR()`, `combineBCR()`, or `combineExpression()`. This function calculates the clonal frequency and proportion, then bins clones into categories based on customizable thresholds. This is useful for categorizing clones prior to downstream analysis or visualization, without needing to attach the data to a single-cell object.

Key Parameter(s) for `clonalBin()`

* `clone.call`: Defines the clonal sequence grouping (`gene`, `nt`, `aa`, or `strict`).
* `chain`: The TCR/BCR chain to use (`both`, `TRA`, `TRB`, `TRG`, `TRD`, `IGH`, `IGL`).
* `group.by`: A column header in the metadata to group the analysis by (e.g., "sample", "Type").
* `proportion`: Whether to use proportion (`TRUE`) or total frequency (`FALSE`) for binning.
* `clone.size`: The bins for grouping based on proportion or frequency.

### Basic Usage with Proportion

By default, `clonalBin()` uses proportion-based binning with the following default bins: `c(Rare = 1e-4, Small = 0.001, Medium = 0.01, Large = 0.1, Hyperexpanded = 1)`.

```{r}
combined.TCR <- clonalBin(combined.TCR, clone.call = "strict")

# Check the cloneSize column
head(combined.TCR[[1]][, c("barcode", "CTstrict", "clonalProportion", "clonalFrequency", "cloneSize")])

# View the distribution of clone sizes
table(combined.TCR[[1]]$cloneSize)
```

### Using Frequency-Based Binning

When using frequency-based binning (`proportion = FALSE`), the `clone.size` values must be integers. If the maximum frequency exceeds the upper bin limit, it will be automatically adjusted.

```{r}
combined.TCR.freq <- combineTCR(contig_list,
                                samples = c("P17B", "P17L", "P18B", "P18L",
                                            "P19B","P19L", "P20B", "P20L"))

combined.TCR.freq <- clonalBin(combined.TCR.freq,
                               clone.call = "strict",
                               proportion = FALSE,
                               clone.size = c(Rare = 1, Small = 5, Medium = 20,
                                              Large = 100, Hyperexpanded = 500))

# Check the frequency-based binning
head(combined.TCR.freq[[1]][, c("barcode", "clonalFrequency", "cloneSize")])
```

### Grouping by a Variable

You can calculate clonal frequency and proportion across samples grouped by a variable (e.g., "Type" or "sample"). This is particularly useful when you want consistent binning across experimental conditions.

```{r}
combined.TCR.grouped <- combineTCR(contig_list,
                                   samples = c("P17B", "P17L", "P18B", "P18L",
                                               "P19B","P19L", "P20B", "P20L"))

combined.TCR.grouped <- addVariable(combined.TCR.grouped,
                                    variable.name = "Type",
                                    variables = rep(c("B", "L"), 4))

combined.TCR.grouped <- clonalBin(combined.TCR.grouped,
                                  clone.call = "strict",
                                  group.by = "Type")

# Check grouped binning
head(combined.TCR.grouped[[1]][, c("barcode", "Type", "clonalProportion", "cloneSize")])
```

## annotateInvariant

The `annotateInvariant()` function enables the identification of mucosal-associated invariant T (`MAIT`) cells and invariant natural killer T (`iNKT`) cells in single-cell sequencing datasets. These specialized T-cell subsets are defined by their characteristic TCR usage, making them distinguishable within single-cell immune profiling data. The function extracts TCR chain information from the provided single-cell dataset and evaluates it against known invariant TCR criteria for either MAIT or iNKT cells. Each cell is assigned a score indicating the presence (1) or absence (0) of the specified invariant T-cell population.

Key Parameter(s) for `annotateInavriant()`

* `type`: Character string specifying the type of invariant T cell to annotate (`MAIT` or `iNKT`).
* `species`: Character string specifying the species (`mouse` or `human`).

```{r, eval = FALSE, tidy = FALSE}
combined <- annotateInvariant(combined, 
                              type = "MAIT", 
                              species = "human")

combined <- annotateInvariant(combined, 
                              type = "iNKT", 
                              species = "human")
```

***

# Basic Clonal Visualizations

## cloneCall: How to call clones.

* `gene` - use the VDJC genes comprising the TCR/Ig  
* `nt` - use the nucleotide sequence of the CDR3 region  
* `aa` - use the amino acid sequence of the CDR3 region  
* `strict` - use the VDJC genes comprising the TCR + the nucleotide sequence of the CDR3 region. This is the [proper definition of clonotype](https://www.ncbi.nlm.nih.gov/pubmed/19568742). For ```combineBCR()``` strict refers to the edit distance clusters + Vgene of the Ig. 

It is important to note that the clonotype is called using essentially the combination of genes or nt/aa CDR3 sequences for both loci. As of this implementation of scRepertoire, clonotype calling is not incorporating small variations within the CDR3 sequences. As such the `gene` approach will be the most sensitive, while the use of `nt` or `aa` is moderately so, and the most specific for clonotypes being `strict`. Additionally, the clonotype call is trying to incorporate both loci, *i.e.*, both **TCRA** and **TCRB** chains and if a single cell barcode has multiple sequences identified (*i.e.*, 2 TCRA chains expressed in one cell). Using the 10x approach, there is a subset of barcodes that only return one of the immune receptor chains. The unreturned chain is assigned an **NA** value.

## clonalQuant: Quantifying Unique Clones

The `clonalQuant()` function is used to explore the clones by returning the total or relative numbers of unique clones.

Key Parameter(s) for `clonalQuant()`

* `scale`: If `TRUE`, converts the output to the relative percentage of unique clones scaled by the total repertoire size; if `FALSE` (default), reports the total number of unique clones.

To visualize the relative percent of unique clones across all chains (`"both"`) using the `strict` clone definition:

```{r tidy = FALSE}
clonalQuant(combined.TCR, 
            cloneCall="strict", 
            chain = "both", 
            scale = TRUE)
```

Another option is to define the visualization by data classes using the `group.by` parameter. Here, we'll use the `"Type"` variable, which was previously added to the combined.TCR list.

```{r}
clonalQuant(combined.TCR, 
            cloneCall="gene", 
            group.by = "Type", 
            scale = TRUE)
```

## clonalAbundance: Distribution of Clones by Size

`clonalAbundance()` allows for the examination of the relative distribution of clones by abundance. It produces a line graph showing the total number of clones at specific frequencies within a sample or group.

Key Parameter(s) for `clonalAbundance()`

* `scale`: If `TRUE`, converts the graphs into density plots to show relative distributions; if `FALSE` (default), displays raw counts.

To visualize the raw clonal abundance using the `gene` clone definition:

```{r tidy = FALSE}
clonalAbundance(combined.TCR, 
                cloneCall = "gene", 
                scale = FALSE)
```

`clonalAbundance()` output can also be converted into a density plot, which may allow for better comparisons between different repertoire sizes, by setting `scale = TRUE`.

```{r}
clonalAbundance(combined.TCR, 
                cloneCall = "gene", 
                scale = TRUE)
```

## clonalLength: Distribution of Sequence Lengths

`clonalLength()` allows you to look at the length distribution of the CDR3 sequences. Importantly, unlike the other basic visualizations, the `cloneCall` can only be `nt` (nucleotide) or `aa` (amino acid). Due to the method of calling clones as outlined previously (e.g., using NA for unreturned chain sequences or multiple chains within a single barcode), the length distribution may reveal a multimodal curve.

To visualize the amino acid length distribution for both chains ("both"):

```{r tidy = FALSE}
clonalLength(combined.TCR, 
             cloneCall="aa", 
             chain = "both") 
```

To visualize the amino acid length distribution for the `TRA` chain, scaled as a density plot:

```{r tidty = FALSE}
clonalLength(combined.TCR, 
             cloneCall="aa", 
             chain = "TRA", 
             scale = TRUE) 
```

## clonalCompare: Clonal Dynamics Between Categorical Variables

`clonalCompare()` allows you to look at clones between samples and changes in dynamics. It is useful for tracking how the proportions of top clones change between conditions.

Key Parameters for `clonalCompare()`

* `samples`: A character vector to isolate specific samples by their list element name.
* `clones`: A character vector of specific clonal sequences to visualize. If used, `top.clones` will be ignored.
*` top.clones`: The top n number of clones to graph, calculated based on the frequency within individual samples.
* `highlight.clones`: A character vector of specific clonal sequences to color; all other clones will be greyed out.
* `relabel.clones`: If `TRUE`, simplifies the legend by labeling isolated clones numerically (e.g., "Clone: 1").
* `graph`: The type of plot to generate; `alluvial` (default) or `area`.
* `proportion`: If `TRUE` (default), the y-axis represents proportional abundance; if `FALSE`, it represents raw clone counts.

To compare the top 10 clones between samples "P17B" and "P17L" using amino acid sequences as an alluvial plot:

```{r tidy = FALSE}
clonalCompare(combined.TCR, 
                  top.clones = 10, 
                  samples = c("P17B", "P17L"), 
                  cloneCall="aa", 
                  graph = "alluvial")
```

We can also choose to highlight specific clones, such as in the case of *"CVVSDNTGGFKTIF_CASSVRRERANTGELFF"* and *"NA_CASSVRRERANTGELFF"* using the `highlight.clones` parameter. In addition, we can simplify the plot to label the clones as "Clone: 1", "Clone: 2", etc., by setting `relabel.clones = TRUE`.

```{r tidy = FALSE}
clonalCompare(combined.TCR, 
              top.clones = 10,
              highlight.clones = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF", "NA_CASSVRRERANTGELFF"),
              relabel.clones = TRUE,
              samples = c("P17B", "P17L"), 
              cloneCall="aa", 
              graph = "alluvial")
```

Alternatively, if we only want to show specific clones, we can use the `clones` parameter.

```{r}
clonalCompare(combined.TCR, 
              clones = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF", "NA_CASSVRRERANTGELFF"),
              relabel.clones = TRUE,
              samples = c("P17B", "P17L"), 
              cloneCall="aa", 
              graph = "alluvial")
```


## clonalScatter: Scatterplot of Two Variables

clonalScatter() organizes two repertoires, quantifies their relative clone sizes, and produces a scatter plot comparing the two samples. Clones are categorized by counts into singlets or expanded, either exclusively present or shared between the selected samples.

Key Parameter(s) for `clonalScatter()`

* `x.axis`, `y.axis`: Names of the list elements or meta data variable to place on the x-axis and y-axis.
* `dot.size`: Specifies how dot size is determined; `total` (default) displays the total number of clones between the x- and y-axis, or a specific list element name for size calculation.
* `graph`: The type of graph to display; `proportion` for the relative proportion of clones (default) or `count` for the total count of clones by sample.

To compare samples "P18B" and "P18L" based on `gene` clone calls, with dot size representing the total number of clones, and plotting clone proportions:

```{r tidy = FALSE}
clonalScatter(combined.TCR, 
              cloneCall ="gene", 
              x.axis = "P18B", 
              y.axis = "P18L",
              dot.size = "total",
              graph = "proportion")
```

***

# Visualizing Clonal Dynamics

## clonalHomeostasis: Examining Clonal Space

By examining the clonal space, we effectively look at the relative space occupied by clones at specific proportions. Another way to think about this would be to consider the total immune receptor sequencing run as a measuring cup. In this cup, we will fill liquids of different viscosity—or different numbers of clonal proportions. Clonal space homeostasis asks what percentage of the cup is filled by clones in distinct proportions (or liquids of different viscosity, to extend the analogy). The proportional cut points are set under the `cloneSize` variable in the function and can be adjusted.

Default `cloneSize` Bins

* **Rare**: 0 to 0.0001
* **Small**: 0.0001 to 0.001
* **Medium**: 0.001 to 0.01
* **Large**: 0.01 to 0.1
* **Hyperexpanded**: 0.1 to 1

To visualize clonal homeostasis using gene clone calls with default `cloneSize` bins:

```{r tidy = FALSE}
clonalHomeostasis(combined.TCR, 
                  cloneCall = "gene")

```

We can reassign the proportional cut points for `cloneSize`, which can drastically alter the visualization and analysis 

```{r tidy = FALSE}
clonalHomeostasis(combined.TCR, 
                  cloneCall = "gene",
                  cloneSize = c(Rare = 0.001, Small = 0.01, Medium = 0.1, 
                                Large = 0.3, Hyperexpanded = 1))
```

In addition, we can use the `group.by` parameter to look at the relative proportion of clones between groups, such as by tissue type. First, ensure the "Type" variable is added to `combined.TCR`:

```{r tidy = FALSE}
combined.TCR <- addVariable(combined.TCR, 
                            variable.name = "Type", 
                            variables = rep(c("B", "L"), 4))
```

Now, visualize clonal homeostasis grouped by "Type":

```{r tidy = FALSE}
clonalHomeostasis(combined.TCR, 
                  group.by = "Type",
                  cloneCall = "gene")
```

`clonalHomeostasis()` provides an assessment of how different "sizes" of clones (based on their proportional abundance) contribute to the overall repertoire. This visualization helps to identify shifts in repertoire structure, such as expansion of large clones in response to infection or contraction in chronic conditions, offering insights into immune system activity and health.

## clonalProportion: Examining Space Occupied by Ranks of Clones

Like clonal space homeostasis, `clonalProportion()` also categorizes clones into separate bins. The key difference is that instead of looking at the relative proportion of the clone to the total, `clonalProportion()` ranks the clones by their total count or frequency of occurrence and then places them into predefined bins.

The `clonalSplit` parameter represents the ranking of clonotypes. For example, 1:10 refers to the top 10 clonotypes in each sample. The default bins are set under the `clonalSplit` variable and can be adjusted. 

Default `clonalSplit` Bins

* **10** (top 1-10 clones)
* **100**: (top 11-100 clones)
* **1000**: (top 101-1000 clones)
* **10000**: (top 1001-10000 clones)
* **30000**: (top 10001-30000 clones)
* **100000**: (top 30001-100000 clones)
 
To visualize the clonal proportion using `gene` clone calls with default `clonalSplit` bins:

```{r tidy = FALSE}
clonalProportion(combined.TCR, 
                 cloneCall = "gene") 
```

To visualize clonal proportion using `nt` (nucleotide) clone calls with custom `clonalSplit` bins:

```{r}
clonalProportion(combined.TCR, 
                 cloneCall = "nt",
                 clonalSplit = c(1, 5, 10, 100, 1000, 10000)) 
```

`clonalProportion()` complements `clonalHomeostasis()` by providing a perspective on how the "richest" (most abundant) clones contribute to the overall repertoire space. By segmenting clones into rank-based bins, it helps identify whether a few highly expanded clones or a larger number of moderately expanded clones dominate the immune response, offering distinct insights into repertoire structure and dynamics.

***

# Summarizing Repertoires

## percentGeneUsage

Gene quantification and visualization has been redesigned to offer a more robust and translatable function under `percentGeneUsage()`. We have maintained the functionality of the previous functions for gene-level quantification under the aliases `vizGenes()`, `percentGenes()`, and `percentVJ()`.

### vizGenes: Flexible Gene Usage Visualization

The `vizGenes()` function offers a highly adaptable approach to visualizing the relative usage of TCR or BCR genes. It acts as a versatile alias for `percentGeneUsage()`, allowing for comparisons across different chains, scaling of values, and selection between bar charts and heatmaps.

* `x.axis`: Specifies the gene segment to display along the x-axis (e.g., "TRBV", "TRBD", "IGKJ").
* `y.axis`
  * Another gene segment (e.g., "TRAV", "TRBJ") for paired gene analysis. When x.axis and y.axis are both gene segments, vizGenes() internally calls percentGeneUsage() with genes = c(x.axis, y.axis), resulting in a heatmap.
  * A categorical variable (e.g., "sample", "orig.ident") to visualize gene usage across different groups. When y.axis is a categorical variable, vizGenes() maps it to the group.by parameter of percentGeneUsage(), creating facets for each category.
* `plot`: Determines the visualization type:
  * "barplot": Ideal for visualizing the distribution of a single gene segment.
  * "heatmap": Suitable for single gene usage (with a group.by or y.axis categorical variable) or for paired gene analysis.
* `summary.fun`: (Inherited from `percentGeneUsage`) Defines the statistic to display: "percent" (default), "proportion", or "count". This implicitly handles the scaling of values.

```{r tidy = FALSE}
vizGenes(combined.TCR,
         x.axis = "TRBV",
         y.axis = NULL, # No specific y-axis variable, will group all samples
         plot = "barplot",
         summary.fun = "proportion") 
```

This plot shows the proportion of each *TRBV* gene segment observed across the entire combined.TCR dataset. Since `y.axis` is NULL, samples are grouped by the list element of the combined.TCR data.

`vizGenes()` is particularly useful for examining gene pairings. Let's look at the differences in *TRBV* and *TRBJ* usage between peripheral blood and lung samples from your dataset. We'll subset combined.TCR for this analysis.

```{r tidy = FALSE}
# Peripheral Blood Samples
vizGenes(combined.TCR[c("P17B", "P18B", "P19B", "P20B")],
         x.axis = "TRBV",
         y.axis = "TRBJ",
         plot = "heatmap",
         summary.fun = "percent") # Display percentages

# Lung Samples
vizGenes(combined.TCR[c("P17L", "P18L", "P19L", "P20L")],
         x.axis = "TRBV",
         y.axis = "TRBJ",
         plot = "heatmap",
         summary.fun = "percent") # Display percentages
```

In these examples, by providing both `x.axis` and `y.axis` as gene segments ("TRBV" and "TRBJ"), `vizGenes()` automatically performs a paired gene analysis, generating a heatmap where the intensity reflects the percentage of each V-J pairing.

Beyond V-J pairings within a single chain, `vizGenes()` can also visualize gene usage across different chains. For instance, to examine *TRBV* and *TRAV* pairings for patient P17's samples:

```{r tidy = FALSE}
vizGenes(combined.TCR[c("P17B", "P17L")],
         x.axis = "TRBV",
         y.axis = "TRAV",
         plot = "heatmap",
         summary.fun = "count") 
```

### percentGenes: Quantifying Single Gene Usage

The `percentGenes()` function is a specialized alias for `percentGeneUsage()` designed to quantify the usage of a single V, D, or J gene locus for a specified immune receptor chain. By default, it returns a heatmap visualization.

Key Parameters for `percentGenes()`:

* `chain`: Specifies the immune receptor chain (e.g., "TRB", "TRA", "IGH", "IGL"). 
* `gene`: Indicates the gene locus to quantify: "Vgene", "Dgene", or "Jgene".
* `group.by`, `order.by`, `summary.fun`, `exportTable`, `palette`: These parameters are directly passed to `percentGeneUsage()`.

To quantify and visualize the percentage of TRBV gene usage across your samples:

```{r}
percentGenes(combined.TCR,
             chain = "TRB",
             gene = "Vgene",
             summary.fun = "percent")
```

This generates a heatmap showing the percentage of each TRBV gene segment within each sample, allowing for easy visual comparison of gene usage profiles across your samples.

The raw data returned by `percentGenes()` (when `exportTable` = TRUE) can be a powerful input for further downstream analysis, such as dimensionality reduction techniques like Principal Component Analysis. This allows you to summarize the complex gene usage patterns and identify samples with similar or distinct repertoires.

```{r}
df.genes <- percentGenes(combined.TCR,
                         chain = "TRB",
                         gene = "Vgene",
                         exportTable = TRUE,
                         summary.fun = "proportion") 

# Performing PCA on the gene usage matrix
pc <- prcomp(t(df.genes) )

# Getting data frame to plot from
df_plot <- as.data.frame(cbind(pc$x[,1:2], colnames(df.genes)))
colnames(df_plot) <- c("PC1", "PC2", "Sample")
df_plot$PC1 <- as.numeric(df_plot$PC1)
df_plot$PC2 <- as.numeric(df_plot$PC2)

ggplot(df_plot, aes(x = PC1, y = PC2)) +
  geom_point(aes(fill = Sample), shape = 21, size = 5) +
  guides(fill=guide_legend(title="Samples")) +
  scale_fill_manual(values = hcl.colors(nrow(df_plot), "inferno")) +
  theme_classic() +
  labs(title = "PCA of TRBV Gene Usage")
```

This PCA plot visually clusters samples based on their *TRBV* gene usage profiles, helping to identify underlying patterns or relationships between samples.

### percentVJ: Quantifying V-J Gene Pairings

The `percentVJ()` function is another specialized alias for `percentGeneUsage()`, specifically designed to quantify the proportion or percentage of V and J gene segments paired together within individual repertoires. It always produces a heatmap visualization.

Key Parameters for `percenVJ()`:

* `chain`: Specifies the immune receptor chain (e.g., **"TRB"**, **"TRA"**, **"IGH"**, **"IGL"**). This dictates which V and J gene segments are analyzed (e.g., **TRBV** and **TRBJ** for `chain` = "TRB")

```{r}
percentVJ(combined.TCR[1:2],
          chain = "TRB",
          summary.fun = "percent")
```

This heatmap displays the percentage of each *TRBV-TRBJ* pairing for each sample, providing a detailed view of the V-J recombination landscape.

Similar to `percentGenes()`, the quantitative output of `percentVJ()` can be used for dimensionality reduction to summarize V-J pairing patterns across samples.

```{r}
df.vj <- percentVJ(combined.TCR,
                   chain = "TRB",
                   exportTable = TRUE,
                   summary.fun = "proportion") # Export proportions for PCA

# Performing PCA on the V-J pairing matrix
pc.vj <- prcomp(t(df.vj))

# Getting data frame to plot from
df_plot_vj <- as.data.frame(cbind(pc.vj$x[,1:2], colnames(df.vj)))
colnames(df_plot_vj) <- c("PC1", "PC2", "Sample")
df_plot_vj$PC1 <- as.numeric(df_plot_vj$PC1)
df_plot_vj$PC2 <- as.numeric(df_plot_vj$PC2)

# Plotting the PCA results
ggplot(df_plot_vj, aes(x = PC1, y = PC2)) +
  geom_point(aes(fill = Sample), shape = 21, size = 5) +
  guides(fill=guide_legend(title="Samples")) +
  scale_fill_manual(values = hcl.colors(nrow(df_plot_vj), "inferno")) +
  theme_classic() +
  labs(title = "PCA of TRBV-TRBJ Gene Pairings")
```

## percentAA: Amino Acid Composition by Residue

Quantify the proportion of amino acids along the CDR3 sequence with `percentAA()`. By default, the function will pad the sequences with NAs up to the maximum of `aa.length`. Sequences longer than aa.length will be removed before visualization (default `aa.length` = 20). 

Key Parameter(s) for `percentAA()`

* `aa.length`: The maximum length of the CDR3 amino acid sequence to consider.

To visualize the relative percentage of amino acids at each position of the CDR3 sequences for the TRB chain, up to a length of 20 amino acids:

```{r, message = FALSE, tidy = FALSE}
percentAA(combined.TCR, 
          chain = "TRB", 
          aa.length = 20)
```

This plot displays the relative proportion of each amino acid at different positions across the CDR3 sequence for the specified chain. It provides insights into the amino acid composition and variability at each residue, which can be indicative of functional constraints or selection pressures on the CDR3

## positionalEntropy: Entropy across CDR3 Sequences

We can also quantify the level of entropy/diversity across amino acid residues along the CDR3 sequence. `positionalEntropy()` combines the quantification by residue of `percentAA()` with diversity calculations. Positions without variance will have a value reported as 0 for the purposes of comparison.

Key Parameter(s) for `positionalEntropy()`

* `method`  
  * `shannon` - Shannon Index
  * `inv.simpson` - Inverse Simpson Index
  * `gini.simpson` - Gini-Simpson Index
  * `norm.entropy` - Normalized Entropy 
  * `pielou` - Pielou's Evenness
  * `hill1`, `hill2`, `hill3` - Hill Numbers 
  
To visualize the normalized entropy across amino acid residues for the TRB chain, up to a length of 20 amino acids:

```{r tidy = FALSE}
positionalEntropy(combined.TCR, 
                  chain = "TRB", 
                  aa.length = 20)
```

The plot generated by `positionalEntropy()` illustrates the diversity or entropy at each amino acid position within the CDR3 sequence. Higher entropy values indicate greater variability in amino acid usage at that position, suggesting less selective pressure or more promiscuous binding, while lower values suggest conserved positions critical for structural integrity or antigen recognition.

## positionalProperty: Amino Acid Properties across CDR3 Sequence

Like ```positionalEntropy()```, we can also examine a series of amino acid properties along the CDR3 sequences using ```positionalProperty()```. Important differences for ```positionalProperty()``` is dropping NA values as they would void the mean calculation and displaying a ribbon with the 95% confidence interval surrounding the mean value for the selected properties.

Key Parameter(s) for `positionalEntropy()`

* `method`
  * `atchleyFactors`: [citation](https://pubmed.ncbi.nlm.nih.gov/15851683/)
  * `crucianiProperties`: [citation](https://analyticalsciencejournals.onlinelibrary.wiley.com/doi/abs/10.1002/cem.856)
  * `FASGAI`: [citation](https://pubmed.ncbi.nlm.nih.gov/18318694/)
  * `kideraFactors`: [citation](https://link.springer.com/article/10.1007/BF01025492)
  * `MSWHIM`: [citation](https://pubs.acs.org/doi/10.1021/ci980211b)
  * `ProtFP`: [citation](https://pubmed.ncbi.nlm.nih.gov/24059694/)
  *  `stScales`: [citation](https://pubmed.ncbi.nlm.nih.gov/19373543/)
  *  `tScales`: [citation](https://www.sciencedirect.com/science/article/abs/pii/S0022286006006314)
  *  `VHSE`: [citation](https://pubmed.ncbi.nlm.nih.gov/15895431/)
  *  `zScales`: [citation](https://pubmed.ncbi.nlm.nih.gov/9651153/)
  
To examine the Atchley factors of amino acids across the CDR3 sequence for the first two samples:

```{r tidy = FALSE}
positionalProperty(combined.TCR[c(1,2)], 
                  chain = "TRB", 
                  aa.length = 20, 
                  method = "atchleyFactors") + 
  scale_color_manual(values = hcl.colors(5, "inferno")[c(2,4)])
```

This function provides a detailed view of how physicochemical properties of amino acids change along the CDR3 sequence. By visualizing the mean property value and its confidence interval, one can identify positions with distinct characteristics or significant variations between groups, offering insights into structural and functional aspects of the CDR3.

## percentKmer: Motif Quantification 

Another quantification of the composition of the CDR3 sequence is to define motifs by sliding across the amino acid or nucleotide sequences at set intervals, resulting in substrings or kmers.

Key Parameter(s) for `percentKmer()`:

* `cloneCall`: Defines the clonal sequence grouping; only accepts `aa` (amino acids) or `nt` (nucleotides)
* `motif.length`: The length of the kmer to analyze.
* `top.motifs`: Displays the n most variable motifs as a function of median absolute deviation.

To visualize the percentage of the top 25 most variable 3-mer amino acid motifs for the TRB chain:

```{r tidy = FALSE}
percentKmer(combined.TCR, 
            cloneCall = "aa",
            chain = "TRB", 
            motif.length = 3, 
            top.motifs = 25)
```

To perform the same analysis but for nucleotide motifs:

```{r}
percentKmer(combined.TCR, 
            cloneCall = "nt",
            chain = "TRB", 
            motif.length = 3, 
            top.motifs = 25)
```

The heatmaps generated by `percentKmer()` illustrate the relative composition of frequently occurring k-mer motifs across samples or groups. This analysis can highlight recurrent sequence patterns in the CDR3, which may be associated with specific antigen recognition, disease states, or processing mechanisms. By examining the most variable motifs, researchers can identify key sequence features that differentiate immune repertoires.

***

# Comparing Clonal Diversity and Overlap

## clonalDiversity: Clonal Diversity Quantification

Diversity can be measured for samples or by other variables. `clonalDiversity()` calculates a specified diversity metric for groups within a dataset. To control for variations in library size, the function can perform bootstrapping with downsampling. It resamples each group to the size of the smallest group and calculates the diversity metric across multiple iterations, returning the mean values.

### How `clonalDiversity()` Handles Downsampling and Bootstrapping

* It determines the number of clones in the smallest group.
* For each group, it performs `n.boots` iterations (default = 100).
* In each iteration, it randomly samples the clones (with replacement) down to the size of the smallest group.
* It calculates the selected diversity metric on this downsampled set.
* The final reported diversity value is the mean of the results from all bootstrap iterations.
* This process can be disabled by setting `skip.boots = TRUE`.

### Available Diversity Metrics (`metric`)

| **Metric** | **What it measures** | **Typical scale** | **Higher value indicates…** |
| :------------- | :---------------------------------------------------------------------------------------------------------------- | :---------------- | :--------------------------------------------------------------------------------- |
| `shannon`      | Joint *richness + evenness* of clonotypes; moderately sensitive to rare clones (entropy of the frequency distribution). | 0 → ∞             | **More distinct clonotypes** *and* a more even clone-size distribution.            |
| `inv.simpson`  | Evenness-weighted diversity (inverse Simpson’s D); heavily penalises dominance by a few large clones.             | 1 → ∞             | **Few highly dominant clones**; diversity spread across many mid-abundance clones. |
| `norm.entropy` | Shannon entropy divided by ln *S* (where *S* = no. clonotypes); removes library-size dependence.                   | 0 → 1             | Same interpretation as `shannon`, but **comparable across samples of different depths**. |
| `gini.simpson` | 1 – Simpson’s D; probability that two reads drawn at random come from **different** clonotypes.                     | 0 → 1             | **Greater heterogeneity** (higher chance the two reads represent distinct clones). |
| `chao1`        | Estimated *richness only* (number of clonotypes), correcting for unseen singletons/doubletons.                    | 0 → ∞             | **More unique clonotypes**, regardless of their relative abundances.               |
| `ACE`          | Abundance-based Coverage Estimator—alternative richness estimator robust when many clones are rare.               | 0 → ∞             | **More unique clonotypes**, especially when the repertoire contains many low-frequency clones. |
| `gini`         | Inequality of clone size distribution (Gini Coefficient); measures how unevenly clone sizes are distributed.      | 0 → 1             | **Greater inequality**; a few highly abundant clones dominate the repertoire.      |
| `d50`          | Repertoire dominance; how many of the top clones are needed to constitute 50% of the total library.               | 1 → ∞             | **Less dominance** by top clones; the repertoire is more evenly distributed.       |
| `hill0`        | Richness (Hill number q=0); the total number of unique clonotypes observed in the sample.                         | 0 → ∞             | **More unique clonotypes** (greater raw richness).                                 |
| `hill1`        | Effective number of abundant clonotypes (Hill number q=1); the exponential of Shannon entropy.                    | 1 → ∞             | **Higher diversity** of common clonotypes, balancing richness and evenness.        |
| `hill2`        | Effective number of dominant clonotypes (Hill number q=2); equivalent to the inverse Simpson index.              | 1 → ∞             | **Higher diversity** among the most dominant clonotypes in the repertoire.       |


Key Parameters for `clonalDiversity()`

* `metric`: The diversity metric to calculate.
* `x.axis`: An additional metadata variable to group samples along the x-axis.
* `return.boots`: If `TRUE`, returns all bootstrap values instead of the mean.
* `skip.boots`: If `TRUE`, disables downsampling and bootstrapping.
* `n.boots`: The number of bootstrap iterations to perform (default is 100).

To calculate Shannon diversity, grouped by sample:

```{r tidy = FALSE}
clonalDiversity(combined.TCR, 
                cloneCall = "gene")
```

There are two options for grouping in `clonalDiversity()`: `group.by` and `x.axis`.

* `group.by`: Reorganizes the clone information into new groups that the calculation will be based on.
* `x.axis`: Keeps the organization of the clone information the same, but plots along the x-axis for improved visibility or grouping.

First, add a "Patient" variable to combined.TCR

```{r tidy = FALSE}
combined.TCR <- addVariable(combined.TCR, 
                            variable.name = "Patient", 
                             variables = c("P17", "P17", "P18", "P18", 
"P19","P19", "P20", "P20"))
```

Now, calculate Inverse Simpson diversity, grouped by "Patient":

```{r tidy = FALSE}
clonalDiversity(combined.TCR, 
                cloneCall = "gene", 
                group.by = "Patient", 
                metric = "inv.simpson")
```

Calculate Inverse Simpson diversity with "Patient" on the x-axis, keeping the original grouping:

```{r tidy = FALSE}
clonalDiversity(combined.TCR, 
                cloneCall = "gene", 
                x.axis = "Patient", 
                metric = "inv.simpson")
```

`clonalDiversity()` functions in quantifying and comparing the diversity of immune repertoires across different samples or experimental conditions. By implementing bootstrapping and offering a wide range of diversity metrics, it provides robust and comparable measures of clonal richness and evenness, which are fundamental for understanding immune responses and disease states

## clonalRarefaction: Sampling-based Extrapolation

`clonalRarefaction()` uses Hill numbers to estimate rarefaction, or estimating species richness, based on the abundance of clones across groupings. The underlying rarefaction calculation uses the observed receptor abundance to compute diversity. This function relies on the `iNEXT` R package.

### Hill Numbers and Their Interpretation

* `0`: Species richness
* `1`: Shannon Diversity
* `2`: Simpson Diversity

### plot.type Options

* `1`: Sample-size-based rarefaction/extrapolation curve
* `2`: Sample completeness curve
* `3`: Coverage-based rarefaction/extrapolation curve

Key Parameter(s) for `clonalRarefaction()`

* `plot.type`: The type of plot to generate.
* `hill.numbers`: The Hill numbers to be plotted (0, 1, or 2).
* `n.boots`: The number of bootstrap replicates used to derive confidence intervals (default is 20).

This function relies on the [iNEXT](https://cran.r-project.org/web/packages/iNEXT/index.html) with the accompanying [manuscript](http://chao.stat.nthu.edu.tw/wordpress/paper/120_pdf_appendix.pdf). Like the other wrapping functions in scRepertoire, please cite the original work. The sample completeness curve (**plot.type** = 2), may not show full sample coverage due to the size/diversity of the input data.

### Rarefaction using Species Richness (q = 0)

```{r, message=FALSE, tidy = FALSE}
clonalRarefaction(combined.TCR,
                  plot.type = 1,
                  hill.numbers = 0,
                  n.boots = 2)

clonalRarefaction(combined.TCR,
                  plot.type = 2,
                  hill.numbers = 0,
                  n.boots = 2)

clonalRarefaction(combined.TCR,
                  plot.type = 3,
                  hill.numbers = 0,
                  n.boots = 2)
```

### Rarefaction using Shannon Diversity (q = 1)

```{r tidy = FALSE}
clonalRarefaction(combined.TCR,
                  plot.type = 1,
                  hill.numbers = 1,
                  n.boots = 2)

clonalRarefaction(combined.TCR,
                  plot.type = 2,
                  hill.numbers = 1,
                  n.boots = 2)

clonalRarefaction(combined.TCR,
                  plot.type = 3,
                  hill.numbers = 1,
                  n.boots = 2)
```

`clonalRarefaction()` provides robust estimates of immune repertoire diversity, allowing for comparison of richness and effective diversity across samples, even with varying sequencing depths. By visualizing rarefaction and extrapolation curves, researchers can assess the completeness of their sampling and make fair comparisons of diversity, which is essential for understanding immune complexity.

## clonalSizeDistribution: Modeling Clonal Composition

Another method for modeling the repertoire distribution is a discrete gamma-GPD spliced threshold model, proposed by [Koch et al.](https://pubmed.ncbi.nlm.nih.gov/30485278/) The spliced model models the repertoire and allows for the application of a power law distribution for larger clonal-expanded sequences and a Poisson distribution for smaller clones. After fitting the models, repertoires can be compared using Euclidean distance. If using this function, please read/cite [Koch et al.](https://pubmed.ncbi.nlm.nih.gov/30485278/) and check out the [powerTCR](https://bioconductor.org/packages/release/bioc/html/powerTCR.html) R package.

Key Parameter(s) for ```clonalSizeDistribution()```

* `method`: The agglomeration method for hierarchical clustering (e.g., "ward.D2").

To model the clonal size distribution using amino acid clone calls and hierarchical clustering with the "ward.D2" method:

```{r tidy = FALSE}
clonalSizeDistribution(combined.TCR, 
                       cloneCall = "aa", 
                       method= "ward.D2")
```

`clonalSizeDistribution()` offers a sophisticated approach to modeling immune repertoire composition, distinguishing between the distribution of rare and expanded clones. By applying a spliced statistical model, it provides a more accurate representation of the repertoire's underlying clonal architecture, enabling robust comparisons and a deeper understanding of immune system dynamics.

## clonalOverlap: Exploring Sequence Overlap

If you are interested in measures of similarity between the samples loaded into scRepertoire, using `clonalOverlap()` can assist in the visualization. The underlying `clonalOverlap()` calculation varies by the `method` parameter.

### `method` Parameters for `clonalOverlap()`

* `overlap` - Overlap coefficient 
* `morisita` - Morisita's overlap index
* `jaccard` - Jaccard index
* `cosine` - Cosine similarity
* `raw` - Exact number of overlapping clones

To calculate and visualize the Morisita overlap index using strict clone calls:

```{r tidy = FALSE}
clonalOverlap(combined.TCR, 
              cloneCall = "strict", 
              method = "morisita")
```

To calculate and visualize the raw number of overlapping clones using strict clone calls:

```{r tidy = FALSE}
clonalOverlap(combined.TCR, 
              cloneCall = "strict", 
              method = "raw")
```

`clonalOverlap()` is a tool for assessing the similarity and shared clonotypes between different immune repertoires. By offering various quantitative methods and a clear heatmap visualization, it allows researchers to identify the degree of overlap, providing insights into shared immune responses, cross-reactivity, or the impact of different treatments on clonal composition.

***

# Combining Clones and Single-Cell Objects

The data in the scRepertoire package is derived from a [study](https://pubmed.ncbi.nlm.nih.gov/33622974/) of acute respiratory stress disorder in the context of bacterial and COVID-19 infections. The internal single cell data (`scRep_example()`) built in to scRepertoire is randomly sampled 500 cells from the fully integrated Seurat object to minimize the package size. We will use both Seurat and Single-Cell Experiment (SCE) with scater to perform further visualizations in tandem. 

## Preprocessed Single-Cell Object

```{r tidy = FALSE}
scRep_example <- get(data("scRep_example"))

#Making a Single-Cell Experiment object
sce <- Seurat::as.SingleCellExperiment(scRep_example)

#Adding patient information
scRep_example$Patient <- substr(scRep_example$orig.ident, 1,3)

#Adding type information
scRep_example$Type <- substr(scRep_example$orig.ident, 4,4)
```

## Note on Dimensional Reduction

In single-cell RNA sequencing workflows, dimensional reduction is typically performed by first identifying highly variable features. These features are then used directly for UMAP/tSNE projection or as inputs for principal component analysis. The same approach is commonly applied to clustering as well.

However, in immune-focused datasets, VDJ genes from TCR and BCR are often among the most variable genes. This variability arises naturally due to clonal expansion and diversity within lymphocytes. As a result, UMAP projections and clustering outcomes may be influenced by clonal information rather than broader transcriptional differences across cell types.

To mitigate this issue, a common strategy is to exclude VDJ genes from the set of highly variable features before proceeding with clustering and dimensional reduction. We introduce a set of functions that facilitate this process by removing VDJ-related genes from either a Seurat Object or a vector of gene names (useful for SCE-based workflows).

### Functions to Exclude VDJ Genes

* `quietVDJgenes()` – Removes both TCR and BCR VDJ genes.
* `quietTCRgenes()` – Removes only TCR VDJ genes.
* `quietBCRgenes()` – Removes only BCR VDJ genes, but retains BCR VDJ pseudogenes in the variable features.

Let's first check the top 10 variable features in the scRep_example Seurat object before any removal:

```{r tidy = FALSE}
# Check the first 10 variable features before removal
VariableFeatures(scRep_example)[1:10]
```

Now, we'll remove TCR VDJ genes from the scRep_example object:

```{r tidy = FALSE}
# Remove TCR VDJ genes
scRep_example <- quietTCRgenes(scRep_example)

# Check the first 10 variable features after removal
VariableFeatures(scRep_example)[1:10]
```

By applying these functions, you can ensure that clustering and dimensional reduction are driven by broader transcriptomic differences across cell types rather than being skewed by the inherent variability due to clonal expansion. This provides a more accurate representation of cellular heterogeneity independent of clonal lineage.

## Preprocessed Single-Cell Object

The data in the `scRepertoire` package is derived from a [study](https://pubmed.ncbi.nlm.nih.gov/33622974/) of acute respiratory stress disorder in the context of bacterial and COVID-19 infections. The internal single cell data (`scRep_example()`) built in to scRepertoire is randomly sampled 500 cells from the fully integrated Seurat object to minimize the package size. However, for the purpose of the vignette we will use the full single-cell object with 30,000 cells. We will use both Seurat and Single-Cell Experiment (SCE) with `scater` to perform further visualizations in tandem. 

## combineExpression

After processing the contig data into clones via `combineBCR()` or `combineTCR()`, we can add the clonal information to the single-cell object using `combineExpression()`. 

**Importantly**, the major requirement for the attachment is matching contig cell barcodes and barcodes in the row names of the metadata of the Seurat or Single-Cell Experiment object. If these do not match, the attachment will fail. We suggest making changes to the single-cell object barcodes for ease of use.

### Calculating `cloneSize`

Part of `combineExpression()` is calculating the clonal frequency and proportion, placing each clone into groups called `cloneSize`. The default `cloneSize` argument uses the following bins: `c(Rare = 1e-4, Small = 0.001, Medium = 0.01, Large = 0.1, Hyperexpanded = 1)`, which can be modified to include more/less bins or different names.

Clonal frequency and proportion are dependent on the repertoires being compared. You can modify the calculation using the `group.by` parameter, such as grouping by a Patient variable. If `group.by` is not set, `combineExpression()` will calculate clonal frequency, proportion, and `cloneSize` as a function of individual sequencing runs. Additionally, `cloneSize` can use the frequency of clones when `proportion = FALSE`. 

Key Parameter(s) for `combineExpression()`

* `input.data`: The product of `combineTCR()`, `combineBCR()`, or a list containing both.
* `sc.data`: The Seurat or Single-Cell Experiment (SCE) object to attach the clonal data to.
* `proportion`: If `TRUE` (default), calculates the proportion of the clone; if `FALSE`, calculates total frequency.
* `cloneSize`: Bins for grouping based on proportion or frequency. If proportion is `FALSE` and cloneSize bins are not set high enough, the upper limit will automatically adjust.
* `filterNA`: Method to subset the Seurat/SCE object of barcodes without clone information.
* `addLabel`: Adds a label to the frequency header, useful for trying multiple group.by variables or recalculating frequencies after subsetting.

We can look at the default cloneSize groupings using the Single-Cell Experiment object we just created, with group.by set to the sample variable used in combineTCR():

```{r tidy = FALSE}
sce <- combineExpression(combined.TCR, 
                         sce, 
                         cloneCall="gene", 
                         group.by = "sample", 
                         proportion = TRUE)

#Define color palette 
colorblind_vector <- hcl.colors(n=7, palette = "inferno", fixup = TRUE)

plotUMAP(sce, colour_by = "cloneSize") +
    scale_color_manual(values=rev(colorblind_vector[c(1,3,5,7)]))
```

Alternatively, if we want `cloneSize` to be based on the frequency of the clone, we can set proportion = FALSE and will need to change the `cloneSize` bins to integers. If we haven't inspected our clone data, setting the upper limit of the clonal frequency might be difficult; `combineExpression()` will automatically adjust the upper limit to fit the distribution of the frequencies.

```{r tidy = FALSE}
scRep_example <- combineExpression(combined.TCR, 
                                   scRep_example, 
                                   cloneCall="gene", 
                                   group.by = "sample", 
                                   proportion = FALSE, 
                                   cloneSize=c(Single=1, Small=5, Medium=20, Large=100, Hyperexpanded=500))

# Bug in Seurat 5.3.1 no longer handles NA for DimPlot
# Will Update in Future Commits

# Seurat::DimPlot(scRep_example, group.by = "cloneSize") +
#    scale_color_manual(values=rev(colorblind_vector[c(1,3,4,5,7)]))
```

### Combining both TCR and BCR 

If you have both TCR and BCR enrichment data, or wish to include information for both gamma-delta and alpha-beta T cells, you can create a single list containing the outputs of `combineTCR()` and `combineBCR()` and then use `combineExpression()`. 

**Major Note**: If there are duplicate barcodes (e.g., if a cell has both Ig and TCR information), the immune receptor information will not be added for those cells. It might be worth checking cluster identities and removing incongruent barcodes in the products of `combineTCR()` and `combineBCR()`. As an anecdote, the  [testing data](https://support.10xgenomics.com/single-cell-vdj/datasets/6.0.1/SC5v2_Melanoma_5Kcells_Connect_single_channel_SC5v2_Melanoma_5Kcells_Connect_single_channel) used to improve this function showed 5-6% barcode overlap.

```{r, eval=FALSE, tidy = FALSE}
#This is an example of the process, which will not be evaluated during knit
TCR <- combineTCR(...)
BCR <- combineBCR(...)
list.receptors <- c(TCR, BCR)


seurat <- combineExpression(list.receptors, 
                            seurat, 
                            cloneCall="gene", 
                            proportion = TRUE)
```

`combineExpression()` is a core function in `scRepertoire` that bridges the immune repertoire data with single-cell gene expression data. It enriches your Seurat or SCE object with crucial clonal information, including calculated frequencies and proportions, allowing for integrated analysis of cellular identity and clonal dynamics. Its flexibility in defining `cloneSize` and handling various grouping scenarios makes it adaptable to diverse experimental designs, even accommodating the integration of both TCR and BCR data.

*** 

# Visualizations for Single-Cell Objects

## clonalOverlay 

Using the dimensional reduction graphs as a reference, `clonalOverlay()` generates an overlay of the positions of clonally-expanded cells. It highlights areas of high clonal frequency or proportion on your UMAP or tSNE plots.

Key Parameters for clonalOverlay()

* `sc.data`: The single-cell object after` combineExpression()`.
* `reduction`: The dimensional reduction to visualize (e.g., "umap", "pca"). Default is "pca".
* `cut.category`: The metadata variable to use for filtering the overlay (e.g., "clonalFrequency" or "clonalProportion").
* `cutpoint`: The lowest clonal frequency or proportion to include in the contour plot.
* `bins`: The number of contours to draw.

`clonalOverlay()` can be used to look across all cells or faceted by a metadata variable using `facet.by`. The overall dimensional reduction will be maintained as we facet, while the contour plots will adjust based on the `facet.by` variable. The coloring of the dot plot is based on the active identity of the single-cell object.

This visualization was authored by Dr. Francesco Mazziotta and inspired by Drs. Carmona and Andreatta and their work with [ProjectTIL](https://github.com/carmonalab/ProjecTILs), a pipeline for annotating T cell subtypes.

```{r tidy = FALSE}
clonalOverlay(scRep_example, 
              reduction = "umap", 
              cutpoint = 1, 
              bins = 10, 
              facet.by = "Patient") + 
              guides(color = "none")
```

## clonalNetwork 

Similar to `clonalOverlay()`, `clonalNetwork()` visualizes the network interaction of clones shared between clusters along the single-cell dimensional reduction. This function shows the relative proportion of clones flowing from a starting node, with the ending node indicated by an arrow

### Filtering Options for clonalNetwork()

* `filter.clones`:
  * Select a number to isolate the clones comprising the top n number of cells (e.g., `filter.clones = 2000`).
  * Select `min` to scale all groups to the size of the minimum group.
* `filter.identity`: For the identity chosen to visualize, show the "to" and "from" network connections for a single group.
* `filter.proportion`: Remove clones from the network that comprise less than a certain proportion of clones in groups.
* `filter.graph`: Remove reciprocal edges from one half of the graph, allowing for cleaner visualization.
                          
Now, visualize the clonal network with no specific identity filter:

```{r tidy = FALSE}
#ggraph needs to be loaded due to issues with ggplot
library(ggraph)

#No Identity filter
clonalNetwork(scRep_example, 
              reduction = "umap", 
              group.by = "seurat_clusters",
              filter.clones = NULL,
              filter.identity = NULL,
              cloneCall = "aa")
```

We can look at the clonal relationships relative to a single cluster using the `filter.identity` parameter. For example, focusing on Cluster 3:

```{r, tidy = FALSE}
#Examining Cluster 3 only
clonalNetwork(scRep_example, 
              reduction = "umap", 
              group.by = "seurat_clusters",
              filter.identity = 3,
              cloneCall = "aa")
```

You can also use the `exportClones` parameter to quickly get clones that are shared across multiple identity groups, along with the total number of clone copies in the dataset.

```{r tidy = FALSE}
shared.clones <- clonalNetwork(scRep_example, 
                               reduction = "umap", 
                               group.by = "seurat_clusters",
                               cloneCall = "aa", 
                               exportClones = TRUE)
head(shared.clones)
```

## highlightClones

The `highlightClones()` function allows you to specifically visualize the distribution of particular clonal sequences on your single-cell dimensional reduction plots. This helps in tracking the location and expansion of clones of interest.

Key Parameters for `highlightClones()`

* `cloneCall`: The type of sequence to use for highlighting (e.g., "aa", "nt", "strict").
* `sequence`: A character vector of the specific clonal sequences to highlight.

To highlight the most prominent amino acid sequences: *CAERGSGGSYIPTF_CASSDPSGRQGPRWDTQYF* and *CARKVRDSSYKLIF_CASSDSGYNEQFF*:

```{r tidy = FALSE}
scRep_example <- highlightClones(scRep_example, 
                    cloneCall= "aa", 
                    sequence = c("CAERGSGGSYIPTF_CASSDPSGRQGPRWDTQYF", 
                                 "CARKVRDSSYKLIF_CASSDSGYNEQFF"))

# Bug in Seurat 5.3.1 no longer handles NA for DimPlot
# Will Update in Future Commits

# Seurat::DimPlot(scRep_example, group.by = "highlight") + 
#   guides(color=guide_legend(nrow=3,byrow=TRUE)) + 
#   ggplot2::theme(plot.title = element_blank(), 
#                 legend.position = "bottom")
```

## clonalOccupy

`clonalOccupy()` visualizes the count of cells by cluster, categorized into specific clonal frequency ranges. It uses the cloneSize metadata variable (generated by `combineExpression()`) to plot the number of cells within each clonal expansion designation, using a secondary variable like cluster. Credit for the idea goes to Drs. Carmona and Andreatta.

Key Parameters for `clonalOccupy()`
* `x.axis`: The variable in the metadata to graph along the x-axis (e.g., "seurat_clusters", "ident").
* `label`: If TRUE, includes the number of clones in each category by x.axis variable.
* `proportion`: If TRUE, converts the stacked bars into relative proportions.
* `na.include`: If TRUE, visualizes NA values.

To visualize the count of cells by `seurat_clusters` based on `cloneSize` groupings:

```{r tidy = FALSE}
clonalOccupy(scRep_example, 
              x.axis = "seurat_clusters")
```

To visualize the proportion of cells by `ident` (active identity), without labels:

```{r}
clonalOccupy(scRep_example, 
             x.axis = "ident", 
             proportion = TRUE, 
             label = FALSE)
```

## alluvialClones

After the metadata has been modified with clonal information, `alluvialClones()` allows you to look at clones across multiple categorical variables, enabling the examination of the interchange between these variables. Because this function produces a graph with each clone arranged by called stratification, it may take some time depending on the size of the repertoire.

Key Parameters for `alluvialClones()`

* `y.axes`: The columns that will separate the proportional visualizations.
* `color`: The column header or clone(s) to be highlighted.
* `facet`: The column label to separate facets.
* `alpha`: The column header to have gradated opacity.
* `top.clones`: Show only the top N clones by frequency.
* `min.freq`: Minimum frequency threshold for displaying flows.
* `highlight.clones`: Character vector of specific clone sequences to highlight.
* `highlight.color`: Color to use for highlighted clones (default: "red").

### Basic Alluvial Plot

To visualize clonal flow across "Patient", "ident", and "Type", highlighting specific amino acid clones:

```{r tidy = FALSE}
alluvialClones(scRep_example,
               clone.call = "aa",
               y.axes = c("Patient", "ident", "Type"),
               color = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF", "NA_CASSVRRERANTGELFF")) +
    scale_fill_manual(values = c("grey", colorblind_vector[3]))
```

To visualize clonal flow across "Patient", "ident", and "Type", coloring by "ident":

```{r tidy = FALSE}
alluvialClones(scRep_example,
               clone.call = "gene",
               y.axes = c("Patient", "ident", "Type"),
               color = "ident")
```

### Filtering to Top Clones

When working with large repertoires, it can be helpful to focus on the most frequent clones. The `top.clones` parameter allows you to filter to the top N clones:

```{r tidy = FALSE}
alluvialClones(scRep_example,
               clone.call = "aa",
               y.axes = c("Patient", "ident"),
               top.clones = 25,
               color = "ident")
```

### Highlighting Specific Clones

Use `highlight.clones` to call out specific clones of interest while showing the context of other clones:

```{r tidy = FALSE}
alluvialClones(scRep_example,
               clone.call = "aa",
               y.axes = c("Patient", "ident", "Type"),
               highlight.clones = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF"),
               highlight.color = "red")
```

### Visual Customization

Fine-tune the appearance with parameters like `stratum.width`, `flow.alpha`, and `label.size`:

```{r tidy = FALSE}
alluvialClones(scRep_example,
               clone.call = "gene",
               y.axes = c("Patient", "ident"),
               color = "ident",
               stratum.width = 0.3,
               flow.alpha = 0.7,
               label.size = 3)
```

`alluvialClones()` provides a visual representation of clonal distribution and movement across multiple categorical annotations. It is particularly effective for tracking how specific clones or clonal groups transition between different states, tissues, or cell types, offering a dynamic perspective on immune repertoire evolution and function.

## getCirclize and vizCirclize

Like alluvial graphs, we can also visualize the interconnection of clusters using chord diagrams from the [circlize R package](https://jokergoo.github.io/circlize_book/book/). scRepertoire provides two functions for this:

* `getCirclize()` - Generates data for use with the circlize package
* `vizCirclize()` - A convenient wrapper that handles the circlize plotting automatically

These functions calculate the relative number of unique and shared clones based on the `group.by` variable using the product of `combineExpression()`. If using the downstream `circlize` R package, please read and cite the following [manuscript](https://pubmed.ncbi.nlm.nih.gov/24930139/).

Key Parameters for `getCirclize()` and `vizCirclize()`

* `group.by`: A column header (or vector of column headers for hierarchical grouping).
* `method`: The method for calculating link values: `"unique"` (default), `"jaccard"`, or `"overlap"`.
* `proportion`: If `TRUE`, normalizes the relationship by proportion; if `FALSE` (default), uses unique clone counts.
* `include.self`: If `TRUE`, includes counting clones within a single group comparison.
* `min.shared`: Minimum number of shared clones to include a link.

### Quick Visualization with vizCirclize

The simplest way to create a chord diagram is using `vizCirclize()`:

```{r tidy = FALSE}
library(circlize)
library(scales)

vizCirclize(scRep_example,
            group.by = "seurat_clusters",
            clone.call = "aa")
```

### Manual Control with getCirclize

For more control over the visualization, use `getCirclize()` to generate the data and then plot with circlize directly. By default, `getCirclize()` returns a simple data frame:

```{r tidy = FALSE}
circles <- getCirclize(scRep_example,
                       group.by = "seurat_clusters")

#Just assigning the normal colors to each cluster
grid.cols <- hue_pal()(length(unique(scRep_example$seurat_clusters)))
names(grid.cols) <- unique(scRep_example$seurat_clusters)

#Graphing the chord diagram
chordDiagram(circles, self.link = 1, grid.col = grid.cols)
```

### Multi-Level Hierarchical Grouping

A powerful feature is the ability to create hierarchical groupings by passing a vector of column names to `group.by`. This creates compound sector labels:

```{r tidy = FALSE}
vizCirclize(scRep_example,
            group.by = c("Patient", "Type"),
            clone.call = "aa",
            min.shared = 1)
```

### Directional Flow

Show the direction of clonal sharing with arrows:

```{r tidy = FALSE}
vizCirclize(scRep_example,
            group.by = "seurat_clusters",
            clone.call = "aa",
            directional = TRUE)
```

### Different Similarity Methods

Use different methods for calculating link values:

```{r tidy = FALSE}
# Jaccard similarity
circles.jaccard <- getCirclize(scRep_example,
                               group.by = "seurat_clusters",
                               method = "jaccard",
                               include.metadata = TRUE)

# The include.metadata option returns sector statistics
head(circles.jaccard$sector.stats)
```

### Subsetting and Proportion

This can also be used if we want to explore just the lung-specific T cells by subsetting the single-cell object. For the sake of this vignette, we can also look at setting `proportion = TRUE` to get a scaled output.

```{r}
subset <- subset(scRep_example, Type == "L")

circles <- getCirclize(subset, group.by = "ident", proportion = TRUE)

grid.cols <- scales::hue_pal()(length(unique(subset@active.ident)))
names(grid.cols) <- levels(subset@active.ident)

chordDiagram(circles,
             self.link = 1,
             grid.col = grid.cols,
             directional = 1,
             direction.type = "arrows",
             link.arr.type = "big.arrow")
```

`getCirclize()` and `vizCirclize()` facilitate the creation of visually striking and informative chord diagrams to represent shared clonal relationships between distinct groups within your single-cell data. By providing a flexible way to quantify and format clonal overlap, they enable researchers to effectively illustrate complex clonal connectivity patterns, which are crucial for understanding immune communication and migration.

***

# Quantifying Clonal Bias

## StartracDiversity

From the excellent work by [Zhang et al. (2018, Nature)](https://www.nature.com/articles/s41586-018-0694-x), the authors introduced new methods for looking at clones by cellular origins and cluster identification. We strongly recommend you read and cite their publication when using this function.

To use the ```StartracDiversity()``` function, you need the output of the ```combineExpression()``` function and a column in your metadata that specifies the tissue of origin. 

### Indices Output from `StartracDiversity()`

*  `expa` - Clonal Expansion. Measures the degree of clonal proliferation within a given cell cluster. It is calculated as `1 - normalized Shannon entropy`, where a higher value indicates that a few clones dominate the cluster.  
*  `migr` - Cross-tissue Migration. Quantifies the movement of clonal T cells between different tissues, as defined by the `type` parameter. This index is based on the entropy of a single clonotype's distribution across the specified tissues.  
*  `tran` - State Transition. Measures the developmental transition of T cell clones between different functional clusters. This index is calculated from the entropy of a single clonotype's distribution across the cell clusters identified in the data.

Key Parameters for `StartracDiversity()`

* `type`: The variable in the metadata that provides tissue type.
* `group.by`: A column header in the metadata to group the analysis by (e.g., "sample", "treatment"). 

By default, `StartracDiversity()` will calculate all three indices (`expa`, `migr`, and `tran`) for each cluster and group you define. This provides a comprehensive overview of the clonal dynamics in your dataset. The output is a three-paneled plot, with each panel representing one index.

In the example data, `type` corresponds to the "Type" column, which includes "P" (peripheral blood) and "L" (lung) classifiers. The analysis is grouped by the "Patient" column.

```{r}
# Calculate and plot all three STARTRAC indices
StartracDiversity(scRep_example, 
                  type = "Type", 
                  group.by = "Patient")
```

### Calculating a Single Index

If you're only interested in one aspect of clonal dynamics, you can specify it using the `index` parameter. For example, to only calculate and plot clonal expansion:

```{r}
# Calculate and plot only the clonal expansion index
StartracDiversity(scRep_example, 
                  type = "Type", 
                  group.by = "Patient",
                  index = "expa")
```

### Pairwise Migration Analysis

Another feature of `StartracDiversity()` is the ability to perform pairwise comparisons. To specifically quantify the migration between two tissues (e.g., Lung vs. Periphery), set `index = "migr"` and tell the function which column to use for the comparison with `pairwise = "Type"`.

```{r}
# # Calculate pairwise migration between tissues
StartracDiversity(scRep_example, 
                  type = "Type", 
                  group.by = "Patient",
                  index = "migr",
                  pairwise = "Type")
```

## clonalBias

A new metric proposed by [Massimo et al](https://pubmed.ncbi.nlm.nih.gov/35829695/), ```clonalBias()```, like STARTRAC, is a clonal metric that seeks to quantify how individual clones are skewed towards a specific cellular compartment or cluster. A clone bias of `1` indicates that a clone is composed of cells from a single compartment or cluster, while a clone bias of `0` matches the background subtype distribution. Please read and cite the linked manuscript if using `clonalBias()`

Key Parameter(s) for `clonalBias()`

* `group.by`: A column header in the metadata that bias will be based on.
* `split.by`: The variable to use for calculating the baseline frequencies (e.g., "Type" for lung vs peripheral blood comparison)
* `n.boots`: Number of bootstraps to downsample.
* `min.expand`: Clone frequency cut-off for the purpose of comparison (default = 10).

Here we calculate and plot clonal bias using `aa` clone calls, splitting by "Patient" and grouping by "seurat_clusters", with a minimum expansion of 5 and 10 bootstraps:

```{r, message = FALSE, tidy = FALSE}
clonalBias(scRep_example, 
           cloneCall = "aa", 
           split.by = "Patient", 
           group.by = "seurat_clusters",
           n.boots = 10, 
           min.expand =5)
```

***

# Clustering by Edit Distance

## clonalCluster: Cluster by Sequence Similarity

The `clonalCluster()` function provides a powerful method to group clonotypes based on sequence similarity. It calculates the edit distance between CDR3 sequences and uses this information to build a network, identifying closely related clusters of T or B cell receptors. This functionality allows for a more nuanced definition of a "clone" that extends beyond identical sequence matches.

### Core Concepts

The clustering process follows these key steps:

* **Sequence Selection**: The function selects either the nucleotide (`sequence = "nt"`) or amino acid (`sequence = "aa"`) CDR3 sequences for a specified chain.
* **Distance Calculation**: It calculates the edit distance between all pairs of sequences. By default, it also requires sequences to share the same V gene (`use.v = TRUE`).
* **Network Construction**: An edge is created between any two sequences that meet the similarity threshold, forming a network graph.
* **Clustering**: A graph-based clustering algorithm is run to identify connected components or communities within the network. By default, it identifies all directly or indirectly connected sequences as a single cluster (`cluster.method = "components"`).
* **Output**: The function can either add the resulting cluster IDs to the input object, return an `igraph` object for network analysis, or export a sparse adjacency matrix.

### Understanding the `threshold` Parameter

The behavior of the threshold parameter is critical for controlling cluster granularity:

* **Normalized Similarity (threshold < 1)**: When the threshold is a value between 0 and 1 (e.g., `0.85`), it represents the normalized edit distance (Levenshtein distance / mean sequence length). A higher value corresponds to a stricter similarity requirement. This is useful for comparing sequences of varying lengths.
* **Raw Edit Distance (threshold >= 1)**: When the threshold is a whole number (e.g., 2), it represents the maximum raw edit distance allowed. A lower value corresponds to a stricter similarity requirement. This is useful when you want to allow a specific number of mutations.

Key Parameter(s) for ```clonalCluster()```

* `sequence`: Specifies whether to cluster based on `aa` (amino acid) or `nt` (nucleotide) sequences.
* `threshold`: The similarity threshold for clustering. Values `< 1` are normalized similarity, while values `>= 1` are raw edit distance.
* `group.by`: A column header in the metadata or lists to group the analysis by (e.g., "sample", "treatment"). If `NULL`, clusters are calculated across all sequences.
* `use.V`: If `TRUE`, sequences must share the same V gene to be clustered together.
* `use.J`: If `TRUE`, sequences must share the same J gene to be clustered together.
* `cluster.method`: The clustering algorithm to use. Defaults to `components`, which finds connected subgraphs.
* `cluster.prefix`: A character prefix to add to the cluster names (e.g., "cluster.").
* `exportGraph`: If `TRUE`, returns an igraph object of the sequence network.
* `exportAdjMatrix`: If `TRUE`, returns a sparse adjacency matrix (dgCMatrix) of the network.

### Demonstrating Basic Use

To run clustering on the first two samples for the TRA chain, using amino acid sequences with a normalized similarity threshold of 0.85:

```{r tidy = FALSE}
# Run clustering on the first two samples for the TRA chain
sub_combined <- clonalCluster(combined.TCR[c(1,2)], 
                              chain = "TRA", 
                              sequence = "aa", 
                              threshold = 0.85)

# View the new cluster column
head(sub_combined[[1]][, c("barcode", "TCR1", "TRA.Cluster")])
```

### Demonstrating Clustering with a Single-Cell Object

You can calculate clusters based on specific metadata variables within a single-cell object by using the `group.by` parameter. This is useful for analyzing clusters on a per-sample or per-patient basis without subsetting the data first.

First, add patient and type information to the `scRep_example` Seurat object:

```{r tidy = FALSE}
#Adding patient information
scRep_example$Patient <- substr(scRep_example$orig.ident, 1,3)

#Adding type information
scRep_example$Type <- substr(scRep_example$orig.ident, 4,4)
```

Now, run clustering on the `scRep_example` Seurat object, grouping calculations by "Patient":

```{r}
# Run clustering, but group calculations by "Patient"
scRep_example <- clonalCluster(scRep_example, 
                               chain = "TRA", 
                               sequence = "aa", 
                               threshold = 0.85, 
                               group.by = "Patient")

#Define color palette 
num_clusters <- length(unique(na.omit(scRep_example$TRA.Cluster)))
cluster_colors <- hcl.colors(n = num_clusters, palette = "inferno")

# Bug in Seurat 5.3.1 no longer handles NA for DimPlot
# Will Update in Future Commits

# DimPlot(scRep_example, group.by = "TRA.Cluster") +
#  scale_color_manual(values = cluster_colors, na.value = "grey") + 
#  NoLegend()
```

### Returning an igraph Object:

Instead of modifying the input object, ```clonalCluster()``` can export the underlying network structure for advanced analysis. Set `exportGraph = TRUE` to get an igraph object consisting of the networks of barcodes by the indicated clustering scheme. 

```{r tidy = FALSE}
set.seed(42)
#Clustering Patient 19 samples
igraph.object <- clonalCluster(combined.TCR[c(5,6)],
                               chain = "TRB",
                               sequence = "aa",
                               group.by = "sample",
                               threshold = 0.85, 
                               exportGraph = TRUE)

#Setting color scheme
col_legend <- factor(igraph::V(igraph.object)$group)
col_samples <- hcl.colors(2,"inferno")[as.numeric(col_legend)]
color.legend <- factor(unique(igraph::V(igraph.object)$group))
sample.vertices <- V(igraph.object)[sample(length(igraph.object), 500)]

subgraph.object <- induced_subgraph(igraph.object, vids = sample.vertices)
V(subgraph.object)$degrees <- igraph::degree(subgraph.object)
edge_alpha_color <- adjustcolor("gray", alpha.f = 0.3)

#Plotting
plot(subgraph.object,
     layout = layout_nicely(subgraph.object),
     vertex.label = NA,
     vertex.size = sqrt(igraph::V(subgraph.object)$degrees), 
     vertex.color = col_samples[sample.vertices],
     vertex.frame.color = "white", 
     edge.color = edge_alpha_color,
     edge.arrow.size = 0.05,
     edge.curved = 0.05, 
     margin = -0.1)
legend("topleft", 
       legend = levels(color.legend), 
       pch = 16, 
       col = unique(col_samples), 
       bty = "n")
```

### Returning a Sparse Adjacency Matrix

For computational applications, you can export a sparse adjacency matrix using `exportAdjMatrix = TRUE`. This matrix represents the connections between all barcodes in the input data, with the edit distance that meet the threshold in places of connection.

```{r}
# Generate the sparse matrix
adj.matrix <- clonalCluster(combined.TCR[c(1,2)],
                            chain = "TRB",
                            exportAdjMatrix = TRUE)

# View the dimensions and a snippet of the matrix
dim(adj.matrix)
print(adj.matrix[1:10, 1:10])
```

### Using Both Chains

You can analyze the combined network of both TRA/TRB or IGH/IGL chains by setting `chain = "both"`. This will create a single cluster column named `Multi.Cluster`.

```{r}
# Cluster using both TRB and TRA chains simultaneously
clustered_both <- clonalCluster(combined.TCR[c(1,2)], 
                                chain = "both")

# View the new "Multi.Cluster" column
head(clustered_both[[1]][, c("barcode", "TCR1", "TCR2", "Multi.Cluster")])
```

### Using Different Clustering Algorithms

While the default `cluster.method = "components"` is robust, you can use other algorithms from the igraph package, such as `walktrap` or `louvain`, to potentially uncover different community structures.

```{r}
# Cluster using the walktrap algorithm
graph_walktrap <- clonalCluster(combined.TCR[c(1,2)],
                                cluster.method = "walktrap",
                                exportGraph = TRUE)

# Compare the number of clusters found
length(unique(V(graph_walktrap)$cluster))
```

Overall, `clonalCluster()` is a versatile function for defining and analyzing clonal relationships based on sequence similarity. It allows researchers to move beyond exact sequence matches, providing a more comprehensive understanding of clonal families. The ability to customize parameters like `threshold`, `chain` selection, and `group.by` ensures adaptability to diverse research questions. Furthermore, the option to export `igraph` objects or sparse adjacency matrices provides advanced users with the tools for in-depth network analysis.

***

# Conclusion  

This has been a general overview of the capabilities of scRepertoire from the initial processing and visualization to attach to the mRNA expression values in a single-cell object. If you have any questions, comments, or suggestions, please visit the [GitHub repository](https://github.com/BorchLab/scRepertoire).

### Session Info

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalCluster.R
\name{clonalCluster}
\alias{clonalCluster}
\title{Cluster clones by sequence similarity}
\usage{
clonalCluster(
  input.data,
  chain = "TRB",
  sequence = "aa",
  threshold = 0.85,
  group.by = NULL,
  dist.type = NULL,
  dist.mat = NULL,
  normalize = "length",
  gap.open = NULL,
  gap.extend = NULL,
  cluster.method = "components",
  cluster.prefix = "cluster.",
  use.V = TRUE,
  use.J = FALSE,
  export.adj.matrix = NULL,
  export.graph = NULL,
  dist_type = NULL,
  dist_mat = NULL,
  gap_open = NULL,
  gap_extend = NULL,
  exportAdjMatrix = NULL,
  exportGraph = NULL
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}},
\code{\link[=combineBCR]{combineBCR()}} or \code{\link[=combineExpression]{combineExpression()}}.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{sequence}{Clustering based on either \code{aa} or \code{nt} sequences.}

\item{threshold}{The similarity threshold. If < 1, treated as normalized
similarity (higher is stricter). If >= 1, treated as raw edit distance
(lower is stricter).}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, clusters will be calculated across
all sequences.}

\item{dist.type}{The distance metric to use. Options: \code{"levenshtein"} (default),
\code{"hamming"}, \code{"damerau"} (allows transpositions), \code{"nw"} (Needleman-Wunsch),
or \code{"sw"} (Smith-Waterman).}

\item{dist.mat}{The substitution matrix to use for alignment-based metrics
(\code{"nw"} or \code{"sw"}). Options: \code{"BLOSUM45"}, \code{"BLOSUM50"}, \code{"BLOSUM62"},
\code{"BLOSUM80"} (default), \code{"BLOSUM100"}, \code{"PAM30"}, \code{"PAM40"}, \code{"PAM70"}, \code{"PAM120"},
\code{"PAM250"}, or \code{"identity"}.}

\item{normalize}{Method for normalizing distances. Options: \code{"none"},
\code{"maxlen"} (divide by max sequence length), or \code{"length"} (default, divide
by mean sequence length). If \code{threshold < 1}, this controls how the
similarity is calculated.}

\item{gap.open}{Penalty for opening a gap in alignment metrics (default: -10).}

\item{gap.extend}{Penalty for extending a gap in alignment metrics (default: -1).}

\item{cluster.method}{The clustering algorithm to use. Defaults to \code{"components"},
which finds connected subgraphs.}

\item{cluster.prefix}{A character prefix to add to the cluster names (e.g.,
"cluster.").}

\item{use.V}{If \code{TRUE}, sequences must share the same V gene to be
clustered together.}

\item{use.J}{If \code{TRUE}, sequences must share the same J gene to be
clustered together.}

\item{export.adj.matrix}{If \code{TRUE}, the function returns a sparse
adjacency matrix (\code{dgCMatrix}) of the network.}

\item{export.graph}{If \code{TRUE}, the function returns an \code{igraph}
object of the sequence network.}

\item{dist_type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{dist.type} instead.}

\item{dist_mat}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{dist.mat} instead.}

\item{gap_open}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{gap.open} instead.}

\item{gap_extend}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{gap.extend} instead.}

\item{exportAdjMatrix}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.adj.matrix} instead.}

\item{exportGraph}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.graph} instead.}
}
\value{
Depending on the export parameters, one of the following:
\itemize{
\item An amended \code{input.data} object with a new metadata column containing cluster IDs (default).
\item An \code{igraph} object if \code{export.graph = TRUE}.
\item A sparse \code{dgCMatrix} object if \code{export.adj.matrix = TRUE}.
}
}
\description{
This function clusters TCRs or BCRs based on the edit distance or alignment
score of their CDR3 sequences. It can operate on either nucleotide (\code{nt})
or amino acid (\code{aa}) sequences and can optionally enforce that clones share
the same V and/or J genes. The output can be the input object with an added
metadata column for cluster IDs, a sparse adjacency matrix, or an \code{igraph}
graph object representing the cluster network.
}
\details{
The clustering process is as follows:
\enumerate{
\item The function retrieves the relevant chain data from the input object.
\item It calculates the distance between all sequences within each group
(or across the entire dataset if \code{group.by} is \code{NULL}).
\item An edge list is constructed, connecting sequences that meet the similarity
\code{threshold}.
\item The \code{threshold} parameter behaves differently based on its value:
\itemize{
\item \strong{\code{threshold} < 1 (e.g., 0.85):} Interpreted as a \emph{normalized}
distance. A higher value means greater similarity is required.
\item \strong{\code{threshold} >= 1 (e.g., 2):} Interpreted as a maximum \emph{raw} edit
distance. A lower value means greater similarity is required.
}
\item \strong{Distance Metrics:}
\itemize{
\item \strong{Levenshtein/Hamming/Damerau:} Standard edit distance calculations.
\item \strong{Alignment (NW/SW):} If \code{dist.type} is "nw" (Needleman-Wunsch) or
"sw" (Smith-Waterman), alignment scores are calculated using the
specified substitution matrix (\code{dist.mat}). These scores are converted
to a distance-like metric for clustering.
}
\item An \code{igraph} graph is built from the edge list.
\item A clustering algorithm is run on the graph (default: connected components).
}
}
\examples{
# Getting the combined contigs
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B","P19L", "P20B", "P20L"))

# Standard Levenshtein clustering (85\% similarity)
sub_combined <- clonalCluster(combined[c(1,2)],
                              chain = "TRA",
                              sequence = "aa",
                              threshold = 0.85)

# Alignment-based clustering using BLOSUM80
sub_combined_nw <- clonalCluster(combined[c(1,2)],
                                 chain = "TRA",
                                 dist.type = "nw",
                                 dist.mat = "BLOSUM80",
                                 threshold = 0.85)

# Export the graph object instead
graph_obj <- clonalCluster(combined[c(1,2)],
                           chain = "TRA",
                           export.graph = TRUE)

}
\concept{Visualizing_Clones}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/percentGeneUsage.R
\name{percentGeneUsage}
\alias{percentGeneUsage}
\alias{vizGenes}
\alias{percentGenes}
\alias{percentVJ}
\title{Summarizes and Visualizes Gene Usage}
\usage{
percentGeneUsage(
  input.data,
  chain = "TRB",
  genes = "TRBV",
  group.by = NULL,
  order.by = NULL,
  summary.fun = c("percent", "proportion", "count"),
  plot.type = "heatmap",
  export.table = NULL,
  palette = "inferno",
  exportTable = NULL,
  ...
)

vizGenes(
  input.data,
  x.axis = "TRBV",
  y.axis = NULL,
  group.by = NULL,
  plot = "heatmap",
  order.by = NULL,
  summary.fun = c("percent", "proportion", "count"),
  export.table = NULL,
  palette = "inferno",
  exportTable = NULL
)

percentGenes(
  input.data,
  chain = "TRB",
  gene = "Vgene",
  group.by = NULL,
  order.by = NULL,
  export.table = NULL,
  summary.fun = c("percent", "proportion", "count"),
  palette = "inferno",
  exportTable = NULL
)

percentVJ(
  input.data,
  chain = "TRB",
  group.by = NULL,
  order.by = NULL,
  summary.fun = c("percent", "proportion", "count"),
  export.table = NULL,
  palette = "inferno",
  exportTable = NULL
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}}, \code{\link[=combineBCR]{combineBCR()}},
or \code{\link[=combineExpression]{combineExpression()}}.}

\item{chain}{The TCR/BCR chain to use. Accepted values: \code{TRA}, \code{TRB}, \code{TRG},
\code{TRD}, \code{IGH}, \code{IGL} (for both light chains)}

\item{genes}{A character vector specifying the gene loci to analyze.
Can be a single gene e.g., "TRBV" or "IGHJ" or a pair for genes analysis
(e.g., c("TRBV", "TRAV"), or "TRBV", "TRBJ").}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed as
by list element or active identity in the case of single-cell objects.}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{summary.fun}{Character string choosing the summary statistic -
\code{"percent"} (default), \code{"proportion"}, or \code{"count"}.}

\item{plot.type}{The type of plot to return: \code{"heatmap"} (default for paired loci,
also available for single loci), or \code{"barplot"} (for single loci).}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.
\link[grDevices:palettes]{hcl.pals}.}

\item{...}{Additional arguments passed to the ggplot theme}

\item{x.axis}{Gene segments to separate the x-axis, such as \code{TRAV},
\code{TRBD}, \code{IGKJ}.}

\item{y.axis}{Variable to separate the y-axis, can be both categorical
or other gene gene segments, such as \code{TRAV}, \code{TRBD}, \code{IGKJ}.}

\item{plot}{The type of plot to return - heatmap or barplot.}

\item{gene}{\code{Vgene}, \code{Dgene} or \code{Jgene}}
}
\value{
A ggplot object displaying a heatmap or bar plot of gene usage.
If \code{exportTable = TRUE}, a matrix or data frame of the raw data is returned.
}
\description{
This function quantifies and visualizes the usage of V, D, or J genes,
or gene pairings across T or B cell clones.
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))
                                    
# Visualize single gene (TRBV) usage as a heatmap, grouped by sample
percentGeneUsage(combined,
                 genes = "TRBV",
                 group.by = "sample",
                 plot.type = "heatmap",
                 summary.fun = "percent")

# Visualize single gene (TRBV) usage as a barplot, grouped by sample
percentGeneUsage(combined,
                 genes = "TRBV",
                 group.by = "sample",
                 plot.type = "barplot",
                 summary.fun = "count")

# Visualize paired gene (TRBV-TRBJ) usage as a heatmap
percentGeneUsage(combined[1:2],
                 genes = c("TRBV", "TRBJ"),
                 group.by = "sample",
                 plot.type = "heatmap",
                 summary.fun = "proportion")

# Export the raw data table for single gene usage
trbv_usage_table <- percentGeneUsage(combined,
                                     genes = "TRBV",
                                     group.by = "sample",
                                     export.table = TRUE,
                                     summary.fun = "count")

# Export the raw data table for paired gene usage
trbv_trbj_usage_table <- percentGeneUsage(combined,
                                          genes = c("TRBV", "TRBJ"),
                                          group.by = "sample",
                                          export.table = TRUE,
                                          summary.fun = "percent")


# Visualize paired gene (TRAV-TRAJ) usage as a heatmap
vizGenes(combined[1:2],
         x.axis = "TRAV",
         y.axis = "TRAJ",
         group.by = "sample",
         summary.fun = "count")

# Visualize cross-chain gene pairing (TRBV-TRAV)
vizGenes(combined[1:2],
         x.axis = "TRBV",
         y.axis = "TRAV",
         group.by = "sample",
         summary.fun = "percent")


# Quantify and visualize TRA V-gene usage as a heatmap
percentGenes(combined,
             chain = "TRA",
             gene = "Vgene",
             group.by = "sample",
             summary.fun = "percent")

# Quantify TRA J-gene usage and export the table
traj_usage_table <- percentGenes(combined,
                                 chain = "TRA",
                                 gene = "Jgene",
                                 group.by = "sample",
                                 export.table = TRUE,
                                 summary.fun = "count")

# Quantify and visualize TRB V-J gene pairings as a heatmap
percentVJ(combined[1:2],
          chain = "TRB",
          group.by = "sample",
          summary.fun = "percent")

# 2. Quantify TRA V-J gene pairings and export the table
trav_traj_table <- percentVJ(combined,
                             chain = "TRA",
                             group.by = "sample",
                             export.table = TRUE,
                             summary.fun = "proportion")
}
\concept{Summarize_Repertoire}
\concept{Visualizing_Clones}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quietVDJgenes.R
\name{getHumanIgPseudoGenes}
\alias{getHumanIgPseudoGenes}
\title{Get Human Immunoglobulin pseudogenes}
\usage{
getHumanIgPseudoGenes()
}
\value{
Character vector of human immunoglobulin pseudogenes.
}
\description{
This function returns a character vector of human immunoglobulin
pseudogenes. These are also the genes that are removed from the
variable gene list in \code{\link[=quietVDJgenes]{quietVDJgenes()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadContigs.R
\name{loadContigs}
\alias{loadContigs}
\title{Load Immune Receptor Sequencing Contigs}
\usage{
loadContigs(input, format = "10X")
}
\arguments{
\item{input}{A directory path containing contig files or a list/data frame
of pre-loaded contig data.}

\item{format}{A string specifying the data format. Must be one of:
\code{auto}, \verb{10X}, \code{AIRR}, \code{BD}, \code{Dandelion}, \code{JSON}, \code{MiXCR}, \code{ParseBio},
\code{TRUST4}, \code{WAT3R}, or \code{Immcantation}. If "auto", the function attempts
automatic format detection.}
}
\value{
A list of contigs formatted for use with \code{\link[=combineTCR]{combineTCR()}} or
\code{\link[=combineBCR]{combineBCR()}}. Rows containing only NA values (aside from the barcode)
are dropped.
}
\description{
This function loads and processes contig data from various single-cell
immune receptor sequencing formats. It reads data from a directory
(recursively) or from an already loaded list/data frame, transforms it to a
common structure, and returns a list of contigs ready for downstream analysis
with \code{\link[=combineTCR]{combineTCR()}} or \code{\link[=combineBCR]{combineBCR()}}.

Supported file formats and their expected file names:
\itemize{
\item \verb{10X}: "filtered_contig_annotations.csv"
\item \code{AIRR}: "airr_rearrangement.tsv"
\item \code{BD}: "Contigs_AIRR.tsv"
\item \code{Dandelion}: "all_contig_dandelion.tsv"
\item \code{Immcantation}: "_data.tsv" (or similar)
\item ``JSON`: ".json"
\item \code{ParseBio}: "barcode_report.tsv"
\item \code{MiXCR}: "clones.tsv"
\item \code{TRUST4}: "barcode_report.tsv"
\item \code{WAT3R}: "barcode_results.csv"
}
}
\examples{
TRUST4 <- read.csv("https://www.borch.dev/uploads/contigs/TRUST4_contigs.csv")
contig.list <- loadContigs(TRUST4, format = "TRUST4")

BD <- read.csv("https://www.borch.dev/uploads/contigs/BD_contigs.csv")
contig.list <- loadContigs(BD, format = "BD")

WAT3R <- read.csv("https://www.borch.dev/uploads/contigs/WAT3R_contigs.csv")
contig.list <- loadContigs(WAT3R, format = "WAT3R")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalOccupy.R
\name{clonalOccupy}
\alias{clonalOccupy}
\title{Plot cloneSize by Variable in Single-Cell Objects}
\usage{
clonalOccupy(
  sc.data,
  x.axis = "ident",
  label = TRUE,
  facet.by = NULL,
  order.by = NULL,
  proportion = FALSE,
  na.include = FALSE,
  export.table = NULL,
  palette = "inferno",
  exportTable = NULL,
  ...
)
}
\arguments{
\item{sc.data}{The single-cell object after \code{\link[=combineExpression]{combineExpression()}}}

\item{x.axis}{The variable in the meta data to graph along the x.axis.}

\item{label}{Include the number of clone in each category by x.axis variable}

\item{facet.by}{The column header used for faceting the graph}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{proportion}{Convert the stacked bars into relative proportion}

\item{na.include}{Visualize NA values or not}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
Stacked bar plot of counts of cells by clone frequency group
}
\description{
View the count of clones frequency group in Seurat or SCE object
meta data after \code{\link[=combineExpression]{combineExpression()}}. The visualization
will take the new meta data variable \code{cloneSize} and
plot the number of cells with each designation using a secondary
variable, like cluster. Credit to the idea goes to Drs. Carmona
and Andreatta and their work with \href{https://github.com/carmonalab/ProjecTILs}{ProjectTIL}.
}
\examples{
# Getting the combined contigs
combined <- combineTCR(contig_list,
                        samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))

# Getting a sample of a Seurat object
scRep_example <- get(data("scRep_example"))

# Using combineExpresion()
scRep_example <- combineExpression(combined, scRep_example)

# Using clonalOccupy
clonalOccupy(scRep_example, x.axis = "ident")
table <- clonalOccupy(scRep_example, x.axis = "ident", export.table = TRUE)

}
\concept{SC_Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quietVDJgenes.R
\name{quietVDJgenes}
\alias{quietVDJgenes}
\alias{quietTCRgenes}
\alias{quietTCRgenes.default}
\alias{quietTCRgenes.Seurat}
\alias{quietBCRgenes}
\alias{quietBCRgenes.default}
\alias{quietBCRgenes.Seurat}
\title{Remove TCR and BCR genes from variable gene results}
\usage{
quietVDJgenes(input.data, ...)

quietTCRgenes(input.data, ...)

\method{quietTCRgenes}{default}(input.data, ...)

\method{quietTCRgenes}{Seurat}(input.data, assay = NULL, ...)

quietBCRgenes(input.data, ...)

\method{quietBCRgenes}{default}(input.data, ...)

\method{quietBCRgenes}{Seurat}(input.data, assay = NULL, ...)
}
\arguments{
\item{input.data}{Single-cell object in Seurat format or vector of variable genes
to use in reduction}

\item{...}{Reserved for future arguments}

\item{assay}{The Seurat assay slot to use to remove immune receptor genes
from, NULL value will default to the default assay}
}
\value{
Seurat object or vector list with TCR genes removed.
}
\description{
Most single-cell workflows use highly-expressed and highly-variable
genes for the initial calculation of PCA and subsequent dimensional
reduction. This function will remove the TCR and/or BCR genes from the
variable features in a Seurat object or from a vector (potentially
generated by the Bioconductor scran workflow).
}
\examples{
example <- quietVDJgenes(scRep_example)
scRep <- quietTCRgenes(scRep_example)
ibex_example <- quietBCRgenes(scRep_example)

}
\author{
Nicky de Vrij, Nikolaj Pagh, Nick Borcherding, Qile Yang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alluvialClones.R
\name{alluvialClones}
\alias{alluvialClones}
\title{Alluvial Plotting for Single-Cell Object}
\usage{
alluvialClones(
  sc.data,
  clone.call = NULL,
  chain = "both",
  y.axes = NULL,
  color = NULL,
  facet = NULL,
  alpha = NULL,
  top.clones = NULL,
  min.freq = 0,
  highlight.clones = NULL,
  highlight.color = "red",
  stratum.width = 0.2,
  flow.alpha = 0.5,
  show.labels = TRUE,
  label.size = 2,
  order.strata = NULL,
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{sc.data}{The product of \code{\link[=combineExpression]{combineExpression()}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{y.axes}{The columns that will separate the proportional visualizations.}

\item{color}{The column header or clone(s) to be highlighted.}

\item{facet}{The column label to separate.}

\item{alpha}{The column header to have gradated opacity.}

\item{top.clones}{Show only the top N clones by frequency. If \code{NULL} (default),
show all clones.}

\item{min.freq}{Minimum frequency threshold for displaying flows. Clones
appearing fewer than this many times are filtered out.}

\item{highlight.clones}{Character vector of specific clone sequences to highlight.
These clones will be colored distinctly while others are shown in gray.}

\item{highlight.color}{Color to use for highlighted clones (default: "red").}

\item{stratum.width}{Width of the stratum bars (default: 0.2).}

\item{flow.alpha}{Transparency of the flows (default: 0.5). Highlighted clones
use full opacity.}

\item{show.labels}{If \code{TRUE} (default), display stratum labels.}

\item{label.size}{Text size for stratum labels (default: 2).}

\item{order.strata}{Named list specifying the order of levels within each stratum.
Names should match column names in y.axes.}

\item{export.table}{If \code{TRUE}, returns a data frame of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object visualizing categorical distribution of clones, or a
data.frame if \code{export.table = TRUE}.
}
\description{
View the proportional contribution of clones by Seurat or SCE object
meta data after \code{\link[=combineExpression]{combineExpression()}}. The visualization
is based on the ggalluvial package, which requires the aesthetics
to be part of the axes that are visualized. Therefore, alpha, facet,
and color should be part of the the axes you wish to view or will
add an additional stratum/column to the end of the graph.
}
\examples{
# Getting the combined contigs
combined <- combineTCR(contig_list,
                        samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))

# Getting a sample of a Seurat object
scRep_example <- get(data("scRep_example"))

# Using combineExpresion()
scRep_example <- combineExpression(combined, scRep_example)
scRep_example$Patient <- substring(scRep_example$orig.ident, 1,3)

# Using alluvialClones()
alluvialClones(scRep_example,
                   clone.call = "gene",
                   y.axes = c("Patient", "ident"),
                   color = "ident")

# Show only top 50 most frequent clones
alluvialClones(scRep_example,
                   clone.call = "aa",
                   y.axes = c("Patient", "ident"),
                   top.clones = 50)

# Highlight specific clones
alluvialClones(scRep_example,
                   clone.call = "aa",
                   y.axes = c("Patient", "ident"),
                   highlight.clones = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF"))

}
\concept{SC_Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/positionalEntropy.R
\name{positionalEntropy}
\alias{positionalEntropy}
\title{Examining the Diversity of Amino Acids by Position}
\usage{
positionalEntropy(
  input.data,
  chain = "TRB",
  group.by = NULL,
  order.by = NULL,
  aa.length = 20,
  method = "norm.entropy",
  export.table = NULL,
  palette = "inferno",
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}},
\code{\link[=combineBCR]{combineBCR()}}, or \code{\link[=combineExpression]{combineExpression()}}}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed as
by list element or active identity in the case of single-cell objects.}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{aa.length}{The maximum length of the CDR3 amino acid sequence.}

\item{method}{The method to calculate the entropy/diversity -
\code{"shannon"}, \code{"inv.simpson"}, \code{"gini.simpson"}, \code{"norm.entropy"},
\code{"pielou"}, \code{"hill0"}, \code{"hill1"}, \code{"hill2"}}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object displaying entropy or diversity by amino acid position.
If \code{exportTable = TRUE}, a matrix of the raw data is returned.
}
\description{
This function the diversity amino acids along the residues of the CDR3
amino acid sequence. Please see \code{\link[=clonalDiversity]{clonalDiversity()}} for more information
on the underlying methods for diversity/entropy calculations. Positions
without variance will have a value reported as 0 for the purposes of comparison.
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))

# Using positionalEntropy()
positionalEntropy(combined, 
                  chain = "TRB", 
                  aa.length = 20)
                  
}
\concept{Summarize_Repertoire}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotateInvariant.R
\name{annotateInvariant}
\alias{annotateInvariant}
\title{Annotate invariant T cells (MAIT or iNKT) in single-cell TCR data}
\usage{
annotateInvariant(
  input.data,
  type = c("MAIT", "iNKT"),
  species = c("mouse", "human")
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}} or \code{\link[=combineExpression]{combineExpression()}}.}

\item{type}{Character specifying the type of invariant cells to
annotate (\code{MAIT} or \code{iNKT}).}

\item{species}{Character specifying the species
('mouse' or 'human').}
}
\value{
A single-cell object or list with the corresponding annotation
scores (0 or 1) added.
}
\description{
The \code{\link[=annotateInvariant]{annotateInvariant()}} function identifies potential mucosal-associated
invariant T (\verb{MAIT``) cells or invariant natural killer T (}iNKT`) cells from
single-cell sequencing datasets based on their characteristic  TCR usage.
It extracts TCR chain information from the provided single-cell
data, checks it against known invariant T-cell receptor criteria for either
MAIT or iNKT cells, and returns a score indicating the presence (1) or
absence (0) of these invariant cell populations for each individual cell.
The function supports data from mouse and human samples, providing a
convenient method to annotate specialized T-cell subsets within single-cell
analyses.
}
\examples{
# Getting the combined contigs
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))

# Getting a sample of a Seurat object
scRep_example <- get(data("scRep_example"))

# Using combineExpresion()
scRep_example <- combineExpression(combined, scRep_example)

# Using annotateInvariant()
annotateInvariant(input.data = scRep_example, type = "MAIT", species = "human")
annotateInvariant(input.data = scRep_example, type = "iNKT", species = "human")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scRepertoire-package.R
\docType{package}
\name{scRepertoire-package}
\alias{scRepertoire}
\alias{scRepertoire-package}
\title{scRepertoire: A toolkit for single-cell immune receptor profiling}
\description{
scRepertoire is a toolkit for processing and analyzing single-cell T-cell receptor (TCR) and immunoglobulin (Ig). The scRepertoire framework supports use of 10x, AIRR, BD, MiXCR, TRUST4, and WAT3R single-cell formats. The functionality includes basic clonal analyses, repertoire summaries, distance-based clustering and interaction with the popular Seurat and SingleCellExperiment/Bioconductor R single-cell workflows.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://www.borch.dev/uploads/scRepertoire/}
  \item Report bugs at \url{https://github.com/BorchLab/scRepertoire/issues}
}

}
\author{
\strong{Maintainer}: Nick Borcherding \email{ncborch@gmail.com}

Authors:
\itemize{
  \item Qile Yang \email{qile.yang@berkeley.edu}
  \item Ksenia Safina \email{safina@broadinstitute.org}
}

Other contributors:
\itemize{
  \item Justin Reimertz \email{justin.reimertz@childrens.harvard.edu} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{expression2List}
\alias{expression2List}
\title{\strong{DEPRECATED} Take the meta data in seurat/SCE and place it into a list}
\usage{
expression2List(sc, ..., force = FALSE)
}
\arguments{
\item{sc}{output of \code{\link[=combineExpression]{combineExpression()}}.}

\item{...}{previously the \code{group} or \code{split.by} argument, indicating the
column header to group the new list by. This should strictly be one argument
and is an ellipsis for backwards compatibility. Everything after the first
argument is ignored.}

\item{force}{logical. If not \code{TRUE} (default), a deprecation error will
be thrown. Otherwise the function will run but not guaranteed to be stable.}
}
\value{
list derived from the meta data of single-cell object with
elements divided by the group parameter
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

Allows users to perform more fundamental measures of clonotype analysis
using the meta data from the seurat or SCE object. For Seurat objects the
active identity is automatically added as "cluster". Remaining grouping
parameters or SCE or Seurat objects must appear in the meta data.

This function is deprecated as of version 2 due to the confusion it caused
to many users. Users are encouraged to remain with the abstraction barrier
of combined single cell objects and the outputs of \code{\link[=combineTCR]{combineTCR()}} and
\code{\link[=combineBCR]{combineBCR()}} for all functions.

We discourage the use of this function, but if you have to use it, set the
\code{force} argument to \code{TRUE}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalOverlap.R
\name{clonalOverlap}
\alias{clonalOverlap}
\title{Examining the clonal overlap between groups or samples}
\usage{
clonalOverlap(
  input.data,
  clone.call = NULL,
  method = c("overlap", "morisita", "jaccard", "cosine", "raw"),
  chain = "both",
  group.by = NULL,
  order.by = NULL,
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}},
\code{\link[=combineBCR]{combineBCR()}}, or \code{\link[=combineExpression]{combineExpression()}}}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{method}{The method to calculate the \code{overlap}, \code{morisita},
\code{jaccard}, \code{cosine} indices or \code{raw} for the base numbers}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed
by list element or active identity in the case of single-cell objects.}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object visualizing clonal overlap or a data.frame if
\code{exportTable = TRUE}.
}
\description{
This functions allows for the calculation and visualizations of
various overlap metrics for clones. The methods include overlap
coefficient (\code{overlap}), Morisita's overlap index
(\code{morisita}), Jaccard index (\code{jaccard}), cosine
similarity (\code{cosine}) or the exact number of clonal
overlap (\code{raw}).
}
\details{
The formulas for the indices are as follows:

\strong{Overlap Coefficient:}
\deqn{overlap = \frac{\sum \min(a, b)}{\min(\sum a, \sum b)}}

\strong{Raw Count Overlap:}
\deqn{raw = \sum \min(a, b)}

\strong{Morisita Index:}
\deqn{morisita = \frac{\sum a b}{(\sum a)(\sum b)}}

\strong{Jaccard Index:}
\deqn{jaccard = \frac{\sum \min(a, b)}{\sum a + \sum b - \sum \min(a, b)}}

\strong{Cosine Similarity:}
\deqn{cosine = \frac{\sum a b}{\sqrt{(\sum a^2)(\sum b^2)}}}

Where:
\itemize{
\item{\eqn{a} and \eqn{b} are the abundances of species \eqn{i} in groups A and B, respectively.}
}
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))

# Using clonalOverlap()
clonalOverlap(combined,
              clone.call = "aa",
              method = "jaccard")

}
\concept{Visualizing_Clones}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalSizeDistribution.R
\name{clonalSizeDistribution}
\alias{clonalSizeDistribution}
\title{Plot powerTCR Clustering Based on Clonal Size}
\usage{
clonalSizeDistribution(
  input.data,
  clone.call = NULL,
  chain = "both",
  method = "ward.D2",
  threshold = 1,
  group.by = NULL,
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}},
\code{\link[=combineBCR]{combineBCR()}}, or \code{\link[=combineExpression]{combineExpression()}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{method}{The clustering parameter for the dendrogram.}

\item{threshold}{Numerical vector containing the thresholds
the grid search was performed over.}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed as
by list element or active identity in the case of single-cell objects.}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object visualizing dendrogram of clonal size distribution
or a data.frame if \code{exportTable = TRUE}.
}
\description{
This function produces a hierarchical clustering of clones by sample
using discrete gamma-GPD spliced threshold model. If using this
model please read and cite powerTCR (more info available at
\href{https://pubmed.ncbi.nlm.nih.gov/30485278/}{PMID: 30485278}).
}
\details{
The probability density function (pdf) for the \strong{Generalized Pareto Distribution (GPD)} is given by:
\deqn{f(x|\mu, \sigma, \xi) = \frac{1}{\sigma} \left( 1 + \xi \left( \frac{x - \mu}{\sigma} \right) \right)^{-\left( \frac{1}{\xi} + 1 \right)}}

Where:
\itemize{
\item{\eqn{\mu} is a location parameter}
\item{\eqn{\sigma > 0} is a scale parameter}
\item{\eqn{\xi} is a shape parameter}
\item{\eqn{x \ge \mu} if \eqn{\xi \ge 0} and \eqn{\mu \le x \le \mu - \sigma/\xi} if \eqn{\xi < 0}}
}

The probability density function (pdf) for the \strong{Gamma Distribution} is given by:
\deqn{f(x|\alpha, \beta) = \frac{x^{\alpha-1} e^{-x/\beta}}{\beta^\alpha \Gamma(\alpha)}}

Where:
\itemize{
\item{\eqn{\alpha > 0} is the shape parameter}
\item{\eqn{\beta > 0} is the scale parameter}
\item{\eqn{x \ge 0}}
\item{\eqn{\Gamma(\alpha)} is the gamma function of \eqn{\alpha}}
}
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B","P19L", "P20B", "P20L"))

# Using clonalSizeDistribution()
clonalSizeDistribution(combined,
                       clone.call = "strict",
                       method="ward.D2")

}
\author{
Hillary Koch
}
\concept{Visualizing_Clones}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalProportion.R
\name{clonalProportion}
\alias{clonalProportion}
\title{Plot the Clonal Space Occupied by Specific Clones}
\usage{
clonalProportion(
  input.data,
  clonal.split = NULL,
  clone.call = NULL,
  chain = "both",
  group.by = NULL,
  order.by = NULL,
  export.table = NULL,
  palette = "inferno",
  clonalSplit = NULL,
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}},
\code{\link[=combineBCR]{combineBCR()}}, or \code{\link[=combineExpression]{combineExpression()}}.}

\item{clonal.split}{The cut points for the specific clones, default = c(10,
100, 1000, 10000, 30000, 1e+05)}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed
by list element or active identity in the case of single-cell objects.}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}}

\item{clonalSplit}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clonal.split} instead.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object dividing space occupied by ranks of clones or a
data.frame if \code{exportTable = TRUE}.
}
\description{
This function calculates the relative clonal space occupied by the
clones. The grouping of these clones is based on the parameter
\code{clonalSplit}, at default, \code{clonalSplit} will group the clones
into bins of 1:10, 11:100, 101:1001, etc. To adjust the clones
selected, change the numbers in the variable split. If a matrix output
for the data is preferred, set \code{exportTable} = TRUE.
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))
# Using clonalProportion()
clonalProportion(combined, clone.call = "gene")

}
\concept{Visualizing_Clones}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combineContigs.R
\name{combineBCR}
\alias{combineBCR}
\title{Combine B Cell Receptor Contig Data}
\usage{
combineBCR(
  input.data,
  samples = NULL,
  ID = NULL,
  chain = "IGH",
  sequence = "nt",
  dist.type = NULL,
  dist.mat = NULL,
  normalize = "length",
  gap.open = NULL,
  gap.extend = NULL,
  call.related.clones = TRUE,
  group.by = NULL,
  threshold = 0.85,
  cluster.method = "components",
  use.V = TRUE,
  use.J = TRUE,
  remove.na = NULL,
  remove.multi = NULL,
  filter.multi = NULL,
  filter.nonproductive = NULL,
  removeNA = NULL,
  removeMulti = NULL,
  filterMulti = NULL,
  filterNonproductive = NULL,
  dist_type = NULL,
  dist_mat = NULL,
  gap_open = NULL,
  gap_extend = NULL
)
}
\arguments{
\item{input.data}{List of filtered contig annotations or outputs from
\code{\link[=loadContigs]{loadContigs()}}.}

\item{samples}{A character vector of sample labels. Must be the same length
as the input list.}

\item{ID}{An optional character vector for additional sample identifiers.}

\item{chain}{The chain to use for clustering when \code{call.related.clones = TRUE}.
Passed to \code{clonalCluster()}. Default is \code{"IGH"}.}

\item{sequence}{The sequence type (\code{"nt"} or \code{"aa"}) to use for clustering.
Passed to \code{clonalCluster()}. Default is \code{"nt"}.}

\item{dist.type}{The distance metric to use. Options: \code{"levenshtein"} (default),
\code{"hamming"}, \code{"damerau"}, \code{"nw"} (Needleman-Wunsch), or \code{"sw"} (Smith-Waterman).}

\item{dist.mat}{The substitution matrix to use for alignment-based metrics
(\code{"nw"} or \code{"sw"}). Options include \code{"BLOSUM62"}, \code{"PAM30"}, etc.}

\item{normalize}{Method for normalizing distances. Options: \code{"none"} (default),
\code{"maxlen"}, or \code{"length"}.}

\item{gap.open}{Penalty for opening a gap in alignment metrics (default: -10).}

\item{gap.extend}{Penalty for extending a gap in alignment metrics (default: -1).}

\item{call.related.clones}{Logical. If \code{TRUE}, uses \code{clonalCluster()} to
identify related clones based on sequence similarity. If \code{FALSE}, defines
clones by the exact V-gene and CDR3 amino acid sequence.}

\item{group.by}{The column header used for to group clones.
If (`NULL``), clusters will be calculated across samples.}

\item{threshold}{The similarity threshold passed to \code{clonalCluster()} if
\code{call.related.clones = TRUE}. See \code{?clonalCluster} for details.}

\item{cluster.method}{The clustering algorithm to use. Defaults to \code{"components"},
which finds connected subgraphs.}

\item{use.V}{Logical. If \code{TRUE}, sequences must share the same V gene to be
clustered together.}

\item{use.J}{Logical. If \code{TRUE}, sequences must share the same J gene to be
clustered together.}

\item{remove.na}{This will remove any chain without values.}

\item{remove.multi}{Logical. If \code{TRUE}, removes cells that have more than
one distinct heavy or light chain after processing.}

\item{filter.multi}{Logical. If \code{TRUE}, filters multi-chain cells to retain
only the most abundant IGH and IGL/IGK chains.}

\item{filter.nonproductive}{Logical. If \code{TRUE}, removes non-productive contigs
from the analysis.}

\item{removeNA}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{remove.na} instead.}

\item{removeMulti}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{remove.multi} instead.}

\item{filterMulti}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{filter.multi} instead.}

\item{filterNonproductive}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{filter.nonproductive} instead.}

\item{dist_type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{dist.type} instead.}

\item{dist_mat}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{dist.mat} instead.}

\item{gap_open}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{gap.open} instead.}

\item{gap_extend}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{gap.extend} instead.}
}
\value{
A list of data frames, where each data frame represents a sample.
Each row corresponds to a unique cell barcode, with columns detailing the
BCR chains and the assigned clone ID.
}
\description{
This function consolidates a list of BCR sequencing results to the level
of the individual cell barcodes. Using the samples and ID parameters,
the function will add the strings as prefixes to prevent issues with
repeated barcodes. The resulting new barcodes will need to match the
Seurat or SCE object in order to use, \code{\link[=combineExpression]{combineExpression()}}. Unlike
\code{\link[=combineTCR]{combineTCR()}}, combineBCR produces a column \code{CTstrict} based on the
edit distance clustering from \code{\link[=clonalCluster]{clonalCluster()}}. The \code{CTstrict} column
is formatted as \code{Heavy_Light} (underscore-separated) for downstream
compatibility. Connected clones are labeled with \code{cluster.X}, while
unconnected clones (singlets) are labeled with the V gene and CDR3
sequence (e.g., \verb{IGHV3-64.CAKSYS..._IGKV3-15.CQQYSN...}).
}
\examples{
# Data derived from the 10x Genomics intratumoral NSCLC B cells
BCR <- read.csv("https://www.borch.dev/uploads/contigs/b_contigs.csv")
combined <- combineBCR(BCR,
                       samples = "Patient1",
                       threshold = 0.85)

}
\concept{Loading_and_Processing_Contigs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/positionalProperty.R
\name{positionalProperty}
\alias{positionalProperty}
\title{Plot Positional Physicochemical Property Analysis}
\usage{
positionalProperty(
  input.data,
  chain = "TRB",
  group.by = NULL,
  order.by = NULL,
  aa.length = 20,
  method = "atchleyFactors",
  export.table = NULL,
  palette = "inferno",
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}}, \code{\link[=combineBCR]{combineBCR()}}, or
\code{\link[=combineExpression]{combineExpression()}}}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed
by list element or active identity in the case of single-cell objects.}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{aa.length}{The maximum length of the CDR3 amino acid sequence.}

\item{method}{Character string (one of the supported names)
Defaults to \code{"atchleyFactors"}, but includes: \code{"crucianiProperties"},
\code{"FASGAI"}, \code{"kideraFactors"}, \code{"MSWHIM"}, \code{"ProtFP"}, \code{"stScales"},
\code{"tScales"}, \code{"VHSE"}, \code{"zScales"}}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any \link[grDevices:palettes]{hcl.pals}}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object displaying property by amino acid position.
If \code{exportTable = TRUE}, a matrix of the raw data is returned.
}
\description{
This function analyzes the physicochemical properties of amino acids at each
position along the CDR3 sequence. It calculates the mean property value and
the 95\% confidence interval for each position across one or more groups,
visualizing the results as a line plot with a confidence ribbon.
}
\details{
The function uses one of several established physicochemical property scales
to convert amino acid sequences into numerical vectors. More information for
the individual methods can be found at the following citations:

\strong{atchleyFactors:} \href{https://pubmed.ncbi.nlm.nih.gov/15851683/}{citation}

\strong{crucianiProperties:} \href{https://analyticalsciencejournals.onlinelibrary.wiley.com/doi/abs/10.1002/cem.856}{citation}

\strong{FASGAI:} \href{https://pubmed.ncbi.nlm.nih.gov/18318694/}{citation}

\strong{kideraFactors:} \href{https://link.springer.com/article/10.1007/BF01025492}{citation}

\strong{MSWHIM:} \href{https://pubs.acs.org/doi/10.1021/ci980211b}{citation}

\strong{ProtFP:} \href{https://pubmed.ncbi.nlm.nih.gov/24059694/}{citation}

\strong{stScales:} \href{https://pubmed.ncbi.nlm.nih.gov/19373543/}{citation}

\strong{tScales:} \href{https://www.sciencedirect.com/science/article/abs/pii/S0022286006006314}{citation}

\strong{VHSE:} \href{https://pubmed.ncbi.nlm.nih.gov/15895431/}{citation}

\strong{zScales:} \href{https://pubmed.ncbi.nlm.nih.gov/9651153/}{citation}
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))

# Using positionalProperty()
positionalProperty(combined, 
                   chain = "TRB",
                   method = "atchleyFactors", 
                   aa.length = 20)
                   
}
\author{
Florian Bach, Nick Borcherding
}
\concept{Summarize_Repertoire}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exportClones.R
\name{exportClones}
\alias{exportClones}
\title{Export Clonal Data in Various Formats}
\usage{
exportClones(
  input.data,
  format = "paired",
  group.by = NULL,
  write.file = TRUE,
  dir = NULL,
  file.name = "clones.csv"
)
}
\arguments{
\item{input.data}{The product of \code{combineTCR()}, \code{combineBCR()}, or
\code{combineExpression()}.}

\item{format}{The format for exporting clones.
Options are: \code{paired}, \code{airr}, \code{TCRMatch}, \code{tcrpheno}, \code{immunarch}.}

\item{group.by}{The variable in the metadata to use for grouping. If \code{NULL},
data will be grouped by the sample names.}

\item{write.file}{If \code{TRUE} (default), saves the output to a CSV file. If
\code{FALSE}, returns the data frame or list to the R environment.}

\item{dir}{The directory where the output file will be saved.
Defaults to the current working directory.}

\item{file.name}{The name of the file to be saved.}
}
\value{
A data frame or list in the specified format, either returned to the
R environment or saved as a CSV file.
}
\description{
Exports clonal information (gene sequences, amino acids, nucleotides) from
scRepertoire objects into a file or a data frame. The output format can be
tailored for compatibility with different analysis workflows.
}
\details{
The \code{format} parameter determines the structure of the output:
\itemize{
\item{\code{paired}}: Exports a data frame where each row represents a barcode,
with paired chain information (amino acid, nucleotide, genes) in separate
columns.
\item{\code{airr}}: Exports a data frame that adheres to the Adaptive Immune
Receptor Repertoire (AIRR) Community format, with each row representing
a single receptor chain.
\item{\code{TCRMatch}}: Exports a data frame specifically for the TCRMatch
algorithm, containing the TRB chain amino acid sequence and clonal
frequency.
\item{\code{tcrpheno}}: Exports a data frame compatible with the \code{tcrpheno}
pipeline, with TRA and TRB chains in separate columns.
\item{\code{immunarch}}: Exports a list containing a data frame and metadata
formatted for use with the \code{immunarch} package.
}
}
\examples{
\dontrun{
#Making combined contig data
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B", "P19L", "P20B", "P20L"))

# Export as a paired data frame and save to a file
exportClones(combined, format = "paired", file.name = "paired_clones.csv")

# Return an AIRR-formatted data frame to the environment
airr_df <- exportClones(combined, format = "airr", write.file = FALSE)
}
}
\author{
Jonathan Noonan, Nick Borcherding
}
\concept{Loading_and_Processing_Contigs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exportClones.R
\name{._bind_contig_list}
\alias{._bind_contig_list}
\title{Bind a List of Contig Data Frames and Add Grouping Variable}
\usage{
._bind_contig_list(contig_list)
}
\description{
Bind a List of Contig Data Frames and Add Grouping Variable
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getCircilize.R
\name{getCirclize}
\alias{getCirclize}
\title{Generate Data Frame to Plot Chord Diagram}
\usage{
getCirclize(
  sc.data,
  clone.call = NULL,
  group.by = NULL,
  method = c("unique", "abundance", "jaccard", "overlap"),
  proportion = FALSE,
  symmetric = TRUE,
  include.self = TRUE,
  include.metadata = FALSE,
  min.shared = 0,
  top.links = NULL,
  filter.sectors = NULL,
  palette = "inferno",
  cloneCall = NULL
)
}
\arguments{
\item{sc.data}{The single-cell object after \code{\link[=combineExpression]{combineExpression()}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{group.by}{A column header (or vector of column headers for hierarchical
grouping) in the metadata to group the analysis by (e.g., "sample", "treatment").
If \code{NULL}, data will be analyzed by active identity. When multiple columns are
provided, they are combined with "_" separator for multi-level annotations.}

\item{method}{The method for calculating link values: \code{"unique"} (default) counts
unique shared clones, \code{"abundance"} sums clone frequencies, \code{"jaccard"} calculates
Jaccard similarity, \code{"overlap"} calculates overlap coefficient.}

\item{proportion}{Calculate the relationship by unique clones (\code{FALSE}, default)
or normalized by proportion (\code{TRUE}).}

\item{symmetric}{If \code{TRUE} (default), returns symmetric relationships. If \code{FALSE},
returns directional flow showing proportion of source's clones found in destination.}

\item{include.self}{Include counting the clones within a single group.by comparison.}

\item{include.metadata}{If \code{TRUE}, returns a list with links data frame and
sector-level metadata including cell counts, clone counts, and expansion metrics.}

\item{min.shared}{Minimum number of shared clones to include a link (default 0).}

\item{top.links}{Keep only the top N links by value. If \code{NULL} (default), keep all.}

\item{filter.sectors}{Character vector of sectors to include. If \code{NULL}, include all.}

\item{palette}{Colors to use for sector color suggestions - input any
\link[grDevices:palettes]{hcl.pals}.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}
}
\value{
A data frame of shared clones between groups formatted for
\link[circlize:chordDiagram]{chordDiagram}. If \code{include.metadata = TRUE}, returns
a list with \code{links} (the edge data frame), \code{sectors} (sector-level statistics),
and \code{colors} (suggested colors for each sector).
}
\description{
This function will take the meta data from the product of
\code{\link[=combineExpression]{combineExpression()}} and generate a relational data frame to
be used for a chord diagram. Each chord will represent the number of
clones unique and shared across the multiple \code{group.by} variable.
If using the downstream circlize R package, please read and cite the
following \href{https://pubmed.ncbi.nlm.nih.gov/24930139/}{manuscript}.
If looking for more advanced ways for circular visualizations, there
is a great \href{https://jokergoo.github.io/circlize_book/book/}{cookbook}
for the circlize package.
}
\examples{
# Getting the combined contigs
combined <- combineTCR(contig_list,
                        samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))

# Getting a sample of a Seurat object
scRep_example <- get(data("scRep_example"))
scRep_example <- combineExpression(combined,
                                   scRep_example)

# Getting data frame output for Circlize
circles <- getCirclize(scRep_example,
                       group.by = "seurat_clusters")

# Multi-level grouping for hierarchical chord diagrams
scRep_example$Patient <- substring(scRep_example$orig.ident, 1, 3)
circles <- getCirclize(scRep_example,
                       group.by = c("Patient", "seurat_clusters"))

# Get rich output with sector metadata
result <- getCirclize(scRep_example,
                      group.by = "seurat_clusters",
                      include.metadata = TRUE)

}
\author{
Dillon Corvino, Nick Borcherding
}
\concept{SC_Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalOverlay.R
\name{clonalOverlay}
\alias{clonalOverlay}
\title{Visualize Distribution of Clonal Frequency}
\usage{
clonalOverlay(
  sc.data,
  reduction = NULL,
  cut.category = "clonalFrequency",
  cutpoint = 30,
  bins = 25,
  pt.size = 0.5,
  pt.alpha = 1,
  facet.by = NULL,
  ...
)
}
\arguments{
\item{sc.data}{The single-cell object after \code{\link[=combineExpression]{combineExpression()}}.}

\item{reduction}{The dimensional reduction to visualize.}

\item{cut.category}{Meta data variable of the single-cell object to use for
filtering.}

\item{cutpoint}{The overlay cut point to include, this corresponds to the
cut.category variable in the meta data of the single-cell object.}

\item{bins}{The number of contours to the overlay}

\item{pt.size}{The point size for plotting (default is 0.5)}

\item{pt.alpha}{The alpha value for plotting (default is 1)}

\item{facet.by}{meta data variable to facet the comparison}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object visualizing distributions of clones along a
dimensional reduction within the single-cell object
}
\description{
This function allows the user to visualize the clonal expansion by overlaying the
cells with specific clonal frequency onto the dimensional reduction plots in Seurat.
Credit to the idea goes to Drs Andreatta and Carmona and their work with
\href{https://github.com/carmonalab/ProjecTILs}{ProjectTIL}.
}
\examples{
# Getting the combined contigs
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))
# Getting a sample of a Seurat object
scRep_example <- get(data("scRep_example"))

# Using combineExpresion()
scRep_example <- combineExpression(combined, 
                                   scRep_example)

# Using clonalOverlay()
clonalOverlay(scRep_example, 
              reduction = "umap", 
              cutpoint = 3, 
              bins = 5) 

}
\author{
Francesco Mazziotta, Nick Borcherding
}
\concept{SC_Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{contig_list}
\alias{contig_list}
\title{A List of Eight Single-cell TCR Sequencing Runs.}
\description{
A list of 8 \code{filtered_contig_annotations.csv} files
outputted from 10X Cell Ranger. More information on the
data can be found in the following
\href{https://pubmed.ncbi.nlm.nih.gov/33622974/}{manuscript}.
}
\concept{Data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subsetClone.R
\name{subsetClones}
\alias{subsetClones}
\title{Subset The Product of combineTCR() or combineBCR()}
\usage{
subsetClones(input.data, name, variables = NULL)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}} or
\code{\link[=combineBCR]{combineBCR()}}.}

\item{name}{The column header/name to use for subsetting.}

\item{variables}{The values to subset by, must be in the in the variable.}
}
\value{
list of contigs that have been filtered for the name parameter
}
\description{
This function allows for the subsetting of the product of
\code{\link[=combineTCR]{combineTCR()}} or \code{\link[=combineBCR]{combineBCR()}}
by the name of the individual list element.
}
\examples{
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))
subset <- subsetClones(combined, name = "sample", variables = c("P17B"))

}
\concept{Loading_and_Processing_Contigs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/startracDiversity.R
\name{StartracDiversity}
\alias{StartracDiversity}
\title{Calculate Startrac-based Diversity Indices}
\usage{
StartracDiversity(
  sc.data,
  clone.call = NULL,
  chain = "both",
  index = c("expa", "migr", "tran"),
  type = NULL,
  group.by = NULL,
  pairwise = NULL,
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{sc.data}{The single-cell object after \code{\link[=combineExpression]{combineExpression()}}.
For SCE objects, the cluster variable must be in the meta data under
"cluster".}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{index}{A character vector specifying which indices to calculate.
Options: "expa", "migr", "tran". Default is all three.}

\item{type}{The metadata variable that specifies tissue type for migration
analysis.}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed as
by list element or active identity in the case of single-cell objects.}

\item{pairwise}{The metadata column to be used for pairwise comparisons.
Set to the \code{type} variable for pairwise migration or "cluster" for
pairwise transition.}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object visualizing STARTRAC diversity metrics or data.frame if
\code{exportTable = TRUE}.
}
\description{
This function utilizes the STARTRAC approach to calculate T cell
diversity metrics based on the work of Zhang et al. (2018, Nature)
\href{https://pubmed.ncbi.nlm.nih.gov/30479382/}{PMID: 30479382}. It can compute
three distinct indices: clonal expansion (\code{expa}), cross-tissue migration
(\code{migr}), and state transition (\code{tran}).
}
\details{
The function requires a \code{type} variable in the metadata, which specifies the
tissue origin or any other categorical variable for migration analysis.

\strong{Indices:}
\itemize{
\item{\strong{expa (Clonal Expansion):}} Measures the extent of clonal
proliferation within a T cell cluster. It is calculated as
\verb{1 - normalized Shannon entropy}. A higher value indicates greater
expansion of a few clones.
\item{\strong{migr (Cross-Tissue Migration):}} Quantifies the movement of
clonal T cells across different tissues (as defined by the \code{type}
parameter). It is based on the entropy of a clonotype's distribution
across tissues.
\item{\strong{tran (State Transition):}} Measures the developmental
transition of clonal T cells between different functional clusters.
It is based on the entropy  of a clonotype's distribution across
clusters.
}

\strong{Pairwise Analysis:}
The \code{pairwise} parameter enables the calculation of migration or transition
between specific pairs of tissues or clusters, respectively.
\itemize{
\item{For migration (\code{index = "migr"}), set \code{pairwise} to the \code{type} column
(e.g., \code{pairwise = "Type"}).}
\item{For transition (\code{index = "tran"}), set \code{pairwise} to \code{"cluster"}.}
}
}
\examples{
# Getting the combined contigs
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))

# Getting a sample of a Seurat object
scRep_example  <- get(data("scRep_example"))
scRep_example  <- combineExpression(combined, scRep_example)
scRep_example$Patient <- substring(scRep_example$orig.ident,1,3)
scRep_example$Type <- substring(scRep_example$orig.ident,4,4) 

# Calculate a single index (expansion)
StartracDiversity(scRep_example, 
                  type = "Type", 
                  group.by = "Patient",
                  index = "expa")
                  
# Calculate pairwise transition 
StartracDiversity(scRep_example, 
                  type = "Type", 
                  group.by = "Patient",
                  index = "tran",
                  pairwise = "cluster") 

}
\author{
Liangtao Zheng
}
\concept{SC_Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalLength.R
\name{clonalLength}
\alias{clonalLength}
\title{Plot the Distribution of Sequence Lengths}
\usage{
clonalLength(
  input.data,
  clone.call = NULL,
  chain = "both",
  group.by = NULL,
  order.by = NULL,
  scale = FALSE,
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}},
\code{\link[=combineBCR]{combineBCR()}}, or \code{\link[=combineExpression]{combineExpression()}}}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{nt} (CDR3 nucleotide sequence) or \code{aa} (CDR3 amino acid sequence)}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed
by list element or active identity in the case of single-cell objects.}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{scale}{Converts the graphs into density plots in order to show
relative distributions.}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object visualizing the distributions by length, or a data.frame if
\code{export.table = TRUE}.
}
\description{
This function displays either the nucleotide \code{nt} or amino
acid \code{aa} sequence length. The sequence length visualized
can be selected using the chains parameter, either the combined clone
(both chains) or across all single chains. Visualization can either
be a histogram or if \code{scale} = TRUE, the output will
be a density plot. Multiple sequencing runs can be group together
using the group.by parameter.
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list,
                        samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))
clonalLength(combined, clone.call = "aa", chain = "both")

}
\concept{Visualizing_Clones}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalBias.R
\name{clonalBias}
\alias{clonalBias}
\title{Calculate Clonal Bias Towards a Cluster or Compartment}
\usage{
clonalBias(
  sc.data,
  clone.call = NULL,
  split.by = NULL,
  group.by = NULL,
  n.boots = 20,
  min.expand = 10,
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{sc.data}{The single-cell object after \code{\link[=combineExpression]{combineExpression()}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{split.by}{The variable to use for calculating the baseline frequencies.
For example, "Type" for lung vs peripheral blood comparison}

\item{group.by}{A column header in the metadata that bias will be based on.}

\item{n.boots}{number of bootstraps to downsample.}

\item{min.expand}{clone frequency cut off for the purpose of comparison.}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
ggplot scatter plot with clone bias
}
\description{
The metric seeks to quantify how individual clones are skewed towards
a specific cellular compartment or cluster. A clone bias of \code{1} -
indicates that a clone is composed of cells from a single
compartment or cluster, while a clone bias of \code{0} - matches the
background subtype distribution. Please read and cite the following
\href{https://pubmed.ncbi.nlm.nih.gov/35829695/}{manuscript}
if using \code{\link[=clonalBias]{clonalBias()}}.
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list,
                        samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))

# Getting a sample of a Seurat object
scRep_example <- get(data("scRep_example"))

# Using combineExpresion()
scRep_example <- combineExpression(combined, scRep_example)
scRep_example$Patient <- substring(scRep_example$orig.ident,1,3)

# Using clonalBias()
clonalBias(scRep_example,
              clone.call = "aa",
              split.by = "Patient",
              group.by = "seurat_clusters",
              n.boots = 5,
              min.expand = 2)


}
\concept{SC_Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalAbundance.R
\name{clonalAbundance}
\alias{clonalAbundance}
\title{Plot the Relative Abundance of Clones}
\usage{
clonalAbundance(
  input.data,
  clone.call = NULL,
  chain = "both",
  scale = FALSE,
  group.by = NULL,
  order.by = NULL,
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}},
\code{\link[=combineBCR]{combineBCR()}}, or \code{\link[=combineExpression]{combineExpression()}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{scale}{Converts the graphs into density plots in order to show
relative distributions.}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed
by list element or active identity in the case of single-cell objects.}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object visualizing clonal abundance by group, or a
data.frame if \code{export.table = TRUE}.
}
\description{
Displays the number of clones at specific frequencies by sample
or group. Visualization can either be a line graph (
\code{scale} = FALSE) using calculated numbers or density
plot (\code{scale} = TRUE). Multiple sequencing runs can
be group together using the group parameter. If a matrix
output for the data is preferred, set
\code{export.table} = TRUE.
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list,
                        samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))

# Using clonalAbundance()
clonalAbundance(combined,
                clone.call = "gene",
                scale = FALSE)

}
\author{
Nick Borcherding, Justin Reimertz
}
\concept{Visualizing_Clones}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/percentAA.R
\name{percentAA}
\alias{percentAA}
\title{Plot Relative Amino Acid Composition by Position}
\usage{
percentAA(
  input.data,
  chain = "TRB",
  group.by = NULL,
  order.by = NULL,
  aa.length = 20,
  export.table = NULL,
  palette = "inferno",
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}}, \code{\link[=combineBCR]{combineBCR()}}, or
\code{\link[=combineExpression]{combineExpression()}}.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed
by list element or active identity in the case of single-cell objects.}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{aa.length}{The maximum length of the CDR3 amino acid sequence.}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any \link[grDevices:palettes]{hcl.pals}.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object visualizing amino acid by proportion or a data.frame if
\code{exportTable = TRUE}.
}
\description{
This function the proportion of amino acids along the residues
of the CDR3 amino acid sequence.
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))

# Using percentAA()
percentAA(combined, 
          chain = "TRB", 
          aa.length = 20)
}
\concept{Summarize_Repertoire}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalRarefaction.R
\name{clonalRarefaction}
\alias{clonalRarefaction}
\title{Calculate rarefaction based on the abundance of clones}
\usage{
clonalRarefaction(
  input.data,
  clone.call = NULL,
  chain = "both",
  group.by = NULL,
  plot.type = 1,
  hill.numbers = 0,
  n.boots = 20,
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}}, \code{\link[=combineBCR]{combineBCR()}}, or
\code{\link[=combineExpression]{combineExpression()}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed
by list element or active identity in the case of single-cell objects.}

\item{plot.type}{sample-size-based rarefaction/extrapolation curve
(\code{type = 1}); sample completeness curve (\code{type = 2});
coverage-based rarefaction/extrapolation curve (\code{type = 3}).}

\item{hill.numbers}{The Hill numbers to be plotted out
(0 - species richness, 1 - Shannon, 2 - Simpson)}

\item{n.boots}{The number of bootstrap replicates used to derive confidence
intervals for the diversity estimates. More replicates can provide a more
reliable measure of statistical variability.}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\description{
This functions uses the Hill numbers of order q: species richness (\code{q = 0}),
Shannon diversity (\code{q = 1}), the exponential of Shannon entropy and Simpson
diversity (\code{q = 2}, the inverse of Simpson concentration) to compute diversity
estimates for rarefaction and extrapolation. The function relies on the
\code{\link[iNEXT:iNEXT]{iNEXT::iNEXT()}} R package. Please read and cite the
\href{https://besjournals.onlinelibrary.wiley.com/doi/10.1111/2041-210X.12613}{manuscript}
if using this function. The input into the iNEXT calculation is abundance,
incidence-based calculations are not supported.
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))
                                    
# Using clonalRarefaction()
clonalRarefaction(combined[c(1,2)], clone.call = "gene", n.boots = 3)


}
\concept{Visualizing_Clones}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getContigDoublets.R
\name{getContigDoublets}
\alias{getContigDoublets}
\title{Get Contig Doublets}
\usage{
getContigDoublets(tcrOutput, bcrOutput)
}
\arguments{
\item{tcrOutput}{Output of \code{\link[=combineTCR]{combineTCR()}}. A list of data.frames containing
TCR contig information, each dataframe must have a \code{barcode} column.}

\item{bcrOutput}{Output of \code{\link[=combineBCR]{combineBCR()}}. A list of data.frames containing
BCR contig information, each dataframe must have a \code{barcode} column.}
}
\value{
A dataframe of barcodes that exist in both the TCR and BCR data, with
columns from both sets of data. There will be an additional column
\code{contigType} of type factor with levels 'TCR' and 'BCR' indicating the
origin of the contig - this will be the new first column.

If there are no doublets, the returned
data.frame will have the same colnames but no rows.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

This function identifies potential doublets by finding common barcodes
between TCR and BCR outputs. It extracts unique barcodes from each list
of dataframes, finds the intersection of the barcodes, and joins the
resulting data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/percentKmer.R
\name{percentKmer}
\alias{percentKmer}
\title{Analyze K-mer Motif Composition}
\usage{
percentKmer(
  input.data,
  chain = "TRB",
  clone.call = NULL,
  group.by = NULL,
  order.by = NULL,
  motif.length = 3,
  min.depth = 3,
  top.motifs = 30,
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}},
\code{\link[=combineBCR]{combineBCR()}}, or \code{\link[=combineExpression]{combineExpression()}}}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{nt} (CDR3 nucleotide sequence) or \code{aa} (CDR3 amino acid sequence).}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed as
by list element or active identity in the case of single-cell objects.}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{motif.length}{The length of the kmer to analyze}

\item{min.depth}{Minimum count a motif must reach to be retained in the
output (\verb{>= 1}). \strong{Default:} \code{3}.}

\item{top.motifs}{Return the n most variable motifs as a function of
median absolute deviation}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object displaying a heatmap of motif percentages.
If \code{exportTable = TRUE}, a matrix of the raw data is returned.
}
\description{
This function calculates and visualizes the frequency of k-mer
motifs for either nucleotide (nt) or amino acid (aa) sequences. It produces
a heatmap showing the relative composition of the most variable motifs across
samples or groups.
}
\details{
The function first calculates k-mer frequencies for each sample/group.
By default, it then identifies the 30 most variable motifs based on the Median
Absolute Deviation (MAD) across all samples and displays their frequencies
in a heatmap.
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))

# Using percentKmer()
percentKmer(combined, 
            chain = "TRB", 
            motif.length = 3)

}
\concept{Summarize_Repertoire}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exportClones.R
\name{._split_and_pad}
\alias{._split_and_pad}
\title{Split String and Pad to a Fixed-Width Matrix}
\usage{
._split_and_pad(x, split, n_cols)
}
\description{
Split String and Pad to a Fixed-Width Matrix
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combineContigs.R
\name{combineTCR}
\alias{combineTCR}
\title{Combine T Cell Receptor Contig Data}
\usage{
combineTCR(
  input.data,
  samples = NULL,
  ID = NULL,
  remove.na = NULL,
  remove.multi = NULL,
  filter.multi = NULL,
  filter.nonproductive = NULL,
  removeNA = NULL,
  removeMulti = NULL,
  filterMulti = NULL,
  filterNonproductive = NULL
)
}
\arguments{
\item{input.data}{List of filtered contig annotations or
outputs from \code{\link[=loadContigs]{loadContigs()}}.}

\item{samples}{The labels of samples (recommended).}

\item{ID}{The additional sample labeling (optional).}

\item{remove.na}{This will remove any chain without values.}

\item{remove.multi}{This will remove barcodes with greater than 2 chains.}

\item{filter.multi}{This option will allow for the selection of the 2
corresponding chains with the highest expression for a single barcode.}

\item{filter.nonproductive}{This option will allow for the removal of
nonproductive chains if the variable exists in the contig data. Default
is set to TRUE to remove nonproductive contigs.}

\item{removeNA}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{remove.na} instead.}

\item{removeMulti}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{remove.multi} instead.}

\item{filterMulti}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{filter.multi} instead.}

\item{filterNonproductive}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{filter.nonproductive} instead.}
}
\value{
List of clones for individual cell barcodes
}
\description{
This function consolidates a list of TCR sequencing results to
the level of  the individual cell barcodes. Using the \code{samples} and
\code{ID} parameters, the function will add the strings as prefixes to
prevent issues with repeated  barcodes. The resulting new barcodes will
need to match the Seurat or SCE object in order to use,
\code{\link[=combineExpression]{combineExpression()}}. Several levels of filtering exist -
\code{remove.na}, \code{remove.multi}, or \code{filter.multi} are parameters
that control how the function deals with barcodes with multiple chains
recovered.
}
\examples{
combined <- combineTCR(contig_list,
                        samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))

}
\concept{Loading_and_Processing_Contigs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalBin.R
\name{clonalBin}
\alias{clonalBin}
\title{Bin Clones by Frequency or Proportion}
\usage{
clonalBin(
  input.data,
  clone.call = NULL,
  chain = "both",
  group.by = NULL,
  proportion = TRUE,
  clone.size = NULL,
  cloneCall = NULL,
  cloneSize = NULL
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}}, \code{\link[=combineBCR]{combineBCR()}}, or
\code{\link[=combineExpression]{combineExpression()}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL}
(for both light chains), \code{both}.}

\item{group.by}{A column header in the metadata to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed
by list element.}

\item{proportion}{Whether to use proportion (\code{TRUE}) or total
frequency (\code{FALSE}) of the clone for binning.}

\item{clone.size}{The bins for the grouping based on proportion or frequency.
If proportion is \code{FALSE} and the clone.size values are not set high enough
based on frequency, the upper limit of clone.size will be automatically
updated.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{cloneSize}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.size} instead.}
}
\value{
A list of data frames with clonal frequency, clonal proportion,
and cloneSize columns added.
}
\description{
This function adds a clonal grouping variable (\code{cloneSize}) to the output
of \code{\link[=combineTCR]{combineTCR()}}, \code{\link[=combineBCR]{combineBCR()}}, or \code{\link[=combineExpression]{combineExpression()}}. It calculates
the clonal frequency and proportion, then bins clones into categories based
on customizable thresholds. This is useful for categorizing clones prior to
downstream analysis or visualization.
}
\examples{
# Getting the combined contigs
combined <- combineTCR(contig_list,
                        samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))

# Adding clonal bins with default settings (proportion-based)
combined <- clonalBin(combined)

# Adding clonal bins based on frequency
combined <- clonalBin(combined,
                      proportion = FALSE,
                      clone.size = c(Rare = 1, Small = 5, Medium = 20,
                                     Large = 100, Hyperexpanded = 500))

# Using a custom grouping variable
combined <- addVariable(combined,
                        variable.name = "Type",
                        variables = rep(c("B", "L"), 4))
combined <- clonalBin(combined, group.by = "Type")

}
\author{
Nick Borcherding
}
\concept{Clonal_Analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalDiversity.R
\name{clonalDiversity}
\alias{clonalDiversity}
\title{Calculate Clonal Diversity}
\usage{
clonalDiversity(
  input.data,
  clone.call = NULL,
  metric = "shannon",
  chain = "both",
  group.by = NULL,
  order.by = NULL,
  x.axis = NULL,
  export.table = NULL,
  palette = "inferno",
  n.boots = 100,
  return.boots = FALSE,
  skip.boots = FALSE,
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}},
\code{\link[=combineBCR]{combineBCR()}}, or \code{\link[=combineExpression]{combineExpression()}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{metric}{The diversity metric to calculate. Must be a single string from
the list of available metrics (see Details).}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed
by list element or active identity in the case of single-cell objects.}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{x.axis}{An additional metadata variable to group samples along the x-axis.}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}.}

\item{n.boots}{The number of bootstrap iterations to perform (default is 100).}

\item{return.boots}{If \code{TRUE}, returns all bootstrap values instead of the
mean. Automatically enables \code{export.table}.}

\item{skip.boots}{If \code{TRUE}, disables downsampling and bootstrapping. The
metric will be calculated on the full dataset for each group. Defaults to \code{FALSE}.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object visualizing the diversity metric, or a data.frame if
\code{export.table = TRUE}.
}
\description{
This function calculates a specified diversity metric for samples or groups
within a dataset. To control for variations in library size, the function
can perform bootstrapping with downsampling. It resamples each group to the
size of the smallest group and calculates the diversity metric across
multiple iterations, returning the mean value.
}
\details{
The function operates by first splitting the dataset by the specified \code{group.by}
variable.

\strong{Downsampling and Bootstrapping:}
To make a fair comparison between groups of different sizes, diversity metrics
often require normalization. This function implements this by downsampling.
\enumerate{
\item It determines the number of clones in the smallest group.
\item For each group, it performs \code{n.boots} iterations (default = 100).
\item In each iteration, it randomly samples the clones (with replacement) down to
the size of the smallest group.
\item It calculates the selected diversity metric on this downsampled set.
\item The final reported diversity value is the mean of the results from all
bootstrap iterations.
}

This process can be disabled by setting \code{skip.boots = TRUE}.

Available Diversity Metrics (metric):
The function uses a registry of metrics imported from the immApex package.
You can select one of the following:
\itemize{
\item{\code{"shannon"}: Shannon's Entropy. See \code{\link[immApex]{shannon_entropy}}.}
\item{\code{"inv.simpson"}: Inverse Simpson Index. See \code{\link[immApex]{inv_simpson}}.}
\item{\code{"gini.simpson"}: Gini-Simpson Index. See \code{\link[immApex]{gini_simpson}}.}
\item{\code{"norm.entropy"}: Normalized Shannon Entropy. See \code{\link[immApex]{norm_entropy}}.}
\item{\code{"pielou"}: Pielou's Evenness (same as norm.entropy). See \code{\link[immApex]{pielou_evenness}}.}
\item{\code{"ace"}: Abundance-based Coverage Estimator. See \code{\link[immApex]{ace_richness}}.}
\item{\code{"chao1"}: Chao1 Richness Estimator. See \code{\link[immApex]{chao1_richness}}.}
\item{\code{"gini"}: Gini Coefficient for inequality. See \code{\link[immApex]{gini_coef}}.}
\item{\code{"d50"}: The number of top clones making up 50\% of the library. See \code{\link[immApex]{d50_dom}}.}
\item{\code{"hill0"}, \code{"hill1"}, \code{"hill2"}: Hill numbers of order 0, 1, and 2. See \code{\link[immApex]{hill_q}}.}
}
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B","P19L", "P20B", "P20L"))

# Calculate Shannon diversity, grouped by sample
clonalDiversity(combined,
                clone.call = "gene",
                metric = "shannon")

# Calculate Inverse Simpson without bootstrapping
clonalDiversity(combined,
                clone.call = "aa",
                metric = "inv.simpson",
                skip.boots = TRUE)

}
\author{
Andrew Malone, Nick Borcherding, Nathan Vanderkraan
}
\concept{Visualizing_Clones}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createHTOContigList.R
\name{createHTOContigList}
\alias{createHTOContigList}
\title{Deconvolute Contig Information from Multiplexed Experiments}
\usage{
createHTOContigList(contig, sc.data, group.by = NULL, multi.run = NULL)
}
\arguments{
\item{contig}{The filtered contig annotation file from multiplexed experiment}

\item{sc.data}{The Seurat or Single-Cell Experiment object.}

\item{group.by}{One or more meta data headers to create the contig
list based on. If more than one header listed, the function combines
them into a single variable.}

\item{multi.run}{If using integrated single-cell object, the meta data
variable that indicates the sequencing run.}
}
\value{
Returns a list of contigs as input for \code{\link[=combineBCR]{combineBCR()}}
or \code{\link[=combineTCR]{combineTCR()}}
}
\description{
This function reprocess and forms a list of contigs for downstream analysis
in scRepertoire, \code{\link[=createHTOContigList]{createHTOContigList()}} take the filtered contig
annotation output and the single-cell RNA object to create the list.
If using an integrated single-cell object, it is recommended to split the
object by sequencing run and remove extra prefixes and suffixes on the
barcode before using \code{\link[=createHTOContigList]{createHTOContigList()}}. Alternatively,
the variable \code{multi.run} can be used to separate a list of contigs
by a meta data variable. This may have issues with the repeated barcodes.
}
\examples{
\dontrun{
filtered.contig <- read.csv(".../Sample/outs/filtered_contig_annotations.csv")

contig.list <- createHTOContigList(contig = filtered.contig, 
                                   sc.data = Seurat.Obj, 
                                   group.by = "HTO_maxID")
}

}
\concept{Loading_and_Processing_Contigs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalScatter.R
\name{clonalScatter}
\alias{clonalScatter}
\title{Scatter Plot Comparing Clones Across Two Samples}
\usage{
clonalScatter(
  input.data,
  clone.call = NULL,
  x.axis = NULL,
  y.axis = NULL,
  chain = "both",
  dot.size = "total",
  group.by = NULL,
  graph = "proportion",
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}},
\code{\link[=combineBCR]{combineBCR()}}, or \code{\link[=combineExpression]{combineExpression()}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{x.axis}{name of the list element to appear on the x.axis.}

\item{y.axis}{name of the list element to appear on the y.axis.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{dot.size}{either total or the name of the list element to
use for size of dots.}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed
by list element or active identity in the case of single-cell objects.}

\item{graph}{graph either the clonal "proportion" or "count".}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object visualizing clonal dynamics between two groupings or
a data.frame if \code{exportTable = TRUE}.
}
\description{
This function produces a scatter plot directly comparing
the specific clones between two samples. The clones will
be categorized by counts into singlets or expanded, either
exclusive or shared between the selected samples.
}
\examples{
#Making combined contig data
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))

# Using clonalScatter()
clonalScatter(combined, 
              x.axis = "P17B", 
              y.axis = "P17L",
              graph = "proportion")

}
\concept{Visualizing_Clones}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/highlightClones.R
\name{highlightClones}
\alias{highlightClones}
\title{Highlighting Specific Clones}
\usage{
highlightClones(sc.data, clone.call = NULL, sequence = NULL, cloneCall = NULL)
}
\arguments{
\item{sc.data}{The single-cell object to attach after
\code{\link[=combineExpression]{combineExpression()}}}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{sequence}{The specific sequence or sequence to highlight}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}
}
\value{
Single-cell object object with new meta data column
for indicated clones
}
\description{
Use a specific clonal sequence to highlight on top of the dimensional
reduction in single-cell object.
}
\examples{
# Getting the combined contigs
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))

# Getting a sample of a Seurat object
scRep_example  <- get(data("scRep_example"))

# Using combineExpresion()
scRep_example  <- combineExpression(combined, 
                                    scRep_example)

# Using highlightClones()
scRep_example   <- highlightClones(scRep_example,
                                   clone.call= "aa",
                                   sequence = c("CVVSDNTGGFKTIF_CASSVRRERANTGELFF"))

}
\concept{SC_Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalNetwork.R
\name{clonalNetwork}
\alias{clonalNetwork}
\title{Visualize Clonal Network in Dimensional Reductions}
\usage{
clonalNetwork(
  sc.data,
  clone.call = NULL,
  chain = "both",
  reduction = "umap",
  group.by = "ident",
  filter.clones = NULL,
  filter.identity = NULL,
  filter.proportion = NULL,
  filter.graph = FALSE,
  export.clones = NULL,
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportClones = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{sc.data}{The single-cell object after \code{\link[=combineExpression]{combineExpression()}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{reduction}{The name of the dimensional reduction of the
single-cell object.}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). This will be the nodes overlaid onto the graph.}

\item{filter.clones}{Use to select the top n clones (e.g., \verb{filter.clones`**` = 2000) or n of clones based on the minimum number of all the comparators (e.g., `filter.clone} = "min").}

\item{filter.identity}{Display the network for a specific level of the
indicated identity.}

\item{filter.proportion}{Remove clones from the network below a specific
proportion.}

\item{filter.graph}{Remove the reciprocal edges from the half of the graph,
allowing for cleaner visualization.}

\item{export.clones}{Exports a table of clones that are shared
across multiple identity groups and ordered by the total number
of clone copies.}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices:palettes]{hcl.pals}.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportClones}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.clones} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
ggplot object
}
\description{
This function generates a network based on clonal
proportions of an indicated identity and then superimposes
the network onto a single-cell object dimensional reduction
plot.
}
\examples{
\dontrun{
# Getting the combined contigs
combined <- combineTCR(contig_list,
                        samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))

# Getting a sample of a Seurat object
scRep_example  <- get(data("scRep_example"))

# Using combineExpresion()
scRep_example  <- combineExpression(combined, scRep_example)

# Using clonalNetwork()
clonalNetwork(scRep_example,
              reduction = "umap",
              group.by = "seurat_clusters")
}

}
\concept{SC_Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combineExpression.R
\name{combineExpression}
\alias{combineExpression}
\title{Adding Clonal Information to Single-Cell Object}
\usage{
combineExpression(
  input.data,
  sc.data,
  clone.call = NULL,
  chain = "both",
  group.by = NULL,
  proportion = TRUE,
  filter.na = NULL,
  clone.size = NULL,
  add.label = NULL,
  cloneCall = NULL,
  cloneSize = NULL,
  filterNA = NULL,
  addLabel = NULL
)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}}, \code{\link[=combineBCR]{combineBCR()}} or a list of
both c(\code{\link[=combineTCR]{combineTCR()}}, \code{\link[=combineBCR]{combineBCR()}}).}

\item{sc.data}{The Seurat or Single-Cell Experiment (SCE) object to attach}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL}
(for both light chains), \code{both}.}

\item{group.by}{A column header in lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, will be based on the list element.}

\item{proportion}{Whether to proportion (\code{TRUE}) or total
frequency (\code{FALSE}) of the clone based on the group.by variable.}

\item{filter.na}{Method to subset Seurat/SCE object of barcodes without
clone information}

\item{clone.size}{The bins for the grouping based on proportion or frequency.
If proportion is \code{FALSE} and the clone.sizes are not set high enough
based on frequency, the upper limit of clone.sizes will be automatically
updated.}

\item{add.label}{This will add a label to the frequency header, allowing
the user to try multiple group.by variables or recalculate frequencies after
subsetting the data.}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{cloneSize}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.size} instead.}

\item{filterNA}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{filter.na} instead.}

\item{addLabel}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{add.label} instead.}
}
\value{
Single-cell object with clone information added to meta data
information
}
\description{
This function adds the immune receptor information to the Seurat or
SCE object to the meta data. By default this function also calculates
the frequencies and proportion of the clones by sequencing
run (\code{group.by} = NULL). To change how the frequencies/proportions
are calculated, select a column header for the \code{group.by} variable.
Importantly, before using \code{\link[=combineExpression]{combineExpression()}} ensure the
barcodes of the single-cell object object match the barcodes in the output
of the \code{\link[=combineTCR]{combineTCR()}} or \code{\link[=combineBCR]{combineBCR()}}.
}
\examples{
# Getting the combined contigs
combined <- combineTCR(contig_list, 
                        samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))

# Getting a sample of a Seurat object
scRep_example <- get(data("scRep_example"))

# Using combineExpresion()
scRep_example <- combineExpression(combined, scRep_example)

}
\concept{SC_Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clonalCompare.R
\name{clonalCompare}
\alias{clonalCompare}
\title{Compare Clonal Abundance Across Variables}
\usage{
clonalCompare(
  input.data,
  clone.call = NULL,
  chain = "both",
  samples = NULL,
  clones = NULL,
  top.clones = NULL,
  highlight.clones = NULL,
  relabel.clones = FALSE,
  group.by = NULL,
  order.by = NULL,
  graph = "alluvial",
  proportion = TRUE,
  export.table = NULL,
  palette = "inferno",
  cloneCall = NULL,
  exportTable = NULL,
  ...
)
}
\arguments{
\item{input.data}{The product of \code{\link{combineTCR}},
\code{\link{combineBCR}}, or \code{\link{combineExpression}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{chain}{The TCR/BCR chain to use. Use \code{both} to include both chains
(e.g., TRA/TRB). Accepted values: \code{TRA}, \code{TRB}, \code{TRG}, \code{TRD}, \code{IGH}, \code{IGL},
\code{IGK}, \code{Light} (for both light chains), or \code{both} (for TRA/B and Heavy/Light).}

\item{samples}{The specific samples to isolate for visualization.}

\item{clones}{The specific clonal sequences of interest}

\item{top.clones}{The top number of clonal sequences per group.
(e.g., top.clones = 5)}

\item{highlight.clones}{Clonal sequences to highlight, if present,
all other clones returned will be grey}

\item{relabel.clones}{Simplify the legend of the graph by returning
clones that are numerically indexed}

\item{group.by}{A column header in the metadata or lists to group the analysis
by (e.g., "sample", "treatment"). If \code{NULL}, data will be analyzed
by list element or active identity in the case of single-cell objects.}

\item{order.by}{A character vector defining the desired order of elements
of the \code{group.by} variable. Alternatively, use \code{alphanumeric} to sort groups
automatically.}

\item{graph}{The type of plot to generate. Accepted values are \code{alluvial}
(default) or \code{area}}

\item{proportion}{If \code{TRUE} (default), the y-axis will represent the
proportional abundance of clones. If \code{FALSE}, the y-axis will represent
raw clone counts.`}

\item{export.table}{If \code{TRUE}, returns a data frame or matrix of the results
instead of a plot.}

\item{palette}{Colors to use in visualization - input any
\link[grDevices]{hcl.pals}}

\item{cloneCall}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{clone.call} instead.}

\item{exportTable}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{export.table} instead.}

\item{...}{Additional arguments passed to the ggplot theme}
}
\value{
A ggplot object visualizing proportions of clones by groupings, or a
data.frame if \code{export.table = TRUE}.
}
\description{
This function visualizes the relative abundance of specific clones across
different samples or groups. It is useful for tracking how the proportions
of top clones change between conditions. The output can be an alluvial plot
to trace clonal dynamics or an area plot to show compositional changes.
}
\examples{
# Making combined contig data
combined <- combineTCR(contig_list,
                       samples = c("P17B", "P17L", "P18B", "P18L",
                                   "P19B","P19L", "P20B", "P20L"))

# Using clonalCompares()
clonalCompare(combined,
              top.clones = 5,
              samples = c("P17B", "P17L"),
              clone.call = "aa")

}
\concept{Visualizing_Clones}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addVariable.R
\name{addVariable}
\alias{addVariable}
\title{Adding Variables After combineTCR() or combineBCR()}
\usage{
addVariable(input.data, variable.name = NULL, variables = NULL)
}
\arguments{
\item{input.data}{The product of \code{\link[=combineTCR]{combineTCR()}} or \code{\link[=combineBCR]{combineBCR()}}.}

\item{variable.name}{A character string that defines the new variable to add.}

\item{variables}{A character vector defining the desired column value for
each list element. Must match the length of the product of \code{\link[=combineTCR]{combineTCR()}} or
\code{\link[=combineBCR]{combineBCR()}}.}
}
\value{
input.data list with the variable column added to each element.
}
\description{
This function adds variables to the product of \code{\link[=combineTCR]{combineTCR()}},
or \code{\link[=combineBCR]{combineBCR()}} to be used in later visualizations.
For each element, the function will add a column (labeled by
\code{variable.name}) with the variable. The length of the
\code{variables} parameter needs to match the length of the
combined object.
}
\examples{
combined <- combineTCR(contig_list, 
                       samples = c("P17B", "P17L", "P18B", "P18L", 
                                    "P19B","P19L", "P20B", "P20L"))
combined <- addVariable(combined, 
                        variable.name = "Type", 
                        variables = rep(c("B", "L"), 4))

}
\concept{Loading_and_Processing_Contigs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vizCirclize.R
\name{vizCirclize}
\alias{vizCirclize}
\title{Visualize Clonal Relationships as a Chord Diagram}
\usage{
vizCirclize(
  sc.data,
  clone.call = NULL,
  group.by = NULL,
  method = c("unique", "jaccard", "overlap"),
  proportion = FALSE,
  directional = FALSE,
  self.link = 1,
  include.self = TRUE,
  transparency = 0.5,
  link.visible = TRUE,
  annotate.sectors = TRUE,
  min.shared = 0,
  palette = "inferno",
  sector.colors = NULL,
  export.table = FALSE
)
}
\arguments{
\item{sc.data}{The single-cell object after \code{\link[=combineExpression]{combineExpression()}}.}

\item{clone.call}{Defines the clonal sequence grouping. Accepted values
are: \code{gene} (VDJC genes), \code{nt} (CDR3 nucleotide sequence), \code{aa} (CDR3 amino
acid sequence), or \code{strict} (VDJC + nt). A custom column header can also be used.}

\item{group.by}{A column header (or vector of column headers for hierarchical
grouping) in the metadata to group the analysis by.}

\item{method}{The method for calculating link values: \code{"unique"} (default) counts
unique shared clones, \code{"jaccard"} calculates Jaccard similarity,
\code{"overlap"} calculates overlap coefficient.}

\item{proportion}{Calculate the relationship by unique clones (\code{FALSE}, default)
or normalized by proportion (\code{TRUE}).}

\item{directional}{If \code{TRUE}, show directional arrows on chords. Default is \code{FALSE}.}

\item{self.link}{How to handle self-links. \code{1} = show as loops, \code{2} = show as
parallel lines. Default is \code{1}.}

\item{include.self}{Include self-links (clones within a single group). Default is \code{TRUE}.}

\item{transparency}{Transparency of the chord links (0-1). Default is \code{0.5}.}

\item{link.visible}{If \code{FALSE}, hide the chord links and show only sectors.}

\item{annotate.sectors}{If \code{TRUE} (default), display sector names.}

\item{min.shared}{Minimum number of shared clones to include a link (default 0).}

\item{palette}{Colors to use for sectors - input any \link[grDevices:palettes]{hcl.pals}.}

\item{sector.colors}{Named vector of colors for specific sectors. Overrides palette.}

\item{export.table}{If \code{TRUE}, returns the data instead of plotting.}
}
\value{
Invisibly returns the circlize data list. If circlize is not installed
or \code{export.table = TRUE}, returns the data that would be used for plotting.
}
\description{
This function creates a chord diagram visualization of shared clones between
groups using the circlize package. It provides a convenient wrapper around
\code{\link[=getCirclize]{getCirclize()}} that handles the circlize plotting code automatically.
}
\details{
This function requires the circlize package to be installed. If circlize is
not available, the function will return the data that would be used for plotting
and provide instructions for manual plotting.

The chord diagram shows relationships between groups (sectors) where the width
of each chord represents the number or proportion of shared clones between
the connected groups.
}
\examples{
\dontrun{
# Getting the combined contigs
combined <- combineTCR(contig_list,
                        samples = c("P17B", "P17L", "P18B", "P18L",
                                    "P19B","P19L", "P20B", "P20L"))

# Getting a sample of a Seurat object
scRep_example <- get(data("scRep_example"))
scRep_example <- combineExpression(combined,
                                   scRep_example)

# Simple chord diagram
vizCirclize(scRep_example, group.by = "seurat_clusters")

# Directional chord diagram with arrows
vizCirclize(scRep_example,
            group.by = "seurat_clusters",
            directional = TRUE)

# Multi-level grouping
scRep_example$Patient <- substring(scRep_example$orig.ident, 1, 3)
vizCirclize(scRep_example,
            group.by = c("Patient", "seurat_clusters"))
}

}
\seealso{
\code{\link[=getCirclize]{getCirclize()}} for generating the underlying data
}
\author{
Nick Borcherding
}
\concept{SC_Functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{scRep_example}
\alias{scRep_example}
\title{A Seurat Object of 500 Single T cells,}
\description{
The object is compatible with \code{contig_list} and the TCR
sequencing data can be added with \code{combineExpression}.  The data is
from 4 patients with acute respiratory distress, with samples taken
from both the lung and peripheral blood. More information on the
data can be found in the following
\href{https://pubmed.ncbi.nlm.nih.gov/33622974/}{manuscript}.
}
\concept{Data}
